/***
 *  $Id: ResidueIndex.hpp 770 2009-07-26 19:37:20Z zola $
 **
 *  File: ResidueIndex.hpp
 *  Created: Dec 17, 2006
 *
 *  Department of Electrical & Computer Engineering
 *  Iowa State University of Science & Technology
 *  Ames, IA 50011, USA
 *
 *  Author: Jaroslaw Zola <zola@iastate.edu>
 *  Please see attached LICENSE file.
 */

#ifndef JZ_RESIDUE_INDEX_HPP
#define JZ_RESIDUE_INDEX_HPP

#include "config.hpp"
#include "cache_manager.hpp"
#include "util_mpi.hpp"
#include <iostream>
#include <vector>
#include <string.h>

#if USE_RMA

class GlobalStorage {
public:    
    void init(int first_S, int last_S, int nseq) {
	win = false;

	fS = first_S;
	lS = last_S;
	
	BSIZE = 0;
	BUFF_ = 0;
	OFF_ = new unsigned int[nseq];
	LEN_ = new unsigned int[nseq];

	memset(OFF_, 0, nseq * sizeof(unsigned int));
	memset(LEN_, 0, nseq * sizeof(unsigned int));
    } // init

    void reallocate(int s) {
 	BSIZE += s;	
 	BUFF_ = (int*)realloc(BUFF_, BSIZE * sizeof(int));
 	memset(BUFF_ + BSIZE - s, 0, s * sizeof(int));	
    } // reallocate

    void free() {
	delete[] LEN_;
	delete[] OFF_;

	if (win == true) {
	    MPI_Win_free(&G_STOR_WIN);
	    MPI_Free_mem(BUFF_);
	}
	else ::free(BUFF_);
	
	BSIZE = 0;
	BUFF_ = 0;
	OFF_ = LEN_ = 0;
    } // free

    void create_mpi_win(MPI_Comm comm) {
	int* tmp = 0;
	MPI_Alloc_mem(BSIZE * sizeof(int), MPI_INFO_NULL, &tmp);
	
	memcpy(tmp, BUFF_, BSIZE * sizeof(int));
	::free(BUFF_);
	
	BUFF_ = tmp;

	MPI_Win_create(BUFF_, BSIZE * sizeof(int), sizeof(int), MPI_INFO_NULL,
		       comm, &G_STOR_WIN);

	win = true;
    } // create_mpi_win

    bool win;

    unsigned int BSIZE;
    int* BUFF_;
    unsigned int* OFF_; // offset for given sequence
    unsigned int* LEN_; // lenght of total sequence data (all residues)

    MPI_Win G_STOR_WIN;

    int fS, lS; // first, and one-past-last sequence index
    std::vector<int> seq2host;

}; // GlobalSorage


extern GlobalStorage G_STORAGE;


class ResCol {
public:
    ResCol() : values_(0) { }

    int& operator[](int n) { return values_[n]; }

    void allocate(int n) {
	this->free();
	values_ = (int*)malloc(n * sizeof(int));
    } // allocate

    void reallocate(int n) { 
	values_ = (int*)realloc(values_, n * sizeof(int)); 
    } // reallocate

    void free() { 
	::free(values_);
	values_ = 0;
    } // free
    
    int* values_;
    
}; // class ResCol

class ResRow {
public:
    ResRow() : size_(0), column_(0) { }
    
    ResCol& operator[](int n) { return column_[n]; }

    void allocate(int n) {
	this->free();
	size_ = n;
	column_ = new ResCol[n];
    } // allocate

    void free() {
	for (int i = 0; i < size_; ++i) column_[i].free();
	delete[] column_;
	size_ = 0;
	column_ = 0;
    } // free
    
    int size_;
    ResCol* column_;

}; // class ResRow


// HERE GOES ALL RMA 

class ResIdxSeq {
public:
    ResIdxSeq() : storage(0) { }

    int* operator[](int n) { return storage + storage[n]; }
    
    int* storage;
    int* state;

}; // ResIdxSeq

class ResIdx {
public:
    ResIdx() : miss(0), tot(0), rank(-1), mpi_env(0), cache_size(0), cache() { }

    ResIdx(int) 
	: miss(0), tot(0), rank(-1), mpi_env(0), cache_size(0), cache() { }
        
    operator bool() const { return !cache.empty(); }
    
    void get(int n, ResIdxSeq& seq) {
	int* addr = 0;
	mem_block mb(0, 0);

	int host = G_STORAGE.seq2host[n];

	int len = G_STORAGE.LEN_[n];
	int off = G_STORAGE.OFF_[n];

//	++tot;

	if (host == rank) { addr = G_STORAGE.BUFF_ + off; }
	else { // remote access
	    bool res = cache.find(n, mb);

	    if (res == true) {
		(*mb.first) = 1;
		addr = mb.second;
	    }
	    else {
//		++miss;

		// if miss read remote
		addr = new int[len];

		MPI_Win_lock(MPI_LOCK_SHARED, host, 0, G_STORAGE.G_STOR_WIN);
		MPI_Get(addr, len, MPI_INT, host, off, len, MPI_INT,
			G_STORAGE.G_STOR_WIN);
		MPI_Win_unlock(host, G_STORAGE.G_STOR_WIN);	       

		mb.first = new int(1);
		mb.second = addr;

		// 8 is overhead
		cache.insert(n, mb, 8 + len * sizeof(int));
	    }
	} // remote access

	seq.state = mb.first;
	seq.storage = addr;
    } // get

    void alloc(int n) {
	rank = mpi_env->rank();
	cache.clear();
	cache = cache_manager<int, mem_block>(cache_size);

	miss = 0;
	tot = 0;
    } // alloc

    void free() { 
	cache.clear();
// 	std::cout << "rank: " << mpi_env->rank()
// 		  << " tot: " << tot << " miss: " << miss 
// 		  << " hit rate: " << (1.0 * (tot - miss)) / tot << std::endl;
    }

    int miss;
    int tot;
    
    int rank;
    const MPI_Env* mpi_env;

    typedef std::pair<int*, int*> mem_block; // state, address

    unsigned long int cache_size;
    cache_manager<int, mem_block> cache;

}; // class ResIdx

#endif // USE_RMA

#endif // JZ_RESIDUE_INDEX_HPP
