/***
 *  $Id: time_log.hpp 700 2006-11-16 18:43:43Z zola $
 **
 *  File: time_log.hpp
 *  Developed: Aug 20, 2005
 *
 *  Institute of Computer & Information Sciences
 *  Czestochowa University of Technology
 *  Dabrowskiego 73
 *  42-200 Czestochowa, Poland
 *  tel/fax: +48 (0....)34 3250589
 *
 *  Author: Jaroslaw Zola <zola@icis.pcz.pl>
 *  Copyright (c) 2005 Jaroslaw Zola <zola@icis.pcz.pl>
 *  For copyright details please contact the author.
 */

#ifndef YAT_TIME_LOG_HPP
#define YAT_TIME_LOG_HPP

#include <iomanip>
#include <iostream>
#include <string>
#include <vector>
#include <sys/time.h>


/** Yet Another Toolkit namespace.
 */
namespace yat {

  /** Class that performs logging of the execution times.
   */
  class TimeLog {
  public:
      /**
       */
      typedef unsigned long int size_type;
      

      explicit TimeLog(const std::string& name = "", size_type n = 1024)
	  : name_(name), start_(0.0), stop_(0.0) { 
	  
	  points_.reserve(n);
	  times_.reserve(n);

      } // TimeLog


      /** @return the number of check-points.
       */
      size_type size() const { return points_.size(); }

      /** @return the total time of execution (in seconds).
       */
      double total() const { return stop_ - start_; }


      double at_point(size_type n) const {
	  return (n < 1) ? (times_[0] - start_) : (times_[n] - times_[n-1]);
      } // at_point

      double to_point(size_type n) const { return times_[n]; }


      void start() { start_ = get_time_(); }
      
      void stop() { stop_ = get_time_(); }


      void add(const std::string id = "") { 
	  times_.push_back(get_time_());
	  points_.push_back(id);
      } // add

      
      void print(std::ostream& out) const {
	  out << std::fixed << std::setprecision(6);
	  out << name_ << std::endl;
	  out << points_.size() << std::endl;
	  
	  size_type n = points_.size();

	  for (size_type i = 0; i < n; ++i) {
	      if (!points_[i].empty()) {
		  out << points_[i] << " ";
	      }
	      else out << ". ";
	      out << at_point(i) << std::endl;
	  }

	  out << std::endl;
	  
	  size_type tt = static_cast<size_type>(total());
	  size_type ht = tt / 3600;
	  size_type mt = (tt % 3600) / 60;
	  size_type st = (tt % 3600) % 60;

	  out << "Total" << std::endl;
	  out << total() << std::endl;
	  out << ht << ":" << mt << ":" << st << std::endl;
      } // print


  private:
      timeval t_;

      double get_time_() {
	  gettimeofday(&t_, 0);
	  return t_.tv_sec + (0.000001 * t_.tv_usec);
      } // get_time_


      std::string name_;

      double start_;
      double stop_;

      std::vector<std::string> points_;
      std::vector<double> times_;

      size_type pos_;      

  }; // class TimeLog

} // namespace yat

#endif // YAT_TIME_LOG_HPP
