/***
 *  $Id: JZ_ConstraintList.cpp 770 2009-07-26 19:37:20Z zola $
 **
 *  File: JZ_ConstraintList.cpp
 *  Created: Dec 11, 2006
 *
 *  Department of Electrical & Computer Engineering
 *  Iowa State University of Science & Technology
 *  Ames, IA 50011, USA
 *
 *  Author: Jaroslaw Zola <zola@iastate.edu>
 *  Please see attached LICENSE file.
 */

#include "JZ_ConstraintList.hpp"

struct bucket_comp {
    
    bool operator()(const FakeEntry& lhs, const FakeEntry& rhs) const {
	int l = (lhs.entry[0] + lhs.entry[1]) % p;
	int r = (rhs.entry[0] + rhs.entry[1]) % p;

	return l < r;
    }

    int p;
}; // struct bucket_comp

void par_sort_library(TC_ConstraintList* CL, const MPI_Env& mpi_env) {
    int rank = mpi_env.rank();
    unsigned int nseq = CL->S->nseq;

    std::vector<int> pack_size(mpi_env.size(), 0);
    std::vector<int> bucket_size(mpi_env.size() * mpi_env.size(), 0);

    int recv_size = 0;
    short int* recv_buf = 0;

    std::vector<int> send_disp(mpi_env.size() + 1, 0);
    std::vector<int> recv_count(mpi_env.size(), 0);
    std::vector<int> recv_disp(mpi_env.size() + 1, 0);

    if (mpi_env.am_I_root()) {
	std::cout << "\nDISTRIBUTING LIBRARY.....\n";

	// collect sizes of buckets
	MPI_Allgather(&pack_size[0], mpi_env.size(), MPI_INT,
		      &bucket_size[0], mpi_env.size(), MPI_INT, mpi_env.comm());

	// receive data
	for (int i = 0; i < mpi_env.size(); ++i) {
	    recv_size += bucket_size[i * mpi_env.size() + rank];

	    recv_count[i] = bucket_size[i * mpi_env.size() + rank];
	    recv_disp[i + 1] = recv_disp[i] + recv_count[i];
	}

	recv_buf = (short int*)malloc(sizeof(short int) * recv_size);

	MPI_Alltoallv(CL->L.tab, &pack_size[0], &send_disp[0], MPI_SHORT,
		      recv_buf, &recv_count[0], &recv_disp[0], MPI_SHORT,
		      mpi_env.comm());

	std::cout << ".....[Done]\n";
    }
    else { // WORKER
	for (int i = 0; i < CL->ne; ++i) {
	    int pos = (CL->L[i][0] + CL->L[i][1]) % mpi_env.size();
	    pack_size[pos] += FakeArr::entry_size;
	}

	// collect sizes of buckets
	MPI_Allgather(&pack_size[0], mpi_env.size(), MPI_INT,
		      &bucket_size[0], mpi_env.size(), MPI_INT, mpi_env.comm());

	for (int i = 0; i < mpi_env.size(); ++i) {
	    recv_size += bucket_size[i * mpi_env.size() + rank];

	    send_disp[i + 1] = send_disp[i] + pack_size[i];

	    recv_count[i] = bucket_size[i * mpi_env.size() + rank];
	    recv_disp[i + 1] = recv_disp[i] + recv_count[i];
	}

	// sort buckets
	bucket_comp cmp; cmp.p = mpi_env.size();
	FakeEntry* arr = (FakeEntry*)(CL->L.tab);
	std::sort(arr, arr + CL->ne, cmp);

	// send/receive buckets
	recv_buf = (short int*)malloc(sizeof(short int) * recv_size);

	MPI_Alltoallv(CL->L.tab, &pack_size[0], &send_disp[0], MPI_SHORT,
		      recv_buf, &recv_count[0], &recv_disp[0], MPI_SHORT,
		      mpi_env.comm());
    } // WORKER

    // set library
    CL->L.free();
    CL->L.tab = recv_buf;
    CL->L.size = recv_size;
    CL->ne = recv_size / FakeArr::entry_size;
    CL->max_L_len = CL->ne;
    CL->seq_indexed = 0;
    CL->residue_indexed = 0;

    // compres library
    TC_compact_list(CL, 0, CL->ne, "default");
} // par_sort_library


TC_ConstraintList*
JZ_master_compile_cl(TC_ConstraintList* CL,
		     const std::vector<std::string>& in_files,
		     const char* weight, const MPI_Env& mpi_env) {
    if (!CL)
	return CL;

    enum {
	JOB_REQ, JOB_RES
    };

    std::vector<TC_Method *> methods;
    TC_Method *meth = 0;
    unsigned int job_buf_size = 0;
    char *cbuffer = 0;
    const unsigned int meth_buf_size = 10 * (sizeof(int) + 100) + 4 * sizeof(int) + 2 * (sizeof(int) + 1000);
    int *ibuffer = 0;
    unsigned int job_res[3];
    MPI_Status status;


    bool no_jobs = true;
    std::vector<std::string>::const_reverse_iterator rit = in_files.rbegin(), rend = in_files.rend();

    for (; rit != rend; ++rit)
	if ((*rit)[0] == 'X') {
	    CL = TC_read_constraint_list(CL, rit->c_str() + 1, "matrix", 0);
	    if (CL->M)
		break;
	}

    if (!CL->M) {
	std::vector<const std::string *> files;
	char id;
	const char *m_name = 0, *m_file = 0;

	std::cout << "\nCOMPILING PRIMARY LIBRARY.....\n";

	methods.reserve(in_files.size());
	files.reserve(in_files.size());

	for (rit = in_files.rbegin(); rit != rend; ++rit) {
	    id = (*rit)[0];

	    if (id == 'M') {
		m_name = rit->c_str();
		m_file = TC_method_name2method_file(m_name + 1);

		if (!m_file) {
		    if (AR_check_file_exists(m_name))
			m_file = m_name;
		    else if (AR_check_file_exists(m_name + 1))
			m_file = m_name + 1;
		    else {
			std::cerr << '\n' << PROGRAM << ": WARNING: '" << m_name << "' is not a valid method.\n";
			continue;
		    }
		}

		meth = TC_method_file2tc_method(m_file);

		if (!meth->executable[0]) {
		    std::cerr << '\n' << PROGRAM << ": WARNING: '" << m_name << "' is not a valid method.\n";
		    AR_free(meth);
		    continue;
		}

		if (!TC_strm(weight, "default"))
		    strncpy(meth->weight, weight, 100 - 1);

		methods.push_back(meth);
	    }
	    else if (id == 'A' || id == 'L') {
		files.push_back(&(*rit));
	    }
	} // for (rit)

	std::vector<const std::string *>::const_iterator it_f = files.begin();

	if (!methods.empty()) {
	    std::vector<AR_lib_job> job_list;
	    std::vector<int> job_cnt(methods.size());

	    AR_method_list2job_list(methods, CL->S, CL->DM, CL->lib_list, job_list, job_cnt);
	    std::sort(job_list.begin(), job_list.end(), AR_job_cmp(job_cnt));

	    if (!job_list.empty()) {
		no_jobs = false;

		const unsigned int N_m = methods.size();
		const unsigned int N_j = job_list.size();
		const unsigned int w_size = mpi_env.size() - 1; // without master
		unsigned int chunk_size = static_cast<unsigned int>(ceil(N_j / (2.0 * w_size)));

		job_buf_size = chunk_size * 3 + 1;

		{ // temporary block
		    unsigned int tmp_buf[2];

		    tmp_buf[0] = N_m;
		    tmp_buf[1] = job_buf_size;

		    MPI_Bcast(tmp_buf, 2, MPI_UNSIGNED, mpi_env.get_root(), mpi_env.comm());

		    cbuffer = new char[meth_buf_size];
		    int pack_pos = 0;
		    int len;

		    for (unsigned int i = 0; i < N_m; ++i) {
			meth = methods[i];

			len = strlen(meth->executable) + 1;
			MPI_Pack(&len, 1, MPI_INT, cbuffer, meth_buf_size, &pack_pos, mpi_env.comm());
			MPI_Pack(meth->executable, len, MPI_CHAR, cbuffer, meth_buf_size, &pack_pos, mpi_env.comm());

			len = strlen(meth->executable2) + 1;
			MPI_Pack(&len, 1, MPI_INT, cbuffer, meth_buf_size, &pack_pos, mpi_env.comm());
			MPI_Pack(meth->executable2, len, MPI_CHAR, cbuffer, meth_buf_size, &pack_pos, mpi_env.comm());

			len = strlen(meth->in_flag) + 1;
			MPI_Pack(&len, 1, MPI_INT, cbuffer, meth_buf_size, &pack_pos, mpi_env.comm());
			MPI_Pack(meth->in_flag, len, MPI_CHAR, cbuffer, meth_buf_size, &pack_pos, mpi_env.comm());

			len = strlen(meth->in_flag2) + 1;
			MPI_Pack(&len, 1, MPI_INT, cbuffer, meth_buf_size, &pack_pos, mpi_env.comm());
			MPI_Pack(meth->in_flag2, len, MPI_CHAR, cbuffer, meth_buf_size, &pack_pos, mpi_env.comm());

			len = strlen(meth->out_flag) + 1;
			MPI_Pack(&len, 1, MPI_INT, cbuffer, meth_buf_size, &pack_pos, mpi_env.comm());
			MPI_Pack(meth->out_flag, len, MPI_CHAR, cbuffer, meth_buf_size, &pack_pos, mpi_env.comm());

			len = strlen(meth->aln_mode) + 1;
			MPI_Pack(&len, 1, MPI_INT, cbuffer, meth_buf_size, &pack_pos, mpi_env.comm());
			MPI_Pack(meth->aln_mode, len, MPI_CHAR, cbuffer, meth_buf_size, &pack_pos, mpi_env.comm());

			len = strlen(meth->out_mode) + 1;
			MPI_Pack(&len, 1, MPI_INT, cbuffer, meth_buf_size, &pack_pos, mpi_env.comm());
			MPI_Pack(meth->out_mode, len, MPI_CHAR, cbuffer, meth_buf_size, &pack_pos, mpi_env.comm());

			len = strlen(meth->seq_type) + 1;
			MPI_Pack(&len, 1, MPI_INT, cbuffer, meth_buf_size, &pack_pos, mpi_env.comm());
			MPI_Pack(meth->seq_type, len, MPI_CHAR, cbuffer, meth_buf_size, &pack_pos, mpi_env.comm());

			len = strlen(meth->weight) + 1;
			MPI_Pack(&len, 1, MPI_INT, cbuffer, meth_buf_size, &pack_pos, mpi_env.comm());
			MPI_Pack(meth->weight, len, MPI_CHAR, cbuffer, meth_buf_size, &pack_pos, mpi_env.comm());

			len = strlen(meth->matrix) + 1;
			MPI_Pack(&len, 1, MPI_INT, cbuffer, meth_buf_size, &pack_pos, mpi_env.comm());
			MPI_Pack(meth->matrix, len, MPI_CHAR, cbuffer, meth_buf_size, &pack_pos, mpi_env.comm());

			MPI_Pack(&(meth->gop), 1, MPI_INT, cbuffer, meth_buf_size, &pack_pos, mpi_env.comm());
			MPI_Pack(&(meth->gep), 1, MPI_INT, cbuffer, meth_buf_size, &pack_pos, mpi_env.comm());

			MPI_Pack(&(meth->minid), 1, MPI_INT, cbuffer, meth_buf_size, &pack_pos, mpi_env.comm());
			MPI_Pack(&(meth->maxid), 1, MPI_INT, cbuffer, meth_buf_size, &pack_pos, mpi_env.comm());

			len = strlen(meth->param) + 1;
			MPI_Pack(&len, 1, MPI_INT, cbuffer, meth_buf_size, &pack_pos, mpi_env.comm());
			MPI_Pack(meth->param, len, MPI_CHAR, cbuffer, meth_buf_size, &pack_pos, mpi_env.comm());

			len = strlen(meth->param1) + 1;
			MPI_Pack(&len, 1, MPI_INT, cbuffer, meth_buf_size, &pack_pos, mpi_env.comm());
			MPI_Pack(meth->param1, len, MPI_CHAR, cbuffer, meth_buf_size, &pack_pos, mpi_env.comm());

#if DEBUG_MPI
			std::cout << "\nBCASTING DESC. OF '" << meth->executable << "' METHOD\n";
#endif
			MPI_Bcast(cbuffer, pack_pos, MPI_PACKED, mpi_env.get_root(), mpi_env.comm());
			pack_pos = 0;
		    } // for (i)
		} // temporary block

		delete[] cbuffer;

		ibuffer = new int[job_buf_size];

		int to_send = N_j, to_recv = 0;
		unsigned int curr_job = 0;

		// send initial jobs (last in map is always root)
		unsigned int i;

		for (i = 0; i < w_size; ++i) {
		    AR_send_jobs(mpi_env.map(i), mpi_env, chunk_size, curr_job, job_list, ibuffer);
		    to_send -= chunk_size;
		    ++to_recv;

		    if (to_send < 1)
			break;
		}

		chunk_size = static_cast<unsigned int>(to_send / (2.0 * w_size));
		if (chunk_size == 0)
		    ++chunk_size;


		// continue scheduling
		do {
		    if (it_f != files.end()) {
			CL = TC_read_constraint_list(CL, (*it_f)->c_str(), 0, weight);
			++it_f;
		    }

		    MPI_Recv(job_res, 3, MPI_UNSIGNED, MPI_ANY_SOURCE, LIB_JOB_TAG + 1, mpi_env.comm(), &status);
		    --to_recv;

		    // request
		    if (job_res[0] == JOB_REQ) {
			if (to_send > 0) {
			    AR_send_jobs(job_res[1], mpi_env, chunk_size, curr_job, job_list, ibuffer);
			    to_send -= chunk_size;
			    ++to_recv;

			    chunk_size = static_cast<unsigned int>(to_send / w_size);
			    if (chunk_size == 0)
				++chunk_size;
			}
			else {
			    curr_job = 0;
			    AR_send_jobs(job_res[1], mpi_env, 0, curr_job, job_list, ibuffer);
			}
		    }
		}
		while ((to_send > 0) || (to_recv > 0));

		// if some there are more hosts than chunks
		++i;
		curr_job = 0;
		
		for (; i < w_size; ++i) {
		    AR_send_jobs(mpi_env.map(i), mpi_env, 0, curr_job, job_list, ibuffer);
		}
	       	       
		delete[] ibuffer;
	    }
	} // if (!methods.empty())

	std::cout << ".....[Done]\n";

	for (; it_f != files.end(); ++it_f) {
	    CL = TC_read_constraint_list(CL, (*it_f)->c_str(), 0, weight);
	}


	par_sort_library(CL, mpi_env);
    
    
    } // if (!CL->M)


    if (no_jobs) {
	unsigned int msg[2] = { 0, 0 };
	MPI_Bcast(msg, 2, MPI_UNSIGNED, mpi_env.get_root(), mpi_env.comm());
    }


    std::vector<TC_Method *>::iterator it = methods.begin();

    for (; it != methods.end(); ++it) {
//	TC_free_constraint_list((*it)->PW_CL);
	delete (*it)->PW_CL;
	AR_free(*it);
    }

    return CL;

} // JZ_master_compile_cl


TC_ConstraintList*
JZ_worker_compile_cl(TC_ConstraintList* CL,
		     const std::vector<std::string>& in_files,
		     const char* weight, const MPI_Env& mpi_env) {

    if (!CL)
	return CL;

    enum {
	JOB_REQ, JOB_RES
    };

    std::vector<TC_Method *> methods;
    TC_Method *meth = 0;
    unsigned int job_buf_size = 0;
    char *cbuffer = 0;
    const unsigned int meth_buf_size = 10 * (sizeof(int) + 100) + 4 * sizeof(int) + 2 * (sizeof(int) + 1000);
    std::vector<TC_ConstraintList *> libs;
    int *ibuffer = 0;
    unsigned int job_res[3];
    MPI_Status status;
    

    { // temporary block
	unsigned int tmp_buf[2];

	MPI_Bcast(tmp_buf, 2, MPI_UNSIGNED, mpi_env.get_root(), mpi_env.comm());

	methods.resize(tmp_buf[0]);
	job_buf_size = tmp_buf[1];

	if (methods.size() > 0)
	    cbuffer = new char[meth_buf_size];
	int pack_pos = 0;
	int len;

	for (unsigned int i = 0; i < methods.size(); ++i) {
	    MPI_Bcast(cbuffer, meth_buf_size, MPI_PACKED, mpi_env.get_root(), mpi_env.comm());

	    meth = static_cast<TC_Method *>(AR_allocate(1, sizeof(TC_Method)));

	    MPI_Unpack(cbuffer, meth_buf_size, &pack_pos, &len, 1, MPI_INT, mpi_env.comm());
	    MPI_Unpack(cbuffer, meth_buf_size, &pack_pos, meth->executable, len, MPI_CHAR, mpi_env.comm());

	    MPI_Unpack(cbuffer, meth_buf_size, &pack_pos, &len, 1, MPI_INT, mpi_env.comm());
	    MPI_Unpack(cbuffer, meth_buf_size, &pack_pos, meth->executable2, len, MPI_CHAR, mpi_env.comm());

	    MPI_Unpack(cbuffer, meth_buf_size, &pack_pos, &len, 1, MPI_INT, mpi_env.comm());
	    MPI_Unpack(cbuffer, meth_buf_size, &pack_pos, meth->in_flag, len, MPI_CHAR, mpi_env.comm());

	    MPI_Unpack(cbuffer, meth_buf_size, &pack_pos, &len, 1, MPI_INT, mpi_env.comm());
	    MPI_Unpack(cbuffer, meth_buf_size, &pack_pos, meth->in_flag2, len, MPI_CHAR, mpi_env.comm());

	    MPI_Unpack(cbuffer, meth_buf_size, &pack_pos, &len, 1, MPI_INT, mpi_env.comm());
	    MPI_Unpack(cbuffer, meth_buf_size, &pack_pos, meth->out_flag, len, MPI_CHAR, mpi_env.comm());

	    MPI_Unpack(cbuffer, meth_buf_size, &pack_pos, &len, 1, MPI_INT, mpi_env.comm());
	    MPI_Unpack(cbuffer, meth_buf_size, &pack_pos, meth->aln_mode, len, MPI_CHAR, mpi_env.comm());

	    MPI_Unpack(cbuffer, meth_buf_size, &pack_pos, &len, 1, MPI_INT, mpi_env.comm());
	    MPI_Unpack(cbuffer, meth_buf_size, &pack_pos, meth->out_mode, len, MPI_CHAR, mpi_env.comm());

	    MPI_Unpack(cbuffer, meth_buf_size, &pack_pos, &len, 1, MPI_INT, mpi_env.comm());
	    MPI_Unpack(cbuffer, meth_buf_size, &pack_pos, meth->seq_type, len, MPI_CHAR, mpi_env.comm());

	    MPI_Unpack(cbuffer, meth_buf_size, &pack_pos, &len, 1, MPI_INT, mpi_env.comm());
	    MPI_Unpack(cbuffer, meth_buf_size, &pack_pos, meth->weight, len, MPI_CHAR, mpi_env.comm());

	    MPI_Unpack(cbuffer, meth_buf_size, &pack_pos, &len, 1, MPI_INT, mpi_env.comm());
	    MPI_Unpack(cbuffer, meth_buf_size, &pack_pos, meth->matrix, len, MPI_CHAR, mpi_env.comm());

	    MPI_Unpack(cbuffer, meth_buf_size, &pack_pos, &(meth->gop), 1, MPI_INT, mpi_env.comm());
	    MPI_Unpack(cbuffer, meth_buf_size, &pack_pos, &(meth->gep), 1, MPI_INT, mpi_env.comm());

	    MPI_Unpack(cbuffer, meth_buf_size, &pack_pos, &(meth->minid), 1, MPI_INT, mpi_env.comm());
	    MPI_Unpack(cbuffer, meth_buf_size, &pack_pos, &(meth->maxid), 1, MPI_INT, mpi_env.comm());

	    MPI_Unpack(cbuffer, meth_buf_size, &pack_pos, &len, 1, MPI_INT, mpi_env.comm());
	    MPI_Unpack(cbuffer, meth_buf_size, &pack_pos, meth->param, len, MPI_CHAR, mpi_env.comm());

	    MPI_Unpack(cbuffer, meth_buf_size, &pack_pos, &len, 1, MPI_INT, mpi_env.comm());
	    MPI_Unpack(cbuffer, meth_buf_size, &pack_pos, meth->param1, len, MPI_CHAR, mpi_env.comm());

	    methods[i] = meth;
	    pack_pos = 0;
	} // for (i)
    } // end of temporary block

    delete[] cbuffer;

    if (!methods.empty()) { // if methods
	libs.resize(methods.size(), 0);
	std::vector<TC_ConstraintList *>::iterator it = libs.begin();

	for (; it != libs.end(); ++it)
	    *it = TC_duplicate_constraint_list_soft(CL);

	ibuffer = new int[job_buf_size];

	do {
	    // get jobs
	    MPI_Recv(ibuffer, job_buf_size, MPI_INT, mpi_env.get_root(), LIB_JOB_TAG, mpi_env.comm(), &status);
#if DEBUG_MPI
	    std::cout << "\nNODE " << mpi_env.rank() << " RECIVED LIB JOBS FROM MASTER\n";
#endif

	    // and process them
	    unsigned int to_proc;

	    to_proc = ibuffer[0];
	    if (to_proc == 0) break;

	    AR_process_jobs(to_proc, ibuffer + 1, methods, libs);

	    // send request for next jobs
	    job_res[0] = JOB_REQ;
	    job_res[1] = mpi_env.rank();
	    job_res[2] = 0;

	    MPI_Send(job_res, 3, MPI_UNSIGNED, mpi_env.get_root(), LIB_JOB_TAG + 1, mpi_env.comm());
#if DEBUG_MPI
	    std::cout << "\nNODE " << mpi_env.rank() << " SENT REQUEST FOR LIB JOBS TO MASTER\n";
#endif
	}
	while (true);

	delete[] ibuffer;


	// COMPACT AND SORT
	for (it = libs.begin(); it != libs.end(); ++it) {
	    TC_compact_list(*it, 0, (*it)->ne, "best");
	    CL = TC_merge_constraint_list(*it, CL);
	    (*it)->L.free();
	    TC_free_constraint_list(*it);
	}
	
	par_sort_library(CL, mpi_env);

	
    } // if methods

    std::vector<TC_Method *>::iterator it = methods.begin();
    for (; it != methods.end(); ++it) {
//	TC_free_constraint_list((*it)->PW_CL);
	delete (*it)->PW_CL;
	AR_free(*it);
    }

    return CL;
} // JZ_worker_compile_cl


TC_ConstraintList*
JZ_par_compile_constraint_list(TC_ConstraintList* CL,
			       const std::vector<std::string>& in_files,
			       const char* weight, const MPI_Env& mpi_env) {
    
    if (mpi_env.am_I_root()) {
	CL = JZ_master_compile_cl(CL, in_files, weight, mpi_env);
    }
    else {
	CL = JZ_worker_compile_cl(CL, in_files, weight, mpi_env);
    }
    
    return CL;
} // JZ_par_compile_constraint_list


TC_ConstraintList* 
JZ_evaluate_constraint_list_reference(TC_ConstraintList* CL,
				      const MPI_Env& mpi_env) {
    int a, s1, s2, r1, r2, w;

    if (CL->M) {
	CL->max_value = CL->max_ext_value = 20;
	CL->nomatch = (CL->nomatch * CL->normalise) / CL->max_ext_value;
	return CL;
    }

    CL->max_value = CL->max_ext_value = 0;

    // find max_value
    for (a = 0; a < CL->ne; a++) {
	s1 = CL->L[a][SEQ1];
	s2 = CL->L[a][SEQ2];
	r1 = CL->L[a][R1];
	r2 = CL->L[a][R2];
	w = CL->L[a][WE];
	if (w == TC_UNDEFINED);
	else  CL->max_value = MAX(w, CL->max_value);
    }

    // find max_ext_value
    int nseq = CL->S->nseq;
    int* res_count = 0;
    int* red_count = 0;

    for (int i = 0; i < nseq; ++i) {
	int len =  CL->S->len[i];

	res_count = (int*)realloc(res_count, (len + 1) * sizeof(int));
	red_count = (int*)realloc(red_count, (len + 1) * sizeof(int));
	memset(res_count, 0, (len + 1) * sizeof(int));
	
	for (int j = 0; j < CL->ne; ++j) {
	    s1 = CL->L[j][SEQ1];
	    s2 = CL->L[j][SEQ2];
	    r1 = CL->L[j][R1];
	    r2 = CL->L[j][R2];
	    w = CL->L[j][WE];
	    if ((s1 == i) && (w != TC_UNDEFINED)) {
		res_count[r1] += w;
	    }
	    else if ((s2 == i) && (w != TC_UNDEFINED)) {
		res_count[r2] += w;
	    }
	    if (i < s1) break;
	}
	
	MPI_Allreduce(res_count, red_count, len + 1, MPI_INT, MPI_SUM,
		      mpi_env.comm());

	CL->max_ext_value = std::max(CL->max_ext_value,
				     *(std::max_element(red_count,
							red_count + len + 1)));
    }

    free(res_count);
    free(red_count);

    CL->max_ext_value = MAX(1, CL->max_ext_value);
    CL->nomatch = (CL->nomatch * CL->normalise) / CL->max_ext_value;

    // set max_value
    // THIS VALUE SEEMS TO BE NEVER USED
    // WE KEEP IT FOR COMPATIBILITY PURPOUSE
    int mval = CL->max_value;
    MPI_Allreduce(&mval, &CL->max_value, 1, MPI_INT, MPI_MAX,
		  mpi_env.comm());

    return CL;
} // JZ_evaluate_constraint_list_reference


TC_ConstraintList* JZ_re_weight_constraint_list(TC_ConstraintList* CL,
						const TC_Weights *W, 
						const MPI_Env& mpi_env) {
    float* weight;
    float w;
    int sA, sB;

    if (!CL->L) return CL;

    weight = W->SEQ_W;

    for (int i = 0; i < CL->ne; ++i) {
	sA = CL->L[i][SEQ1];
	sB = CL->L[i][SEQ2];
	w = MIN(weight[sA], weight[sB]);
	CL->L[i][WE] = (int)(w * CL->L[i][WE]);
    }

    CL = JZ_evaluate_constraint_list_reference(CL, mpi_env);    

    return CL;
} // JZ_re_weight_constraint_list


TC_ConstraintList* JZ_weight_constraint_list(TC_ConstraintList* CL,
					     const char *seq_weight,
					     const MPI_Env& mpi_env) {
    if (CL->ne == 0) return CL;

    TC_Weights* W = TC_compute_tcoffee_weight(CL);

    CL = JZ_re_weight_constraint_list(CL, W, mpi_env);
    CL->W = W;

//     std::cout << "RANK: " << mpi_env.rank() << " max_val: " << CL->max_value
// 	      << " max_ext_val: " << CL->max_ext_value << std::endl;

    return CL;
} // JZ_weight_constraint_list


void broadcast_DM(TC_ConstraintList* CL, const MPI_Env& cl_env) {
    int nseq = CL->S->nseq;
    int* isnd_buf = new int[nseq * nseq];

    if (cl_env.am_I_root() == true) {
	for (int i = 0; i < nseq; ++i) {
	    memcpy(isnd_buf + i * nseq, CL->DM->similarity_matrix[i],
		   nseq * sizeof(int));
	}

	MPI_Bcast(isnd_buf, nseq * nseq, MPI_INT,
		  cl_env.get_root(), cl_env.comm());
    }
    else {
	CL->DM = AR_declare_distance_matrix(nseq);

	MPI_Bcast(isnd_buf, nseq * nseq, MPI_INT, 
		  cl_env.get_root(), cl_env.comm());

	for (int i = 0; i < nseq; ++i) {
 	    memcpy(CL->DM->similarity_matrix[i], isnd_buf + i * nseq, 
		   nseq * sizeof(int));
	}
    }
    
    delete[] isnd_buf;
} // broadcast_DM
