/***
 *  $Id: ResidueIndex.cpp 708 2007-01-08 22:47:08Z zola $
 **
 *  File: ResidueIndex.cpp
 *  Created: Dec 17, 2006
 *
 *  Department of Electrical & Computer Engineering
 *  Iowa State University of Science & Technology
 *  Ames, IA 50011, USA
 *
 *  Author: Jaroslaw Zola <zola@iastate.edu>
 *  Please see attached LICENSE file.
 */

#include "ResidueIndex.hpp"

#include "globals.hpp"
#include "util_mem.hpp"
#include "TC_Sequences.hpp"
#include "TC_ConstraintList.hpp"


#if USE_RMA

GlobalStorage G_STORAGE;


int** TC_clean_residue_pair_hasch(int s1, int r1, int s2, int r2, 
				  ResIdxSeq& seq1, int** hasch,
				  TC_ConstraintList* CL) {
    int a, t_s, t_r;
   
    if (!hasch)	return hasch;

    int* res1 = seq1[r1];

    for (a = 1; a < res1[0]; a += 3) {
	t_s = res1[a];
	t_r = res1[a + 1];
	hasch[t_s][t_r] = 0;
    }
    
    hasch[s1][r1] = hasch[s2][r2] = 0;

    return hasch;
} // TC_clean_residue_pair_hasch


TC_ConstraintList* AR_index_res_constraint_list(TC_ConstraintList* CL, 
						int field) {

    if (CL->residue_indexed && CL->residue_field == field) return CL;

    CL->residue_index.alloc(0);

    int nseq = CL->S->nseq;

    int rank = CL->residue_index.mpi_env->rank();
    int wsize = CL->residue_index.mpi_env->size();
    int root = CL->residue_index.mpi_env->get_root();

    // MAP SEQUENCES TO HOST    
    G_STORAGE.seq2host.resize(nseq, -1);

    int first_S, last_S;
    first_S = last_S = -1;

    int chunk = nseq / wsize;

    if (chunk < 1) {
	for (int i = 0; i < nseq; ++i) {
	    first_S = i;
	    last_S = i + 1;
	    G_STORAGE.seq2host[i] = i;
	}
    }
    else {
	int pos = 0;
	for (int i = 0; i < wsize; ++i) {
	    if (rank == i) first_S = pos;

	    if (i == root) {
		for (int j = 0; j < chunk + (nseq % wsize); ++j) {
		    G_STORAGE.seq2host[pos] = i;
		    ++pos;
		}
	    }
	    else {
		for (int j = 0; j < chunk; ++j) {
		    G_STORAGE.seq2host[pos] = i;
		    ++pos;
		}
	    }
	    
	    if (rank == i) last_S = pos;
	}
    } // if chunk < 1

    G_STORAGE.init(first_S, last_S, nseq);


    std::vector<unsigned int> TMP_OFF_(nseq, 0);
    std::vector<unsigned int> TMP_LEN_(nseq, 0);

    
    // CREATE INDEX
    int s1, r1, s2, r2, w;

    ResRow row;

    for (int i = 0; i < nseq; ++i) {
	int len = CL->S->len[i];
	
	row.allocate(len + 1);
	for (int j = 0; j < len + 1; ++j) {
	    row[j].allocate(1);
	    row[j][0] = 1;
	}

	for (int j = 0; j < CL->ne; ++j) {
	    s1 = CL->L[j][SEQ1];
	    s2 = CL->L[j][SEQ2];
	    r1 = CL->L[j][R1];
	    r2 = CL->L[j][R2];
	    w = CL->L[j][WE];

	    int ir1 = -1;
	    int is2 = -1;
	    int ir2 = -1;
	    
	    if (s1 == i) {
		ir1 = r1;
		is2 = s2;
		ir2 = r2;
	    }
	    else if (s2 == i) {
		ir1 = r2;
		is2 = s1;
		ir2 = r1;
	    }
	    
	    if (ir1 != -1) {
 		row[ir1][0] += 3;
 		int nsize = row[ir1][0];
 		row[ir1].reallocate(nsize);
 		row[ir1][nsize - 3] = is2;
 		row[ir1][nsize - 2] = ir2;
 		row[ir1][nsize - 1] = w;
	    }
	    
	    if (i < s1) break;
	} // j < ne

	// ASSEMBLY OF THE INDEX
	// create send buffer
	int tot_len = 0;
	for (int j = 0; j < len + 1; ++j) tot_len += row[j][0];

	int pos = 0;
	int* snd_buf = new int[tot_len];

	for (int j = 0; j < len + 1; ++j) {
	    memcpy(snd_buf + pos, row[j].values_,
		   row[j][0] * sizeof(int));
	    pos += row[j][0];
	    row[j].free();
	}
	
	// gather partial indexes
	std::vector<int> lenghts(wsize, 0);
	std::vector<int> displ(wsize + 1, 0);

	MPI_Gather(&tot_len, 1, MPI_INT, &lenghts[0], 1, MPI_INT,
		   G_STORAGE.seq2host[i], CL->residue_index.mpi_env->comm());

	int recv_size = 0;
	int* recv_buf = 0;

	for (int j = 1; j < wsize + 1; ++j) {
	    recv_size += lenghts[j - 1];
	    displ[j] = displ[j - 1] + lenghts[j - 1];
	}

	if (rank == G_STORAGE.seq2host[i]) recv_buf = new int[recv_size];

	MPI_Gatherv(snd_buf, tot_len, MPI_INT,
		    recv_buf, &lenghts[0], &displ[0], MPI_INT,
		    G_STORAGE.seq2host[i], CL->residue_index.mpi_env->comm());
	
	// place recived data into global buffer
	if (rank == G_STORAGE.seq2host[i]) { // main process
	    std::vector<int> tmp_displ(displ);

	    // set global info
	    unsigned int idx = G_STORAGE.BSIZE;

	    // allocate sequence header (residues description)
	    G_STORAGE.reallocate(len + 1);
	    TMP_OFF_[i] = idx;
	    
	    int disp_val = 0;
	    int res_disp = 0;

	    int* tmp_off = new int[len + 2];
	    memset(tmp_off, 0, (len + 2) * sizeof(int));

	    // for every residue
	    tot_len = 0;
	    for (int j = 0; j < len + 1; ++j) {
		res_disp = 1; // at least one value stored

		// for each host
		for (int k = 0; k < wsize; ++k) {
		    // check total lenght of data for residue
		    disp_val = recv_buf[tmp_displ[k]];
		    res_disp += (disp_val - 1);
		    // shift to the position of next res
		    tmp_displ[k] += disp_val;
		} // k < wsize
		tot_len += res_disp;
		G_STORAGE.BUFF_[idx + j] = res_disp;
	    } // j < len + 1

	    // compute offset
	    for (int j = 1; j < len + 2; ++j) {
		tmp_off[j] = G_STORAGE.BUFF_[idx + j - 1] + tmp_off[j - 1];
	    }

	    // shift to data region
	    for (int j = 0; j < len + 1; ++j) {
		G_STORAGE.BUFF_[idx + j] = tmp_off[j] + len + 1;
	    }

	    TMP_LEN_[i] = len + 1 + tot_len;
	    G_STORAGE.reallocate(tot_len);

	    // place residue data
	    int buf_pos = idx + len + 1;
	    
	    // first index is always empty
	    G_STORAGE.BUFF_[buf_pos] = 1;
	    for (int k = 0; k < wsize; ++k) ++displ[k];
	    ++buf_pos;

	    for (int j = 1; j < len + 1; ++j) {
		if ((tmp_off[j + 1] - tmp_off[j]) == 1) {
		    G_STORAGE.BUFF_[buf_pos] = 1;
		    for (int k = 0; k < wsize; ++k) ++displ[k];
		    ++buf_pos;
		}
		else {
		    G_STORAGE.BUFF_[buf_pos] = tmp_off[j + 1] - tmp_off[j];
		    ++buf_pos;

		    for (int k = 0; k < wsize; ++k) {
			disp_val = recv_buf[displ[k]];

			if (1 < disp_val) {
			    memcpy(G_STORAGE.BUFF_ + buf_pos, 
				   recv_buf + displ[k] + 1,
				   (disp_val - 1) * sizeof(int));
			    buf_pos += (disp_val - 1);
			}
			
			displ[k] += disp_val;
		    }
		}
	    } // j < len + 1

	    delete[] tmp_off;
	} // if rank = 

	delete[] snd_buf;
	delete[] recv_buf;

	row.free();
    } // i < nseq
    
    // reduce offsets and and lenghts
    MPI_Allreduce(&TMP_OFF_[0], G_STORAGE.OFF_, nseq, MPI_UNSIGNED, MPI_SUM,
		  CL->residue_index.mpi_env->comm());

    MPI_Allreduce(&TMP_LEN_[0], G_STORAGE.LEN_, nseq, MPI_UNSIGNED, MPI_SUM,
		  CL->residue_index.mpi_env->comm());

    G_STORAGE.create_mpi_win(CL->residue_index.mpi_env->comm());

    CL->residue_indexed = 1;
    CL->residue_field = field;

    return CL;
} // AR_index_res_constraint_list


int TC_residue_pair_extended_list(TC_ConstraintList* CL, 
				  int s1, int r1,
				  int s2, int r2) {
    static int** hasch;
    static int max_len;

    static ResIdxSeq seq1;
    static ResIdxSeq seq2;

    if (r1 <= 0 || r2 <= 0)  return 0;

    int i_score = 0;
    double score = 0;
    double max_score = 0;
    double max_val = 0;
    int a, t_s, t_r;

    int field = CL->weight_field;

    // this can be useful with RMA...
    // if (CACHE.find(s1, r1, s2, r2, i_score)) return score;


    if (!hasch || max_len != (CL->S)->max_len) {
	max_len = (CL->S)->max_len;
	if (hasch) TC_free_int(hasch);
	hasch = TC_declare_int((CL->S)->nseq, (CL->S)->max_len + 1);
    }


    CL = AR_index_res_constraint_list(CL, field);


    CL->residue_index.get(s1, seq1);
    CL->residue_index.get(s2, seq2);

    int* res1 = seq1[r1];
    int* res2 = seq2[r2];
    
    /* Check matches for R1 in the indexed lib */
    hasch[s1][r1] = FORBIDEN;

    for (a = 1; a < res1[0]; a += 3) {
	t_s = res1[a];
	t_r = res1[a + 1];
	int w = res1[a + 2];
	hasch[t_s][t_r] = w;
	max_score += w;
    }

    /* Check Matches for r1 <-> r2 in the indexed lib */
    for (a = 1; a < res2[0]; a += 3) {
	t_s = res2[a];
	t_r = res2[a + 1];
	int w = seq2[r2][a + 2];

	if (hasch[t_s][t_r]) {
	    if (hasch[t_s][t_r] == FORBIDEN) {
		score += w;
		max_score += w;
	    }
	    else {
		double delta = MIN(hasch[t_s][t_r], w);
		score += delta;
		max_score -= hasch[t_s][t_r];
		max_score += delta;
		max_val = MAX(max_val, delta);
	    }
	}
	else max_score += w;
    }
    
    max_score -= hasch[s2][r2];
    TC_clean_residue_pair_hasch(s1, r1, s2, r2, seq1, hasch, CL);

    if (max_score == 0) score = 0;

    else if (CL->normalise) {
	score = ((score * CL->normalise) / max_score) * SCORE_K;
	if (max_val >= 1000 && CL->overweight == ON) {
	    score *= max_val / (double)CL->normalise;
	}
    }

    i_score = (int)score;

    // this can be useful with RMA...
    // CACHE.insert(s1, r1, s2, r2, i_score);

    if (seq1.state) {
	if (*seq1.state == 2) {
	    delete seq1.state;
	    seq1.state = 0;
	    delete[] seq1.storage;
	    seq1.storage = 0;
	}
	else (*seq1.state) = 0;
    }

    if (seq2.state) {
	if (*seq2.state == 2) {
	    delete seq2.state;
	    seq2.state = 0;	    
	    delete[] seq2.storage;
	    seq2.storage = 0;
	}
	else (*seq2.state) = 0;
    }

    return i_score;
} // TC_residue_pair_extended_list



int TC_extend_residue_pair(TC_ConstraintList* CL, 
			   int s1, int r1, int s2, int r2) {

    std::cout << "WHOOOPS: this part has been not implemented yet!!!\n"
	      << __FILE__ << ":" << __LINE__ << std::endl;

    abort();
} // TC_extend_residue_pair

int TC_fast_get_dp_cost(TC_Alignment* A, 
			int** pos1, int ns1, int* list1, int col1, 
			int** pos2, int ns2, int* list2, int col2, 
			TC_ConstraintList* CL) {

    std::cout << "WHOOOPS: this part has been not implemented yet!!!\n"
	      << __FILE__ << ":" << __LINE__ << std::endl;

    abort();
    
    return -1;
} // TC_fast_get_dp_cost

#endif // USE_RMA
