/*
 * File: TC_Alignment.cpp
 * Created: 06.06.2006
 * Last Modified: 11.09.2006
 * Author: Adrian Rospondek
 */


#include <iostream>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "TC_Alignment.hpp"
#include "TC_ConstraintList.hpp"
#include "TC_Profile.hpp"
#include "TC_Sequences.hpp"
#include "dp_drivers.hpp"
#include "globals.hpp"
#include "input.hpp"
#include "output.hpp"
#include "util.hpp"
#include "util_algo.hpp"
#include "util_io.hpp"
#include "util_mem.hpp"
#include "util_seq.hpp"
#include "util_str.hpp"



TC_Alignment *TC_declare_aln(TC_Sequences *S)
{
	TC_Alignment *LA;
	int a;

	/* order:
	 * [x][0]= which is the xth seq of aln
	 * [x][1]= how many deleted residues before the first one */

	LA = (TC_Alignment *)AR_allocate(1, sizeof(TC_Alignment));
	if (S == 0) {
		LA->declared_len = MAX_LEN_ALN;
		LA->max_n_seq = MAX_N_SEQ;
	}
	else {
		LA->declared_len = 2 * S->max_len + 1;
		LA->max_n_seq = S->nseq + 1;
	}
	LA->S = S;
	LA->seq_comment = TC_declare_char(LA->max_n_seq, STRING);
	LA->aln_comment = TC_declare_char(LA->max_n_seq, STRING);
	LA->seq_al = TC_declare_char(LA->max_n_seq, LA->declared_len);
	LA->name = TC_declare_char(LA->max_n_seq, STRING);
	LA->file = TC_declare_char(LA->max_n_seq, STRING);
	LA->tree_order = TC_declare_char(LA->max_n_seq, STRING);
	LA->order = TC_declare_int(LA->max_n_seq, 2);
	LA->score_seq = (int *)AR_allocate(LA->max_n_seq, sizeof(int));

	for (a = 0; a < LA->max_n_seq; a++)
		LA->order[a][0] = a;
	LA->len_aln = 0;
	LA->score_aln = 0;
	LA->len = (int *)AR_allocate(LA->max_n_seq, sizeof(int));
	if (S && S->name)
		for (a = 0; a < S->nseq; a++)
			strncpy(LA->name[a], S->name[a], STRING - 1);

	return LA;
}

TC_Alignment *TC_declare_aln2(int nseq, int len)
{
	TC_Sequences *S;
	TC_Alignment *A;

	S = (TC_Sequences *)AR_allocate(1, sizeof(TC_Sequences));
	S->nseq = nseq;
	S->max_len = len;
	A = TC_declare_aln(S);
	A->S = 0;
	AR_free(S);

	return A;
}

TC_Alignment *TC_realloc_aln(TC_Alignment *A, int new_len)
{
	if (A == 0)
		A = TC_declare_aln(0);

	return TC_realloc_aln2(A, A->max_n_seq, new_len);
}

TC_Alignment *TC_realloc_aln2(TC_Alignment *A, int n_nseq, int n_len)
{
	int a;
	int len, nseq;
	int delta_len, delta_nseq;

	if (A == 0)
		A = TC_declare_aln(0);

	n_len++;
	n_nseq++;
	len = A->declared_len;
	nseq = A->max_n_seq;
	n_len = MAX(len, n_len);
	n_nseq = MAX(nseq, n_nseq);
	delta_nseq = MAX(0, n_nseq - nseq);
	delta_len = MAX(0, n_len - len);

	if (delta_nseq <= 0 && delta_len <= 0)
		return A;
	else {
		A->len = (int *)AR_reallocate(A->len, n_nseq, sizeof(int));
		for (a = nseq; a < n_nseq; a++)
			A->len[a] = 0;

		A->declared_len = n_len;
		A->max_n_seq = n_nseq;
		A->seq_comment = TC_new_realloc_char(A->seq_comment, n_nseq, -1);
		A->aln_comment = TC_new_realloc_char(A->aln_comment, n_nseq, -1);
		A->name = TC_new_realloc_char(A->name, n_nseq, -1);
		A->file = TC_new_realloc_char(A->file, n_nseq, -1);
		A->tree_order = TC_new_realloc_char(A->tree_order, n_nseq, -1);
		A->seq_al = TC_new_realloc_char(A->seq_al, n_nseq, n_len);
		A->order = TC_new_realloc_int(A->order, n_nseq, -1);
		if (A->seq_cache)
			A->seq_cache = TC_new_realloc_int(A->seq_cache, n_nseq, n_len);
		if (A->cdna_cache)
			A->cdna_cache = TC_new_realloc_int(A->cdna_cache, n_nseq, n_len);

		A->score_seq = (int *)AR_reallocate(A->score_seq, n_nseq, sizeof(int));
		for (a = nseq; a < n_nseq; a++)
			A->score_seq[a] = 0;
	}

	return A;
}

TC_Alignment *TC_copy_aln(const TC_Alignment *A, TC_Alignment *B)
{
	int a, b;

	if (A == 0) {
		TC_free_aln(B);
		return 0;
	}
	TC_free_aln(B);

	if (A->S && A->nseq > (A->S)->nseq) {
		B = TC_declare_aln2(A->nseq, MAX((A->S)->max_len + 1, A->len_aln + 1));
		B->S = A->S;
	}
	else
		B = TC_declare_aln((A->S));

	/* SIZES */
	B->max_len = A->max_len;
	B->min_len = A->min_len;
	B->declared_len = A->declared_len;
	B->max_n_seq = A->max_n_seq;
	B->nseq = A->nseq;
	B->len_aln = A->len_aln;

	/* Sequence Information */
	if ((A->S) == 0) {
		AR_free(B->len);
		B->len = (int *)AR_allocate(A->max_n_seq, sizeof(int));
	}
	TC_ga_memcpy_int(A->len, B->len, B->nseq);
	B->seq_comment = TC_copy_char(A->seq_comment, B->seq_comment, -1, -1);
	B->aln_comment = TC_copy_char(A->aln_comment, B->aln_comment, -1, -1);
	B->name = TC_copy_char(A->name, B->name, -1, -1);
	B->file = TC_copy_char(A->file, B->file, -1, -1);
	B->tree_order = TC_copy_char(A->tree_order, B->tree_order, -1, -1);
	B->expanded_order = A->expanded_order;

	TC_free_char(B->seq_al);
	B->seq_al = TC_declare_char(B->max_n_seq, B->declared_len);
	for (a = 0; a < A->max_n_seq; a++) {
		for (b = 0; b < A->declared_len; b++)
			B->seq_al[a][b] = A->seq_al[a][b];
	}

	B->order = TC_copy_int(A->order, B->order, -1, -1);
	B->S = A->S;
	if (A->seq_cache)
		B->seq_cache = TC_copy_int(A->seq_cache, B->seq_cache, -1, -1);
	if (A->cdna_cache)
		B->cdna_cache = TC_copy_int(A->cdna_cache, B->cdna_cache, -1, -1);
	B->P = TC_copy_profile(A->P);

	/* Score */
	if ((A->S) == 0) {
		AR_free(B->score_seq);
		B->score_seq = (int *)AR_allocate(A->max_n_seq, sizeof(int));
	}
	TC_ga_memcpy_int(A->score_seq, B->score_seq, B->nseq);
	B->score_aln = A->score_aln;
	B->score = A->score;

	B->cpu = A->cpu;
	B->finished = A->finished;

	/* Output Options */
	B->output_res_num = A->output_res_num;
	B->residue_case = A->residue_case;
	B->expand = A->expand;

	B->CL = A->CL;
	B->random_tag = A->random_tag;

	/* Make the function Recursive */
	if (A->A)
		B->A = TC_copy_aln(A->A, 0);
	else
		B->A = 0;

	return B;
}

/* Does not free the A->S field (sequences of A) */
TC_Sequences *TC_free_aln(TC_Alignment *LA)
{
	TC_Sequences *S;

	if (LA == 0)
		return 0;

	S = LA->S;
	TC_free_char(LA->file);
	TC_free_char(LA->seq_al);
	TC_free_int(LA->seq_cache);
	TC_free_int(LA->cdna_cache);
	TC_free_char(LA->name);
	TC_free_char(LA->tree_order);
	TC_free_char(LA->seq_comment);
	TC_free_char(LA->aln_comment);
	TC_free_int(LA->order);
	AR_free(LA->score_seq);
	AR_free(LA->len);
	TC_free_profile(LA->P);
	if (LA->A) {
		TC_free_aln(LA->A);
		LA->A = 0;
	}
	AR_free(LA);

	return S;
}

TC_Alignment *TC_seq2aln(TC_Sequences *S, TC_Alignment *A, int rm_gap)
{
	int a;

	A = TC_realloc_aln2(A, S->nseq, S->max_len + 1);
	for (a = 0; a < S->nseq; a++)
		strncpy(A->file[a], S->file[a], STRING - 1);
	A->nseq = S->nseq;
	A->max_len = S->max_len;
	A->min_len = S->min_len;

	for (a = 0; a < S->nseq; a++) {
		A->order[a][0] = a;
		A->order[a][1] = 0;
		strncpy(A->seq_comment[a], S->seq_comment[a], STRING - 1);
		strncpy(A->aln_comment[a], S->aln_comment[a], STRING - 1);

		strncpy(A->name[a], S->name[a], STRING - 1);
		strncpy(A->seq_al[a], S->seq[a], A->declared_len - 1);

		TC_ungap(A->seq_al[a]);
		A->len[a] = strlen(A->seq_al[a]);

		if (rm_gap == 0)
			strncpy(A->seq_al[a], S->seq[a], A->declared_len - 1);

	}
	TC_padd_aln(A);
	A->S = S;

	return A;
}

TC_Alignment *TC_stack_aln(TC_Alignment *A, TC_Alignment *B)
{
	int max_len = 0, max_nseq = 0;
	int a, b;
	
	if (B == 0)
		return A;
	if (A == 0)
		return B;
	
	max_nseq = A->nseq + B->nseq;
	for (a = 0; a < A->nseq; a++)
		max_len = MAX(strlen(A->seq_al[a]), max_len);
	for (a = 0; a < B->nseq; a++)
		max_len = MAX(strlen(B->seq_al[a]), max_len);
	
	A = TC_realloc_aln2(A, max_nseq, max_len + 1);
	
	for (a = A->nseq, b = 0; b < B->nseq; b++, a++) {
		strncpy(A->seq_comment[a], B->seq_comment[b], STRING - 1);
		strncpy(A->aln_comment[a], B->aln_comment[b], STRING - 1);
		strncpy(A->seq_al[a], B->seq_al[b], A->declared_len - 1);
		strncpy(A->name[a], B->name[b], STRING - 1);
		strncpy(A->file[a], B->file[b], STRING - 1);
		A->order[a][0] = B->order[b][0];
		A->order[a][1] = B->order[b][1];
		A->score_seq[a] = B->score_seq[b];
		A->len[a] = B->len[b];
	}
	A->len_aln = MAX(A->len_aln, B->len_aln);
	A->nseq = A->nseq + B->nseq;
	A->score_aln = A->score_aln + B->score_aln;
	A->finished = A->finished + B->finished;
	
	return A;
}

TC_Alignment *TC_get_aln_type(TC_Alignment *A)
{
	if (A->S != 0 && (A->S)->type[0] == '\0')
		A->S = TC_get_sequence_type(A->S);
	else if (A->S == 0) {
		A->S = TC_aln2seq(A);
		A->S = TC_get_sequence_type(A->S);
	}

	return A;
}

static TC_Alignment *TC_aln2random_order(TC_Alignment *A)
{
	int **list;
	char **name_list;
	int a, max;

	max = 100 * A->nseq;
	TC_vsrand(0);

	list = TC_declare_int(A->nseq, 2);
	name_list = TC_declare_char(A->nseq, 100);

	for (a = 0; a < A->nseq; a++) {
		list[a][0] = a;
		list[a][1] = rand() % max;
	}
	TC_sort_int(list, 1, 0, A->nseq - 1);

	for (a = 0; a < A->nseq; a++)
		strncpy(name_list[a], A->name[list[a][0]], 100 - 1);
	A = TC_reorder_aln(A, name_list, A->nseq);

	TC_free_char(name_list);
	TC_free_int(list);

	return A;
}

TC_Alignment *TC_reorder_aln(TC_Alignment *A, char **name, int nseq)
{
	int a, sn;
	int n = 0;
	int *tpp_int;
	TC_Alignment *BUF;

	if (name == 0)
		return TC_aln2random_order(A);

	BUF = TC_copy_aln(A, 0);
	for (a = 0; a < nseq; a++) {
		sn = TC_name_is_in_list(name[a], BUF->name, A->nseq, STRING);
		if (sn == -1);
		else {
			SWAPP(A->order[n], BUF->order[sn], tpp_int);
			strncpy(A->name[n], BUF->name[sn], STRING - 1);
			strncpy(A->seq_al[n], BUF->seq_al[sn], A->declared_len - 1);
			strncpy(A->seq_comment[n], BUF->seq_comment[sn], STRING - 1);
			n++;
		}
	}

	for (a = n; a < A->nseq; a++)
		A->name[a][0] = A->seq_al[a][0] = '\0';
	A->nseq = n;

	if (A->A)
		A->A = TC_reorder_aln(A->A, name, nseq);
	TC_free_aln(BUF);

	return A;
}

TC_Alignment *TC_padd_aln(TC_Alignment *A)
{
	A->seq_al = TC_padd_strings(A->seq_al, A->nseq, '-');
	A->len_aln = strlen(A->seq_al[0]);

	return A;
}

TC_Alignment *TC_ungap_aln(TC_Alignment *A)
{
	return TC_ungap_aln_n(A, 100);
}

/**
 * @brief Remove all the columns of gap-only within an alignment
 */
TC_Alignment *TC_ungap_aln_n(TC_Alignment *A, int p)
{
	int a, b, c;
	int t;

	if (A->nseq == 0)
		return A;

	for (c = 0, a = 0; a < A->len_aln; a++) {
		for (t = 0, b = 0; b < A->nseq; b++)
			t += TC_is_gap(A->seq_al[b][a]);
		if ((t * 100) / A->nseq >= p || (t == A->nseq && p == 100) || (t && p == 1));
		else {
			for (b = 0; b < A->nseq; b++)
				A->seq_al[b][c] = A->seq_al[b][a];
			c++;
		}
	}
	for (b = 0; b < A->nseq; b++)
		A->seq_al[b][c] = '\0';
	A->len_aln = c;

	return A;
}

TC_Alignment *TC_ungap_sub_aln(TC_Alignment *A, int ns, int *ls)
{
	int len;
	int a, b, c, t;

	len = strlen(A->seq_al[ls[0]]);

	for (c = 0, a = 0; a < len; a++) {
		for (t = 0, b = 0; b < ns; b++)
			t += TC_is_gap(A->seq_al[ls[b]][a]);
		if (t == ns);
		else {
			for (b = 0; b < ns; b++)
				A->seq_al[ls[b]][c] = A->seq_al[ls[b]][a];
			c++;
		}
	}
	for (b = 0; b < ns; b++)
		A->seq_al[ls[b]][c] = '\0';

	return A;
}

/**
 * @brief Remove all the columns of gap-only within an alignment.
 */
void TC_compress_aln(TC_Alignment *A)
{
	int a, b, c, d;

	for (c = 0, a = 0; a < A->len_aln; a++) {
		for (b = 0, d = 0; b < A->nseq; b++)
			if (A->seq_al[b][a] != '-') {
				d = 1;
				break;
			}
		if (d == 0);
		else {
			for (b = 0; b < A->nseq; b++)
				A->seq_al[b][c] = A->seq_al[b][a];
			++c;
		}
	}
	A->len_aln = c;

	for (a = 0; a < A->nseq; a++)
		A->seq_al[a][c] = '\0';
}

TC_Alignment *TC_trim_aln_with_seq(TC_Alignment *S, TC_Alignment *P)
{
	static int seqindex;
	TC_Alignment *A, *R;
	int a, b, c;
	
	P = TC_aln2profile(P);
	S = TC_aln2profile(S);
	
	A = TC_align_two_aln(S, P, "blosum62mt", -8, -1, "myers_miller_pair_wise");
	for (a = 0; a < A->nseq; a++)
		sprintf(A->name[a], "tmpname_%d", seqindex++);
	
	R = TC_copy_aln(A, 0);
	for (c = 0, a = 0; a < A->len_aln; a++) {
		if (TC_is_gap(A->seq_al[0][a]));
		else {
			for (b = 0; b < A->nseq; b++)
				R->seq_al[b][c] = A->seq_al[b][a];
			c++;
		}
	}
	for (a = 0; a < A->nseq; a++)
		R->seq_al[a][c] = '\0';
	R->len_aln = c;
	R->S = TC_aln2seq(R);
	
	TC_free_aln(S);
	TC_free_aln(P);
	TC_free_aln(A);
	
	return R;
}

TC_Alignment *TC_extract_sub_aln(const TC_Alignment *B, int nseq, int *list)
{
	TC_Alignment *A = 0;
	int a, b, n, seq;

	A = TC_declare_aln2(nseq, B->len_aln + 1);
	for (n = 0, a = 0; a < nseq; a++) {
		seq = list[a];
		if (seq == -1)
			continue;
		else
			n++;
		strncpy(A->seq_comment[a], B->seq_comment[seq], STRING - 1);
		strncpy(A->aln_comment[a], B->aln_comment[seq], STRING - 1);
		strncpy(A->name[a], B->name[seq], STRING - 1);
		
		for (b = 0; b <= B->len_aln; b++)
			A->seq_al[a][b] = B->seq_al[seq][b];
		A->order[a][0] = B->order[seq][0];
		A->order[a][1] = B->order[seq][1];
		A->score_seq[a] = B->score_seq[seq];
		A->len[a] = B->len[seq];
	}
	A->nseq = n;
	A->len_aln = B->len_aln;
	
	return A;
}

/**
 * @brief Establishes the correspondance between every (1..N+1) residue of each aligned sequence
 * and its correspondance in S.
 * A->seq_cache[a][b] = x means that residue b of aligned sequence a corresponds to residue x
 * of the sequence with the same index in S,
 * A->seq_cache[a][b] = 0 means there is no correspondance.
 * a is the index of the sequence.
 * Applying this function is needed for turning an alignment into a constraint list.
 */
TC_Alignment *TC_fix_aln_seq(TC_Alignment *A, const TC_Sequences *S)
{
	TC_Alignment *B;
	char *buf1, *buf2;
	int g0, g1, nr0, nr1;
	int id, tot;
	int a, b, c;

	if (S == 0)
		return A;

	A->seq_cache = TC_declare_int(S->nseq, MAX(A->len_aln + 1, S->max_len + 1));

	for (a = 0; a < S->nseq; a++)
		for (b = 0; b < A->len_aln; b++)
			A->seq_cache[a][b] = -1;

	buf1 = buf2 = 0;
	for (a = 0; a < S->nseq; a++) {
		for (b = 0; b < A->nseq; b++) {
			if (TC_strm(S->name[a], A->name[b])) {
				A->order[b][0] = a;
				
				AR_free(buf1);
				buf1 = (char *)AR_allocate(A->len_aln + 1, sizeof(char));
				strncpy(buf1, A->seq_al[b], A->len_aln);
				TC_ungap(buf1);
				TC_upper_string(buf1);

				AR_free(buf2);
				int l = strlen(S->seq[a]);
				buf2 = (char *)AR_allocate(l + 1, sizeof(char));
				strncpy(buf2, S->seq[a], l);
				TC_ungap(buf2);
				TC_upper_string(buf2);

				if (TC_strm(buf1, buf2)) {
					for (c = 0; c < S->len[a]; c++)
						A->seq_cache[a][c + 1] = c + 1;
				}
				else {
					B = TC_align_two_sequences(buf2, buf1, "blosum62mt", -4, -1,
						"myers_miller_pair_wise");
					for (id = 0, tot = 0, nr0 = 0, nr1 = 0, c = 0; c < B->len_aln; c++) {
						g0 = TC_is_gap(B->seq_al[0][c]);
						g1 = TC_is_gap(B->seq_al[1][c]);
						nr0 += 1 - g0;
						nr1 += 1 - g1;
						if (!g0 && !g1) {
							++tot;
							id += (B->seq_al[0][c] == B->seq_al[1][c]) ? 1 : 0;
							A->seq_cache[a][nr1] = nr0;
						}
						else if (g0 && !g1) {
							A->seq_cache[a][nr1] = 0;
						}
					}
					if (((id * 100) / tot) < 20) {
						const char *name = S->name[a];

						std::cerr << '\n' << PROGRAM << ": ERROR: Two different sequences have the same name: '" << name << "'.\n";
						std::cerr << "\tIf '" << name << "' is a PDB ID, make sure it identifies the right chain (A, B, 1, 2, ...).\n"
								  << "\tChain number or index must be added to the PDB id (i.e. 1gowA).\n"
								  << "\tIf You want to use '" << name << "' anyway, rename it with a non-PDB identifier "
								  << "such as seq_" << name << ".\n";
						exit(EXIT_FAILURE);
					}
					TC_free_sequence(B->S);
					TC_free_aln(B);
				}
			}
		}
	}
	AR_free(buf1);
	AR_free(buf2);

	return A;
}

/**
 * @brief Expands the profiles within an alignment.
 */
TC_Alignment *TC_expand_aln(TC_Alignment *A)
{
	TC_Alignment *MAIN = 0, *SUB = 0;
	TC_Alignment *profile;
	int *list;
	int n_sub_seq = 0;
	int new_nseq = 0;
	int a, b, d, e;
	
	if (!A)
		return A;
	
	list = (int *)AR_allocate(A->nseq, sizeof(int));
	for (a = 0; a < A->nseq; a++) {
		profile = TC_seq2R_template_profile(A->S, A->order[a][0]);
		if (profile && profile->expand) {
			new_nseq += profile->nseq;
		}
		else {
			new_nseq++;
			list[n_sub_seq++] = a;
		}
	}
	
	if (n_sub_seq == A->nseq) {
		AR_free(list);
		return A;
	}
	else if (n_sub_seq == 0) {
		MAIN = TC_copy_aln(A, MAIN);
		MAIN->nseq = 0;
	}
	else {
		MAIN = TC_extract_sub_aln(A, n_sub_seq, list);
	}
	AR_free(list);
	
	for (a = 0; a < A->nseq; a++) {
		profile = TC_seq2R_template_profile(A->S, A->order[a][0]);
		if (profile && profile->expand) {
			SUB = TC_copy_aln(profile, SUB);
			SUB = TC_realloc_aln2(SUB, SUB->nseq, A->len_aln + 1);
			for (e = 0, b = 0; b <= A->len_aln; b++) {
				if (TC_is_gap(A->seq_al[a][b])) {
					for (d = 0; d < SUB->nseq; d++)
						SUB->seq_al[d][b] = '-';
				}
				else {
					for (d = 0; d < SUB->nseq; d++)
						SUB->seq_al[d][b] = profile->seq_al[d][e];
					e++;
				}
			}
			MAIN = TC_stack_aln(MAIN, SUB);
		}
	}
	TC_free_aln(SUB);
	TC_free_aln(A);
	
	return MAIN;
}

TC_Alignment *TC_update_aln_random_tag(TC_Alignment *A)
{
	static int tag;

	if (!A)
		return A;

	A->random_tag = ++tag;
	return A;
}

int TC_analyse_aln_column(TC_Alignment *B, int col)
{
	static char *mat;
	static int ng_cw_star;
	static char **cw_star;
	static int ng_cw_col;
	static char **cw_col;
	static int ng_cw_dot;
	static char **cw_dot;
	char r = ' ';
	int a, b, c;
	int *cw_star_count;
	int *cw_col_count;
	int *cw_dot_count;

	if (!B->S)
		B = TC_get_aln_type(B);

	if (!B->S || !(B->S)->type)
		B = TC_get_aln_type(B);

	if (!mat)
		mat = (char *)AR_allocate(STRING, sizeof(char));

	if (!ng_cw_star) {
		cw_star = TC_make_group_aa(&ng_cw_star, strcpy(mat, "idmat"));
		cw_col = TC_make_group_aa(&ng_cw_col, strcpy(mat, "clustalw_col"));
		cw_dot = TC_make_group_aa(&ng_cw_dot, strcpy(mat, "clustalw_dot"));
	}

	cw_star_count = (int *)AR_allocate(ng_cw_star, sizeof(int));
	cw_col_count = (int *)AR_allocate(ng_cw_col, sizeof(int));
	cw_dot_count = (int *)AR_allocate(ng_cw_dot, sizeof(int));

	for (a = 0; a < B->nseq; a++) {
		c = tolower(B->seq_al[a][col]);
		if (TC_is_gap(c)) {
			r = ' ';
			break;
		}
		for (b = 0; b < ng_cw_star; b++)
			cw_star_count[b] += TC_is_in_set(c, cw_star[b]);
		for (b = 0; b < ng_cw_col; b++)
			cw_col_count[b] += TC_is_in_set(c, cw_col[b]);
		for (b = 0; b < ng_cw_dot; b++)
			cw_dot_count[b] += TC_is_in_set(c, cw_dot[b]);
	}

	if (!TC_is_gap(c) && r == ' ')
		for (b = 0; b < ng_cw_star; b++)
			if (cw_star_count[b] == B->nseq) {
				r = '*';
				break;
			}
	if (!TC_is_gap(c) && r == ' ' && !(TC_strm((B->S)->type, "DNA")))
		for (b = 0; b < ng_cw_col; b++)
			if (cw_col_count[b] == B->nseq) {
				r = ':';
				break;
			}
	if (!TC_is_gap(c) && r == ' ' && !(TC_strm((B->S)->type, "DNA")))
		for (b = 0; b < ng_cw_dot; b++)
			if (cw_dot_count[b] == B->nseq) {
				r = '.';
				break;
			}

	AR_free(cw_star_count);
	AR_free(cw_col_count);
	AR_free(cw_dot_count);

	return r;
}

/*
 *##### with two parameter only: TC_Alignment *A, int n_nseq
 * Turns A into pos, a matrix where each residue is replace by its indice
 * according to the complete sequence.
 * The indices in pos are computed using A->order[x][1] that contains the indice
 * of the first residue of seq x of A
 * n_nseq MUST not be null
 * 
 *##### with more than two param:
 *      int **TC_aln2pos_simple(TC_Alignment *A, int n_nseq, int *ns, int **ls)
 * n_nseq must be set to 0 for the param 3 and four to be read
 * ns[x] = number seq in group
 * ls[x] = list of the sequences in group x (size = ns[x])
 * The computation of the indices is only carried out on the scpecified residues.
 *
 *##### IMPORTANT
 * in pos, the numbering of the residues goes from 1 to L:
 *      pos[0][0] = 3, means that the first position of the first sequence in the alignmnet
 *      contains residue #3 from sequence A->order[0][0];
 */
int **TC_aln2pos_simple(const TC_Alignment *A, int n_nseq, ...)
{
	va_list ap;
	int **T;
	int *list = 0;
	int *ns = 0;
	int **ls = 0;
	int max_nseq;
	int n_len = 0;
	int a, b, c, p, g, l;

	if (A == 0) {
		return 0;
	}
	else {
		if (n_nseq > 0) {
			list = (int *)AR_allocate(n_nseq, sizeof(int));
			for (a = 0; a < n_nseq; a++)
				list[a] = a;
		}
		else {
			va_start(ap, n_nseq);
			ns = va_arg(ap, int *);
			ls = va_arg(ap, int **);
			va_end(ap);
			list = (int *)AR_allocate(ns[0] + ns[1], sizeof(int));
			n_nseq = 0;
			for (a = 0; a < ns[0]; a++)
				list[n_nseq++] = ls[0][a];
			for (a = 0; a < ns[1]; a++)
				list[n_nseq++] = ls[1][a];

		}
		max_nseq = MAX(TC_read_array_size(A->order, sizeof(int *)), TC_return_max_int(A->order,
				TC_read_array_size(A->order, sizeof(int *)), 0)) + 1;
		n_len = TC_get_longest_string(A->seq_al, A->max_n_seq, 0, 0) + 1;

		T = TC_declare_int(max_nseq, n_len);
		for (c = 0; c < n_nseq; c++) {
			a = list[c];
			l = strlen(A->seq_al[a]);

			for (p = A->order[a][1], b = 0; b < l; b++) {
				g = 1 - TC_is_gap(A->seq_al[a][b]);
				p += g;
				T[a][b] = (g == 1) ? p : -(1 + p);
				if (A->seq_al[a][b] == UNDEFINED_RESIDUE)
					T[a][b] = 0;
				if (A->seq_cache && T[a][b] > 0) {
					T[a][b] = A->seq_cache[A->order[a][0]][T[a][b]];
				}
			}
		}
		AR_free(list);
	}

	return T;
}

char *TC_aln2cons_seq(const TC_Alignment *A, int ns, int *ls, int n_groups, char **group_list)
{
	char *seq;
	int *group;
	int best_group = 0;
	int aa_group = 0;
	int len;
	int clean_ls = 0;
	int a, b, c;
	
	len = strlen(A->seq_al[ls[0]]);
	seq = (char *)AR_allocate(len + 1, sizeof(char));
	
	if (ns == 0) {
		ns = A->nseq;
		ls = (int *)AR_allocate(A->nseq, sizeof(int));
		for (a = 0; a < A->nseq; a++)
			ls[a] = a;
		clean_ls = 1;
	}
	
	if (!group_list) {
		group_list = TC_declare_char(26, 2);
		for (a = 0; a < 26; a++)
			group_list[a][0] = a + 'a';
		n_groups = 26;
		aa_group = 1;
	}
	
	for (a = 0; a < len; a++) {
		group = (int *)AR_allocate(n_groups + 1, sizeof(int));
		for (best_group = 0, b = 0; b < ns; b++) {
			if (!TC_is_gap(A->seq_al[ls[b]][a])) {
				for (c = 0; c < n_groups; c++)
					if (TC_is_in_set(tolower(A->seq_al[ls[b]][a]), group_list[c])) {
						group[c]++;
						best_group = (group[c] > group[best_group]) ? c : best_group;
					}
			}
			seq[a] = group_list[best_group][0];
		}
		AR_free(group);
	}
	seq[a] = '\0';
	
	if (aa_group)
		TC_free_char(group_list);
	if (clean_ls)
		AR_free(ls);
	
	return seq;
}

char *TC_aln2cons_seq_mat(const TC_Alignment *A, const char *mat_name)
{
	return TC_sub_aln2cons_seq_mat(A, A->nseq, 0, mat_name);
}

char *TC_sub_aln2cons_seq_mat(const TC_Alignment *A, int ns, int *ls, const char *mat_name)
{
	char *seq, r1, r2;
	int **mat;
	int len;
	int naa;
	int score = 0, best_score, best_r;
	int a, b, c, s;
	
	mat = TC_read_matrice(mat_name);
	len = strlen(A->seq_al[(ls == 0) ? 0 : ls[0]]);
	
	seq = (char *)AR_allocate(len + 1, sizeof(char));
	for (a = 0; a < len; a++) {
		for (b = 0; b < 20; b++) {
			r1 = AA_ALPHABET[b];
			for (naa = 0, score = 0, c = 0; c < ns; c++) {
				s = (ls == 0) ? c : ls[c];
				if (ls && ls[c] == -1)
					continue;
				else if (TC_is_gap(A->seq_al[s][a]))
					continue;
				else {
					naa++;
					r2 = A->seq_al[s][a];
					score += mat[r1 - 'A'][r2 - 'A'];
				}
			}
			if (naa == 0)
				best_r = '-';
			if (b == 0 || score > best_score) {
				best_score = score;
				best_r = r1;
			}
		}
		seq[a] = best_r;
	}
	TC_free_int(mat);
	
	return seq;
}

/**
 * @brief Counts the number of residues in each column of an alignment (Prot/NA).
 */
int **TC_aln2count_mat(const TC_Alignment *A)
{
	int **freq_mat;
	int alp_size;
	int a, b, x;
	
	alp_size = sizeof(AA_ALPHABET);
	freq_mat = TC_declare_int(alp_size + 2, A->len_aln);
	
	for (a = 0; a < A->len_aln; a++) {
		for (b = 0; b < A->nseq; b++) {
			if (TC_is_gap(A->seq_al[b][a]))
				freq_mat[alp_size][a]++;
			else {
				x = tolower(A->seq_al[b][a]);
				freq_mat[x - 'a'][a]++;
				freq_mat[alp_size + 1][a]++;
			}
		}
	}
	
	return freq_mat;
}

int **TC_aln2count_mat2(const TC_Alignment *A)
{
	return TC_sub_aln2count_mat2(A, 0, 0);
}

int **TC_sub_aln2count_mat2(const TC_Alignment *A, int ns, int *l_s)
{
	char r;
	int used[1000];
	int **count;
	int l_s_flag = 0;
	int len;
	int us;
	int a, b;

	/*
	 * count[x][0] = N symbols in column
	 * count[x][1] = Total_size of line
	 * count[x][2] = Gap frequency
	 * count[x][n] = symbol n
	 * count[x][n+1] = N occurence symbol n
	 * count[x][n+2] = N frequence symbol n*100
	 *
	 * special multi-channeling
	 * count[x][count[x][1]] = Nseq
	 * count[x][count[x][1]+s] = residue col x, sequence s
	 */

	for (a = 0; a < 1000; a++)
		used[a] = 0;
	if (ns == 0) {
		ns = A->nseq;
		l_s = (int *)AR_allocate(A->nseq, sizeof(int));
		for (a = 0; a < A->nseq; a++)
			l_s[a] = a;
		l_s_flag = 1;
	}
	len = strlen(A->seq_al[l_s[0]]);

	count = TC_declare_int(len, 100 + ns + 2);
	for (a = 0; a < len; a++) {
		for (us = ns, b = 0; b < ns; b++) {
			r = tolower(A->seq_al[l_s[b]][a]);

			if (TC_is_gap(r))
				us--;
			else if (used[r]) {
				count[a][used[r] * 3 + 1]++;
			}
			else {
				used[r] = ++count[a][0];
				count[a][used[r] * 3] = r;
				count[a][used[r] * 3 + 1]++;
			}
		}
		count[a][1] = count[a][0] * 3 + 2;
		count[a][2] = (A->nseq - us) * 100 / A->nseq;

		for (b = 3; b < count[a][1]; b += 3) {
			count[a][b + 2] = (count[a][b + 1] * 100) / us;
			used[count[a][b]] = 0;
		}
	}

	if (l_s_flag)
		AR_free(l_s);

	return count;
}

int **TC_seq2aln_pos(const TC_Alignment *A, int *ns, int **l_s)
{
	int **code;
	int a, b, c, d, l, p, g;

	l = MAX(strlen(A->seq_al[l_s[0][0]]), strlen(A->seq_al[l_s[1][0]]));
	code = TC_declare_int((A->S)->nseq, l + 1);

	for (c = 0; c < 2; c++) {
		l = strlen(A->seq_al[l_s[c][0]]);
		for (d = 0; d < ns[c]; d++) {
			a = A->order[l_s[c][d]][0];
			for (p = 0, b = 0; b < l; b++) {
				g = TC_is_gap(A->seq_al[l_s[c][d]][b]);
				if (!g) {
					p++;
					code[a][p] = b + 1;
				}
			}
		}
	}

	return code;
}

int TC_output_format_aln(const char *format, const TC_Alignment *inA, const TC_Alignment *inEA,
	const char *name)
{
	TC_Alignment *A = 0;
	/*TC_Alignment *EA = 0;*/

	A = TC_copy_aln(inA, 0);
	/*EA = TC_copy_aln(inEA, 0);*/

	A = TC_expand_aln(A);
	/*EA = TC_expand_number_aln(inA, EA);*/

	if (A && A->expanded_order)
		A = TC_reorder_aln(A, A->expanded_order, A->nseq);
	/*if (EA && EA->expanded_order)
		EA = TC_reorder_aln(EA, EA->expanded_order, EA->nseq);*/

	AR_main_output_aln(A, format, name);

	TC_free_aln(A);
	/*TC_free_aln(EA);*/

	return 1;
}

#if DEBUG_TREE
void print_A(TC_Alignment* A) {
    std::cout << "A->max_len: " << A->max_len << std::endl;
    std::cout << "A->min_len: " << A->min_len << std::endl;
    std::cout << "A->declared_len: " << A->declared_len << std::endl;
    std::cout << "A->max_n_seq: " << A->max_n_seq << std::endl;
    std::cout << "A->nseq: " << A->nseq << std::endl;
    std::cout << "A->len_aln: " << A->len_aln << std::endl;

    std::cout << "A->seq_al: " << A->seq_al << std::endl;
    for (int i = 0; i < A->max_n_seq; ++i) {
	std::cout << " " << i << ": " << A->seq_al[i] << std::endl;
    }

    std::cout << "A->order: " << A->order << std::endl;
    for (int i = 0; i < A->S->nseq; ++i) {
	std::cout << " " << i << ": " 
		  << A->order[i][0] << " " << A->order[i][1] << std::endl;
    }

    std::cout << "A->score_seq: " << A->score_seq << std::endl;
    for (int i = 0; i < A->S->nseq; ++i) {
	std::cout << " " << i << ": " << A->score_seq[i] << std::endl;
    }

    std::cout << "A->score_res: " << A->score_res << std::endl;
    std::cout << "A->score_aln: " << A->score_aln << std::endl;
    std::cout << "A->score: " << A->score << std::endl;
    std::cout << "A->file: " << A->file << std::endl;
    std::cout << "A->seq_comment: " << A->seq_comment << std::endl;
    std::cout << "A->aln_comment: " << A->aln_comment << std::endl;
    std::cout << "A->name: " << A->name << std::endl;
    std::cout << "A->expanded_order: " << A->expanded_order << std::endl;
    std::cout << "A->tree_order: " << A->tree_order << std::endl;
    std::cout << "A->P: " << A->P << std::endl; 
    std::cout << "A->S: " << A->S << std::endl;
    std::cout << "A->CL: " << A->CL << std::endl;
    std::cout << "A->seq_cache: " << A->seq_cache << std::endl;
    std::cout << "A->cdna_cache: " << A->cdna_cache << std::endl;
    std::cout << "A->cpu: " << A->cpu << std::endl;
    std::cout << "A->finished: " << A->finished << std::endl;
    std::cout << "A->output_res_num: " << A->output_res_num << std::endl;
    std::cout << "A->residue_case: " << A->residue_case << std::endl;
    std::cout << "A->expand: " << A->expand << std::endl;
    std::cout << "A->used: " << A->used << std::endl;
    std::cout << "A->num: " << A->num << std::endl;
    std::cout << "A->pos: " << A->pos << std::endl;
    std::cout << "A->A: " << A->A << std::endl;
    std::cout << "A->random_tag: " << A->random_tag << std::endl;
} // print_A
#endif
