/*
 * File: TC_ConstraintList.cpp
 * Created: 06.06.2006
 * Last Modified: 13.10.2006
 * Author: Adrian Rospondek
 */


#include <algorithm>
#include <iostream>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "AR_LibJob.hpp"
#include "JZ_ConstraintList.hpp"
#include "TC_Alignment.hpp"
#include "TC_ConstraintList.hpp"
#include "TC_DistanceMatrix.hpp"
#include "TC_Method.hpp"
#include "TC_Sequences.hpp"
#include "TC_Weights.hpp"
#include "config.hpp"
#include "dp_sim.hpp"
#include "evaluate.hpp"
#include "globals.hpp"
#include "input.hpp"
#include "util.hpp"
#include "util_algo.hpp"
#include "util_io.hpp"
#include "util_mem.hpp"
#include "util_mpi.hpp"
#include "util_seq.hpp"
#include "util_str.hpp"




TC_ConstraintList *TC_declare_constraint_list(TC_Sequences *S, short int **L, int ne, int **M)
{
    TC_ConstraintList *CL = new TC_ConstraintList();
//    CL = (TC_ConstraintList *)AR_allocatate(1, sizeof(TC_ConstraintList));

    CL->S = S;
    CL->M = M;
    // CL->L = L;
    CL->L = 0; // L is always 0
    CL->ne = ne;
    // CL->chunk = 65536;
    CL->chunk = 16;
    CL->weight_field = WE;
    CL->matrices_list = TC_declare_char(20, 20);
    if (S)
	CL->seq_for_quadruplet = (int *)AR_allocate(S->nseq, sizeof(int));
    CL->cpu = 1;

#if USE_RMA
    CL->residue_index = ResIdx();
#endif

    return CL;
}

static TC_ConstraintList *TC_copy_constraint_list(const TC_ConstraintList *CL, int mode)
{
    TC_ConstraintList *NCL;
    TC_Sequences *S;
    unsigned int a, b;

    /* Sequences */
    S = (mode == HARD_COPY) ? TC_duplicate_sequence(CL->S) : CL->S;

    if (mode == HARD_COPY)
	NCL = TC_declare_constraint_list(S, 0, 0, 0);
    else {
//	NCL = (TC_ConstraintList *)AR_allocate(1, sizeof(TC_ConstraintList));
	NCL = new TC_ConstraintList();
	//NCL[0] = CL[0];
    }

    NCL->S = S;
    NCL->copy_mode = mode;
    if (mode == SOFT_COPY)
	NCL->pCL = CL;

    NCL->STRUC_LIST = (mode == HARD_COPY) ? TC_duplicate_sequence(CL->STRUC_LIST) : CL->STRUC_LIST;
    strncpy(NCL->align_pdb_param_file, CL->align_pdb_param_file, 100 - 1);
    strncpy(NCL->align_pdb_hasch_mode, CL->align_pdb_hasch_mode, 100 - 1);
    NCL->W = (mode == HARD_COPY) ? TC_duplicate_weights(CL->W) : CL->W;
    NCL->DM = (mode == HARD_COPY) ? TC_duplicate_distance_matrix(CL->DM) : CL->DM;
    NCL->ktupDM = (mode == HARD_COPY) ? TC_duplicate_distance_matrix(CL->ktupDM) : CL->ktupDM;

    if (mode == HARD_COPY && CL->translation) {
	NCL->translation = (int *)AR_allocate((CL->S)->nseq, sizeof(int));
	for (a = 0; a < (CL->S)->nseq; a++)
	    NCL->translation[a] = CL->translation[a];
    }
    else
	NCL->translation = CL->translation;

    /* Packing Sequence: To use with domain analysis */
    NCL->packed_seq_lu = (mode == HARD_COPY) ? TC_duplicate_int(CL->packed_seq_lu, -1, -1)
	: CL->packed_seq_lu;

    /* DATA */
    if (mode == HARD_COPY) {
	for (a = 0; a < CL->ne; a++)
	    for (b = 0; b < CL->entry_len; b++)
		TC_vwrite_clist(NCL, a, b, TC_vread_clist(CL, a, b));
    }
    else
	NCL->L = CL->L;
    if (mode == HARD_COPY)
	NCL->M = TC_copy_int(CL->M, NCL->M, -1, -1);
    else
	NCL->M = CL->M;

    /* List Information */
    NCL->ne = CL->ne;

    /* Normalisation information */
    NCL->normalise = CL->normalise;
    NCL->overweight = CL->overweight;
    NCL->max_ext_value = CL->max_ext_value;
    NCL->max_value = CL->max_value;

    /* Pair wise alignment method */
    NCL->pw_parameters_set = CL->pw_parameters_set;
    NCL->gop = CL->gop;
    NCL->f_gop = CL->f_gop;
    NCL->gep = CL->gep;
    NCL->f_gep = CL->f_gep;
    NCL->nomatch = CL->nomatch;
    NCL->TG_MODE = CL->TG_MODE;
    NCL->F_TG_MODE = CL->F_TG_MODE;
    strncpy(NCL->dp_mode, CL->dp_mode, 100 - 1);
    NCL->maximise = CL->maximise;
    strncpy(NCL->matrix_for_aa_group, CL->matrix_for_aa_group, 100 - 1);
    NCL->diagonal_threshold = CL->diagonal_threshold;
    NCL->ktup = CL->ktup;
    NCL->use_fragments = CL->use_fragments;
    NCL->fasta_step = CL->fasta_step;
    NCL->lalign_n_top = CL->lalign_n_top;
    NCL->sw_min_dist = CL->sw_min_dist;
    NCL->matrices_list = (mode == HARD_COPY) ? TC_duplicate_char(CL->matrices_list, -1, -1)
	: CL->matrices_list;
    NCL->n_matrices = CL->n_matrices;
    strncpy(NCL->distance_matrix_mode, CL->distance_matrix_mode, 100 - 1);
    strncpy(NCL->tree_mode, CL->tree_mode, 100 - 1);
    NCL->tree_aln = (mode == HARD_COPY) ? TC_copy_aln(CL->tree_aln, 0) : CL->tree_aln;

    /* Functions used for dynamic programming and Evaluation */
    NCL->profile_mode = CL->profile_mode;
    strncpy(NCL->profile_comparison, CL->profile_comparison, 100 - 1);
    NCL->get_dp_cost = CL->get_dp_cost;
    NCL->evaluate_residue_pair = CL->evaluate_residue_pair;
    NCL->pair_wise = CL->pair_wise;

    NCL->weight_field = CL->weight_field;
    NCL->max_n_pair = CL->max_n_pair;

    /* Split parameters */
    NCL->split = CL->split;
    NCL->split_nseq_thres = CL->split_nseq_thres;
    NCL->split_score_thres = CL->split_score_thres;

    strncpy(NCL->evaluate_mode, CL->evaluate_mode, 100 - 1);

    /* Functions for hiding forbiden pairs of residues */
    /* Copy only for soft_copy */
    if (mode == SOFT_COPY)
	NCL->forbiden_pair_list = CL->forbiden_pair_list;

    /* Extention properties: */
    NCL->nseq_for_quadruplet = CL->nseq_for_quadruplet;
    if (mode == HARD_COPY && CL->seq_for_quadruplet) {
	NCL->seq_for_quadruplet = (int *)AR_allocate(S->nseq, sizeof(int));
	for (a = 0; a < S->nseq; a++)
	    NCL->seq_for_quadruplet[a] = CL->seq_for_quadruplet[a];
    }
    else if (mode == SOFT_COPY)
	NCL->seq_for_quadruplet = CL->seq_for_quadruplet;
    /* Extention properties: Do only a soft copy */
    if (mode == SOFT_COPY) {
	NCL->extend_jit = CL->extend_jit;
	NCL->extend_threshold = CL->extend_threshold;
	strncpy(NCL->extend_clean_mode, CL->extend_clean_mode, 100 - 1);
	strncpy(NCL->extend_compact_mode, CL->extend_compact_mode, 100 - 1);
    }

    /* Lookup table parameteres */
    NCL->chunk = CL->chunk;

    /* MISC */
    NCL->cpu = CL->cpu;

    return NCL;
}

/**
 * @brief Duplicate everything in the constraint list.
 */
TC_ConstraintList *TC_duplicate_constraint_list(const TC_ConstraintList *CL)
{
    return TC_copy_constraint_list(CL, HARD_COPY);
}

/**
 * @brief Duplication that does not copy the long lists.
 */
TC_ConstraintList *TC_duplicate_constraint_list_soft(const TC_ConstraintList *CL)
{
    return TC_copy_constraint_list(CL, SOFT_COPY);
}

static TC_ConstraintList *TC_duplicate_constraint_list4lib_computation(const TC_ConstraintList *CL)
{
    TC_ConstraintList *SCL;

    SCL = new TC_ConstraintList();
//    SCL = (TC_ConstraintList *)AR_allocate(1, sizeof(TC_ConstraintList));
    SCL[0] = CL[0];
    SCL->S = CL->S;
    SCL->L = 0;
    SCL->max_L_len = 0;
    SCL->M = 0;
    SCL->ne = 0;
    SCL->residue_indexed = 0;
    SCL->residue_index = 0;

    return SCL;
}

TC_Sequences *TC_free_constraint_list(TC_ConstraintList *CL)
{
    TC_Sequences *S;
    const TC_ConstraintList *pCL;
    int a, b;

    /* Prepare the selective freeing of the CL data structure:
     * If the CL has been obtain from copy, every pointer that is identical to the parent CL (CL->pCL)
     * will not be saved. */
    if (!CL)
	return 0;
    else
	S = CL->S;

    if (CL->copy_mode == SOFT_COPY && !CL->pCL) {
//	AR_free(CL);
	delete CL;
	return S;
    }
    else if (CL->copy_mode == SOFT_COPY) {
	pCL = CL->pCL;
	CL->L = 0;
	if (CL->M == pCL->M)
	    CL->M = 0;
	if (CL->matrices_list == pCL->matrices_list)
	    CL->matrices_list = 0;
	if (CL->STRUC_LIST == pCL->STRUC_LIST)
	    CL->STRUC_LIST = 0;
	if (CL->W == pCL->W)
	    CL->W = 0;
	if (CL->DM == pCL->DM)
	    CL->DM = 0;
	if (CL->ktupDM == pCL->ktupDM)
	    CL->ktupDM = 0;
	if (CL->translation == pCL->translation)
	    CL->translation = 0;
	if (CL->seq_for_quadruplet == pCL->seq_for_quadruplet)
	    CL->seq_for_quadruplet = 0;
    }

    /* End of selective freeing of the CL data structure */
    if (CL->L) CL->L.free();
//	TC_free_short_int(CL->L);
	
    if (CL->M)
	TC_free_int(CL->M);
    if (CL->matrices_list)
	TC_free_char(CL->matrices_list);
    if (CL->start_index)
	TC_free_int(CL->start_index);
    if (CL->end_index)
	TC_free_int(CL->end_index);

    CL->M = 0;
    CL->matrices_list = 0;
    CL->start_index = 0;
    CL->end_index = 0;
    
#if USE_RMA
    if (CL->residue_index) CL->residue_index.free();
#else
    if (CL->residue_index) {
	for (a = 0; a < (CL->S)->nseq; a++) {
	    for (b = 0; b <= (CL->S)->len[a]; b++) AR_free(CL->residue_index[a][b]);
	    AR_free(CL->residue_index[a]);
	}
	AR_free(CL->residue_index);
	CL->residue_index = 0;
    }
#endif

    if (CL->STRUC_LIST)
	TC_free_sequence(CL->STRUC_LIST);
    if (CL->W)
	TC_free_weights(CL->W);
    TC_free_distance_matrix(CL->DM);
    TC_free_distance_matrix(CL->ktupDM);
    if (CL->translation)
	AR_free(CL->translation);
    if (CL->seq_for_quadruplet)
	AR_free(CL->seq_for_quadruplet);

//    AR_free(CL);
    delete CL;

    return S;
}

void TC_free_constraint_list_full(TC_ConstraintList *CL)
{
    TC_free_sequence(TC_free_constraint_list(CL));
}

static void TC_free_constraint_list4lib_computation(TC_ConstraintList *CL)
{
    // TC_free_short_int(CL->L);
    CL->L.free();
    TC_free_int(CL->M);
//    AR_free(CL);
    delete CL;
}

short int TC_vwrite_clist(TC_ConstraintList *CL, int a, int b, short int x)
{
    CL->seq_indexed = 0;
    CL->residue_indexed = 0;

    if (!CL->M) {
	if (a >= CL->max_L_len) {
//	    if (!(CL->L)) {
		// (CL->L) = TC_declare_short_int(CL->chunk, CL->entry_len);
//		CL->L.allocate(CL->chunk, CL->entry_len);
//	    }
//	    else {
		// (CL->L) = TC_realloc_short_int((CL->L), CL->max_L_len, CL->entry_len, CL->chunk, 0);
		CL->L.reallocate(CL->chunk, CL->entry_len);
		CL->max_L_len += CL->chunk;
//	    }
	}
	(CL->L)[a][b] = x;
    }

    return x;
}

int TC_vread_clist(const TC_ConstraintList *CL, int a, int b)
{
    if (a >= CL->ne)
	return UNDEFINED_2;
    else if (CL->L)
	return CL->L[a][b];
    else if (CL->M)
	return CL->M[a][b];
    else
	return UNDEFINED_2;
}

short int *TC_extract_entry(short int *entry, int pos, const TC_ConstraintList *CL)
{
    unsigned int a;

    if (entry == 0)
	entry = (short int *)AR_allocate(CL->entry_len, sizeof(short int));

    for (a = 0; a < CL->entry_len; a++)
	entry[a] = TC_vread_clist(CL, pos, a);

    return entry;
}

static TC_ConstraintList *TC_insert_entry2list(const short int *entry, int pos, TC_ConstraintList *CL)
{
    unsigned int a;

    for (a = 0; a < CL->entry_len; a++)
	TC_vwrite_clist(CL, pos, a, entry[a]);

    return CL;
}

TC_ConstraintList *TC_add_entry2list(const short int *entry, TC_ConstraintList *CL)
{
    return TC_insert_entry2list(entry, CL->ne++, CL);
}

static int *TC_seqpair2weight(int s1, int s2, const TC_Alignment *A, const char *weight_mode, int *weight)
{
    int *col;
    int a, c, ref_weight;

    if (!weight)
	weight = (int *)AR_allocate(MAX(2, A->len_aln), sizeof(int));

    weight[0] = FORBIDEN;
    if (weight_mode == 0 || strcmp(weight_mode, "no") == 0 || TC_is_number(weight_mode)) {
	if (TC_is_number(weight_mode))
	    ref_weight = atoi(weight_mode);
	else
	    ref_weight = 1;
	weight[1] = ref_weight;
    }
    else if (strncmp(weight_mode, "len", 3) == 0) {
	weight[1] = A->len_aln;
    }
    else if (TC_strnm(weight_mode, "sim", 3) || TC_strm(weight_mode, "default")) {
	ref_weight = TC_get_seq_sim(A->seq_al[s1], A->seq_al[s2], "-", weight_mode + 3);
	weight[1] = ref_weight;
    }
    else if (TC_strnm(weight_mode, "subset", 6)) {
	ref_weight = TC_get_seq_sim(A->seq_al[s1], A->seq_al[s2], "-", 0);
	weight[1] = ref_weight;
    }
    else if (strncmp(weight_mode, "winsim", 6) == 0) {
	weight = TC_get_seq_winsim(A->seq_al[s1], A->seq_al[s2], "-", weight_mode + 6, weight);
    }
    else if (strncmp(weight_mode, "cdna", 4) == 0) {
	ref_weight = TC_get_seq_sim(A->seq_al[s1], A->seq_al[s2], "-", weight_mode + 4);
	col = (int *)AR_allocate(A->len_aln + 1, sizeof(int));

	if (A->cdna_cache)
	    for (a = 0; a <= A->len_aln; a++)
		col[a] = A->cdna_cache[0][a];
	else
	    for (a = 0; a <= A->len_aln; a++)
		col[a] = 1;
	for (c = 0; c < A->len_aln; c++)
	    weight[c] = ref_weight * col[c];

	AR_free(col);
    }
    else if (TC_strm(weight_mode, "pdb")) {
	//if (!(A->CL) || !(A->CL)->T) {
	std::cerr << '\n' << PROGRAM << ": ERROR: Could not find the PDB structure.\n";
	//}
    }
    else {
	std::cerr << '\n' << PROGRAM << ": ERROR: Weight mode '" << weight_mode << "' is unknown.\n";
    }

    return weight;
}

TC_ConstraintList *TC_aln2constraint_list(const TC_Alignment *A, TC_ConstraintList *CL, const char *in_weight_mode)
{
    TC_ConstraintList *CLB = 0;
    const char *weight_mode = 0;
    char *alp = 0;
    char *p;
    int *weight = 0;
    int fixed_nres1, fixed_nres2;
    int do_pdb = 0;
    int pdb_weight;
    int set_misc;
    int s1, s2;
    int a, b, c, nres1, nres2;

    weight_mode = (!in_weight_mode || TC_strm(in_weight_mode, "default")) ? "sim" : in_weight_mode;

    if (!A)
	return CL;

    if (!CL) {
	TC_Sequences *S;

	S = TC_aln2seq(A);
	CL = TC_declare_constraint_list(S, 0, 0, 0);
	CL->S = S;
    }
    CLB = A->CL;

    do_pdb = (strstr(weight_mode, "pdb")) ? 1 : 0;
    p = (char*)strstr(weight_mode, "_subset_");
    if (p) {
	alp = (char*)strchr(weight_mode, '_') + 1;
	p[0] = '\0';
    }

    for (a = 0; a < A->nseq - 1; a++) {
	for (set_misc = 0, b = a + 1; b < A->nseq; b++) {
	    s1 = TC_name_is_in_list(A->name[a], (CL->S)->name, (CL->S)->nseq, 100);
	    s2 = TC_name_is_in_list(A->name[b], (CL->S)->name, (CL->S)->nseq, 100);

	    if (s1 != -1 && s2 != -1) {
		int use_pair;

		weight = TC_seqpair2weight(a, b, A, weight_mode, weight);

		for (nres1 = A->order[a][1], nres2 = A->order[b][1], c = 0; c < A->len_aln; c++) {
		    nres1 += !TC_is_gap(A->seq_al[a][c]);
		    nres2 += !TC_is_gap(A->seq_al[b][c]);

		    if (TC_strm(weight_mode, "pdb")) {
			pdb_weight = MAX(0, (CLB->evaluate_residue_pair) (CLB, 0, nres1, 1, nres2));
		    }

		    use_pair = 1;
		    use_pair = use_pair && !TC_is_gap(A->seq_al[a][c]);
		    use_pair = use_pair && !TC_is_gap(A->seq_al[b][c]);
		    use_pair = use_pair && A->seq_al[b][c] != UNDEFINED_RESIDUE;
		    use_pair = use_pair && A->seq_al[a][c] != UNDEFINED_RESIDUE;
		    use_pair = use_pair && !(do_pdb && pdb_weight == 0);
		    if (alp)
			use_pair = use_pair && TC_is_in_set(A->seq_al[b][c], alp) && TC_is_in_set(A->seq_al[a][c], alp);

		    if (use_pair) {
			fixed_nres1 = (!A->seq_cache) ? nres1 : A->seq_cache[s1][nres1];
			fixed_nres2 = (!A->seq_cache) ? nres2 : A->seq_cache[s2][nres2];

			if (fixed_nres1 == -1 || fixed_nres2 == -1) {
			    std::cerr << '\n' << PROGRAM << ": PB: Sequence '" << A->name[a] << "', Residue " << nres1
				      << " : Cache= " << fixed_nres1;
			    std::cerr << '\n' << PROGRAM << ": PB: Sequence '" << A->name[b] << "', Residue " << nres2
				      << " : Cache= " << fixed_nres2 << '\n';

			    exit(EXIT_FAILURE);
			}

			if (fixed_nres1 && fixed_nres2) {
			    TC_vwrite_clist(CL, CL->ne, SEQ1, s1);
			    TC_vwrite_clist(CL, CL->ne, SEQ2, s2);
			    TC_vwrite_clist(CL, CL->ne, R1, fixed_nres1);
			    TC_vwrite_clist(CL, CL->ne, R2, fixed_nres2);

			    if (do_pdb) {
				TC_vwrite_clist(CL, CL->ne, WE, pdb_weight);
			    }
			    else {
				TC_vwrite_clist(CL, CL->ne, WE, (weight[0] == FORBIDEN) ? weight[1] : weight[c]);
			    }

			    TC_vwrite_clist(CL, CL->ne, CONS, 1);

			    if (!set_misc) {
				TC_vwrite_clist(CL, CL->ne, MISC, A->len_aln);
				set_misc = 1;
			    }
			    else {
				TC_vwrite_clist(CL, CL->ne, MISC, 0);
			    }

			    CL->ne++;
			}
		    }
		}
	    }
	}
    }
    AR_free(weight);

    if (A->A)
	return TC_aln2constraint_list(A->A, CL, weight_mode);
    else
	return CL;
}

TC_ConstraintList *TC_filter_list(TC_ConstraintList *CL, int start, int len, int T)
{
    int a;
    int field;

    if (T == 0)
	return CL;

    field = WE;
    if (T > 0)
	field = WE;
    else if (T < 0) {
	field = CONS;
	T = -T;
    }

    for (a = start; a < len; a++)
	if (TC_vread_clist(CL, a, field) <= T)
	    TC_vwrite_clist(CL, a, SEQ1, -1);

    CL = TC_sort_constraint_list_inv(CL, 0, CL->ne);
    CL = TC_sort_constraint_list(CL, 0, CL->ne);

    return CL;
}

static TC_ConstraintList *TC_re_weight_constraint_list(TC_ConstraintList *CL, const TC_Weights *W)
{
    float *weight;
    float w;
    int sA, sB;
    int a;

    if (!CL->L)
	return CL;

    weight = W->SEQ_W;
    for (a = 0; a < CL->ne; a++) {
	sA = CL->L[a][SEQ1];
	sB = CL->L[a][SEQ2];
	w = MIN(weight[sA], weight[sB]);
	CL->L[a][WE] = (int)(w * CL->L[a][WE]);
    }
    CL = TC_evaluate_constraint_list_reference(CL);

    return CL;
}

TC_ConstraintList *TC_weight_constraint_list(TC_ConstraintList *CL, const char *seq_weight)
{
    TC_Weights *W;

    if (CL->ne == 0)
	return CL;
    else if (TC_strm(seq_weight, "t_coffee"))
	W = TC_compute_tcoffee_weight(CL);
    else if (AR_check_file_exists(seq_weight)) {
	W = TC_read_seq_weight((CL->S)->name, (CL->S)->nseq, seq_weight);
    }
    else {
	int a;

	W = TC_declare_weights((CL->S)->nseq);
	strncpy(W->mode, "no_seq_weight", FILENAMELEN - 1);
	for (a = 0; a < (CL->S)->nseq; a++) {
	    strncpy(W->seq_name[a], (CL->S)->name[a], 200 - 1);
	    W->SEQ_W[a] = 1;
	}
	CL->W = W;

	return CL;
    }
    CL = TC_re_weight_constraint_list(CL, W);
    CL->W = W;

//     std::cout << "max_val: " << CL->max_value
// 	      << " max_ext_val: " << CL->max_ext_value << std::endl;

    return CL;
}

TC_ConstraintList *TC_evaluate_constraint_list_reference(TC_ConstraintList *CL)
{
    static short int *entry;
    int ***max_res;
    int a, b, c, s1, s2, r1, r2, w;

    if (CL->M)
	CL->max_value = CL->max_ext_value = 20;
    else {
	CL->max_value = CL->max_ext_value = 0;
	max_res = (int ***)AR_allocate((CL->S)->nseq, sizeof(int **));

	for (a = 0; a < (CL->S)->nseq; a++) {
	    max_res[a] = (int **)AR_allocate(strlen((CL->S)->seq[a]) + 1, sizeof(int *));
	    for (b = 0; b <= (CL->S)->len[a]; b++) {
		max_res[a][b] = (int *)AR_allocate((CL->S)->nseq + 1, sizeof(int));
	    }
	}

	for (a = 0; a < CL->ne; a++) {
	    entry = TC_extract_entry(entry, a, CL);
	    s1 = entry[SEQ1];
	    s2 = entry[SEQ2];
	    r1 = entry[R1];
	    r2 = entry[R2];
	    w = entry[WE];
	    if (w == TC_UNDEFINED);
	    else {
		max_res[s1][r1][s2] += w;
		max_res[s2][r2][s1] += w;
		CL->max_value = MAX(w, CL->max_value);
	    }
	}

	for (a = 0; a < (CL->S)->nseq; a++)
	    for (b = 1; b <= (CL->S)->len[a]; b++) {
		for (c = 0; c < (CL->S)->nseq; c++)
		    max_res[a][b][(CL->S)->nseq] += max_res[a][b][c];
		CL->max_ext_value = MAX(max_res[a][b][c], CL->max_ext_value);
	    }

	for (a = 0; a < (CL->S)->nseq; a++) {
	    for (b = 0; b <= (CL->S)->len[a]; b++)
		AR_free(max_res[a][b]);
	    AR_free(max_res[a]);
	}
	AR_free(max_res);

	CL->max_ext_value = MAX(1, CL->max_ext_value);
    }

    if (CL->normalise) {
	CL->nomatch = (CL->nomatch * CL->normalise) / CL->max_ext_value;
    }

    return CL;
}

TC_ConstraintList *TC_sort_constraint_list(TC_ConstraintList *CL, int start, int len)
{
    CL = TC_sort_constraint_list_on_n_fields(CL, start, len, 0, CL->entry_len);

    return CL;
}

TC_ConstraintList *TC_sort_constraint_list_on_n_fields(TC_ConstraintList *CL, int start, int len,
						       int first_field, int n_fields)
{
    if (CL->L) {
//	TC_hsort_list_array((void **)(CL->L.tab) + start, len, sizeof(short int), CL->entry_len, first_field, n_fields);
//	TC_hsort_list_array((void **)(CL->L) + start, len, sizeof(short int), CL->entry_len, first_field, n_fields);
	FakeEntry* arr = (FakeEntry*)CL->L.tab;
	std::sort(arr + start, arr + start + len);
    }
    return CL;
}

static TC_ConstraintList *TC_invert_constraint_list(TC_ConstraintList *CL, const int start, const int len)
{
    int a, b;
    unsigned int c;
    int tp;

    for (a = start, b = start + len - 1; a <= b; ++a, --b) {
	for (c = 0; c < CL->entry_len; ++c) {
	    tp = TC_vread_clist(CL, a, c);
	    TC_vwrite_clist(CL, a, c, TC_vread_clist(CL, b, c));
	    TC_vwrite_clist(CL, b, c, tp);
	}
    }

    return CL;
}

TC_ConstraintList *TC_sort_constraint_list_inv(TC_ConstraintList *CL, int start, int len)
{
    CL = TC_sort_constraint_list(CL, start, len);
    CL = TC_invert_constraint_list(CL, start, len);

    if (start + len == CL->ne) {
	while (TC_vread_clist(CL, CL->ne - 1, SEQ1) == -1)
	    CL->ne--;
    }

    return CL;
}

TC_ConstraintList *TC_compact_list(TC_ConstraintList *CL, const int start, const int len, const char *compact_mode)
{
    int rs1, rs2, rr1, rr2, s1, s2, r1, r2;
    int x, ra, a;

    if (!CL || !CL->L || len == 0 || TC_strm3(compact_mode, "no", "No", "NO"))
	return CL;
//    else if (TC_strm4(compact_mode, "default", "shrink", "shrink_best", "shrink_worst")) {
	// reorganise list to make ordering
        // we use this for bucket sort
	for (a = start; a < (start + len); ++a) {
	    s1 = TC_vread_clist(CL, a, SEQ1);
	    s2 = TC_vread_clist(CL, a, SEQ2);
	    r1 = TC_vread_clist(CL, a, R1);
	    r2 = TC_vread_clist(CL, a, R2);

	    if (s1 > s2 || (s1 == s2 && r1 > r2)) {
		TC_vwrite_clist(CL, a, SEQ1, s2);
		TC_vwrite_clist(CL, a, SEQ2, s1);
		TC_vwrite_clist(CL, a, R1, r2);
		TC_vwrite_clist(CL, a, R2, r1);
	    }
	}
//    }

    TC_sort_constraint_list(CL, start, len);

    rs1 = TC_vread_clist(CL, start, SEQ1);
    rs2 = TC_vread_clist(CL, start, SEQ2);
    rr1 = TC_vread_clist(CL, start, R1);
    rr2 = TC_vread_clist(CL, start, R2);
    ra = start;

    if (rs1 == rs2 && rr1 == rr2)
	TC_vwrite_clist(CL, start, SEQ1, -1);

    for (a = start + 1; a < (start + len); ++a) {
	s1 = TC_vread_clist(CL, a, SEQ1);
	s2 = TC_vread_clist(CL, a, SEQ2);
	r1 = TC_vread_clist(CL, a, R1);
	r2 = TC_vread_clist(CL, a, R2);

	if (s1 == s2 && r1 == r2)
	    TC_vwrite_clist(CL, a, SEQ1, -1);
	else if (s1 == rs1 && s2 == rs2 && r1 == rr1 && r2 == rr2) {
	    x = TC_vread_clist(CL, ra, WE);

	    if (TC_strm2(compact_mode, "default", "mirror_sum"))
		TC_vwrite_clist(CL, ra, WE, TC_vread_clist(CL, a, WE) + x);
	    else if (TC_strm2(compact_mode, "best", "shrink_best"))
		TC_vwrite_clist(CL, ra, WE, MAX(TC_vread_clist(CL, a, WE), x));
	    else if (TC_strm2(compact_mode, "worst", "shrink_worst"))
		TC_vwrite_clist(CL, ra, WE, MIN(TC_vread_clist(CL, a, WE), x));

	    if (!TC_strm(compact_mode, "shrink")) {
		TC_vwrite_clist(CL, ra, CONS, TC_vread_clist(CL, ra, CONS) + TC_vread_clist(CL, a, CONS));
		TC_vwrite_clist(CL, ra, MISC, TC_vread_clist(CL, ra, MISC) + TC_vread_clist(CL, a, MISC));
	    }

	    TC_vwrite_clist(CL, a, SEQ1, -1);
	}
	else {
	    rs1 = s1;
	    rs2 = s2;
	    rr1 = r1;
	    rr2 = r2;
	    ra = a;
	}
    }

    TC_sort_constraint_list_inv(CL, 0, CL->ne);
    TC_sort_constraint_list(CL, 0, CL->ne);

    // do we need this??? (2 ints removed from entry)
    if (TC_strm3(compact_mode, "consPwe", "wePcons", "cons")) {
	for (a = start; a < (start + len); a++) {
	    if (TC_strm2(compact_mode, "consPwe", "wePcons"))
		TC_vwrite_clist(CL, a, WE, TC_vread_clist(CL, a, WE) * TC_vread_clist(CL, a, CONS));
	    else if (TC_strm(compact_mode, "cons"))
		TC_vwrite_clist(CL, a, WE, TC_vread_clist(CL, a, CONS));
	}
    }

    return CL;
}

static TC_ConstraintList *TC_fast_merge_constraint_list(const TC_ConstraintList *SL,
							TC_ConstraintList *ML)
{
    short int *entry = 0;
    int a;

    for (a = 0; a < SL->ne; a++) {
	entry = TC_extract_entry(entry, a, SL);
	TC_add_entry2list(entry, ML);
    }
    AR_free(entry);

    return ML;
}

static TC_ConstraintList *TC_slow_merge_constraint_list(const TC_ConstraintList *SL,
							TC_ConstraintList *ML)
{
    TC_Sequences *S1, *S2;
    short int *entry = 0;
    int **name_index;
    int **seq_index;
    int a, s1, s2;

    if (!ML)
	return (TC_ConstraintList *)SL;

    S1 = SL->S;
    S2 = ML->S;
    name_index = TC_index_seq_name(S1, S2);
    seq_index = TC_index_seq_res(S1, S2, name_index);

    for (a = 0; a < SL->ne; a++) {
	entry = TC_extract_entry(entry, a, SL);
	s1 = entry[SEQ1];
	s2 = entry[SEQ2];
	if (S1 == S2)
	    TC_add_entry2list(entry, ML);
	else if (name_index[s1][0] == -1 || name_index[s2][0] == -1 || !seq_index[s1]
		 || !seq_index[s2]);
	else {
	    int r1, r2;

	    r1 = seq_index[s1][entry[R1] - 1];
	    r2 = seq_index[s2][entry[R2] - 1];
	    entry[SEQ1] = name_index[s1][0];
	    entry[SEQ2] = name_index[s2][0];
	    if (r1 != -1 && r2 != -1 && entry[SEQ1] != entry[SEQ2]) {
		entry[R1] = r1 + 1;
		entry[R2] = r2 + 1;
		TC_add_entry2list(entry, ML);
	    }
	}
    }
    TC_free_int(name_index);
    TC_free_int(seq_index);
    AR_free(entry);

    return ML;
}

TC_ConstraintList *TC_merge_constraint_list(const TC_ConstraintList *SL, TC_ConstraintList *ML)
{
    if (!ML)
	return (TC_ConstraintList *)SL;
    else if (SL->S == ML->S)
	return TC_fast_merge_constraint_list(SL, ML);
    else
	return TC_slow_merge_constraint_list(SL, ML);
}

/**
 * @brief Indexes a constraint list.
 * 1 - Checks the flag seq_indexed
 * 2 - if flag set to 0:
 *     CL->start_index[seq1][seq2] indicatse the first position for seq1 Vs seq2
 *     CL->end_index[seq1][seq3]   indicatse the last  position for seq1 Vs seq2
 * Any modif to CL->L should cause the flag 1 to be set to 0.
 */
TC_ConstraintList *TC_index_constraint_list(TC_ConstraintList *CL)
{
    int a, csA, csB, sA, sB;

    if (CL->seq_indexed);
    else {
	if (CL->start_index != 0)
	    TC_free_int(CL->start_index);
	CL->start_index = TC_declare_int((CL->S)->nseq, (CL->S)->nseq);

	if (CL->end_index != 0)
	    TC_free_int(CL->end_index);
	CL->end_index = TC_declare_int((CL->S)->nseq, (CL->S)->nseq);

	csA = TC_vread_clist(CL, 0, SEQ1);
	csB = TC_vread_clist(CL, 0, SEQ2);

	CL->start_index[csA][csB] = 0;
	CL->start_index[csB][csA] = 0;
	for (a = 1; a < CL->ne; a++) {
	    sA = TC_vread_clist(CL, a, SEQ1);
	    sB = TC_vread_clist(CL, a, SEQ2);
	    if (sA != csA || sB != csB) {
		CL->end_index[csA][csB] = a;
		CL->end_index[csB][csA] = a;
		csA = sA;
		csB = sB;
		CL->start_index[csA][csB] = a;
		CL->start_index[csB][csA] = a;
	    }
	}
	CL->end_index[csB][csA] = CL->ne;
	CL->end_index[csA][csB] = CL->ne;
	CL->seq_indexed = 1;
    }
    return CL;
}

/**
 * @brief This function reorganises the content of the CL->L matrix,
 * so that a single look up gives the constraints associated with any residue.
 * 1 - if CL->residue_indexed == 1 return
 * 2 - CL->residue_index[Seq X][Res Y of Seq X][0] = Z
 *   Z = Number of residues matching X(Y)*3+1
 * CL->residue_index[Seq X][Res Y of Seq X][0] = Z
 * CL->residue_index[Seq X][Res Y of Seq X][c+0] = seq W
 * CL->residue_index[Seq X][Res Y of Seq X][c+1] = res V
 * CL->residue_index[Seq X][Res Y of Seq X][c+2] = weight W(V) Vs X(Y)
 *
 * NOTE: Works All right with a sequence against itself.
 * NOTE: Any modification of CL->L should result in residue_indexed to be reset.
 */
#if !USE_RMA
TC_ConstraintList *TC_index_res_constraint_list(TC_ConstraintList *CL, int field)
{
    int a, b, s1, s2, r1, r2, w;

    if (CL->residue_indexed && CL->residue_field == field);
    else {
	if (CL->residue_index) {
	    for (a = 0; a < (CL->S)->nseq; a++)
		for (b = 0; b <= (CL->S)->len[a]; b++) {
		    AR_free(CL->residue_index[a][b]);
		    CL->residue_index[a][b] = (int *)AR_allocate(1, sizeof(int));
		    CL->residue_index[a][b][0] = 1;
		}
	}
	else if (!CL->residue_index) {
	    CL->residue_index = (int ***)AR_allocate((CL->S)->nseq, sizeof(int **));
	    for (a = 0; a < (CL->S)->nseq; a++) {
		CL->residue_index[a] =
		    (int **)AR_allocate(((CL->S)->len[a] + 1), sizeof(int *));
		for (b = 0; b <= (CL->S)->len[a]; b++) {
		    CL->residue_index[a][b] = (int *)AR_allocate(1, sizeof(int));
		    CL->residue_index[a][b][0] = 1;
		}
	    }
	}
	for (a = 0; a < CL->ne; a++) {
	    s1 = TC_vread_clist(CL, a, SEQ1);
	    s2 = TC_vread_clist(CL, a, SEQ2);
	    r1 = TC_vread_clist(CL, a, R1);
	    r2 = TC_vread_clist(CL, a, R2);
	    w = TC_vread_clist(CL, a, field);

	    CL->residue_index[s1][r1][0] += 3;
	    CL->residue_index[s1][r1] =
		(int *)AR_reallocate(CL->residue_index[s1][r1],
				     CL->residue_index[s1][r1][0], sizeof(int));
	    CL->residue_index[s1][r1][CL->residue_index[s1][r1][0] - 3] = s2;
	    CL->residue_index[s1][r1][CL->residue_index[s1][r1][0] - 2] = r2;
	    CL->residue_index[s1][r1][CL->residue_index[s1][r1][0] - 1] = w;

	    CL->residue_index[s2][r2][0] += 3;
	    CL->residue_index[s2][r2] =
		(int *)AR_reallocate(CL->residue_index[s2][r2],
				     CL->residue_index[s2][r2][0], sizeof(int));
	    CL->residue_index[s2][r2][CL->residue_index[s2][r2][0] - 3] = s1;
	    CL->residue_index[s2][r2][CL->residue_index[s2][r2][0] - 2] = r1;
	    CL->residue_index[s2][r2][CL->residue_index[s2][r2][0] - 1] = w;
	}
	CL->residue_indexed = 1;
	CL->residue_field = field;
    }

    return CL;
}
#endif // !USE_RMA


int **TC_main_search_in_list_constraint(int *key, int *p, int k_len, TC_ConstraintList *CL)
{

    int **l = 0;
    int start, end;

    CL = TC_index_constraint_list(CL);

    start = CL->start_index[key[SEQ1]][key[SEQ2]];
    end = CL->end_index[key[SEQ1]][key[SEQ2]];

    if (CL->L) {
	l = (int **)TC_search_in_list_array(key, p, k_len, (void **)((CL->L) + start), end - start);
    }

    return l;
}

void AR_method_list2job_list(const std::vector<TC_Method *> &meth_list, const TC_Sequences *S,
			     const TC_DistanceMatrix *DM, const char *lib_list, std::vector<AR_lib_job> &job_list, std::vector<int> &job_cnt)
{
    std::vector<int *> lib_seqs;

    if (lib_list && *lib_list) {
	int buf_size = 0;
	const char *buf = 0;
	int nl, n;
	char **lines, **list;
	int i, j, k;
	int *tab = 0, x;

	buf = AR_read_file(lib_list, buf_size);
	lines = TC_string2list2(buf, "\n");
	AR_free(const_cast<char *>(buf));

	nl = atoi(lines[0]);
	for (i = 1; i < nl; ++i) {
	    if (TC_is_blank(lines[i]))
		continue;

	    list = TC_string2list(lines[i]);

	    if (TC_is_number(list[1]) && (n = atoi(list[1])) > 1) {
		tab = new int[n + 1];

		tab[0] = n;
		for (j = 2, k = 1; j < n + 2; ++j) {
		    if (TC_is_number(list[j]))
			x = atoi(list[j]);
		    else
			x = TC_name_is_in_list(list[j], S->name, S->nseq, 100);

		    if (-1 < x && x < S->nseq)
			tab[k++] = x;
		}

		if (k == n + 1)
		    lib_seqs.push_back(tab);
		else
		    delete[] tab;
	    }
	    TC_free_char(list);
	}
	TC_free_char(lines);
    }

    std::vector<TC_Method *>::const_iterator it = meth_list.begin();
    const char *aln_mode, *out_mode;
    AR_lib_job job;
    int nj;
    char buf_seq[64];

    for (; it != meth_list.end(); ++it) {
	++job.method;
	nj = job_list.size();

	aln_mode = (*it)->aln_mode;
	out_mode = (*it)->out_mode;

	if (strcmp(aln_mode, "multiple") == 0) {
	    std::cerr << '\n' << PROGRAM << ": WARNING: Method '" << (*it)->executable << "' is a multiple method.\n"
		      << "\tMultiple methods are not supported in current version of " << PROGRAM << ".\n";
	}
	else if (!lib_seqs.empty()) {
	    std::vector<int *>::const_iterator it2 = lib_seqs.begin();
	    const int *seq;

	    for (; it2 != lib_seqs.end(); ++it2) {
		seq = *it2;
		if (seq[0] == 2) {
		    sprintf(buf_seq, "2 %d %d", seq[1], seq[2]);

		    if (TC_check_seq_type(*it, buf_seq, S)) {
			job.seq[0] = seq[1];
			job.seq[1] = seq[2];
			job_list.push_back(job);
		    }
		}
	    }
	}
	else if (strstr(aln_mode, "pairwise")) {
	    bool do_mirror, do_self;
	    int i, j;

	    do_mirror = (strstr(aln_mode, "m_") != 0);
	    do_self = (strstr(aln_mode, "s_") != 0);

	    for (i = 0; i < S->nseq; ++i)
		for (j = (!do_mirror) * i; j < S->nseq; ++j) {
		    if ((i != j || do_self)
			&& TC_is_in_range(DM->similarity_matrix[i][j], (*it)->minid, (*it)->maxid)) {
			sprintf(buf_seq, "2 %d %d", i, j);

			if (TC_check_seq_type(*it, buf_seq, S)) {
			    job.seq[0] = i;
			    job.seq[1] = j;
			    job_list.push_back(job);
			}
		    }
		}
	}

	nj = job_list.size() - nj;
	job_cnt[it - meth_list.begin()] = nj;

	if (nj == 0)
	    std::cout << "\tNo suitable sequences (type: " << (*it)->seq_type << ") for method '"
		      << (*it)->executable << "'.\n";
    } // for (it)

    std::vector<int *>::iterator it2 = lib_seqs.begin();
    for (; it2 != lib_seqs.end(); ++it2)
	delete[] (*it2);
}

void AR_send_jobs(const int recv, const MPI_Env &mpi_env, unsigned int chunk_size, unsigned int &curr_job,
		  const std::vector<AR_lib_job> &job_list, int *ibuf)
{
    if (curr_job + chunk_size > job_list.size())
	chunk_size = job_list.size() - curr_job;

    const unsigned int N = curr_job + chunk_size;
    int i = 0;

    ibuf[i++] = chunk_size;
    for (; curr_job < N; ++curr_job) {
	ibuf[i++] = job_list[curr_job].method;
	ibuf[i++] = job_list[curr_job].seq[0];
	ibuf[i++] = job_list[curr_job].seq[1];
    }

    MPI_Send(ibuf, i, MPI_INT, recv, LIB_JOB_TAG, mpi_env.comm());
#if DEBUG_MPI
    std::cout << "\nMASTER NODE SENT LIB JOBS (" << chunk_size << ") TO NODE " << recv << '\n';
#endif
}

static TC_ConstraintList *TC_method2pw_cl(const TC_Method *M, const TC_ConstraintList *CL)
{
    TC_ConstraintList *PW_CL = 0;
    TC_Sequences *S;
    const char *mode;
    const char *mat = 0;
    const char *group_mat = 0;

    mode = M->executable;
    PW_CL = TC_duplicate_constraint_list_soft(CL);
    PW_CL->pw_parameters_set = 1;
    S = PW_CL->S;

    if (TC_strm(S->type, "PROTEIN")) {
	mat = "blosum62mt";
	group_mat = "vasiliky";
	PW_CL->ktup = 2;
    }
    else if (TC_strm(S->type, "DNA")) {
	mat = "idmat";
	group_mat = "idmat";
	PW_CL->ktup = 5;
    }

    if (M->matrix[0])
	mat = M->matrix;

    PW_CL->M = TC_read_matrice(mat);

    if (M->gop != TC_UNDEFINED) {
	PW_CL->gop = M->gop;
    }
    else {
	PW_CL->gop = (int)(TC_get_avg_matrix_mm(PW_CL->M, AA_ALPHABET) * 10);
    }
    if (M->gep != TC_UNDEFINED)
	PW_CL->gep = M->gep;
    else
	PW_CL->gep = -1;

    if (TC_strm(mode, "slow_pair")) {
	PW_CL->L = 0;
	PW_CL->maximise = 1;
	PW_CL->TG_MODE = 1;
	PW_CL->use_fragments = 0;
	PW_CL->extend_jit = 0;
	strncpy(PW_CL->matrix_for_aa_group, group_mat, 100 - 1);

	strncpy(PW_CL->dp_mode, "myers_miller_pair_wise", 100 - 1);
	PW_CL->evaluate_residue_pair = TC_evaluate_matrix_score;
	PW_CL->get_dp_cost = TC_slow_get_dp_cost;
    }
    else if (TC_strm(mode, "lalign_id_pair")) {
	PW_CL->L = 0;
	PW_CL->maximise = 1;
	PW_CL->TG_MODE = 1;
	PW_CL->use_fragments = 0;
	PW_CL->extend_jit = 0;
	strncpy(PW_CL->matrix_for_aa_group, group_mat, 100 - 1);

	PW_CL->lalign_n_top = CL->lalign_n_top;
	PW_CL->pair_wise = TC_sim_pair_wise_lalign;
	PW_CL->evaluate_residue_pair = TC_evaluate_matrix_score;
	PW_CL->get_dp_cost = TC_slow_get_dp_cost;
    }
    else {
	TC_free_constraint_list(PW_CL);
	PW_CL = 0;
    }

    return PW_CL;
}

static void AR_produce_list(const std::vector<AR_LibJob *> &job_list)
{
    std::vector<AR_LibJob *>::const_iterator it = job_list.begin();
    AR_LibJob *job;
    TC_Method *M;

    for (; it != job_list.end(); ++it) {
	job = *it;
	M = job->param.TCM;

	if (!M->PW_CL)
	    M->PW_CL = TC_method2pw_cl(M, job->param.BASE_CL);
	job->io.CL = TC_duplicate_constraint_list4lib_computation(job->param.BASE_CL);

	job->submit();
	job->retrieve();
	
	job->param.BASE_CL = TC_merge_constraint_list(job->io.CL, job->param.BASE_CL);
	TC_free_constraint_list4lib_computation(job->io.CL);
    }
}

void AR_process_jobs(const unsigned int n_jobs, const int *job_desc, const std::vector<TC_Method *> &methods,
		     const std::vector<TC_ConstraintList *> &libs)
{
    std::vector<AR_LibJob *> job_list(n_jobs, 0);
    AR_LibJob *job = 0;
    int m, s1, s2;

    for (unsigned int i = 0, j = 0; i < n_jobs; ++i) {
	m = job_desc[j++];
	s1 = job_desc[j++];
	s2 = job_desc[j++];

	job = new AR_LibJob;
	job->param.TCM = methods[m];
	job->param.seq_c[0] = s1;
	job->param.seq_c[1] = s2;
	job->param.BASE_CL = libs[m];

	job_list[i] = job;
    }

    AR_produce_list(job_list);

    for (unsigned int i = 0; i < n_jobs; ++i) {
	delete job_list[i];
    }
}

TC_ConstraintList *AR_par_compile_constraint_list(TC_ConstraintList *CL, const std::vector<std::string> &in_files,
						  const char *weight, const MPI_Env &mpi_env)
{
    if (!CL)
	return CL;

    enum {
	JOB_REQ, JOB_RES
    };

    std::vector<TC_Method *> methods;
    TC_Method *meth = 0;

    const unsigned int meth_buf_size = 10 * (sizeof(int) + 100) + 4 * sizeof(int) + 2 * (sizeof(int) + 1000);
    char *cbuffer = 0;
    unsigned int job_buf_size = 0;
    int *ibuffer = 0;
    unsigned int job_res[3];
    MPI_Status status;

    std::vector<TC_ConstraintList *> libs;

    if (mpi_env.am_I_root()) {
	bool no_jobs = true;
	std::vector<std::string>::const_reverse_iterator rit = in_files.rbegin(), rend = in_files.rend();

	for (; rit != rend; ++rit)
	    if ((*rit)[0] == 'X') {
		CL = TC_read_constraint_list(CL, rit->c_str() + 1, "matrix", 0);
		if (CL->M)
		    break;
	    }

	if (!CL->M) {
	    std::vector<const std::string *> files;
	    char id;
	    const char *m_name = 0, *m_file = 0;

	    std::cout << "\nCOMPILING PRIMARY LIBRARY.....\n";

	    methods.reserve(in_files.size());
	    files.reserve(in_files.size());

	    for (rit = in_files.rbegin(); rit != rend; ++rit) {
		id = (*rit)[0];

		if (id == 'M') {
		    m_name = rit->c_str();
		    m_file = TC_method_name2method_file(m_name + 1);

		    if (!m_file) {
			if (AR_check_file_exists(m_name))
			    m_file = m_name;
			else if (AR_check_file_exists(m_name + 1))
			    m_file = m_name + 1;
			else {
			    std::cerr << '\n' << PROGRAM << ": WARNING: '" << m_name << "' is not a valid method.\n";
			    continue;
			}
		    }

		    meth = TC_method_file2tc_method(m_file);

		    if (!meth->executable[0]) {
			std::cerr << '\n' << PROGRAM << ": WARNING: '" << m_name << "' is not a valid method.\n";
			AR_free(meth);
			continue;
		    }

		    if (!TC_strm(weight, "default"))
			strncpy(meth->weight, weight, 100 - 1);

		    methods.push_back(meth);
		}
		else if (id == 'A' || id == 'L') {
		    files.push_back(&(*rit));
		}
	    } // for (rit)

	    std::vector<const std::string *>::const_iterator it_f = files.begin();

	    if (!methods.empty()) {
		std::vector<AR_lib_job> job_list;
		std::vector<int> job_cnt(methods.size());

		AR_method_list2job_list(methods, CL->S, CL->DM, CL->lib_list, job_list, job_cnt);
		std::sort(job_list.begin(), job_list.end(), AR_job_cmp(job_cnt));

		if (!job_list.empty()) {
		    no_jobs = false;

		    const unsigned int N_m = methods.size();
		    const unsigned int N_j = job_list.size();
		    const unsigned int w_size = mpi_env.size() - 1; // without master
		    unsigned int chunk_size = static_cast<unsigned int>(ceil(N_j / (2.0 * w_size)));

		    job_buf_size = chunk_size * 3 + 1;

		    { // temporary block
			unsigned int tmp_buf[2];

			tmp_buf[0] = N_m;
			tmp_buf[1] = job_buf_size;

			MPI_Bcast(tmp_buf, 2, MPI_UNSIGNED, mpi_env.get_root(), mpi_env.comm());

			cbuffer = new char[meth_buf_size];
			int pack_pos = 0;
			int len;

			for (unsigned int i = 0; i < N_m; ++i) {
			    meth = methods[i];

			    len = strlen(meth->executable) + 1;
			    MPI_Pack(&len, 1, MPI_INT, cbuffer, meth_buf_size, &pack_pos, mpi_env.comm());
			    MPI_Pack(meth->executable, len, MPI_CHAR, cbuffer, meth_buf_size, &pack_pos, mpi_env.comm());

			    len = strlen(meth->executable2) + 1;
			    MPI_Pack(&len, 1, MPI_INT, cbuffer, meth_buf_size, &pack_pos, mpi_env.comm());
			    MPI_Pack(meth->executable2, len, MPI_CHAR, cbuffer, meth_buf_size, &pack_pos, mpi_env.comm());

			    len = strlen(meth->in_flag) + 1;
			    MPI_Pack(&len, 1, MPI_INT, cbuffer, meth_buf_size, &pack_pos, mpi_env.comm());
			    MPI_Pack(meth->in_flag, len, MPI_CHAR, cbuffer, meth_buf_size, &pack_pos, mpi_env.comm());

			    len = strlen(meth->in_flag2) + 1;
			    MPI_Pack(&len, 1, MPI_INT, cbuffer, meth_buf_size, &pack_pos, mpi_env.comm());
			    MPI_Pack(meth->in_flag2, len, MPI_CHAR, cbuffer, meth_buf_size, &pack_pos, mpi_env.comm());

			    len = strlen(meth->out_flag) + 1;
			    MPI_Pack(&len, 1, MPI_INT, cbuffer, meth_buf_size, &pack_pos, mpi_env.comm());
			    MPI_Pack(meth->out_flag, len, MPI_CHAR, cbuffer, meth_buf_size, &pack_pos, mpi_env.comm());

			    len = strlen(meth->aln_mode) + 1;
			    MPI_Pack(&len, 1, MPI_INT, cbuffer, meth_buf_size, &pack_pos, mpi_env.comm());
			    MPI_Pack(meth->aln_mode, len, MPI_CHAR, cbuffer, meth_buf_size, &pack_pos, mpi_env.comm());

			    len = strlen(meth->out_mode) + 1;
			    MPI_Pack(&len, 1, MPI_INT, cbuffer, meth_buf_size, &pack_pos, mpi_env.comm());
			    MPI_Pack(meth->out_mode, len, MPI_CHAR, cbuffer, meth_buf_size, &pack_pos, mpi_env.comm());

			    len = strlen(meth->seq_type) + 1;
			    MPI_Pack(&len, 1, MPI_INT, cbuffer, meth_buf_size, &pack_pos, mpi_env.comm());
			    MPI_Pack(meth->seq_type, len, MPI_CHAR, cbuffer, meth_buf_size, &pack_pos, mpi_env.comm());

			    len = strlen(meth->weight) + 1;
			    MPI_Pack(&len, 1, MPI_INT, cbuffer, meth_buf_size, &pack_pos, mpi_env.comm());
			    MPI_Pack(meth->weight, len, MPI_CHAR, cbuffer, meth_buf_size, &pack_pos, mpi_env.comm());

			    len = strlen(meth->matrix) + 1;
			    MPI_Pack(&len, 1, MPI_INT, cbuffer, meth_buf_size, &pack_pos, mpi_env.comm());
			    MPI_Pack(meth->matrix, len, MPI_CHAR, cbuffer, meth_buf_size, &pack_pos, mpi_env.comm());

			    MPI_Pack(&(meth->gop), 1, MPI_INT, cbuffer, meth_buf_size, &pack_pos, mpi_env.comm());
			    MPI_Pack(&(meth->gep), 1, MPI_INT, cbuffer, meth_buf_size, &pack_pos, mpi_env.comm());

			    MPI_Pack(&(meth->minid), 1, MPI_INT, cbuffer, meth_buf_size, &pack_pos, mpi_env.comm());
			    MPI_Pack(&(meth->maxid), 1, MPI_INT, cbuffer, meth_buf_size, &pack_pos, mpi_env.comm());

			    len = strlen(meth->param) + 1;
			    MPI_Pack(&len, 1, MPI_INT, cbuffer, meth_buf_size, &pack_pos, mpi_env.comm());
			    MPI_Pack(meth->param, len, MPI_CHAR, cbuffer, meth_buf_size, &pack_pos, mpi_env.comm());

			    len = strlen(meth->param1) + 1;
			    MPI_Pack(&len, 1, MPI_INT, cbuffer, meth_buf_size, &pack_pos, mpi_env.comm());
			    MPI_Pack(meth->param1, len, MPI_CHAR, cbuffer, meth_buf_size, &pack_pos, mpi_env.comm());

#if DEBUG_MPI
			    std::cout << "\nBCASTING DESC. OF '" << meth->executable << "' METHOD\n";
#endif
			    MPI_Bcast(cbuffer, pack_pos, MPI_PACKED, mpi_env.get_root(), mpi_env.comm());
			    pack_pos = 0;
			} // for (i)
		    } // temporary block

		    delete[] cbuffer;

		    ibuffer = new int[job_buf_size];

		    int to_send = N_j, to_recv = 0;
		    unsigned int curr_job = 0;

		    // send initial jobs (last in map is always root)
		    unsigned int i;

		    for (i = 0; i < w_size; ++i) {
			AR_send_jobs(mpi_env.map(i), mpi_env, chunk_size, curr_job, job_list, ibuffer);
			to_send -= chunk_size;
			++to_recv;

			if (to_send < 1)
			    break;
		    }

		    chunk_size = static_cast<unsigned int>(to_send / (2.0 * w_size));
		    if (chunk_size == 0)
			++chunk_size;

// 					libs.resize(N_m, 0);
// 					std::vector<TC_ConstraintList *>::iterator it = libs.begin();

// 					for (; it != libs.end(); ++it)
// 						*it = TC_declare_constraint_list(CL->S, 0, 0, 0);

		    // continue scheduling
		    do {
			if (it_f != files.end()) {
			    CL = TC_read_constraint_list(CL, (*it_f)->c_str(), 0, weight);
			    ++it_f;
			}

			MPI_Recv(job_res, 3, MPI_UNSIGNED, MPI_ANY_SOURCE, LIB_JOB_TAG + 1, mpi_env.comm(), &status);
			--to_recv;

			// request
			if (job_res[0] == JOB_REQ) {
			    if (to_send > 0) {
				AR_send_jobs(job_res[1], mpi_env, chunk_size, curr_job, job_list, ibuffer);
				to_send -= chunk_size;
				++to_recv;

				chunk_size = static_cast<unsigned int>(to_send / w_size);
				if (chunk_size == 0)
				    ++chunk_size;
			    }
			    else {
				curr_job = 0;
				AR_send_jobs(job_res[1], mpi_env, 0, curr_job, job_list, ibuffer);
				to_recv += N_m;
			    }
			}
			// result
			else {
			    if (job_res[2] > 0) {
				const unsigned int size = job_res[2] * TC_ConstraintList::entry_len;
				short int *ibuf = new short int[size];

				MPI_Recv(ibuf, size, MPI_SHORT, status.MPI_SOURCE, LIB_TAG, mpi_env.comm(), &status);

				const short int *entry = ibuf;
				// TC_ConstraintList *cl = libs[job_res[1]];

				for (unsigned int i = 0; i < job_res[2]; ++i) {
				    CL = TC_add_entry2list(entry, CL);
				    entry += TC_ConstraintList::entry_len;
				}

				delete[] ibuf;

				// Modified by Jaroslaw Zola
				// CL = TC_merge_constraint_list(cl, CL);
				// TC_free_constraint_list(cl);
				// libs[job_res[1]] = TC_declare_constraint_list(CL->S, 0, 0, 0);
			    }
			}
		    }
		    while (to_send > 0 || to_recv > 0);

		    ++i;
		    to_recv = (w_size - i) * N_m;
		    curr_job = 0;

		    for (; i < w_size; ++i)
			AR_send_jobs(mpi_env.map(i), mpi_env, 0, curr_job, job_list, ibuffer);

		    while (to_recv-- > 0)
			MPI_Recv(job_res, 3, MPI_UNSIGNED, MPI_ANY_SOURCE, LIB_JOB_TAG + 1, mpi_env.comm(), &status);

		    delete[] ibuffer;

/*                                       Modified by Jaroslaw Zola
					 Libs are compacted on worker side and merged when received
					 // compact precompiled libs
					 for (it = libs.begin(); it != libs.end(); ++it) {
					 TC_compact_list(*it, 0, (*it)->ne, "best");
					 std::cout << "\tMethod '" << methods[it - libs.begin()]->executable
					 << "'  =>  " << (*it)->ne << " entries\n";
					 }

					 // merge all libs into primary lib
					 for (it = libs.begin(); it != libs.end(); ++it) {
					 CL = TC_merge_constraint_list(*it, CL);
					 TC_free_constraint_list(*it);
					 }
*/
		}
	    } // if (!methods.empty())

	    for (; it_f != files.end(); ++it_f) {
		CL = TC_read_constraint_list(CL, (*it_f)->c_str(), 0, weight);
	    }

	    std::cout << "\tSize Before Compact: " << CL->ne;

	    // compact primary lib
	    TC_compact_list(CL, 0, CL->ne, "default");

	    std::cout << "\n\tTotal Size:          " << CL->ne << "\n.....[Done]\n";
	} // if (!CL->M)

	// DO WE NEED THIS!!!
	// TC_evaluate_constraint_list_reference(CL);

	if (no_jobs) {
	    unsigned int msg[2] = { 0, 0 };

	    MPI_Bcast(msg, 2, MPI_UNSIGNED, mpi_env.get_root(), mpi_env.comm());
	}
    }
    else { // WORKER side
	{ // temporary block
	    unsigned int tmp_buf[2];

	    MPI_Bcast(tmp_buf, 2, MPI_UNSIGNED, mpi_env.get_root(), mpi_env.comm());

	    methods.resize(tmp_buf[0]);
	    job_buf_size = tmp_buf[1];

	    if (methods.size() > 0)
		cbuffer = new char[meth_buf_size];
	    int pack_pos = 0;
	    int len;

	    for (unsigned int i = 0; i < methods.size(); ++i) {
		MPI_Bcast(cbuffer, meth_buf_size, MPI_PACKED, mpi_env.get_root(), mpi_env.comm());

		meth = static_cast<TC_Method *>(AR_allocate(1, sizeof(TC_Method)));

		MPI_Unpack(cbuffer, meth_buf_size, &pack_pos, &len, 1, MPI_INT, mpi_env.comm());
		MPI_Unpack(cbuffer, meth_buf_size, &pack_pos, meth->executable, len, MPI_CHAR, mpi_env.comm());

		MPI_Unpack(cbuffer, meth_buf_size, &pack_pos, &len, 1, MPI_INT, mpi_env.comm());
		MPI_Unpack(cbuffer, meth_buf_size, &pack_pos, meth->executable2, len, MPI_CHAR, mpi_env.comm());

		MPI_Unpack(cbuffer, meth_buf_size, &pack_pos, &len, 1, MPI_INT, mpi_env.comm());
		MPI_Unpack(cbuffer, meth_buf_size, &pack_pos, meth->in_flag, len, MPI_CHAR, mpi_env.comm());

		MPI_Unpack(cbuffer, meth_buf_size, &pack_pos, &len, 1, MPI_INT, mpi_env.comm());
		MPI_Unpack(cbuffer, meth_buf_size, &pack_pos, meth->in_flag2, len, MPI_CHAR, mpi_env.comm());

		MPI_Unpack(cbuffer, meth_buf_size, &pack_pos, &len, 1, MPI_INT, mpi_env.comm());
		MPI_Unpack(cbuffer, meth_buf_size, &pack_pos, meth->out_flag, len, MPI_CHAR, mpi_env.comm());

		MPI_Unpack(cbuffer, meth_buf_size, &pack_pos, &len, 1, MPI_INT, mpi_env.comm());
		MPI_Unpack(cbuffer, meth_buf_size, &pack_pos, meth->aln_mode, len, MPI_CHAR, mpi_env.comm());

		MPI_Unpack(cbuffer, meth_buf_size, &pack_pos, &len, 1, MPI_INT, mpi_env.comm());
		MPI_Unpack(cbuffer, meth_buf_size, &pack_pos, meth->out_mode, len, MPI_CHAR, mpi_env.comm());

		MPI_Unpack(cbuffer, meth_buf_size, &pack_pos, &len, 1, MPI_INT, mpi_env.comm());
		MPI_Unpack(cbuffer, meth_buf_size, &pack_pos, meth->seq_type, len, MPI_CHAR, mpi_env.comm());

		MPI_Unpack(cbuffer, meth_buf_size, &pack_pos, &len, 1, MPI_INT, mpi_env.comm());
		MPI_Unpack(cbuffer, meth_buf_size, &pack_pos, meth->weight, len, MPI_CHAR, mpi_env.comm());

		MPI_Unpack(cbuffer, meth_buf_size, &pack_pos, &len, 1, MPI_INT, mpi_env.comm());
		MPI_Unpack(cbuffer, meth_buf_size, &pack_pos, meth->matrix, len, MPI_CHAR, mpi_env.comm());

		MPI_Unpack(cbuffer, meth_buf_size, &pack_pos, &(meth->gop), 1, MPI_INT, mpi_env.comm());
		MPI_Unpack(cbuffer, meth_buf_size, &pack_pos, &(meth->gep), 1, MPI_INT, mpi_env.comm());

		MPI_Unpack(cbuffer, meth_buf_size, &pack_pos, &(meth->minid), 1, MPI_INT, mpi_env.comm());
		MPI_Unpack(cbuffer, meth_buf_size, &pack_pos, &(meth->maxid), 1, MPI_INT, mpi_env.comm());

		MPI_Unpack(cbuffer, meth_buf_size, &pack_pos, &len, 1, MPI_INT, mpi_env.comm());
		MPI_Unpack(cbuffer, meth_buf_size, &pack_pos, meth->param, len, MPI_CHAR, mpi_env.comm());

		MPI_Unpack(cbuffer, meth_buf_size, &pack_pos, &len, 1, MPI_INT, mpi_env.comm());
		MPI_Unpack(cbuffer, meth_buf_size, &pack_pos, meth->param1, len, MPI_CHAR, mpi_env.comm());

		methods[i] = meth;
		pack_pos = 0;
	    } // for (i)
	} // end of temporary block

	delete[] cbuffer;

	if (!methods.empty()) { // if methods
	    libs.resize(methods.size(), 0);
	    std::vector<TC_ConstraintList *>::iterator it = libs.begin();

	    for (; it != libs.end(); ++it)
		*it = TC_duplicate_constraint_list_soft(CL);

	    ibuffer = new int[job_buf_size];

	    do {
		// get jobs
		MPI_Recv(ibuffer, job_buf_size, MPI_INT, mpi_env.get_root(), LIB_JOB_TAG, mpi_env.comm(), &status);
#if DEBUG_MPI
		std::cout << "\nNODE " << mpi_env.rank() << " RECIVED LIB JOBS FROM MASTER\n";
#endif

		// and process them
		unsigned int to_proc;

		to_proc = ibuffer[0];
		if (to_proc == 0)
		    break;

		AR_process_jobs(to_proc, ibuffer + 1, methods, libs);

		// send request for next jobs
		job_res[0] = JOB_REQ;
		job_res[1] = mpi_env.rank();
		job_res[2] = 0;

		MPI_Send(job_res, 3, MPI_UNSIGNED, mpi_env.get_root(), LIB_JOB_TAG + 1, mpi_env.comm());
#if DEBUG_MPI
		std::cout << "\nNODE " << mpi_env.rank() << " SENT REQUEST FOR LIB JOBS TO MASTER\n";
#endif
	    }
	    while (true);

	    delete[] ibuffer;

	    // compact libs
	    for (it = libs.begin(); it != libs.end(); ++it)
		TC_compact_list(*it, 0, (*it)->ne, "best");

	    // send libs
	    job_res[0] = JOB_RES;

	    int tot_lib_size = 0;

	    for (it = libs.begin(); it != libs.end(); ++it) {
		job_res[1] = it - libs.begin();
		job_res[2] = (*it)->ne;

		tot_lib_size += (*it)->ne;

		MPI_Send(job_res, 3, MPI_UNSIGNED, mpi_env.get_root(), LIB_JOB_TAG + 1, mpi_env.comm());

		if (job_res[2] > 0) {
		    // const short int **L = const_cast<const short int **>((*it)->L);
		    // const short int **L = const_cast<const short int **>((*it)->L.tab);
		    unsigned int size = job_res[2] * TC_ConstraintList::entry_len;

		    short int *sibuffer = new short int[size];
		    short int *pos = sibuffer;

		    for (unsigned int i = 0; i < job_res[2]; ++i)
			pos = std::copy((*it)->L[i], (*it)->L[i] + TC_ConstraintList::entry_len, pos);

		    MPI_Send(sibuffer, size, MPI_SHORT, mpi_env.get_root(), LIB_TAG, mpi_env.comm());
#if DEBUG_MPI
		    std::cout << "\nNODE " << mpi_env.rank() << " SENT PRECOMPILED LIB TO MASTER\n";
#endif

		    delete[] sibuffer;
		}

		(*it)->L.free();
		TC_free_constraint_list(*it);
	    }
	} // if methods
	
    } // worker

    std::vector<TC_Method *>::iterator it = methods.begin();

    for (; it != methods.end(); ++it) {
//	TC_free_constraint_list((*it)->PW_CL);
	delete (*it)->PW_CL;
	AR_free(*it);
    }

    return CL;
}

#if DEBUG_TREE
void print_CL(TC_ConstraintList* CL) {
    std::cout << "CL->ne: " << CL->ne << std::endl;
    std::cout << "CL->normalise: " << CL->normalise << std::endl;
    std::cout << "CL->max_ext_value: " << CL->max_ext_value << std::endl;
    std::cout << "CL->max_value: " << CL-> max_value << std::endl;
    std::cout << "CL->overweight: " << CL->overweight << std::endl;
    std::cout << "CL->pw_parameters_set: " 
	      << CL->pw_parameters_set << std::endl;
    std::cout << "CL->gop: " << CL->gop << std::endl;
    std::cout << "CL->gep: " << CL->gep << std::endl;
    std::cout << "CL->f_gop: " << CL->f_gop << std::endl;
    std::cout << "CL->f_gep: " << CL->f_gop << std::endl;
    std::cout << "CL->nm_gop: " << CL->nm_gop << std::endl;
    std::cout << "CL->nm_gep: " << CL->nm_gop << std::endl;
    std::cout << "CL->nomatch: " << CL->nomatch << std::endl;
    std::cout << "CL->TG_MODE: " << CL->TG_MODE << std::endl;
    std::cout << "CL->F_TG_MODE: " << CL->TG_MODE << std::endl;
    std::cout << "CL->dp_mode: " << CL->dp_mode << std::endl;
    std::cout << "CL->maximise: " << CL->maximise << std::endl;
    std::cout << "CL->diagonal_threshold: " 
	      << CL->diagonal_threshold << std::endl;
    std::cout << "C->ktup: " << CL->ktup << std::endl;
    std::cout << "C->use_fragments: " << CL->use_fragments << std::endl;
    std::cout << "CL->fasta_step: " << CL->fasta_step << std::endl;
    std::cout << "CL->lalign_n_top: " << CL->lalign_n_top << std::endl;
    std::cout << "CL->sw_min_dist: " << CL->sw_min_dist << std::endl;
    std::cout << "CL->n_matrices: " << CL->n_matrices << std::endl;
    std::cout << "CL->tree_mode: " << CL->tree_mode << std::endl;
    std::cout << "CL->W->SEQ_W: " << std::endl;
    if (CL->W != 0) {
	for (int i = 0; i < CL->W->nseq; ++i) {
	    std::cout << " " << CL->W->SEQ_W[i] << std::endl;
	}
    }
} // print_CL

#endif
