/*
 * File: TC_DistanceMatrix.cpp
 * Created: 06.06.2006
 * Last Modified: 11.09.2006
 * Author: Adrian Rospondek
 */


#include <iostream>
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <MPI_ibuffer.hpp>
#include <MPI_obuffer.hpp>
#include "AR_node_time.hpp"
#include "TC_Alignment.hpp"
#include "TC_ConstraintList.hpp"
#include "TC_DistanceMatrix.hpp"
#include "TC_Sequences.hpp"
#include "config.hpp"
#include "dp_drivers.hpp"
#include "evaluate.hpp"
#include "globals.hpp"
#include "input.hpp"
#include "util.hpp"
#include "util_math.hpp"
#include "util_mem.hpp"
#include "util_mpi.hpp"
#include "util_seq.hpp"



TC_DistanceMatrix *AR_declare_distance_matrix(const int size)
{
	TC_DistanceMatrix *DM;

	DM = (TC_DistanceMatrix *)AR_allocate(1, sizeof(TC_DistanceMatrix));
	DM->similarity_matrix = TC_declare_int(size, size);
	DM->distance_matrix = TC_declare_int(size, size);
	DM->score_similarity_matrix = TC_declare_int(size, size);

	return DM;
}

TC_DistanceMatrix *TC_duplicate_distance_matrix(const TC_DistanceMatrix *inDM)
{
	TC_DistanceMatrix *DM;

	if (!inDM)
		return 0;

	DM = (TC_DistanceMatrix *)AR_allocate(1, sizeof(TC_DistanceMatrix));
	DM->similarity_matrix = TC_duplicate_int(inDM->similarity_matrix, -1, -1);
	DM->distance_matrix = TC_duplicate_int(inDM->distance_matrix, -1, -1);
	DM->score_similarity_matrix = TC_duplicate_int(inDM->score_similarity_matrix, -1, -1);

	return DM;
}

void TC_free_distance_matrix(TC_DistanceMatrix *DM)
{
	if (!DM)
		return;

	TC_free_int(DM->similarity_matrix);
	TC_free_int(DM->distance_matrix);
	TC_free_int(DM->score_similarity_matrix);
	AR_free(DM);
}

static inline void AR_send_job(int recv, const MPI_Env &mpi_env, int chunk_size, int row, int col)
{
	int params[3];

	params[0] = chunk_size;
	params[1] = row;
	params[2] = col;

	MPI_Send(params, 3, MPI_INT, recv, CHUNK_TAG + recv, mpi_env.comm());
#if DEBUG_MPI
	std::cout << "\nMASTER NODE SENT JOB (dist. matrix part from cell [" << row << ", " << col
			  << "], size " << chunk_size << ") TO NODE " << recv << '\n';
#endif
}

static void AR_next_chunk(const int chunk_size, const int n, int &r, int &c)
{
	c += chunk_size;
	while (c >= n && r < n) {
		++r;
		c = r + (c - n) + 1;
	}
}

static double AR_calculate(const int size, int row, int col, TC_Alignment *A, TC_ConstraintList *CL,
	const char *mode, empi::MPI_ibuffer &ibuff)
{
	const int nseq = (CL->S)->nseq;
	int ns[2];
	int **l_s;
	int id, score;
	double t0, t;
	double bpps = 0.0; // bp per second

	l_s = TC_declare_int(2, 1);
	ns[0] = ns[1] = 1;

	for (int i = 0; i < size; ++i) {
		l_s[0][0] = row;
		l_s[1][0] = col;

		if (!TC_strm(mode, "aln")) {
			TC_ungap(A->seq_al[row]);
			TC_ungap(A->seq_al[col]);
		}

		t0 = AR_get_time();

		if (TC_strm(mode, "slow")) {
			A->score_aln = TC_pair_wise(A, ns, l_s, CL);
			id = TC_get_seq_sim(A->seq_al[row], A->seq_al[col], "-", 0);

			if (CL->L) {
				score = static_cast<int>(static_cast<float>(A->score_aln) / (A->len_aln * SCORE_K));
				if (CL->normalise)
					score = static_cast<int>(static_cast<float>(score * MAXID) / CL->normalise);
			}
			else if (CL->M)
				score = id;

			if (score > MAXID)
				score = (CL->L) ? TC_sub_aln2sub_aln_score(A, CL, ns, l_s) : id;
		}
		else if (TC_strm2(mode, "fast", "very_fast")) {
			A->score_aln = TC_pair_wise(A, ns, l_s, CL);
			id = TC_get_seq_sim(A->seq_al[row], A->seq_al[col], "-", 0);
			score = id;
		}
		else {
			A->score_aln = TC_pair_wise(A, ns, l_s, CL);
			id = A->score_aln;
			score = id * SCORE_K;
		}

		t = AR_get_time() - t0;
		bpps += (((CL->S)->len[row] * (CL->S)->len[col]) / t);

		ibuff << id, score;

		if (++col == nseq) {
			++row;
			col = row + 1;
		}
	}

	TC_free_int(l_s);

	return bpps / size;
}

static TC_DistanceMatrix *AR_par_seq2distance_matrix(const TC_ConstraintList *CL, const char *mode,
	const MPI_Env &mpi_env, std::vector<AR_node_time> &times)
{
	TC_DistanceMatrix *DM = CL->DM;

	const unsigned int nseq = (CL->S)->nseq;
	const unsigned int N = ((nseq * nseq) - nseq) / 2;
	const unsigned int w_size = mpi_env.size() - 1; // without master
	const unsigned int init_chunk = static_cast<unsigned int>(ceil(N / (2.0 * w_size)));

	const unsigned int buf_size = (init_chunk * 2 + 3) * sizeof(int) + sizeof(double);
	char buffer[buf_size];
	MPI_Status status;

	if (mpi_env.am_I_root()) {
		if (!DM)
			DM = AR_declare_distance_matrix(nseq);

		std::cout << "\nCOMPUTING PAIRWISE SIMILARITY (distance matrix mode: " << mode << ").....\n";

		unsigned int chunk_size = init_chunk;
		unsigned int to_send = N, to_recv = N;
		int row = 0, col = 1;
		std::vector<int> times_n(mpi_env.size(), 0);

		times.resize(mpi_env.size());
		for (unsigned int i = 0; i < times.size(); ++i) {
			times[i].node = i;
			times[i].time = 0.0;
		}

		// send initial job to each machine
		for (unsigned int i = 0; i <= w_size; ++i)
			if (i != mpi_env.get_root()) {
				AR_send_job(i, mpi_env, chunk_size, row, col);
				to_send -= chunk_size;

				if (to_send < 1)
					break;
				AR_next_chunk(chunk_size, nseq, row, col);
			}

		chunk_size = static_cast<unsigned int>(to_send / (2.0 * w_size));
		if (chunk_size == 0)
			++chunk_size;

		empi::MPI_obuffer obuff(buf_size, buffer, mpi_env.comm());

		// send remaining tasks
		do {
			// recive result
			MPI_Recv(buffer, buf_size, MPI_PACKED, MPI_ANY_SOURCE, DM_PART_TAG, mpi_env.comm(), &status);
#if DEBUG_MPI
			std::cout << "\nMASTER NODE RECIVED RESULT (dist. matrix part) FROM NODE "
					  << status.MPI_SOURCE << '\n';
#endif

			// put it to the matrix
			int S, r, c;
			int id, score;

			obuff >> S, r, c;

			for (int i = 0; i < S; ++i) {
				obuff >> id, score;
				DM->similarity_matrix[r][c] = DM->similarity_matrix[c][r] = id;
				DM->distance_matrix[r][c] = DM->distance_matrix[c][r] = MAXID - id;
				DM->score_similarity_matrix[r][c] = DM->score_similarity_matrix[c][r] = score;

/**
				std::cout << '\t' << (CL->S)->name[r] << " & " << (CL->S)->name[c]
						  << "\tidentity= " << id << "%  score= " << score << '\n';
*/

				if (++c == nseq) {
					++r;
					c = r + 1;
				}
			}

			// and write time
			double t;

			obuff >> t;
			++times_n[status.MPI_SOURCE];
			times[status.MPI_SOURCE].time += t;

			obuff.reset();

			to_recv -= S;
			// if it was last result finish loop
			if (to_recv == 0) {
				row = col = 0;
				for (unsigned int i = 0; i <= w_size; ++i) {
					if (i != mpi_env.get_root())
						AR_send_job(i, mpi_env, 0, row, col);
				}
				break;
			}

			// if we still have something to do
			if (to_send > 0) {
				AR_send_job(status.MPI_SOURCE, mpi_env, chunk_size, row, col);
				to_send -= chunk_size;
				AR_next_chunk(chunk_size, nseq, row, col);

				chunk_size = static_cast<unsigned int>((1.0 * to_send) / w_size);
				if (chunk_size == 0)
					++chunk_size;
			}
		}
		while (true);

		for (unsigned int i = 0; i < nseq; ++i) {
			DM->similarity_matrix[i][i] = MAXID;
			DM->distance_matrix[i][i] = 0;
		}

		std::cout << ".....[Done]\n";

		for (unsigned int i = 0; i < times.size(); ++i)
			if (times_n[i])
				times[i].time /= times_n[i];
	}
	else { // WORKER side
		if (!DM)
			DM = static_cast<TC_DistanceMatrix *>(AR_allocate(1, sizeof(TC_DistanceMatrix)));

		TC_ConstraintList *NCL;
		TC_Alignment *B;

		NCL = TC_duplicate_constraint_list_soft(CL);
		NCL->pw_parameters_set = 1;

		B = CL->tree_aln;
		if (!B)
			B = TC_seq2aln(NCL->S, 0, 1);

		if (TC_strm(mode, "very_fast")) {
			strncpy(NCL->dp_mode, "very_fast_pair_wise", 100 - 1);
			NCL->evaluate_residue_pair = TC_evaluate_matrix_score;
			if (TC_strm((NCL->S)->type, "DNA")) {
				NCL->M = TC_read_matrice("idmat");
				NCL->gop = -10;
				NCL->gep = -1;
				NCL->ktup = 6;
			}
			else {
				NCL->M = TC_read_matrice("blosum62mt");
				NCL->gop = (int)(TC_get_avg_matrix_mm(NCL->M, AA_ALPHABET) * 10);
				NCL->gep = -1;
				NCL->ktup = 2;
			}
			NCL->use_fragments = 1;
			NCL->diagonal_threshold = 6;
		}
		else if (TC_strm(mode, "aln")) {
			strncpy(NCL->dp_mode, "precomputed_pair_wise", 100 - 1);
		}
		else if (TC_strm(mode, "fast") || TC_strm(mode, "idscore")) {
			if (TC_strm(mode, "idscore"))
				strncpy(NCL->dp_mode, "idscore_pair_wise", 100 - 1);
			else
				strncpy(NCL->dp_mode, "myers_miller_pair_wise", 100 - 1);
			NCL->evaluate_residue_pair = TC_evaluate_matrix_score;

			if (TC_strm((NCL->S)->type, "DNA")) {
				NCL->M = TC_read_matrice("idmat");
				NCL->gop = -10;
				NCL->gep = -1;
			}
			else {
				NCL->M = TC_read_matrice("blosum62mt");
				NCL->gop = (int)(TC_get_avg_matrix_mm(NCL->M, AA_ALPHABET) * 10);
				NCL->gep = -1;
			}
		}

		int params[3];
		empi::MPI_ibuffer ibuff(buf_size, buffer, mpi_env.comm());

		do {
			// get job
			MPI_Recv(params, 3, MPI_INT, mpi_env.get_root(), CHUNK_TAG + mpi_env.rank(),
					 mpi_env.comm(), &status);
#if DEBUG_MPI
			std::cout << "\nNODE " << mpi_env.rank() << " RECIVED JOB FROM MASTER\n";
#endif

			// received 0 computations means end of loop
			if (params[0] == 0)
				break;

			// calculate given part of distance matrix
			ibuff << params[0], params[1], params[2];

			ibuff << AR_calculate(params[0], params[1], params[2], B, NCL, mode, ibuff);

			// and send results
			MPI_Send(const_cast<void *>(ibuff.data()), ibuff.position(), MPI_PACKED, mpi_env.get_root(), DM_PART_TAG,
					 mpi_env.comm());
#if DEBUG_MPI
			std::cout << "\nNODE " << mpi_env.rank() << " SENT RESULT TO MASTER\n";
#endif

			ibuff.reset();
		}
		while (true);

		TC_free_constraint_list(NCL);
		if (!CL->tree_aln)
			TC_free_aln(B);
	}

	strncpy(DM->mode, mode, 100 - 1);

	return DM;
}

TC_DistanceMatrix *AR_par_compute_distance_matrix(const TC_ConstraintList *CL, const char *in_mode,
	const MPI_Env &mpi_env, std::vector<AR_node_time> &times)
{
	if (!CL || (!in_mode && !CL->distance_matrix_mode))
		return 0;

	const char *mode = in_mode ? in_mode : CL->distance_matrix_mode;
	TC_DistanceMatrix *DM = CL->DM;

	if (!DM || !TC_strm(DM->mode, mode))
		DM = AR_par_seq2distance_matrix(CL, mode, mpi_env, times);

	return DM;
}

int **TC_get_sim_aln_array(const TC_Alignment *A, const char *mode)
{
	int **w;
	int a, b;

	w = TC_declare_int(A->nseq, A->nseq);

	for (a = 0; a < A->nseq - 1; ++a) {
		for (b = a + 1; b < A->nseq; ++b) {
			w[a][b] = w[b][a] = TC_generic_get_seq_sim(A->seq_al[a], A->seq_al[b],
					(A->cdna_cache) ? A->cdna_cache[0] : 0, mode);
		}
	}

	return w;
}

int **TC_get_dist_aln_array(const TC_Alignment *A, const char *mode)
{
	int **w;

	w = TC_get_sim_aln_array(A, mode);

	return TC_sim_array2dist_array(w, MAXID);
}

int **TC_sim_array2dist_array(int **p, int max)
{
	int s1, s2, a, b;

	s1 = TC_read_array_size((void *)p, sizeof(void *));
	s2 = TC_read_array_size((void *)p[0], sizeof(int));

	for (a = 0; a < s1; ++a)
		for (b = 0; b < s2; ++b)
			p[a][b] = max - p[a][b];

	return p;
}
