/*
 * File: TC_Sequences.cpp
 * Created: 06.06.2006
 * Last Modified: 16.09.2006
 * Author: Adrian Rospondek
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "TC_Alignment.hpp"
#include "TC_ConstraintList.hpp"
#include "TC_DistanceMatrix.hpp"
#include "TC_Sequences.hpp"
#include "TC_Templates.hpp"
#include "TC_Weights.hpp"
#include "dp_drivers.hpp"
#include "globals.hpp"
#include "input.hpp"
#include "output.hpp"
#include "util.hpp"
#include "util_io.hpp"
#include "util_mem.hpp"
#include "util_seq.hpp"
#include "util_str.hpp"



TC_Sequences *TC_declare_sequence(int min, int max, int nseq)
{
	TC_Sequences *LS;

	LS = (TC_Sequences *)AR_allocate(1, sizeof(TC_Sequences));
	LS->seq_comment = TC_declare_char(nseq, STRING + 1);
	LS->aln_comment = TC_declare_char(nseq, STRING + 1);
	LS->file = TC_declare_char(nseq, STRING + 1);
	LS->name = TC_declare_char(nseq, STRING + 1);
	LS->seq = TC_declare_char(nseq, max + 1);
	LS->len = (int *)AR_allocate(nseq, sizeof(int));
	LS->max_len = max;
	LS->min_len = min;
	LS->nseq = nseq;
	LS->max_nseq = nseq;
	LS->T = (TC_Templates **)TC_declare_arrayN(2, sizeof(TC_Templates), nseq, 1);

	return LS;
}

TC_Sequences *TC_duplicate_sequence(const TC_Sequences *S)
{
	TC_Sequences *LS;
	int a;

	if (S == 0)
		return 0;

	LS = TC_declare_sequence(S->min_len, S->max_len, S->nseq);
	for (a = 0; a < S->nseq; a++) {
		strncpy(LS->seq_comment[a], S->seq_comment[a], STRING);
		strncpy(LS->aln_comment[a], S->aln_comment[a], STRING);
		strncpy(LS->file[a], S->file[a], STRING);
		strncpy(LS->name[a], S->name[a], STRING);
		strncpy(LS->seq[a], S->seq[a], LS->max_len);
		LS->len[a] = S->len[a];
		LS->T[a][0] = S->T[a][0];
	}
	LS->max_len = S->max_len;
	LS->min_len = S->min_len;
	LS->nseq = S->nseq;

	if (S->W)
		LS->W = TC_duplicate_weights(S->W);
	strncpy(LS->type, S->type, sizeof(LS->type) - 1);
	LS->max_nseq = S->nseq;

	return LS;
}

TC_Sequences *TC_realloc_sequence(TC_Sequences *OUT, int new_nseq, int max_len)
{
	if (new_nseq < OUT->max_nseq)
		return OUT;

	OUT->min_len = MIN(OUT->min_len, max_len);
	OUT->max_len = MAX(OUT->max_len, max_len);
	OUT->seq_comment = TC_new_realloc_char(OUT->seq_comment, new_nseq, STRING + 1);
	OUT->aln_comment = TC_new_realloc_char(OUT->aln_comment, new_nseq, STRING + 1);
	OUT->seq = TC_new_realloc_char(OUT->seq, new_nseq, OUT->max_len + 1);
	OUT->name = TC_new_realloc_char(OUT->name, new_nseq, STRING + 1);
	OUT->file = TC_new_realloc_char(OUT->file, new_nseq, STRING + 1);
	OUT->len = (int *)AR_reallocate(OUT->len, (new_nseq + 1), sizeof(int));
	OUT->T =
		(TC_Templates **)TC_realloc_arrayN(2, OUT->T, sizeof(TC_Templates), new_nseq, 1);
	OUT->max_nseq = new_nseq;

	return OUT;
}

void TC_free_sequence(TC_Sequences *LS)
{
	if (!LS)
		return;

	TC_free_char(LS->file);
	TC_free_char(LS->seq_comment);
	TC_free_char(LS->aln_comment);
	TC_free_char(LS->seq);
	TC_free_char(LS->name);
	TC_free_arrayN((void *)LS->T, 2);
	AR_free(LS->len);
	TC_free_weights(LS->W);
	AR_free(LS);
}

TC_Sequences *TC_fill_sequence_struc(int nseq, char **sequences, char **seq_name)
{
	TC_Sequences *S;
	int shortest, longest;
	int a;

	if (nseq > 1) {
		shortest = TC_get_shortest_string(sequences, nseq, 0, 0);
		longest = TC_get_longest_string(sequences, nseq, 0, 0);
	}
	else if (nseq == 1) {
		shortest = longest = strlen(sequences[0]);
	}
	else {
		return 0;
	}

	S = TC_declare_sequence(shortest, longest, nseq);
	S->nseq = nseq;
	S->seq = TC_copy_char(sequences, S->seq, nseq, -1);
	S->name = TC_copy_char(seq_name, S->name, nseq, -1);

	TC_ungap_array(S->seq, nseq);
	for (a = 0; a < S->nseq; a++)
		S->len[a] = strlen(S->seq[a]);

	return S;
}

TC_Sequences *TC_aln2seq(const TC_Alignment *A)
{
	TC_Sequences *LS;
	int a;
	int maxlen;

	if (!A)
		return 0;
	else if (A->nseq == 0)
		return 0;
	for (maxlen = 0, a = 0; a < A->nseq; a++)
		maxlen = MAX(maxlen, strlen(A->seq_al[a]));

	LS = TC_declare_sequence(maxlen + 1, maxlen + 1, A->nseq);
	LS->nseq = A->nseq;
	for (a = 0; a < LS->nseq; a++) {
		strncpy(LS->file[a], A->file[a], STRING);
		LS->len[a] = A->len[a];
		strncpy(LS->seq[a], A->seq_al[a], LS->max_len);
		if (LS->contains_gap == 0)
			TC_ungap(LS->seq[a]);

		LS->len[a] = strlen(LS->seq[a]);
		strncpy(LS->seq_comment[a], A->seq_comment[a], STRING);
		strncpy(LS->aln_comment[a], A->aln_comment[a], STRING);
		strncpy(LS->name[a], A->name[a], STRING);
	}
	return LS;
}

int TC_seq_name2index(const char *name, const TC_Sequences *S)
{
	return S ? TC_name_is_in_list(name, S->name, S->nseq, MAXNAMES + 1) : -1;
}

TC_Sequences *TC_seq2unique_name_seq(TC_Sequences *S)
{
	int a, b, cnt;

	if (S == 0)
		return S;

	for (a = 0; a < S->nseq - 1; ++a) {
		cnt = 1;
		for (b = a + 1; b < S->nseq; ++b)
			if (TC_strm(S->name[a], S->name[b])) {
				snprintf(S->name[b], STRING + 1, "%s_%d", S->name[a], cnt++);
				std::cerr << '\n' << PROGRAM << ": WARNING: Sequence '" << S->name[a] << "' is duplicated in file '"
						  << S->file[a] << "'.\n\tThe sequence will be renamed.\n";
			}
	}

	return S;
}

/*
 * Expected format for the template file:
 * >seq_name _X_ target_template
 * X: S for sequences, G for genomes (Exoset)
 */
TC_Sequences *AR_seq2template_seq(TC_Sequences *S, const char *template_list,
	std::vector<TC_X_template *> &remote_templates)
{
	// 1: No template
	if ((template_list && template_list[0] == '\0') || TC_strm(template_list, "no_template")) {
		return S;
	}
	// 2: Templates from seqnames (SELF)
	else if (strstr(template_list, "SELF_")) {
		TC_X_template *P;
		const char *p;
		int a;

		for (a = 0; a < S->nseq; a++) {
			p = S->name[a];

			if (strstr(template_list, "_P_")) {
				(S->T[a])->P = P = AR_prepare_P_template(p, p);
				if (P->pdb_id[0] == '\0')
					remote_templates.push_back(P);
			}
			else if(strstr(template_list, "_R_" ))
				(S->T[a])->R = TC_fill_R_template(p, p, S);
			else if (strstr(template_list, "_G_"))
				(S->T[a])->G = TC_fill_G_template(p, p, S);
		}

		return S;
	}
	// 3: Templates named like the sequences (SEQFILE)
	else if (strstr(template_list, "SEQFILE_")) {
		TC_X_template *P;
		const char *p;
		int a;

		p = template_list;
		for (a = 0; a < S->nseq; a++) {
			if (strstr(template_list, "_P_")) {
				(S->T[a])->P = P = AR_prepare_P_template(S->name[a], p);
				if (P->pdb_id[0] == '\0')
					remote_templates.push_back(P);
			}
			else if(strstr(template_list, "_R_" ))
				(S->T[a])->R = TC_fill_R_template(S->name[a], p, S);
			else if (strstr(template_list, "_G_"))
				(S->T[a])->G = TC_fill_G_template(S->name[a], p, S);
		}

		return S;
	}
	// 4: Templates comes in a template_file
	else if (template_list == 0 || TC_format_is_fasta(template_list)) {
		TC_Sequences *T;
		TC_X_template *P;
		const char *p;
		int a, i;

		T = (template_list != 0) ? TC_get_fasta_sequence(template_list) : S;
		for (a = 0; a < T->nseq; a++) {
			if ((i = TC_name_is_in_list(T->name[a], S->name, S->nseq, MAXNAMES)) != -1) {
				if ((p = strstr(T->seq_comment[a], "_P_"))) {
					(S->T[i])->P = P = AR_prepare_P_template(S->name[i], p);
					if (P->pdb_id[0] == '\0')
						remote_templates.push_back(P);
				}
				else if((p = strstr(T->seq_comment[a], "_R_")))
					(S->T[i])->R = TC_fill_R_template(S->name[i], p, S);
				else if ((p = strstr(T->seq_comment[a], "_G_")))
					(S->T[i])->G = TC_fill_G_template(S->name[i], p, S);

				if (T != S)
					strcat(S->seq_comment[i], T->seq_comment[a]);
			}
		}

		if (T != S)
			TC_free_sequence(T);
		
		return S;
	}
	// 5: Templates are generated with a script
	else if (strstr(template_list, "SCRIPT_")) {
		static int ntemp;
		TC_fname *F = 0;
		TC_Alignment *A;
		char outfile[1000];
		char script[1000];
		char *tmp1, *command;

		ntemp++;
		command = (char *)AR_allocate(1000, sizeof(char));
		tmp1 = AR_tmpname();

		A = TC_seq2aln(S, 0, 0);
		TC_output_fasta_seq(tmp1, A);

		strncpy(script, TC_after_strstr(template_list, "SCRIPT_"), 1000 - 1);
		F = TC_parse_fname(S->file[0]);
		snprintf(outfile, 1000, "%s_%d.template_list", F->name, ntemp);
		snprintf(command, 1000, "%s -infile=%s -outfile=%s", script, tmp1, outfile);
		command = TC_substitute(command, "@", " ");
		system(command);

		TC_free_aln(A);
		AR_free(command);
		AR_free(tmp1);

		if (AR_check_file_exists(outfile) && TC_format_is_fasta(outfile)) {
			AR_display_output_filename(std::cout, "TEMPLATE_LIST", "fasta_seq", outfile, STORE);
			S = AR_seq2template_seq(S, outfile, remote_templates);
		}
		else {
			std::cerr << '\n' << PROGRAM << ": WARNING: Could not run " << script << " to find templates.\n";
		}

		return S;
	}

	return S;
}

/**
 * @brief This function fetches potential templates associated with sequences within a profile.
 */
TC_Sequences *AR_profile_seq2template_seq(TC_Sequences *S, const char *template_list,
	std::vector<TC_X_template *> &remote_templates)
{
	TC_Alignment *A;
	int i;

	for (i = 0; i < S->nseq; i++) {
		if (A = TC_seq2R_template_profile(S, i)) {
			A->S = TC_aln2seq(A);
			A->S = AR_seq2template_seq(A->S, template_list, remote_templates);
		}
	}

	return S;
}

TC_Sequences *TC_seq2template_type(TC_Sequences *Seq)
{
	TC_X_template *P = 0;
	TC_X_template *G = 0;
	TC_X_template *R = 0;
	TC_Alignment *A;
	int a, e;
	bool p;

	e = ' ';
	for (a = 0; a < Seq->nseq; ++a) {
		if (!Seq->T[a])
			continue;
		P = TC_seq_has_template(Seq, a, "_P_");
		G = TC_seq_has_template(Seq, a, "_G_");
		R = TC_seq_has_template(Seq, a, "_R_");
		p = P && P->pdb_id[0];

		sprintf((Seq->T[a])->seq_type, "%c%c%c%c", 'S', p ? 'P' : 's', G ? 'G' : e, R ? 'R' : e);

		if (R && (A = TC_seq2R_template_profile(Seq, a))) {
			A->S = TC_seq2template_type(A->S);
		}
	}

	return Seq;
}

int TC_seq2n_X_template(const TC_Sequences *S, const char *type)
{
	int a, n;

	for (n = 0, a = 0; a < S->nseq; ++a) {
		if (TC_strm2(type, "_P_", "_*_") && (S->T[a])->P)
			++n;
		if (TC_strm2(type, "_R_", "_*_") && (S->T[a])->R)
			++n;
		if (TC_strm2(type, "_G_", "_*_") && (S->T[a])->G)
			++n;
	}

	return n;
}

int TC_prf_in_seq(const TC_Sequences *S)
{
	int a;

	if (!S)
		return 0;

	for (a = 0; a < S->nseq; ++a)
		if (TC_seq2R_template_profile(S, a))
			break;

	return (a < S->nseq);
}

TC_X_template *TC_seq_has_template(const TC_Sequences *S, int n, const char *mode)
{
	TC_Templates *T;

	if (!S || !mode)
		return 0;
	else if (n < 0 || n >= S->nseq)
		return 0;
	else if (!(S->T) || !(S->T[n]))
		return 0;

	T = S->T[n];
	if (TC_strm(mode, "_P_"))
		return T->P;
	else if(TC_strm(mode, "_R_"))
		return T->R;
	else if (TC_strm(mode, "_G_"))
		return T->G;
	else
		return 0;
}

TC_Alignment *TC_seq2R_template_profile(const TC_Sequences *S, int n)
{
	TC_X_template *R;

	R = TC_seq_has_template(S, n, "_R_");
	return R ? R->A : 0;
}

char *TC_seq2P_template_file(const TC_Sequences *S, int n)
{
	TC_X_template *P;

	P = TC_seq_has_template(S, n, "_P_");
	return P ? P->template_file : 0;
}

TC_Sequences *TC_get_sequence_type(TC_Sequences *S)
{
	char *buf, *b, *type;
	int tot;
	int a;

	for (tot = 0, a = 0; a < S->nseq; ++a)
		tot += S->len[a];
	buf = (char *)AR_allocate(tot + 1, sizeof(char));

	for (b = buf, a = 0; a < S->nseq; ++a) {
		strcpy(b, S->seq[a]);
		b += S->len[a];
	}
	strncpy(S->type, type = TC_get_string_type(buf), sizeof(S->type) - 1);
	
	AR_free(buf);
	AR_free(type);

	return S;
}

TC_Sequences *TC_ungap_seq(TC_Sequences *S)
{
	int a;

	if (!S)
		return 0;

	TC_ungap(S->seq[0]);
	S->max_len = S->min_len = S->len[0] = strlen(S->seq[0]);
	for (a = 1; a < S->nseq; a++) {
		TC_ungap(S->seq[a]);
		S->len[a] = strlen(S->seq[a]);
		S->max_len = MAX(S->max_len, S->len[a]);
		S->min_len = MIN(S->min_len, S->len[a]);
	}

	return S;
}

TC_Sequences *TC_reorder_seq(TC_Sequences *A, char **name, int nseq)
{
	TC_Sequences *nA;
	int a, sn;

	nA = TC_duplicate_sequence(A);

	for (a = 0; a < nseq; a++) {
		sn = TC_name_is_in_list(name[a], nA->name, nA->nseq, 100);
		if (sn == -1)
			continue;
		if (nA->file)
			strncpy(A->file[a], nA->file[sn], STRING);
		if (nA->seq_comment)
			strncpy(A->seq_comment[a], nA->seq_comment[sn], STRING);
		if (nA->aln_comment)
			strncpy(A->aln_comment[a], nA->aln_comment[sn], STRING);
		strncpy(A->name[a], nA->name[sn], STRING);
		strncpy(A->seq[a], nA->seq[sn], A->max_len);
		A->len[a] = nA->len[sn];
		A->T[a][0] = nA->T[sn][0];
	}
	A->nseq = nseq;

	TC_free_sequence(nA);

	return A;
}

TC_Sequences *TC_merge_seq(const TC_Sequences *IN, TC_Sequences *OUT)
{
	int a;

	if (IN == 0)
		return OUT;
	else if (OUT == 0)
		return TC_duplicate_sequence(IN);
	else {
		if (TC_check_list_for_dup(IN->name, IN->nseq)) {
			std::cerr << '\n' << PROGRAM << ": ERROR: Sequence '" << TC_check_list_for_dup(IN->name, IN->nseq)
					  << "' is duplicated in file '" << IN->file[0] << "'.\n";
			return 0;
		}
		for (a = 0; a < IN->nseq; a++)
			if ((OUT = TC_add_sequence(IN, OUT, a)) == 0)
				return 0;
		
		return OUT;
	}
}

static char *TC_build_consensus(const char *seq1, const char *seq2, const char *dp_mode)
{
	static char *mat;
	TC_Alignment *A;
	char *buf;
	char c1, c2;
	int a;

	if (!mat)
		mat = (char *)AR_allocate(STRING, sizeof(char));

	A = TC_align_two_sequences(seq1, seq2, strcpy(mat, "idmat"), 0, 0, dp_mode);
	buf = (char *)AR_allocate(A->len_aln + 1, sizeof(char));

	for (a = 0; a < A->len_aln; a++) {
		c1 = A->seq_al[0][a];
		c2 = A->seq_al[1][a];
		if (TC_is_gap(c1) && TC_is_gap(c2))
			buf[a] = '-';
		else if (TC_is_gap(c1))
			buf[a] = c2;
		else if (TC_is_gap(c2))
			buf[a] = c1;
		else if (c1 != c2) {
			AR_free(buf);
			buf = 0;
			TC_free_aln(A);
			return 0;
		}
		else
			buf[a] = c1;
	}
	buf[a] = '\0';
	TC_free_sequence(TC_free_aln(A));

	return buf;
}

TC_Sequences *TC_add_sequence(const TC_Sequences *IN, TC_Sequences *OUT, int i)
{
	char *buf;
	int s;

	if (OUT == 0) {
		OUT = TC_duplicate_sequence(IN);
		return OUT;
	}

	// Adds sequence i of IN at the end of OUT
	if ((s = TC_name_is_in_list(IN->name[i], OUT->name, OUT->nseq, STRING)) == -1) {
		OUT = TC_realloc_sequence(OUT, OUT->nseq + 1, IN->len[i]);
		strncpy(OUT->file[OUT->nseq], IN->file[i], STRING);
		strncpy(OUT->seq_comment[OUT->nseq], IN->seq_comment[i], STRING);
		strncpy(OUT->aln_comment[OUT->nseq], IN->aln_comment[i], STRING);
		strncpy(OUT->name[OUT->nseq], IN->name[i], STRING);
		strncpy(OUT->seq[OUT->nseq], IN->seq[i], OUT->max_len);
		OUT->len[OUT->nseq] = IN->len[i];
		OUT->T[OUT->nseq][0] = IN->T[i][0];
		OUT->nseq++;

		return OUT;
	}
	else if (s != -1 && !TC_case_insensitive_strcmp(IN->seq[i], OUT->seq[s])) {
		std::cerr << '\n' << PROGRAM << ": WARNING: Discrepancy: Sequence '" << IN->name[i] << "' in files: '"
				  << IN->file[i] << "' and '" << OUT->file[s] << "'.\n";

		if ((buf = TC_build_consensus(IN->seq[i], OUT->seq[s], "cfasta_pair_wise")) != 0
			|| (buf = TC_build_consensus(IN->seq[i], OUT->seq[s], "myers_miller_pair_wise")) != 0) {
			OUT->len[s] = strlen(buf);
			OUT->max_len = MAX(OUT->max_len, OUT->len[s]);
			OUT->min_len = MIN(OUT->min_len, OUT->len[s]);
			OUT->seq = TC_realloc_char(OUT->seq, -1, -1, OUT->nseq, OUT->max_len + 1);
			strncpy(OUT->seq[s], buf, OUT->max_len);

			AR_free(buf);
			
			return OUT;
		}
		else {
			std::cerr << '\n' << PROGRAM << ": ERROR: Impossible to reconciliate some sequences.\n";
			return 0;
		}
	}
	else {
		return OUT;
	}
}

/**
 * @brief Index the names of S1 according to S2.
 * index[seq1 of S1][0] -> x  if seq1 is the xth sequence of S2
 *                      -> -1 if seq1 is nowhere to be found
 * index[seq1 of S1][1] -> z  if seq1 is the zth sequence within the xth profile of S2
 */
int **TC_index_seq_name(const TC_Sequences *S1, const TC_Sequences *S2)
{
	TC_Alignment *profile;
	int **index;
	int a, b, x, z;

	index = TC_declare_int(S1->nseq, 2);

	for (a = 0; a < S1->nseq; a++) {
		index[a][0] = index[a][1] = -1;
		x = TC_name_is_in_list(S1->name[a], S2->name, S2->nseq, 100);
		if (x != -1) {
			index[a][0] = x;
			index[a][1] = -1;
		}
		for (b = 0; b < S2->nseq; b++) {
			if ((profile = TC_seq2R_template_profile(S2, b))) {
				z = TC_name_is_in_list(S1->name[a], profile->name, profile->nseq, 100);
				if (z != -1) {
					index[a][0] = b;
					index[a][1] = z;
					b = S2->nseq;
				}
			}
		}
	}

	return index;
}

static int *TC_get_res_index(const char *seq0, const char *seq1)
{
	int *coor, a;

	if (!seq0 || !seq1)
		return 0;

	coor = (int *)AR_allocate(strlen(seq0) + 1, sizeof(int));
	if (!TC_strm(seq0, seq1)) {
		TC_Alignment *A;
		int r0, r1, isr0, isr1;
		int l0 = 0, l1 = 0;

		A = TC_align_two_sequences(seq0, seq1, "pam250mt", -5, -1, "myers_miller_pair_wise");
		for (a = 0; a < A->len_aln; a++) {
			r0 = A->seq_al[0][a];
			r1 = A->seq_al[1][a];
			isr0 = !TC_is_gap(r0);
			isr1 = !TC_is_gap(r1);
			l0 += isr0;
			l1 += isr1;
			if (isr0 && isr1)
				coor[l0 - 1] = l1 - 1;
			else if (isr0)
				coor[l0 - 1] = -1;
		}
		TC_free_aln(A);
	}
	else {
		int l0;

		l0 = strlen(seq0);
		for (a = 0; a < l0; a++)
			coor[a] = a;
	}

	return coor;
}

/**
 * @brief Index the residues of S1 according to S2.
 * index[seq1 of S1][z] -> x, where x is the position of residue z of seq1/S1
 * in S2->seq[index[Seq1/S1]]
 */
int **TC_index_seq_res(const TC_Sequences *S1, const TC_Sequences *S2, int **name_index)
{
	TC_Alignment *profile;
	char *seq1, *seq2;
	int **index;
	int a;

	index = (int **)AR_allocate(S1->nseq, sizeof(int *));

	for (a = 0; a < S1->nseq; a++) {
		seq1 = S1->seq[a];
		if (name_index[a][0] == -1)
			seq2 = 0;
		else if (name_index[a][1] == -1)
			seq2 = S2->seq[name_index[a][0]];
		else if ((profile = TC_seq2R_template_profile(S2, name_index[a][0])) != 0)
			seq2 = profile->seq_al[name_index[a][1]];
		index[a] = TC_get_res_index(seq1, seq2);
	}

	return index;
}

FILE *TC_display_sequences_names(const TC_Sequences *S, FILE *fp, int check_pdb_status, int print_templates)
{
	char *r, *buf;
	int max_len;
	int a;

	if (!S || S->nseq == 0) {
		fprintf(fp, "\n%s: ERROR: No sequences read.\n", PROGRAM);
		return fp;
	}

	for (a = 0, max_len = 0; a < S->nseq; a++)
		max_len = MAX(max_len, strlen(S->name[a]));

	fprintf(fp, "\nINPUT SEQUENCES: %d sequence(s) (%s)\n", S->nseq, (S->type[0]) ? S->type : "Unknown type");
	for (a = 0; a < S->nseq; ++a) {
		fprintf(fp, "\tName: %*s, Length: %4d, Input File: %*s, Struct: ", max_len, S->name[a], S->len[a],
				max_len, S->file[a]);
		
		if (check_pdb_status) {
			if ((r = TC_seq2P_template_file(S, a))) {
				fprintf(fp, "Yes (PDBID %s)", buf = TC_get_pdb_id(r));
				AR_free(buf);
			}
			else
				fprintf(fp, "No");
		}
		else
			fprintf(fp, "Unchecked");

		if (print_templates)
			TC_display_sequence_templates(S, a, std::cout);
		
		fprintf(fp, "\n");
	}
	fprintf(fp, "\n");

	return fp;
}

std::ostream &TC_display_sequence_templates(const TC_Sequences *S, int i, std::ostream &out)
{
	TC_display_X_template((S->T[i])->P, out);
	TC_display_X_template((S->T[i])->G, out);
	TC_display_X_template((S->T[i])->R, out);

	return out;
}
