/*
 * File: TC_Templates.cpp
 * Created: 06.06.2006
 * Last Modified: 16.09.2006
 * Author: Adrian Rospondek
 */


#include <string>
#include <string.h>
#include "TC_Alignment.hpp"
#include "TC_Profile.hpp"
#include "TC_Sequences.hpp"
#include "TC_Templates.hpp"
#include "globals.hpp"
#include "input.hpp"
#include "output.hpp"
#include "util.hpp"
#include "util_io.hpp"
#include "util_mem.hpp"
#include "util_pdb.hpp"
#include "util_str.hpp"



static TC_X_template *TC_fill_X_template(const char *name, const char *templ, const char *token)
{
	TC_X_template *X;
	char *k;

	X = (TC_X_template *)AR_allocate(1, sizeof(TC_X_template));

	strncpy(X->seq_name, name, 100 - 1);
	strncpy(X->template_type, token, 5 -1);
	if (k = (char*)strstr(templ, token))
		sscanf(k + strlen(token), "%s", X->template_name);
	else
		strncpy(X->template_name, templ, 100 - 1);

	return X;
}

static TC_X_template *TC_free_X_template(TC_X_template *X)
{
	if (X->S)
		TC_free_sequence(X->S);
	else if (X->A)
		TC_free_aln(X->A);

	AR_free(X);

	return 0;
}

TC_X_template *TC_fill_G_template(const char *name, const char *templ, TC_Sequences *S)
{
	TC_X_template *G;

	G = TC_fill_X_template(name, templ, "_G_");

	if (TC_strm(name, templ))
		strncpy(G->template_file, TC_output_fasta_seqX(0, "w", S, 0, TC_seq_name2index(name, S)), 100 - 1);
	else if (strstr(templ, "SEQFILE_")) {
		TC_Sequences *ST;
		int i;

		ST = TC_main_read_seq(TC_after_strstr(templ, "SEQFILE_G_"));
		i = TC_seq_name2index(name, ST);
		strncpy(G->template_file, TC_output_fasta_seqX(0, "w", ST, 0, i), 100 - 1);
		strncpy(G->template_name, name, 100 - 1);
		TC_free_sequence(ST);
	}
	else
		strncpy(G->template_file, G->template_name, 100 - 1);

	if (!TC_is_seq(G->template_file)) {
		std::cerr << '\n' << PROGRAM << ": WARNING: _G_template '" << templ << "' could not be found.\n";
		TC_free_X_template(G);

		return 0;
	}
	else
		G->S = TC_get_fasta_sequence(G->template_file);

	return G;
}

TC_X_template *TC_fill_P_template(const char *name, const char *templ)
{
	TC_X_template *P;

	P = AR_prepare_P_template(name, templ);

	if (!P->pdb_id[0]) {
		std::cerr << '\n' << PROGRAM << ": WARNING: Could not fill _P_template for sequence '" << templ << "'.\n";
		TC_free_X_template(P);
		P = 0;
	}

	return P;
}

TC_X_template *AR_prepare_P_template(const char *name, const char *templ)
{
	TC_X_template *P;

	P = TC_fill_X_template(name, templ, "_P_");

	std::string pdb(P->template_name);

	P->pdb_id[0] = '\0';
	if (AR_check_pdb_location(pdb) == LOCAL_PDB) {
		char *buf;

		strncpy(P->template_file, pdb.c_str(), 100 - 1);
		strncpy(P->template_name, buf = TC_path2filename(P->template_file), 100 - 1);
		AR_free(buf);

		strncpy(P->pdb_id, buf = TC_get_pdb_id(P->template_file), 5 - 1);
		AR_free(buf);
	}

	return P;
}

void AR_fill_P_template(TC_X_template *P, const char *file_name)
{
	if (P && *(P->template_type) && strstr(P->template_type, "_P_")
		&& file_name && *file_name && TC_is_pdb_file(file_name)) {
		char *buf;

		strncpy(P->template_file, file_name, 100 - 1);
		strncpy(P->pdb_id, buf = TC_get_pdb_id(P->template_file), 5 - 1);
		AR_free(buf);
	}
}

TC_X_template *TC_fill_R_template(const char *name, const char *templ, TC_Sequences *S)
{
	TC_X_template *R;

	R = TC_fill_X_template(name, templ, "_R_");
	if (!TC_is_aln(R->template_name) && !TC_is_seq(R->template_name)) {
		std::cerr << '\n' << PROGRAM << ": WARNING: _R_template '" << R->template_name << "' could not be found.\n";
		TC_free_X_template(R);
		
		return 0;
	}
	else {
		int s;
		TC_Sequences *S1;
		TC_Alignment *A1;

		if ((R->A = TC_main_read_aln(R->template_name, 0)) == 0) {
			std::cerr << '\n' << PROGRAM << ": WARNING: Could not fill _R_template '" << R->template_name << "'.\n";
			TC_free_X_template(R);
			
			return 0;
		}
		
		if (!S)
			strncpy(R->template_file, R->template_name, 100 - 1);
		else {
			s = TC_name_is_in_list(name, S->name, S->nseq, 100);
			if (s != -1) {
				S1 = TC_fill_sequence_struc(1, &S->seq[s], &S->name[s]);
				A1 = TC_seq2aln(S1, 0, RM_GAP);
				R->A = TC_trim_aln_with_seq(A1, R->A);
				strncpy(R->template_file, AR_tmpname(), 100 - 1);
				TC_output_clustal_aln(R->template_file, R->A);
			}
			else
				strncpy(R->template_file, R->template_name, 100 - 1);
		}
		R->A = TC_aln2profile(R->A);
	}
	
	return R;
}

std::ostream &TC_display_X_template(TC_X_template *X, std::ostream &out)
{
	if (X)
		out << "\n\t\t" << TC_template_type2type_name(X->template_type) << ": Template= " << X->template_name
			<< ", File= " << X->template_file;

	return out;
}

const char *TC_template_type2type_name(const char *type)
{
	if (strstr(type, "_P_"))
		return "PDB struc";
	else if (strstr(type, "_S_"))
		return "Sequence";
	else if (strstr(type, "_R_"))
		return "Profile";
	else if (strstr(type, "_G_"))
		return "Genomic";
	else if (strstr(type, "_RB_"))
		return "BlastProf";
	else
		return type;
}
