/*
 * File: TC_Weights.cpp
 * Created: 06.06.2006
 * Last Modified: 11.09.2006
 * Author: Adrian Rospondek
 */


#include <math.h>
#include <string.h>
#include "TC_ConstraintList.hpp"
#include "TC_DistanceMatrix.hpp"
#include "TC_Sequences.hpp"
#include "TC_Weights.hpp"
#include "globals.hpp"
#include "util_io.hpp"
#include "util_mem.hpp"



TC_Weights *TC_declare_weights(int nseq)
{
	TC_Weights *W;

	W = (TC_Weights *)AR_allocate(1, sizeof(TC_Weights));
	W->comments = (char *)AR_allocate(1000, sizeof(char));
	W->nseq = nseq;
	W->mode = (char *)AR_allocate(FILENAMELEN, sizeof(char));
	W->seq_name = TC_declare_char(W->nseq * 2, 200);
	W->PW_SD = TC_declare_float(W->nseq, W->nseq);
	W->PW_ID = TC_declare_float(W->nseq, W->nseq);
	W->SEQ_W = (float *)AR_allocate(W->nseq, sizeof(float));

	return W;
}

TC_Weights *TC_duplicate_weights(const TC_Weights *W)
{
	TC_Weights *NW;
	int a, b;

	NW = TC_declare_weights(W->nseq);
	strncpy(NW->comments, W->comments, 1000 - 1);
	strncpy(NW->mode, W->mode, FILENAMELEN - 1);

	for (a = 0; a < W->nseq; a++) {
		strncpy(NW->seq_name[a], W->seq_name[a], 200 - 1);
		NW->SEQ_W[a] = W->SEQ_W[a];
		for (b = 0; b < W->nseq; b++) {
			NW->PW_SD[a][b] = W->PW_SD[a][b];
			NW->PW_ID[a][b] = W->PW_ID[a][b];
		}
	}

	return NW;
}

TC_Weights *TC_free_weights(TC_Weights *W)
{
	if (!W)
		return 0;

	AR_free(W->comments);
	AR_free(W->mode);
	TC_free_char(W->seq_name);
	TC_free_float(W->PW_SD);
	TC_free_float(W->PW_ID);
	AR_free(W->SEQ_W);
	AR_free(W);

	return 0;
}

TC_Weights *TC_compute_tcoffee_weight(const TC_ConstraintList *CL)
{
	TC_Weights *W;
	float p, d;
	int nseq;
	int a, b;

	if (!CL->L)
		return 0;

	nseq = (CL->S)->nseq;
	W = TC_declare_weights(nseq);
	strncpy(W->mode, "t_coffee", FILENAMELEN - 1);

	for (a = 0; a < nseq; a++) {
		strncpy(W->seq_name[a], (CL->S)->name[a], 200 - 1);
		W->SEQ_W[a] = 1;
	}

	for (a = 0; a < (CL->S)->nseq - 1; a++)
		for (b = a + 1; b < (CL->S)->nseq; b++) {
			if (b == a)
				d = 1;
			else if (!(CL->S)->len[b] || !(CL->S)->len[a])
				d = 1;
			else
				d = ((float)(CL->DM)->similarity_matrix[a][b] / MAXID) * 10;
			p = pow(d, 3);
			W->SEQ_W[a] += p;
			W->SEQ_W[b] += p;
		}

	for (p = 0, b = 0; b < (CL->S)->nseq; b++) {
		if ((CL->S)->len[b] == 0)
			W->SEQ_W[b] = 0;
		else
			W->SEQ_W[b] = 2 / W->SEQ_W[b];
		p += W->SEQ_W[b];
	}
	for (b = 0; b < (CL->S)->nseq; b++)
		W->SEQ_W[b] = W->SEQ_W[b] * ((float)W->nseq / p);

	return W;
}

std::ostream &TC_display_weights(const TC_Weights *W, std::ostream &out)
{
	out << "\nSEQUENCE WEIGHTS (mode: ";

	if (!W) {
		out << "Unweighted)\n\tEvery sequence  Weight: 1.0\n";
		return out;
	}

	std::streamsize prec;

	prec = out.precision(2);
	showpoint(out);

	out << W->mode << ")\n";

	for (int i = 0; i < W->nseq; ++i)
		out << "\tSeq: " << W->seq_name[i] << "  Weight: " << W->SEQ_W[i] << '\n';

	out.precision(prec);
	noshowpoint(out);

	return out;
}
