/*
 * File: dp_drivers.cpp
 * Created: 06.06.2006
 * Last Modified: 19.09.2006
 * Author: Adrian Rospondek
 */


#include <string.h>
#include "TC_Alignment.hpp"
#include "TC_ConstraintList.hpp"
#include "JZ_ConstraintList.hpp"
#include "TC_Sequences.hpp"
#include "dp_drivers.hpp"
#include "dp_fasta_nw.hpp"
#include "dp_gotoh_nw.hpp"
#include "dp_mm_nw.hpp"
#include "dp_sim.hpp"
#include "evaluate.hpp"
#include "globals.hpp"
#include "input.hpp"
#include "util.hpp"
#include "util_mem.hpp"
#include "util_seq.hpp"
#include "util_str.hpp"

#include <mpi.h>


static int g_align_two_seq_keep_case;
static TC_NT_node *g_SNL;


TC_Alignment *TC_align_two_sequences(const char *seq1, const char *seq2, const char *in_matrix,
				     int gop, int gep, const char *align_mode)
{
    static TC_Alignment *A;
    static char *matrix;
    static int **M;
    TC_ConstraintList *CL;
    TC_Sequences *S;
    char **seq_array;
    char **name_array;
    int *ns;
    int **l_s;

    if (!matrix)
	matrix = (char *)AR_allocate(100, sizeof(char));
	
    CL = (TC_ConstraintList *)AR_allocate(1, sizeof(TC_ConstraintList));
    CL->matrices_list = TC_declare_char(10, 10);

    CL->pw_parameters_set = 1;

    if (!TC_strm(matrix, in_matrix)) {
	strncpy(matrix, in_matrix, 100 - 1);
	M = CL->M = TC_read_matrice(matrix);
    }
    else {
	CL->M = M;
    }

    if (strstr(align_mode, "cdna"))
	CL->evaluate_residue_pair = TC_evaluate_cdna_matrix_score;
    else
	CL->evaluate_residue_pair = TC_evaluate_matrix_score;

    CL->get_dp_cost = TC_get_dp_cost;
    CL->extend_jit = 0;
    CL->maximise = 1;
    CL->gop = gop;
    CL->gep = gep;
    CL->TG_MODE = 2;
    strncpy(CL->matrix_for_aa_group, "vasiliky", 100 - 1);
    CL->use_fragments = 0;
    CL->ktup = 3;
    if (!CL->use_fragments)
	CL->diagonal_threshold = 0;
    else
	CL->diagonal_threshold = 6;

    strncpy(CL->dp_mode, align_mode, 100 - 1);

    seq_array = TC_declare_char(2, MAX(strlen(seq1), strlen(seq2)) + 1);
    strcpy(seq_array[0], seq1);
    strcpy(seq_array[1], seq2);
    TC_ungap_array(seq_array, 2);
    if (g_align_two_seq_keep_case != KEEP_CASE)
	TC_string_array_lower(seq_array, 2);

    name_array = TC_declare_char(2, STRING);
    strcpy(name_array[0], "A");
    strcpy(name_array[1], "B");

    ns = (int *)AR_allocate(2, sizeof(int));
    l_s = TC_declare_int(2, 1);
    ns[0] = ns[1] = 1;
    l_s[0][0] = 0;
    l_s[1][0] = 1;

    CL->S = TC_fill_sequence_struc(2, seq_array, name_array);

    A = TC_seq2aln(CL->S, 0, 1);
    TC_ungap(A->seq_al[0]);
    TC_ungap(A->seq_al[1]);

    A->score_aln = TC_pair_wise(A, ns, l_s, CL);

    AR_free(ns);
    TC_free_int(l_s);
    TC_free_char(name_array);
    TC_free_char(seq_array);

    CL->M = 0;
    S = TC_free_constraint_list(CL);
    TC_free_sequence(S);
    A->S = 0;

    return A;
}

void TC_toggle_case_in_align_two_sequences(int value)
{
    g_align_two_seq_keep_case = value;
}

TC_Alignment *TC_align_two_aln(TC_Alignment *A1, TC_Alignment *A2, const char *matrix, int gop,
			       int gep, const char *align_mode)
{
    TC_Alignment *A = 0;
    TC_ConstraintList *CL;
    TC_Sequences *S;
    int *ns;
    int **ls;
    int a;
	
    CL = (TC_ConstraintList *)AR_allocate(1, sizeof(TC_ConstraintList));
    CL->pw_parameters_set = 1;
    CL->M = TC_read_matrice(matrix);
    CL->matrices_list = TC_declare_char(10, 10);
    CL->evaluate_residue_pair = TC_evaluate_matrix_score;
    CL->get_dp_cost = TC_consensus_get_dp_cost;
    CL->normalise = 1;
    CL->extend_jit = 0;
    CL->maximise = 1;
    CL->gop = gop;
    CL->gep = gep;
    CL->TG_MODE = 2;
    strncpy(CL->matrix_for_aa_group, "vasiliky", 100 - 1);
    CL->use_fragments = 0;
    CL->ktup = 5;
    if (!CL->use_fragments)
	CL->diagonal_threshold = 0;
    else
	CL->diagonal_threshold = 6;
    strncpy(CL->dp_mode, align_mode, 100 - 1);
	
    A = TC_copy_aln(A1, A);
    A = TC_stack_aln(A, A2);
    CL->S = TC_fill_sequence_struc(A->nseq, A->seq_al, A->name);
	
    ns = (int *)AR_allocate(2, sizeof(int));
    ls = TC_declare_int(2, A->nseq);
    ns[0] = A1->nseq;
    ns[1] = A2->nseq;
    for (a = 0; a < ns[0]; a++)
	ls[0][a] = a;
    for (a = 0; a < ns[1]; a++)
	ls[1][a] = a + A1->nseq;
	
    A->score_aln = TC_pair_wise(A, ns, ls,CL);
	
    AR_free(ns);
    TC_free_int(ls);
    S = TC_free_constraint_list(CL);
    TC_free_sequence(S);
    A->S = 0;
	
    return A;
}

static int TC_empty_pair_wise(TC_Alignment *A, int *ns, int **l_s, int glocal)
{
    char *string;
    int *l;
    int l0, l1, len;
    int n, a, b;

    if (glocal == GLOBAL) {
	l0 = strlen(A->seq_al[l_s[0][0]]);
	l1 = strlen(A->seq_al[l_s[1][0]]);
	len = MAX(l1, l0);

	if (len == 0)
	    return 0;
	else if (l0 > l1) {
	    n = ns[1];
	    l = l_s[1];
	}
	else if (l0 < l1) {
	    n = ns[0];
	    l = l_s[0];
	}
	string = TC_generate_string(len, '-');
	for (a = 0; a < n; a++)
	    strncpy(A->seq_al[l[a]], string, A->declared_len - 1);
	A->score = A->score_aln = 0;
	A->len_aln = len;
	AR_free(string);
	return 0;
    }
    else if (glocal == LALIGN) {
	A->A = TC_declare_aln(A->S);
	(A->A)->len_aln = 0;
	for (a = 0; a < 2; a++)
	    for (b = 0; b < ns[a]; b++)
		A->seq_al[l_s[a][b]][0] = '\0';
	(A->A)->score_aln = (A->A)->score = 0;
	return 0;
    }
    else
	return 0;
}

int TC_pair_wise(TC_Alignment *A, int *ns, int **l_s, TC_ConstraintList *CL)
{
    TC_pw_func_type function;
    int score;
    int glocal;

    /* Make sure evaluation functions update their cache if needed */
    A = TC_update_aln_random_tag(A);

    if (!CL->pw_parameters_set) {
	std::cerr << '\n' << PROGRAM << ": ERROR: CL::pw_parameters_set must be set in pair_wise.\n";		
	return 0;
    }

    CL->overweight = ON;

    function = TC_get_pair_wise_function(CL->pair_wise, CL->dp_mode, &glocal);
    if (CL->get_dp_cost == 0)
	CL->get_dp_cost = TC_get_dp_cost;

    if (strlen(A->seq_al[l_s[0][0]]) == 0 || strlen(A->seq_al[l_s[1][0]]) == 0)
	score = TC_empty_pair_wise(A, ns, l_s, glocal);
    else
	score = function(A, ns, l_s, CL);
    CL->overweight = OFF;

    return score;
}

/**
 * @return a function and a mode (Global, Local, ...)
 */
TC_pw_func_type TC_get_pair_wise_function(TC_pw_func_type pw, char *dp_mode, int *glocal)
{
    static TC_pw_func_type *pwl;
    static char **dpl;
    static int *dps;
    static int npw = 0;
    int a;

    /* The first time: initialize the list of pairwse functions */
    if (npw == 0) {
	pwl = (TC_pw_func_type *)AR_allocate(10, sizeof(TC_pw_func_type));
	dpl = TC_declare_char(10, 100);
	dps = (int *)AR_allocate(10, sizeof(int));

	pwl[npw] = TC_myers_miller_pair_wise;
	strncpy(dpl[npw], "default", 100 - 1);
	dps[npw] = GLOBAL;
	++npw;

	pwl[npw] = TC_gotoh_pair_wise;
	strncpy(dpl[npw], "gotoh_pair_wise", 100 - 1);
	dps[npw] = GLOBAL;
	++npw;

	pwl[npw] = TC_idscore_pair_wise;
	strncpy(dpl[npw], "idscore_pair_wise", 100 - 1);
	dps[npw] = GLOBAL;
	++npw;

	pwl[npw] = TC_myers_miller_pair_wise;
	strncpy(dpl[npw], "myers_miller_pair_wise", 100 - 1);
	dps[npw] = GLOBAL;
	++npw;

	pwl[npw] = TC_fasta_gotoh_pair_wise;
	strncpy(dpl[npw], "fasta_pair_wise", 100 - 1);
	dps[npw] = GLOBAL;
	++npw;

	pwl[npw] = TC_cfasta_gotoh_pair_wise;
	strncpy(dpl[npw], "cfasta_pair_wise", 100 - 1);
	dps[npw] = GLOBAL;
	++npw;

	pwl[npw] = TC_very_fast_gotoh_pair_wise;
	strncpy(dpl[npw], "very_fast_pair_wise", 100 - 1);
	dps[npw] = GLOBAL;
	++npw;

	pwl[npw] = TC_precomputed_pair_wise;
	strncpy(dpl[npw], "precomputed_pair_wise", 100 - 1);
	dps[npw] = GLOBAL;
	++npw;

	pwl[npw] = TC_sim_pair_wise_lalign;
	strncpy(dpl[npw], "sim_pair_wise_lalign", 100 - 1);
	dps[npw] = LALIGN;
	++npw;
    }

    for (a = 0; a < npw; a++) {
	if ((dp_mode && TC_strm(dpl[a], dp_mode)) || pwl[a] == pw) {
	    pw = pwl[a];
	    if (dp_mode)
		strcpy(dp_mode, dpl[a]);
	    glocal[0] = dps[a];
	    return pw;
	}
    }

    std::cerr << '\n' << PROGRAM << ": ERROR: Unknown dp mode '" << dp_mode << "'.\n";

    return 0;
}

static bool TC_split_condition(int nseq, int score, const TC_ConstraintList *CL)
{
    bool cond1 = true, cond2 = true;

    if (CL->split_nseq_thres)
	cond1 = (nseq <= CL->split_nseq_thres);
    if (CL->split_score_thres)
	cond2 = (score >= CL->split_score_thres);

    return (cond1 && cond2);
}


TC_NT_node* AR_seq_tree_aln(TC_NT_node** T, TC_Alignment*& A,
			    TC_ConstraintList*& CL, const char* seq_weight) {
#if USE_RMA
    CL = JZ_weight_constraint_list(CL, seq_weight,
				   *(CL->residue_index.mpi_env));
#else
    CL = TC_weight_constraint_list(CL, seq_weight);
#endif

    return TC_tree_aln((T[3][0])->left, (T[3][0])->right, A, (CL->S)->nseq, CL);
} // AR_seq_tree_aln


TC_NT_node *TC_tree_aln(TC_NT_node LT, TC_NT_node RT, TC_Alignment *A, int nseq, TC_ConstraintList *CL)
{
    static int n_groups_done;
    TC_NT_node P;
    int nseq2align;
    int *translation;
    int *n_s;
    int **l_s;
    int score;
    int do_split;
    int a, b;

    if (n_groups_done == 0) {
/*
  if (g_SNL)
  AR_free(g_SNL);
  g_SNL = (TC_NT_node *)AR_allocate((CL->S)->nseq, sizeof(TC_NT_node));
*/

	if (CL->translation)
	    AR_free(CL->translation);
	CL->translation = (int *)AR_allocate((CL->S)->nseq, sizeof(int));

	for (a = 0; a < (CL->S)->nseq; ++a) {
	    CL->translation[a] = TC_name_is_in_list((CL->S)->name[a], (CL->S)->name, (CL->S)->nseq, 100);
	}

	if (nseq > 2)
	    std::cout << "\nPROGRESSIVE ALIGNMENT (Tree Based).....\n";
	else
	    std::cout << "\nPAIRWISE ALIGNMENT (No Tree).....";

	n_groups_done = (CL->S)->nseq;
	A = TC_reorder_aln(A, (CL->S)->name, (CL->S)->nseq);
	A->nseq = nseq;
    }

    translation = CL->translation;
    n_s = (int *)AR_allocate(2, sizeof(int));
    l_s = TC_declare_int(2, nseq);

    if (nseq == 2) {
	n_s[0] = n_s[1] = 1;
	l_s[0][0] = TC_name_is_in_list((CL->S)->name[0], (CL->S)->name, (CL->S)->nseq, 100);
	l_s[1][0] = TC_name_is_in_list((CL->S)->name[1], (CL->S)->name, (CL->S)->nseq, 100);

	A->score_aln = score = TC_pair_wise(A, n_s, l_s, CL);
	std::cout << "[Done]\n";

	AR_free(n_s);
	TC_free_int(l_s);
    }
    else {
	if (RT->parent != LT->parent)
	    std::cerr << '\n' << PROGRAM << ": WARNING: Tree Pb.\n";
	else
	    P = RT->parent;

	if (LT->leaf == 1 && RT->leaf == 0) {
	    TC_tree_aln(RT->left, RT->right, A, nseq, CL);
	}

	else if (RT->leaf == 1 && LT->leaf == 0) {
	    TC_tree_aln(LT->left, LT->right, A, nseq, CL);
	}

	else if (RT->leaf == 0 && LT->leaf == 0) {
	    TC_tree_aln(LT->left, LT->right, A, nseq, CL);
	    TC_tree_aln(RT->left, RT->right, A, nseq, CL);
	}

	if (LT->leaf == 1 && RT->leaf == 1) {
    	    print_A(A);
	    print_CL(CL);
	    
	    // 1: Identify the two groups of sequences to align
	    nseq2align = LT->nseq + RT->nseq;

	    n_s[0] = LT->nseq;
	    for (a = 0; a < LT->nseq; ++a)
		l_s[0][a] = translation[LT->lseq[a]];

	    if (LT->nseq == 1)
		LT->group = l_s[0][0];

	    n_s[1] = RT->nseq;
	    for (a = 0; a < RT->nseq; ++a)
		l_s[1][a] = translation[RT->lseq[a]];

	    if (RT->nseq == 1)
		RT->group = l_s[1][0];

	    P->group = n_groups_done++;

// 	    std::cout << "\t" << " " << LT->name << " vs. " << RT->name
// 		      << " Group " << P->group + 1 << ": Group " << LT->group + 1 << " (" << n_s[0]
// 		      << " seq) with Group " << RT->group + 1 << " (" << n_s[1] << " seq)  =>  ";

	    P->score = A->score_aln = score = TC_pair_wise(A, n_s, l_s, CL);

	    // score = TC_sub_aln2sub_aln_score(A, CL, n_s, l_s);
	    // std::cout << "score= " << score << "  len= " << strlen(A->seq_al[l_s[0][0]]) << '\n';

   	    print_A(A);
	    print_CL(CL);

	    if (nseq2align == nseq) {
		for (b = 0, a = 0; a < n_s[0]; ++a, ++b)
		    strncpy(A->tree_order[b], (CL->S)->name[l_s[0][a]], STRING - 1);
		for (a = 0; a < n_s[1]; ++a, ++b)
		    strncpy(A->tree_order[b], (CL->S)->name[l_s[1][a]], STRING - 1);
		
		n_groups_done = 0;

		std::cout << ".....[Done]\n";
	    }
	}

 	if (P->parent)
 	    P->leaf = 1;
 	if (LT->isseq == 0)
 	    LT->leaf = 0;
 	if (RT->isseq == 0)
 	    RT->leaf = 0;

// 	if (RT->isseq) {
// 	    g_SNL[translation[RT->lseq[0]]] = RT;
// 	    RT->score = 100;
// 	}
// 	if (LT->isseq) {
// 	    g_SNL[translation[LT->lseq[0]]] = LT;
// 	    LT->score = 100;
// 	}

//	do_split = TC_split_condition(nseq2align, A->score_aln, CL);
//	if (CL->split && do_split) {
//			for (a = 0; a < P->nseq; a++)
//				g_SNL[CL->translation[P->lseq[a]]] = 0;
//			g_SNL[CL->translation[RT->lseq[0]]] = P;
//	}

	AR_free(n_s);
	TC_free_int(l_s);
    }

//	return g_SNL;
    return 0;
}
