/*
 * File: evaluate.cpp
 * Created: 06.06.2006
 * Last Modified: 11.09.2006
 * Author: Adrian Rospondek
 */


#include <iostream>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include "JZ_residue_cache.hpp"
#include "TC_Alignment.hpp"
#include "TC_ConstraintList.hpp"
#include "TC_DistanceMatrix.hpp"
#include "TC_Profile.hpp"
#include "TC_Sequences.hpp"
#include "evaluate.hpp"
#include "globals.hpp"
#include "input.hpp"
#include "util.hpp"
#include "util_algo.hpp"
#include "util_mem.hpp"
#include "util_seq.hpp"


static ResidueCache CACHE;


/**
 * @brief Evaluates the score for matching residue S1(r1) wit residue S2(r2) using matrix CL->M.
 */
int TC_evaluate_matrix_score(TC_ConstraintList *CL, int s1, int r1, int s2, int r2)
{
	int score = 0;

	if (TC_seq2R_template_profile(CL->S, s1) || TC_seq2R_template_profile(CL->S, s2))
		score = TC_evaluate_aln_profile_score(CL, s1, r1, s2, r2);
	else if (r1 > 0 && r2 > 0) {
		--r1;
		--r2;
		score = CL->M[(CL->S)->seq[s1][r1] - 'A'][(CL->S)->seq[s2][r2] - 'A'] * SCORE_K;
	}

	return score;
}

int TC_evaluate_cdna_matrix_score(TC_ConstraintList *CL, int s1, int r1, int s2, int r2)
{
	char a1, a2;

	if (r1 > 0 && r2 > 0) {
		r1--;
		r2--;

		a1 = TC_translate_dna_codon((CL->S)->seq[s1] + r1, 'x');
		a2 = TC_translate_dna_codon((CL->S)->seq[s2] + r2, 'x');

		if (a1 == 'x' || a2 == 'x')
			return 0;
		else
			return CL->M[a1 - 'A'][a2 - 'A'] * SCORE_K;
	}
	else {
		return 0;
	}
}

/**
 * @brief Generic profile function.
 */
static int TC_generic_evaluate_profile_score(TC_ConstraintList *CL, TC_Alignment *profile1, int s1,
	int r1, TC_Alignment *profile2, int s2, int r2, TC_prf_cost_func_type prf_prf)
{
	static int *dummy;
	int *prf1, *prf2;
	int score;

	if (!dummy) {
		dummy = (int *)AR_allocate(10, sizeof(int));
		dummy[0] = 1;		// Number of amino acid types on colum
		dummy[1] = 5;		// Length of dummy
		dummy[3] = '\0';	// Amino acid
		dummy[4] = 1;		// Number of occurences
		dummy[5] = 100;		// Frequency in the MSA column
	}

	if (r1 > 0 && r2 > 0) {
		--r1;
		--r2;

		prf1 = (profile1) ? (profile1->P)->count2[r1] : 0;
		prf2 = (profile2) ? (profile2->P)->count2[r2] : 0;

		if (!prf1) {
			prf1 = dummy;
			prf1[3] = (CL->S)->seq[s1][r1];
		}
		else if (!prf2) {
			prf2 = dummy;
			prf2[3] = (CL->S)->seq[s2][r2];
		}

		score = (prf_prf ? prf_prf : TC_cw_profile_profile)(prf1, prf2, CL);
	}
	else
		score = 0;
	
	return score;
}

int TC_evaluate_aln_profile_score(TC_ConstraintList *CL, int s1, int r1, int s2, int r2)
{
	return TC_generic_evaluate_profile_score(CL, TC_seq2R_template_profile(CL->S, s1), s1, r1,
		TC_seq2R_template_profile(CL->S, s2), s2, r2, CL->profile_mode);
}


#if !USE_RMA
/**
 * @brief Computes the extended score for aligning residue seq1(r1) Vs seq2(r2).
 * Computes: matrix_score, non extended score, extended score.
 * The extended score depends on the function index_res_constraint_list.
 * This function can compare a sequence with itself.
 * This is the generic function -> works with everything should be gradually phased out.
 */
int TC_extend_residue_pair(TC_ConstraintList *CL, int s1, int r1, int s2, int r2)
{
	static int **hasch;
	static int *entry;
	double score = 0;
	int **r;
	int field;
	int a, t_s, t_r, p;

	field = CL->weight_field;

	if (r1 <= 0 || r2 <= 0)
		return 0;
	else if (!CL->L && CL->M) {
		return TC_evaluate_matrix_score(CL, s1, r1, s2, r2);
	}
	else if (!CL->extend_jit) {
		if (!entry)
			entry = (int *)AR_allocate(TC_ConstraintList::entry_len, sizeof(int));
		entry[SEQ1] = s1;
		entry[SEQ2] = s2;
		entry[R1] = r1;
		entry[R2] = r2;
		r = TC_main_search_in_list_constraint(entry, &p, 4, CL);
		if (r == 0)
			return 0;
		else
			return r[0][field];
	}
	else {
		if (!hasch) {
			hasch = (int **)AR_allocate((CL->S)->nseq, sizeof(int *));
			for (a = 0; a < (CL->S)->nseq; a++)
				hasch[a] = (int *)AR_allocate((CL->S)->len[a] + 1, sizeof(int));
		}

		CL = TC_index_res_constraint_list(CL, field);

		hasch[s1][r1] = 100000;
		for (a = 1; a < CL->residue_index[s1][r1][0]; a += 3) {
			t_s = CL->residue_index[s1][r1][a];
			t_r = CL->residue_index[s1][r1][a + 1];
			hasch[t_s][t_r] = CL->residue_index[s1][r1][a + 2];
		}

		for (a = 1; a < CL->residue_index[s2][r2][0]; a += 3) {
			t_s = CL->residue_index[s2][r2][a];
			t_r = CL->residue_index[s2][r2][a + 1];
			if (hasch[t_s][t_r]) {
				if (field == WE)
					score += MIN(hasch[t_s][t_r], CL->residue_index[s2][r2][a + 2]);

			}
		}
		score = (CL->normalise) ? ((score * CL->normalise) / CL->max_ext_value) : score;
		for (a = 1; a < CL->residue_index[s1][r1][0]; a += 3) {
			t_s = CL->residue_index[s1][r1][a];
			t_r = CL->residue_index[s1][r1][a + 1];
			hasch[t_s][t_r] = 0;
		}
		hasch[s1][r1] = hasch[s2][r2] = 0;

		return (int)(score * SCORE_K);
	}
}

int **TC_clean_residue_pair_hasch(int s1, int r1, int s2, int r2, int **hasch,
					 TC_ConstraintList *CL)
{
	int a, t_s, t_r;

	if (!hasch)
		return hasch;
	for (a = 1; a < CL->residue_index[s1][r1][0]; a += 3) {
		t_s = CL->residue_index[s1][r1][a];
		t_r = CL->residue_index[s1][r1][a + 1];
		hasch[t_s][t_r] = 0;
	}
	hasch[s1][r1] = hasch[s2][r2] = 0;
	return hasch;
}

#endif // !USE_RMA

/**
 * @brief Computes the extended score for aligning residue seq1(r1) Vs seq2(r2).
 * Computes: matrix_score, non extended score, extended score.
 * The extended score depends on the function index_res_constraint_list.
 * This function can compare a sequence with itself.
 * This is self normalized function.
 */
#if !USE_RMA
int TC_residue_pair_extended_list(TC_ConstraintList* CL, int s1, int r1, int s2, int r2)
{
	static int **hasch;
	static int max_len;
	int i_score = 0;
	double score = 0;
	double max_score = 0;
	double max_val = 0;
	int field;
	int a, t_s, t_r;

	field = CL->weight_field;

	if (r1 <= 0 || r2 <= 0)
		return 0;


	if (!hasch || max_len != (CL->S)->max_len) {
		max_len = (CL->S)->max_len;
		if (hasch) TC_free_int(hasch);
		hasch = TC_declare_int((CL->S)->nseq, (CL->S)->max_len + 1);
	}

	CL = TC_index_res_constraint_list(CL, field);

	/* Check matches for R1 in the indexed lib */
	hasch[s1][r1] = FORBIDEN;
	for (a = 1; a < CL->residue_index[s1][r1][0]; a += 3) {
		t_s = CL->residue_index[s1][r1][a];
		t_r = CL->residue_index[s1][r1][a + 1];
		int w = CL->residue_index[s1][r1][a + 2];
		hasch[t_s][t_r] = CL->residue_index[s1][r1][a + 2];
		max_score += CL->residue_index[s1][r1][a + 2];
	}

	/* Check Matches for r1 <-> r2 in the indexed lib */
	for (a = 1; a < CL->residue_index[s2][r2][0]; a += 3) {
		t_s = CL->residue_index[s2][r2][a];
		t_r = CL->residue_index[s2][r2][a + 1];

		if (hasch[t_s][t_r]) {
			if (hasch[t_s][t_r] == FORBIDEN) {
				score += CL->residue_index[s2][r2][a + 2];
				max_score += CL->residue_index[s2][r2][a + 2];
			}
			else {
				double delta;

				delta = MIN(hasch[t_s][t_r], CL->residue_index[s2][r2][a + 2]);
				score += delta;
				max_score -= hasch[t_s][t_r];
				max_score += delta;
				max_val = MAX(max_val, delta);
			}
		}
		else {
			max_score += CL->residue_index[s2][r2][a + 2];
		}
	}

	max_score -= hasch[s2][r2];
	TC_clean_residue_pair_hasch(s1, r1, s2, r2, hasch, CL);

	if (max_score == 0)
		score = 0;
	else if (CL->normalise) {
		score = ((score * CL->normalise) / max_score) * SCORE_K;
		if (max_val >= 1000 && CL->overweight == ON)
			score *= max_val / (double)CL->normalise;
	}

	i_score = (int)score;

	return i_score;
}
#endif // !USE_RMA


int TC_get_dp_cost(TC_Alignment *A, int **pos1, int ns1, int *list1, int col1, int **pos2, int ns2,
	int *list2, int col2, TC_ConstraintList *CL)
{
	int MODE = 0;
	int score;

	if (A == 0)
		return 0;

	if (MODE != 2 || MODE == 0 || (!CL->L && CL->M) || (!CL->L && 0/*CL->T*/) || ns1 == 1 || ns2 == 1)
		score = TC_slow_get_dp_cost(A, pos1, ns1, list1, col1, pos2, ns2, list2, col2, CL);
	else if (MODE == 1 || MODE == 2)
		score = TC_fast_get_dp_cost(A, pos1, ns1, list1, col1, pos2, ns2, list2, col2, CL);
	else
		score = 0;

	return (score == TC_UNDEFINED) ? TC_UNDEFINED : (score - SCORE_K * CL->nomatch);
}


#if !USE_RMA
/* WARNING: WORKS ONLY WITH LIST TO EXTEND */
int TC_fast_get_dp_cost(TC_Alignment *A, int **pos1, int ns1, int *list1, int col1, int **pos2,
	int ns2, int *list2, int col2, TC_ConstraintList *CL)
{
	static int **hasch1;
	static int **hasch2;
	static int **n_hasch1;
	static int **n_hasch2;
	static int **is_in_col1;
	static int **is_in_col2;
	double score = 0;
	int n_gap1 = 0;
	int n_gap2 = 0;
	int s1, rs1, r1, t_r, t_s, s2, rs2, r2;
	int a, b;

	CL = TC_index_res_constraint_list(CL, WE);

	if (!hasch1) {
		hasch1 = TC_declare_int((CL->S)->nseq, (CL->S)->max_len + 1);
		hasch2 = TC_declare_int((CL->S)->nseq, (CL->S)->max_len + 1);
		n_hasch1 = TC_declare_int((CL->S)->nseq, (CL->S)->max_len + 1);
		n_hasch2 = TC_declare_int((CL->S)->nseq, (CL->S)->max_len + 1);
		is_in_col1 = TC_declare_int((CL->S)->nseq, (CL->S)->max_len + 1);
		is_in_col2 = TC_declare_int((CL->S)->nseq, (CL->S)->max_len + 1);
	}

	for (a = 0; a < ns1; a++) {
		rs1 = list1[a];
		s1 = A->order[rs1][0];
		r1 = pos1[rs1][col1];

		if (r1 < 0)
			n_gap1++;
		else {
			is_in_col1[s1][r1] = 1;
			for (b = 1; b < CL->residue_index[s1][r1][0]; b += 3) {
				t_s = CL->residue_index[s1][r1][b];
				t_r = CL->residue_index[s1][r1][b + 1];
				hasch1[t_s][t_r] += CL->residue_index[s1][r1][b + 2];
				n_hasch1[t_s][t_r]++;
			}
		}
	}

	for (a = 0; a < ns2; a++) {
		rs2 = list2[a];
		s2 = A->order[rs2][0];
		r2 = pos2[rs2][col2];

		if (r2 < 0)
			n_gap2++;
		else {
			is_in_col2[s2][r2] = 1;
			for (b = 1; b < CL->residue_index[s2][r2][0]; b += 3) {
				t_s = CL->residue_index[s2][r2][b];
				t_r = CL->residue_index[s2][r2][b + 1];

				hasch2[t_s][t_r] += CL->residue_index[s2][r2][b + 2];
				n_hasch2[t_s][t_r]++;
			}
		}
	}

	if (ns2 < ns1) {
		for (a = 0; a < ns2; a++) {
			rs2 = list2[a];
			s2 = A->order[rs2][0];
			r2 = pos1[rs2][col2];

			if (r2 < 0);
			else {
				for (b = 1; b < CL->residue_index[s2][r2][0]; b += 3) {
					t_s = CL->residue_index[s2][r2][b];
					t_r = CL->residue_index[s2][r2][b + 1];

					if (hasch2[t_s][t_r] && hasch1[t_s][t_r] && !(is_in_col1[t_s][t_r]
							|| is_in_col2[t_s][t_r])) {
						score +=
							MIN(hasch2[t_s][t_r] * (n_hasch1[t_s][t_r]),
							hasch1[t_s][t_r] * (n_hasch2[t_s][t_r]));
					}
					else if (hasch2[t_s][t_r] && is_in_col1[t_s][t_r]) {
						score += hasch2[t_s][t_r] * (n_hasch1[t_s][t_r] + 1);
					}
					else if (hasch1[t_s][t_r] && is_in_col2[t_s][t_r]) {
						score += hasch1[t_s][t_r] * (n_hasch2[t_s][t_r] + 1);
					}
					hasch2[t_s][t_r] = 0;
					n_hasch2[t_s][t_r] = 0;
				}
				hasch2[s2][r2] = 0;
				is_in_col2[s2][r2] = 0;
			}
		}

		for (a = 0; a < ns1; a++) {
			rs1 = list1[a];
			s1 = A->order[rs1][0];
			r1 = pos1[rs1][col1];

			if (r1 < 0);
			else {
				is_in_col1[s1][r1] = 0;
				hasch1[s1][r1] = 0;
				for (b = 1; b < CL->residue_index[s1][r1][0]; b += 3) {
					t_s = CL->residue_index[s1][r1][b];
					t_r = CL->residue_index[s1][r1][b + 1];

					hasch1[t_s][t_r] = 0;
					n_hasch1[t_s][t_r] = 0;
				}
			}
		}
	}
	else {
		for (a = 0; a < ns1; a++) {
			rs1 = list1[a];
			s1 = A->order[rs1][0];
			r1 = pos1[rs1][col1];

			if (r1 < 0);
			else {
				for (b = 1; b < CL->residue_index[s1][r1][0]; b += 3) {
					t_s = CL->residue_index[s1][r1][b];
					t_r = CL->residue_index[s1][r1][b + 1];

					if (hasch1[t_s][t_r] && hasch2[t_s][t_r] && !(is_in_col2[t_s][t_r]
							|| is_in_col1[t_s][t_r])) {
						score +=
							MIN(hasch1[t_s][t_r] * (n_hasch2[t_s][t_r]),
							hasch2[t_s][t_r] * (n_hasch1[t_s][t_r]));
					}
					else if (hasch1[t_s][t_r] && is_in_col2[t_s][t_r]) {
						score += hasch1[t_s][t_r] * (n_hasch2[t_s][t_r] + 1);
					}
					else if (hasch2[t_s][t_r] && is_in_col1[t_s][t_r]) {
						score += hasch2[t_s][t_r] * (n_hasch1[t_s][t_r] + 1);
					}
					hasch1[t_s][t_r] = 0;
					n_hasch1[t_s][t_r] = 0;
				}
				hasch1[s1][r1] = 0;
				is_in_col1[s1][r1] = 0;
			}
		}

		for (a = 0; a < ns2; a++) {
			rs2 = list2[a];
			s2 = A->order[rs2][0];
			r2 = pos1[rs2][col2];

			if (r2 < 0);
			else {
				is_in_col2[s2][r2] = 0;
				hasch1[s2][r2] = 0;
				for (b = 1; b < CL->residue_index[s2][r2][0]; b += 3) {
					t_s = CL->residue_index[s2][r2][b];
					t_r = CL->residue_index[s2][r2][b + 1];

					hasch2[t_s][t_r] = 0;
					n_hasch2[t_s][t_r] = 0;
				}
			}
		}
	}
	score = (score * SCORE_K) / ((ns1 - n_gap1) * (ns2 - n_gap2));
	score = (CL->normalise) ? ((score * CL->normalise) / CL->max_ext_value) : score;

	return (int)(score - SCORE_K * CL->nomatch);
}
#endif // !USE_RMA


int TC_fast_get_dp_cost_2(TC_Alignment *A, int **pos1, int ns1, int *list1, int col1, int **pos2,
	int ns2, int *list2, int col2, TC_ConstraintList *CL)
{
	static int n_pair;
	static int last_ns1;
	static int last_ns2;
	static int last_top1;
	static int last_top2;
	static int **check_list;
	double score = 0;
	int s, s1, s2, r1, r2, rs1, rs2;
	int res_res = 0;
	int a, b, tmp;

	if (last_ns1 == ns1 && last_top1 == list1[0] && last_ns2 == ns2 && last_top2 == list2[0]);
	else {
		last_ns1 = ns1;
		last_ns2 = ns2;
		last_top1 = list1[0];
		last_top2 = list2[0];
		if (check_list)
			TC_free_int(check_list);
		check_list = TC_declare_int((CL->S)->nseq * (CL->S)->nseq, 3);

		for (n_pair = 0, a = 0; a < ns1; a++) {
			s1 = list1[a];
			rs1 = A->order[s1][0];
			for (b = 0; b < ns2; b++, n_pair++) {
				s2 = list2[b];
				rs2 = A->order[s2][0];
				check_list[n_pair][0] = s1;
				check_list[n_pair][1] = s2;
				check_list[n_pair][2] = (!CL->DM) ? 0 : (CL->DM)->similarity_matrix[rs1][rs2];
			}
			TC_sort_int(check_list, 2, 0, n_pair - 1);
		}
	}

	if (!CL->evaluate_residue_pair) {
		std::cerr << '\n' << PROGRAM << ": WARNING: CL::evaluate_residue_pair is not set.\n"
				  << "\tIt will be automatically set to extend_residue_pair.\n";
		CL->evaluate_residue_pair = TC_extend_residue_pair;
	}

	for (a = n_pair - 1; a >= 0; a--) {
		s1 = check_list[a][0];
		rs1 = A->order[s1][0];
		r1 = pos1[s1][col1];

		s2 = check_list[a][1];
		rs2 = A->order[s2][0];
		r2 = pos2[s2][col2];

		if (r1 > 0 && r2 > 0) {
			res_res++;
		}
		if (rs1 > rs2) {
			SWAPP(rs1, rs2, tmp);
			SWAPP(r1, r2, tmp);
		}

		if ((s = (CL->evaluate_residue_pair) (CL, rs1, r1, rs2, r2)) != TC_UNDEFINED)
			score += s;
		else {
			return TC_UNDEFINED;
		}
		if (res_res >= CL->max_n_pair && CL->max_n_pair != 0)
			a = 0;
	}

	score = (res_res == 0) ? 0 : (score / res_res);
	score = score - SCORE_K * CL->nomatch;

	return (int)score;
}

int TC_slow_get_dp_cost(TC_Alignment *A, int **pos1, int ns1, int *list1, int col1, int **pos2,
	int ns2, int *list2, int col2, TC_ConstraintList *CL)
{
	static int *entry;
	double score = 0;
	int s, s1, s2, r1, r2, rs1, rs2;
	int gap_gap = 0;
	int gap_res = 0;
	int res_res = 0;
	int a, b, tmp;

	if (!CL->evaluate_residue_pair) {
		std::cerr << '\n' << PROGRAM << ": WARNING: CL::evaluate_residue_pair is not set.\n"
				  << "\tIt will be automatically set to extend_residue_pair.\n";
		CL->evaluate_residue_pair = TC_extend_residue_pair;
	}

	if (entry == 0)
		entry = (int *)AR_allocate(TC_ConstraintList::entry_len, sizeof(int));

	for (a = 0; a < ns1; a++) {
		for (b = 0; b < ns2; b++) {
			s1 = list1[a];
			rs1 = A->order[s1][0];
			r1 = pos1[s1][col1];
			s2 = list2[b];
			rs2 = A->order[s2][0];
			r2 = pos2[s2][col2];

			if (rs1 > rs2) {
				SWAPP(rs1, rs2, tmp);
				SWAPP(r1, r2, tmp);
			}

			if (r1 == 0 && r2 == 0)
				gap_gap++;
			else if (r1 < 0 || r2 < 0)
				gap_res++;
			else {
				res_res++;
				if ((s = (CL->evaluate_residue_pair) (CL, rs1, r1, rs2, r2)) != TC_UNDEFINED)
					score += s;
				else {
					return TC_UNDEFINED;
				}
			}
		}
	}

	score = (res_res == 0) ? 0 : (score / res_res);

	return (int)(score - SCORE_K * CL->nomatch);
}

/* Works only with matrix. */
int TC_consensus_get_dp_cost(TC_Alignment *A, int **pos1, int ns1, int *list1, int col1, int **pos2,
	int ns2, int *list2, int col2, TC_ConstraintList *CL)
{
	static char *seq1, *seq2;
	static int last_tag;
	
	if (last_tag != A->random_tag) {
		int ls1, ls2, lenls1, lenls2;
		
		last_tag = A->random_tag;
		AR_free(seq1);
		AR_free(seq2);
		
		seq1 = TC_sub_aln2cons_seq_mat(A, ns1, list1, "blosum62mt");
		seq2 = TC_sub_aln2cons_seq_mat(A, ns2, list2, "blosum62mt");
		ls1 = list1[ns1 - 1];
		ls2 = list2[ns2 - 1];
		lenls1 = strlen(A->seq_al[ls1]);
		lenls2 = strlen(A->seq_al[ls2]);
	}
	
	return (CL->M[seq1[col1] - 'A'][seq2[col2] - 'A'] * SCORE_K) - SCORE_K * CL->nomatch;
}

int TC_cw_profile_get_dp_cost(TC_Alignment *A, int **pos1, int ns1, int *list1, int col1,
	int **pos2, int ns2, int *list2, int col2, TC_ConstraintList *CL)
{
	static int last_tag;
	static int **cons1, **cons2;
	int score;

	if (last_tag != A->random_tag) {
		last_tag = A->random_tag;
		TC_free_int(cons1);
		TC_free_int(cons2);
		cons1 = TC_sub_aln2count_mat2(A, ns1, list1);
		cons2 = TC_sub_aln2count_mat2(A, ns2, list2);
	}

	score = TC_cw_profile_profile(cons1[col1], cons2[col2], CL) - SCORE_K * CL->nomatch;

	return score;
}

TC_ConstraintList *TC_choose_extension_mode(const char *extend_mode, TC_ConstraintList *CL)
{
	if (!CL) {
		return CL;
	}
	else if (TC_strm(extend_mode, "triplet") && !CL->M) {
		CL->evaluate_residue_pair = TC_residue_pair_extended_list;
		CL->get_dp_cost = TC_get_dp_cost;
	}
	else if (TC_strm(extend_mode, "fast_triplet") && !CL->M) {
		CL->evaluate_residue_pair = TC_residue_pair_extended_list;
		CL->get_dp_cost = TC_fast_get_dp_cost;
	}
	else if (TC_strm(extend_mode, "very_fast_triplet") && !CL->M) {
		CL->evaluate_residue_pair = TC_residue_pair_extended_list;
		CL->get_dp_cost = TC_fast_get_dp_cost_2;
	}
	else if (TC_strm(extend_mode, "slow_triplet") && !CL->M) {
		CL->evaluate_residue_pair = TC_residue_pair_extended_list;
		CL->get_dp_cost = TC_slow_get_dp_cost;
	}
	else if (TC_strm(extend_mode, "matrix") || CL->M) {
		CL->evaluate_residue_pair = TC_evaluate_matrix_score;
		CL->get_dp_cost = TC_cw_profile_get_dp_cost;
		CL->normalise = 1;
	}
	else {
		std::cerr << '\n' << PROGRAM << ": ERROR: Unknown extend mode '" << extend_mode << "'.\n";
	}

	return CL;
}

int TC_cw_profile_profile(const int *prf1, const int *prf2, TC_ConstraintList *CL)
{
	double score = 0;
	int res1, res2;
	int a, b;

	for (a = 3; a < prf1[1]; a += 3)
		for (b = 3; b < prf2[1]; b += 3) {
			res1 = prf1[a];
			res2 = prf2[b];

			score += prf1[a + 2] * prf2[b + 2] * CL->M[res1 - 'A'][res2 - 'A'];
		}
	score = (score * SCORE_K) / 10000;

	return (int)score;
}

int TC_muscle_profile_profile(const int *prf1, const int *prf2, TC_ConstraintList *CL)
{
	double score = 0, fg1, fg2, fi, fj, m;
	int res1, res2;
	int a, b;

	for (a = 3; a < prf1[1]; a += 3) {
		res1 = prf1[a];
		fi = (double)prf1[a + 2] / 100;

		for (b = 3; b < prf2[1]; b += 3) {
			res2 = prf2[b];
			fj = (double)prf2[b + 2] / 100;
			m = exp((double)CL->M[res1 - 'A'][res2 - 'A']);
			score += m * fi * fj;
		}
	}
	fg1 = (double)prf1[2] / 100;
	fg2 = (double)prf2[2] / 100;
	score = log(score) * (1 - fg1) * (1 - fg2);
	score = (score * SCORE_K);

	return (int)score;
}

int TC_channel_profile_profile(const int *prf1, const int *prf2, TC_ConstraintList *CL)
{
	int score = 0;

	prf1 += prf1[1];
	prf2 += prf2[1];

	if (prf1[0] != prf1[0]) {
		std::cerr << '\n' << PROGRAM << ": ERROR: Inconsistent number of channels (channel_profile_profile).\n";
	}
	else {
		int a, n;

		for (a = 1, n = 0; a <= prf1[0]; a++) {
			if (prf1[a] > 0 && prf2[a] > 0) {
				n++;
				score += CL->M[prf1[a] - 'A'][prf2[a] - 'A'];
			}
		}

		if (n == 0)
			return 0;

		score = (n == 0) ? 0 : (score * SCORE_K) / n;
	}

	return score;
}

TC_prf_cost_func_type TC_get_profile_mode_function(char *name, TC_prf_cost_func_type func)
{
	static unsigned int nfunc = 0;
	static TC_prf_cost_func_type *flist;
	static char **nlist;

	/* The first time: initialize the list of cost functions */
	if (nfunc == 0) {
		flist = (TC_prf_cost_func_type *)AR_allocate(3, sizeof(TC_prf_cost_func_type));
		nlist = TC_declare_char(3, 50);

		flist[nfunc] = TC_cw_profile_profile;
		strncpy(nlist[nfunc], "cw_profile_profile", 50 - 1);
		nfunc++;

		flist[nfunc] = TC_muscle_profile_profile;
		strncpy(nlist[nfunc], "muscle_profile_profile", 50 - 1);
		nfunc++;

		flist[nfunc] = TC_channel_profile_profile;
		strncpy(nlist[nfunc], "channel_profile_profile", 50 - 1);
		nfunc++;
	}

	for (unsigned int a = 0; a < nfunc; ++a) {
		if ((name && TC_strm(nlist[a], name)) || flist[a] == func) {
			if (name)
				strcpy(name, nlist[a]);
			
			return flist[a];
		}
	}
	
	std::cerr << '\n' << PROGRAM << ": ERROR: Unknown profile-profile function '" << name << "'.\n";

	return 0;
}

/* Warning: Does not take gaps into account */
int TC_sub_aln2sub_aln_score(TC_Alignment *A, TC_ConstraintList *CL, int *ns, int **ls)
{
	float score = 0;
	int **pos;
	int a;

	/* Make sure evaluation functions update their cache if needed */
	A = TC_update_aln_random_tag(A);
	pos = TC_aln2pos_simple(A, -1, ns, ls);

	for (a = 0; a < A->len_aln; a++)
		score += CL->get_dp_cost(A, pos, ns[0], ls[0], a, pos, ns[1], ls[1], a, CL);
	TC_free_int(pos);

	score = (int)(score / (A->len_aln * SCORE_K));
	score = (int)(CL->L && CL->normalise) ? ((score * MAXID) / (CL->normalise)) : score;

	return (int)score;
}

static TC_Alignment *TC_main_coffee_evaluate_output2(TC_Alignment *IN, TC_ConstraintList *CL, const char *mode)
{
	// Make sure evaluation functions update their cache if needed
	IN = TC_update_aln_random_tag(IN);

	if (TC_strm(mode, "no"))
		return 0;
	else if (TC_strm(mode, "tcoffee_fast"))
	    return TC_fast_coffee_evaluate_output(IN, CL);
	else if (TC_strm(mode, "tcoffee_slow"))
		return TC_slow_coffee_evaluate_output(IN, CL);
	else if (TC_strm(mode, "matrix"))
		return TC_matrix_evaluate_output(IN, CL);
	else if (TC_strm(mode, "categories"))
		return TC_categories_evaluate_output(IN);
	else if (TC_strm(mode, "sar"))
		return TC_sar_evaluate_output(IN, CL);
	else if (TC_strm(mode, "boxshade"))
		return TC_boxshade_evaluate_output(IN);
	else if (CL->evaluate_residue_pair == TC_evaluate_matrix_score || CL->ne == 0)
		return TC_matrix_evaluate_output(IN, CL);
	else {
		std::cerr << '\n' << PROGRAM << ": ERROR: Unknown mode of alignment evaluation '" << mode << "'.\n";
		return 0;
	}
}

TC_Alignment *TC_main_coffee_evaluate_output(TC_Alignment *IN, TC_ConstraintList *CL, const char *mode)
{
	TC_Alignment *TopS = 0, *LastS = 0, *CurrentS = 0;

	if (IN->A)
		IN = IN->A;

	while (IN) {
		CurrentS = TC_main_coffee_evaluate_output2(IN, CL, mode);

		if (!TopS)
			LastS = TopS = CurrentS;
		else {
			LastS->A = CurrentS;
			LastS = CurrentS;
		}
		IN = IN->A;
	}

	return TopS;
}

TC_Alignment *TC_categories_evaluate_output(const TC_Alignment *IN)
{
	TC_Alignment *OUT = 0;
	float score, nseq2, tot_aln;
	float n;
	int *aa;
	int a, b, r;

	/* Residue x: sum of observed extended X../sum of possible X.. */
	OUT = TC_copy_aln(IN, OUT);
	aa = (int *)AR_allocate(26, sizeof(int));
	nseq2 = IN->nseq * IN->nseq;

	for (tot_aln = 0, a = 0; a < IN->len_aln; a++) {
		for (n = 0, b = 0; b < IN->nseq; b++) {
			r = IN->seq_al[b][a];
			if (TC_is_gap(r))
				++n;
			else {
				aa[tolower(r) - 'a']++;
				++n;
			}
		}
		n = n * n;
		for (score = 0, b = 0; b < 26; b++) {
			score += aa[b] * aa[b];
			aa[b] = 0;
		}
		
		score = (n == 0) ? 0 : score / n;
		tot_aln += score;
		r = (int)(score * 10);
		r = (r >= 10) ? 9 : r;
		OUT->seq_al[OUT->nseq][a] = '0' + r;
	}
	OUT->score_aln = (int)((tot_aln / OUT->len_aln) * 100);
	strncpy(OUT->name[IN->nseq], "cons", STRING - 1);

	AR_free(aa);

	return OUT;
}

TC_Alignment *TC_matrix_evaluate_output(const TC_Alignment *IN, TC_ConstraintList *CL)
{
	TC_Alignment *OUT = 0;
	double **tot_res;
	double **max_res;
	double **tot_seq;
	double **max_seq;
	double **tot_col;
	double **max_col;
	double max_aln = 0;
	double tot_aln = 0;
	int a, b, c, r, s, r1, r2;

	if (!CL->M)
		CL->M = TC_read_matrice("blosum62mt");
	OUT = TC_copy_aln(IN, OUT);

	tot_res = TC_declare_double(IN->nseq, IN->len_aln);
	max_res = TC_declare_double(IN->nseq, IN->len_aln);
	tot_seq = TC_declare_double(IN->nseq, 1);
	max_seq = TC_declare_double(IN->nseq, 1);
	tot_col = TC_declare_double(IN->len_aln, 1);
	max_col = TC_declare_double(IN->len_aln, 1);
	max_aln = tot_aln = 0;

	for (a = 0; a < IN->len_aln; a++) {
		for (b = 0; b < IN->nseq; b++) {
			r1 = tolower(IN->seq_al[b][a]);
			if (TC_is_gap(r1))
				continue;
			r = CL->M[r1 - 'A'][r1 - 'A'];
			for (c = 0; c < IN->nseq; c++) {
				r2 = tolower(IN->seq_al[c][a]);
				if (b == c || TC_is_gap(r2))
					continue;
				s = CL->M[r2 - 'A'][r1 - 'A'];
				s = (s < 0) ? 0 : s;
				tot_res[b][a] += s;
				max_res[b][a] += r;
				tot_col[a][0] += s;
				max_col[a][0] += r;
				tot_seq[b][0] += s;
				max_seq[b][0] += r;
				tot_aln += s;
				max_aln += r;
			}
		}
	}
	for (a = 0; a < IN->nseq; a++) {
		if (!max_seq[a][0])
			continue;
		OUT->score_seq[a] = (int)((max_seq[a][0] * 100) / max_seq[a][0]);
		for (b = 0; b < IN->len_aln; b++) {
			r1 = IN->seq_al[a][b];
			if (TC_is_gap(r1) || !max_res[a][b])
				continue;
			r1 = (int)((tot_res[a][b] * 10) / max_res[a][b]);
			r1 = (r1 >= 10) ? 9 : r1;
			r1 = r1 < 0 ? 0 : r1;
			OUT->seq_al[a][b] = r1 + '0';
		}
	}
	for (a = 0; a < IN->len_aln; a++) {
		r1 = (max_col[a][0] == 0) ? 0 : (int)((tot_col[a][0] * 10) / max_col[a][0]);
		r1 = (r1 >= 10) ? 9 : r1;
		OUT->seq_al[OUT->nseq][a] = r1 + '0';
	}
	strncpy(OUT->name[IN->nseq], "cons", STRING - 1);
	if (max_aln)
		OUT->score_aln = (int)((100 * tot_aln) / max_aln);

	TC_free_double(tot_res);
	TC_free_double(max_res);
	TC_free_double(tot_seq);
	TC_free_double(max_seq);
	TC_free_double(tot_col);
	TC_free_double(max_col);

	return OUT;
}

TC_Alignment *TC_sar_evaluate_output(const TC_Alignment *IN, TC_ConstraintList *CL)
{
	TC_Alignment *OUT = 0;
	double **tot_res;
	double **max_res;
	double **tot_seq;
	double **max_seq;
	double **tot_col;
	double **max_col;
	double max_aln = 0;
	double tot_aln = 0;
	int a, b, c, r, s, r1, r2;

	if (!CL->M)
		CL->M = TC_read_matrice("blosum62mt");
	OUT = TC_copy_aln(IN, OUT);

	tot_res = TC_declare_double(IN->nseq, IN->len_aln);
	max_res = TC_declare_double(IN->nseq, IN->len_aln);
	tot_seq = TC_declare_double(IN->nseq, 1);
	max_seq = TC_declare_double(IN->nseq, 1);
	tot_col = TC_declare_double(IN->len_aln, 1);
	max_col = TC_declare_double(IN->len_aln, 1);
	max_aln = tot_aln = 0;

	for (a = 0; a < IN->len_aln; a++) {
		for (b = 0; b < IN->nseq; b++) {
			r1 = tolower(IN->seq_al[b][a]);
			for (c = 0; c < IN->nseq; c++) {
				r2 = tolower(IN->seq_al[c][a]);
				if (b == c)
					continue;
				if (TC_is_gap(r1) && TC_is_gap(r2))
					s = 0;
				else
					s = (r1 == r2) ? 1 : 0;
				r = 1;
				tot_res[b][a] += s;
				max_res[b][a] += r;
				tot_col[a][0] += s;
				max_col[a][0] += r;
				tot_seq[b][0] += s;
				max_seq[b][0] += r;
				tot_aln += s;
				max_aln += r;
			}
		}
	}
	for (a = 0; a < IN->nseq; a++) {
		if (!max_seq[a][0])
			continue;
		OUT->score_seq[a] = (int)((max_seq[a][0] * 100) / max_seq[a][0]);
		for (b = 0; b < IN->len_aln; b++) {
			r1 = IN->seq_al[a][b];
			if (TC_is_gap(r1) || !max_res[a][b])
				continue;
			r1 = (int)((tot_res[a][b] * 10) / max_res[a][b]);
			r1 = (r1 >= 10) ? 9 : r1;
			r1 = r1 < 0 ? 0 : r1;
			OUT->seq_al[a][b] = r1 + '0';
		}
	}
	for (a = 0; a < IN->len_aln; a++) {
		r1 = (max_col[a][0] == 0) ? 0 : (int)((tot_col[a][0] * 10) / max_col[a][0]);
		r1 = (r1 >= 10) ? 9 : r1;
		OUT->seq_al[OUT->nseq][a] = r1 + '0';
	}
	strncpy(OUT->name[IN->nseq], "cons", STRING - 1);
	if (max_aln)
		OUT->score_aln = (int)((100 * tot_aln) / max_aln);

	TC_free_double(tot_res);
	TC_free_double(max_res);
	TC_free_double(tot_seq);
	TC_free_double(max_seq);
	TC_free_double(tot_col);
	TC_free_double(max_col);

	return OUT;
}

TC_Alignment *TC_boxshade_evaluate_output(const TC_Alignment *IN)
{
	TC_Alignment *OUT = 0;
	int **aa;
	int r, br, bs, a, b;

	OUT = TC_copy_aln(IN, OUT);
	aa = TC_declare_int(26, 2);

	for (a = 0; a < OUT->len_aln; a++) {
		for (b = 0; b < 26; b++) {
			aa[b][1] = 0;
			aa[b][0] = 'a' + b;
		}
		for (b = 0; b < OUT->nseq; b++) {
			r = tolower(OUT->seq_al[b][a]);
			if (!TC_is_gap(r))
				aa[r - 'a'][1]++;
		}
		TC_sort_int(aa, 1, 0, 25);
		bs = (aa[25][1] * 10) / OUT->nseq;
		br = aa[25][0];
		if (bs == 10)
			bs--;
		bs += '0';
		for (b = 0; b < OUT->nseq; b++) {
			r = tolower(OUT->seq_al[b][a]);
			if (r == br && bs > '5')
				OUT->seq_al[b][a] = bs;
		}
		OUT->seq_al[b][a] = bs;
	}
	strncpy(OUT->name[IN->nseq], "cons", STRING - 1);

	TC_free_int(aa);

	return OUT;
}

TC_Alignment *TC_fast_coffee_evaluate_output(TC_Alignment *IN, TC_ConstraintList *CL)
{
	TC_Alignment *OUT = 0;
	double score_col = 0, score_aln = 0, score_res = 0;
	double max_col, max_aln;
	double *max_seq, *score_seq, **seq_count;
	int **pos;
	int local_m;
	int local_nseq;
	short int *entry = 0;
	int a, b, c, m, res, s, s1, s2, r1, r2;

	/* NORMALIZE: with the highest scoring pair found in the multiple alignment */
	seq_count = TC_declare_double((CL->S)->nseq, (CL->S)->max_len + 1);
	max_seq = (double *)AR_allocate(IN->nseq, sizeof(double));
	score_seq = (double *)AR_allocate(IN->nseq, sizeof(double));

	if (!CL->evaluate_residue_pair) {
		std::cerr << '\n' << PROGRAM << ": WARNING: CL::evaluate_residue_pair is not set.\n"
				  << "\tIt will be automatically set to extend_residue_pair.\n";
		CL->evaluate_residue_pair = TC_extend_residue_pair;
	}

	OUT = TC_copy_aln(IN, OUT);
	pos = TC_aln2pos_simple(IN, IN->nseq);

	for (a = 0; a < CL->ne; a++) {
		entry = TC_extract_entry(entry, a, CL);
		seq_count[entry[SEQ1]][entry[R1]]++;
		seq_count[entry[SEQ2]][entry[R2]]++;
	}

	/* 1: Identify the highest scoring pair within the alignment */
	for (m = 0, a = 0; a < IN->len_aln; a++) {
		for (b = 0; b < IN->nseq; b++) {
			s1 = IN->order[b][0];
			r1 = pos[b][a];
			for (c = 0; c < IN->nseq; c++) {
				s2 = IN->order[c][0];
				r2 = pos[c][a];
				if (s1 == s2 && !CL->do_self)
					continue;
				if (s1 < s2)
					s = (CL->evaluate_residue_pair) (CL, s1, r1, s2, r2);
				else
					s = (CL->evaluate_residue_pair) (CL, s2, r2, s1, r1);
				s = (s != TC_UNDEFINED) ? s : 0;
				m = MAX(m, s);
			}
		}
	}

	local_m = m;
	strncpy(OUT->name[IN->nseq], "cons", STRING - 1);

	for (max_aln = 0, score_aln = 0, a = 0; a < IN->len_aln; a++) {
		OUT->seq_al[IN->nseq][a] = NO_COLOR_RESIDUE;
		for (local_nseq = 0, b = 0; b < IN->nseq; b++)
			local_nseq += (pos[b][a] > 0) ? 1 : 0;
		local_m = m * (local_nseq - 1);
		for (max_col = 0, score_col = 0, b = 0; b < IN->nseq; b++) {
			OUT->seq_al[b][a] = NO_COLOR_RESIDUE;
			s1 = IN->order[b][0];
			r1 = pos[b][a];
			if (r1 <= 0)
				continue;
			for (score_res = 0, c = 0; c < IN->nseq; c++) {
				s2 = IN->order[c][0];
				r2 = pos[c][a];
				if ((s1 == s2 && !CL->do_self) || r2 <= 0)
					continue;
				max_col += m;
				max_seq[b] += m;
				max_aln += m;
				if (s1 < s2)
					s = (CL->evaluate_residue_pair) (CL, s1, r1, s2, r2);
				else
					s = (CL->evaluate_residue_pair) (CL, s2, r2, s1, r1);
				s = (s != TC_UNDEFINED) ? s : 0;
				score_res += s;
				score_col += s;
				score_seq[b] += s;
				score_aln += s;
			}
			res = (local_m == 0) ? NO_COLOR_RESIDUE : (int)((score_res * 10) / local_m);
			OUT->seq_al[b][a] = (res == NO_COLOR_RESIDUE) ? res : MIN(res, 9);
		}
		res = (max_col == 0) ? NO_COLOR_RESIDUE : (int)((score_col * 10) / max_col);
		OUT->seq_al[IN->nseq][a] = (res == NO_COLOR_RESIDUE) ? res : MIN(res, 9);
	}

	IN->score_aln = OUT->score_aln = (max_aln == 0) ? 0 : (int)((score_aln * 100) / max_aln);
	for (a = 0; a < OUT->nseq; a++)
		OUT->score_seq[a] = (max_seq[a] == 0) ? 0 : (int)((score_seq[a] * 100) / max_seq[a]);

	TC_free_double(seq_count);
	AR_free(max_seq);
	AR_free(score_seq);
	TC_free_int(pos);
	AR_free(entry);

	return OUT;
}

TC_Alignment *TC_slow_coffee_evaluate_output(TC_Alignment *IN, TC_ConstraintList *CL)
{
	TC_Alignment *OUT = 0;
	double max_score_r, score_r, max;
	double score_col = 0, score_aln = 0;
	double max_score_col, max_score_aln;
	double *max_score_seq, *score_seq;
	int **pos;
	int ***res_extended_weight;
	int n_res_in_col;
	int a, b, c, res, s, s1, s2, r1, r2;

	if (!CL->evaluate_residue_pair) {
		std::cerr << '\n' << PROGRAM << ": WARNING: CL::evaluate_residue_pair is not set.\n"
				  << "\tIt will be automatically set to extend_residue_pair.\n";
		CL->evaluate_residue_pair = TC_extend_residue_pair;
	}
	OUT = TC_copy_aln(IN, OUT);
	pos = TC_aln2pos_simple(IN, IN->nseq);

	max_score_seq = (double *)AR_allocate(IN->nseq, sizeof(double));
	score_seq = (double *)AR_allocate(IN->nseq, sizeof(double));
	res_extended_weight = (int ***)TC_declare_arrayN(3, sizeof(int), (CL->S)->nseq, (CL->S)->max_len + 1, 2);

	max = (CL->normalise) ? (100 * CL->normalise) * SCORE_K : 100;

	for (a = 0; a < IN->len_aln; a++) {
		for (b = 0; b < IN->nseq - 1; b++) {
			s1 = IN->order[b][0];
			r1 = pos[b][a];
			for (c = b + 1; c < IN->nseq; c++) {
				s2 = IN->order[c][0];
				r2 = pos[c][a];
				if (s1 == s2 && !CL->do_self)
					continue;
				else if (r1 <= 0 || r2 <= 0)
					continue;
				else {
					s = (CL->evaluate_residue_pair) (CL, s1, r1, s2, r2);
					res_extended_weight[s1][r1][0] += s * 100;
					res_extended_weight[s2][r2][0] += s * 100;
					res_extended_weight[s1][r1][1] += (int)max;
					res_extended_weight[s2][r2][1] += (int)max;
				}
			}
		}
	}
	strncpy(OUT->name[IN->nseq], "cons", STRING - 1);
	for (max_score_aln = 0, score_aln = 0, a = 0; a < IN->len_aln; a++) {
		OUT->seq_al[IN->nseq][a] = NO_COLOR_RESIDUE;
		for (n_res_in_col = 0, b = 0; b < IN->nseq; b++)
			n_res_in_col += (pos[b][a] > 0) ? 1 : 0;
		for (max_score_col = 0, score_col = 0, b = 0; b < IN->nseq; b++) {
			OUT->seq_al[b][a] = NO_COLOR_RESIDUE;
			s1 = IN->order[b][0];
			r1 = pos[b][a];
			if (r1 <= 0)
				continue;
			else {
				max_score_r = res_extended_weight[s1][r1][1];
				score_r = res_extended_weight[s1][r1][0];
				if (max_score_r == 0 && n_res_in_col > 1)
					res = 0;
				else if (n_res_in_col == 1)
					res = NO_COLOR_RESIDUE;
				else
					res = (int)((score_r * 10) / max_score_r);
				OUT->seq_al[b][a] = (res == NO_COLOR_RESIDUE) ? res : MIN(res, 9);
				max_score_col += max_score_r;
				score_col += score_r;
				max_score_seq[b] += max_score_r;
				score_seq[b] += score_r;
				max_score_aln += max_score_r;
				score_aln += score_r;
			}
			if (max_score_col == 0 && n_res_in_col > 1)
				res = 0;
			else if (n_res_in_col < 2)
				res = NO_COLOR_RESIDUE;
			else
				res = (int)((score_col * 10) / max_score_col);
			OUT->seq_al[IN->nseq][a] = (res == NO_COLOR_RESIDUE) ? res : MIN(res, 9);
		}
	}
	IN->score_aln = OUT->score_aln = (max_score_aln == 0) ? 0 : (int)((score_aln * 100) / max_score_aln);
	for (a = 0; a < OUT->nseq; a++)
		OUT->score_seq[a] = (max_score_seq[a] == 0) ? 0 : (int)((score_seq[a] * 100) / max_score_seq[a]);

	TC_free_int(pos);
	AR_free(max_score_seq);
	AR_free(score_seq);
	TC_free_arrayN((void *)res_extended_weight, 3);

	return OUT;
}

float TC_get_avg_matrix_mm(int **matrix, const char *alphabet)
{
	float naa;
	float gop;
	int l;
	int a, b;

	l = MIN(20, strlen(alphabet));
	for (naa = 0, gop = 0, a = 0; a < l; a++)
		for (b = 0; b < l; b++)
			if (a != b) {
				gop += matrix[alphabet[a] - 'A'][alphabet[b] - 'A'];
				++naa;
			}

	gop /= naa;

	return gop;
}
