/*
 * File: input.cpp
 * Created: 06.06.2006
 * Last Modified: 19.09.2006
 * Author: Adrian Rospondek
 */


#include <iostream>
#include <sstream>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "AR_Config.hpp"
#include "TC_Alignment.hpp"
#include "TC_ConstraintList.hpp"
#include "TC_Method.hpp"
#include "TC_Sequences.hpp"
#include "TC_Templates.hpp"
#include "TC_Weights.hpp"
#include "TC_tree.hpp"
#include "globals.hpp"
#include "input.hpp"
#include "matrices.hpp"
#include "util.hpp"
#include "util_io.hpp"
#include "util_mem.hpp"
#include "util_seq.hpp"
#include "util_str.hpp"



static int g_distance_tree, g_rooted_tree;
static int g_tot_nseq;



static char TC_identify_format(char **fname)
{
	char mode = '?';

	mode = fname[0][0];
	if ((TC_is_in_set(mode, "ALSPRW") && AR_check_file_exists(fname[0] + 1))
		|| (mode == 'X' && TC_is_matrix(fname[0] + 1))
		|| (mode == 'M' && TC_is_method(fname[0] + 1))) {
		++fname[0];
	}
	else {
		/* WARNING: Order matters => internal methods can be confused with files,
		 * must be checked last */
		if (TC_is_lib(fname[0]))
			mode = 'L';
		else if (TC_is_pdb_file(fname[0]))
			mode = 'P';
		else if (TC_is_seq(fname[0]))
			mode = 'S';
		else if (TC_is_aln(fname[0]))
			mode = 'A';
		else if (TC_is_matrix(fname[0]))
			mode = 'X';
		else if (TC_is_method(fname[0]))
			mode = 'M';
		else
			mode = '?';
	}

	return mode;
}

void AR_identify_list_format(std::vector<std::string>::iterator beg, std::vector<std::string>::iterator end)
{
	char buf[FILENAMELEN + 1];
	const char *string, *s;
	char mode;

	for (; beg != end; ++beg) {
		string = s = beg->c_str();
		mode = TC_identify_format(const_cast<char **>(&s));

		if (mode != '?') {
			snprintf(buf, FILENAMELEN + 1, "%c%s", mode, s);
			beg->assign(buf);
		}
		else
			std::cerr << '\n' << PROGRAM << ": ERROR: '" << string << "' could not be recognised.\n";
	}
}

/**
 * @brief Turns 'file' into 'Sfile', 'Afile', ...
 */
char *TC_name2type_name(const char *name)
{
	char *new_name;
	char mode;

	new_name = (char *)AR_allocate(strlen(name) + 2, sizeof(char));
	strcpy(new_name, name);

	if (TC_is_in_set(name[0], "ALSMXPRW") && !AR_check_file_exists(name))
		;
	else {
		mode = TC_identify_format(&new_name);
		sprintf(new_name, "%c%s", mode, name);
	}

	return new_name;
}

int TC_is_seq(const char *name)
{
	char *format;

	if (!AR_check_file_exists(name))
		return 0;

	format = TC_identify_seq_format(name);
	if (!format || format[0] == '\0') {
		AR_free(format);
		return 0;
	}
	else if (strstr(format, "seq")) {
		AR_free(format);
		return 1;
	}
	else
		return 0;
}

int TC_is_aln(const char *name)
{
	char *format;

	if (!AR_check_file_exists(name))
		return 0;

	format = TC_identify_seq_format(name);
	if (!format || format[0] == '\0') {
		AR_free(format);
		return 0;
	}
	else if (strstr(format, "aln")) {
		AR_free(format);
		return 1;
	}
	else
		return 0;
}

int TC_is_lib(const char *name)
{
	return TC_is_lib_01(name);
}

int TC_is_lib_01(const char *name)
{
	if (TC_token_is_in_file(name, "TC_LIB_FORMAT_01"))
		return 1;
	else if (TC_token_is_in_file(name, "T-COFFEE_LIB_FORMAT_01"))
		return 1;
	else if (TC_token_is_in_file(name, "SEQ_1_TO_N"))
		return 1;
	else
		return 0;
}

int TC_is_matrix(const char *name)
{
	int **m;

	if ((m = TC_read_matrice(name)) != 0) {
		TC_free_int(m);
		return 1;
	}
	return 0;
}

int TC_is_blast_matrix(const char *name)
{
	FILE *fp;

	if ((fp = TC_find_token_in_file(name, 0, "BLAST_MATRIX")) != 0) {
		TC_vfclose(fp);
		return 1;
	}
	else
		return 0;
}

int TC_is_clustalw_matrix(const char *name)
{
	FILE *fp;

	if ((fp = TC_find_token_in_file(name, 0, "CLUSTALW_MATRIX")) != 0) {
		TC_vfclose(fp);
		return 1;
	}
	else
		return 0;
}

int TC_is_pavie_matrix(const char *name)
{
	FILE *fp;

	if ((fp = TC_find_token_in_file(name, 0, "PAVIE_MATRIX")) != 0) {
		TC_vfclose(fp);
		return 1;
	}
	else
		return 0;
}

int TC_is_method(const char *file)
{
	if (TC_token_is_in_file(file, "TC_METHOD_FORMAT_01"))
		return 1;
	if (AR_is_preset_method(file))
		return 1;
	return 0;
}

int TC_is_pdb_name(const char *name)
{
	std::ostringstream oss;
	const char *result_file = 0;
	FILE *fp = 0;
	int result;

	result_file = AR_tmpname();

	oss << g_conf.extract_pdb() << " -is_pdb_name " << name << " > " << result_file;
	system(oss.str().c_str());

	fp = TC_vfopen(result_file, "r");
	fscanf(fp, "%d", &result);
	TC_vfclose(fp);

	return (result == 1);
}

int TC_is_pdb_file(const char *name)
{
	FILE *fp;
	int ispdb = 0;

	if (!AR_check_file_exists(name))
		return 0;

	if ((fp = TC_find_token_in_file(name, 0, "HEADER")) != 0) {
		TC_vfclose(fp);
		++ispdb;
	}
	if ((fp = TC_find_token_in_file(name, 0, "SEQRES")) != 0) {
		TC_vfclose(fp);
		++ispdb;
	}
	if ((fp = TC_find_token_in_file(name, 0, "ATOM")) != 0) {
		TC_vfclose(fp);
		++ispdb;
	}

	if (ispdb > 1)
		return 1;
	else
		return 0;
}

int TC_is_single_seq_weight_file(const char *name)
{
	return TC_token_is_in_file(name, "SINGLE_SEQ_WEIGHT_FORMAT_01");
}

/**
 * @brief This function identifies known sequence and alignmnent formats.
 */
char *TC_identify_seq_format(const char *file)
{
	char *format = 0;

	format = (char *)AR_allocate(32, sizeof(char));

	if (!AR_check_file_exists(file)) {
		std::cerr << '\n' << PROGRAM << ": ERROR: File '" << file << "' does not exist.\n";
	}
	else if (TC_is_pdb_file(file))
		strncpy(format, "pdb_struc", 32 - 1);
	else if (TC_format_is_msf(file))
		strncpy(format, "msf_aln", 32 - 1);
	else if (TC_format_is_fasta_seq(file))
		strncpy(format, "fasta_seq", 32 - 1);
	else if (TC_format_is_fasta_aln(file))
		strncpy(format, "fasta_aln", 32 - 1);
	else if (TC_format_is_pir_aln(file))
		strncpy(format, "pir_aln", 32 - 1);
	else if (TC_format_is_pir_seq(file))
		strncpy(format, "pir_seq", 32 - 1);
	else if (TC_format_is_oligo(file))
		strncpy(format, "oligo_aln", 32 - 1);
	else if (TC_format_is_swissprot(file))
		strncpy(format, "swissprot_seq", 32 - 1);
	else if (TC_format_is_saga(file))
		strncpy(format, "clustal_aln", 32 - 1);

	return format;
}

int TC_format_is_fasta(const char *file)
{
	FILE *fp;
	int c;

	if (!AR_check_file_exists(file))
		return 0;

	fp = TC_vfopen(file, "r");
	if (fp == 0)
		return 0;

	while (isspace(c = fgetc(fp)));
	if (c == '>') {
		while ((c = fgetc(fp)) != '\n');
		while ((c = fgetc(fp)) != '>' && c != '*' && c != EOF);
		if (c != '*') {
			TC_vfclose(fp);
			return 1;
		}
	}
	TC_vfclose(fp);

	return 0;
}

int TC_format_is_fasta_seq(const char *file)
{
	TC_Sequences *S;
	int a, l1, l2, l;

	if (TC_format_is_fasta(file)) {
		S = TC_get_fasta_sequence(file);
		l = strlen(S->seq[0]);
		for (a = 0; a < S->nseq; a++)
			if (strlen(S->seq[a]) != l) {
				TC_free_sequence(S);
				return 1;
			}
		for (a = 0; a < S->nseq; a++) {
			l1 = strlen(S->seq[a]);
			TC_ungap(S->seq[a]);
			l2 = strlen(S->seq[a]);
			if (l1 != l2) {
				TC_free_sequence(S);
				return 0;
			}
		}
		TC_free_sequence(S);

		return 1;
	}
	else
		return 0;
}

int TC_format_is_fasta_aln(const char *file)
{
	if (TC_format_is_fasta(file) && !TC_format_is_fasta_seq(file))
		return 1;
	else
		return 0;
}

int TC_format_is_pir(const char *file)
{
	FILE *fp;
	int c;

	fp = TC_vfopen(file, "r");
	while (isspace(c = fgetc(fp)));
	if (c == '>') {
		while ((c = fgetc(fp)) != '\n');
		while ((c = fgetc(fp)) != '>' && c != '*' && c != EOF);
		TC_vfclose(fp);
		if (c == '*')
			return 1;
	}
	TC_vfclose(fp);

	return 0;
}

int TC_format_is_pir_seq(const char *file)
{
	TC_Sequences *S;
	int a, l1, l2;

	if (TC_format_is_fasta(file)) {
		S = TC_get_pir_sequence(file);
		for (a = 0; a < S->nseq; a++) {
			l1 = strlen(S->seq[a]);
			TC_ungap(S->seq[a]);
			l2 = strlen(S->seq[a]);
			if (l1 != l2) {
				TC_free_sequence(S);
				return 0;
			}
		}
		return 1;
	}
	else
		return 0;
}

int TC_format_is_pir_aln(const char *file)
{
	if (TC_format_is_pir(file) && !TC_format_is_pir_seq(file))
		return 1;
	else
		return 0;
}

int TC_format_is_msf(const char *file)
{
	FILE *fp;

	if ((fp = TC_find_token_in_file_nlines(file, 0, "MSF:", 30)) != 0) {
		TC_vfclose(fp);
		return 1;
	}
	else
		return 0;
}

int TC_format_is_swissprot(const char *name)
{
	FILE *fp;

	if (!AR_check_file_exists(name))
		return 0;

	if ((fp = TC_find_token_in_file_nlines(name, 0, "\nID", 10)) != 0
		&& (fp = TC_find_token_in_file(name, 0, "\nSQ")) != 0) {
		TC_vfclose(fp);
		return 1;
	}
	else
		return 0;
}

int TC_format_is_saga(const char *file)
{
	FILE *fp;

	if ((fp = TC_find_token_in_file(file, 0, "SAGA"))) {
		TC_vfclose(fp);
		return 1;
	}
	else if ((fp = TC_find_token_in_file(file, 0, "CLUSTAL"))) {
		TC_vfclose(fp);
		return 1;
	}
	else if ((fp = TC_find_token_in_file(file, 0, "T-COFFEE_MSA"))) {
		TC_vfclose(fp);
		return 1;
	}
	else if ((fp = TC_find_token_in_file(file, 0, "INTERLEAVED_MSA"))) {
		TC_vfclose(fp);
		return 1;
	}
	else
		return 0;
}

int TC_format_is_oligo(const char *file)
{
	FILE *fp;
	char buf[1000];

	buf[0] = '\0';
	fp = TC_vfopen(file, "r");
	fscanf(fp, "%s", buf);
	TC_vfclose(fp);

	if (TC_strm(buf, "ALPHABET"))
		return 1;
	return 0;
}

static TC_Sequences *TC_add_prf2seq(const char *file, TC_Sequences *S)
{
	TC_Sequences *NS;
	char **new_seq;
	
	if (!TC_is_aln(file) && !TC_is_seq(file))
		return S;
	else {
		TC_X_template *R;
		TC_Alignment *A;
		
		R = TC_fill_R_template(file, file, S);
		if (!R)
			return S;

		A = R->A;
		A->expand = 1;
		
		new_seq = TC_declare_char(1, A->len_aln + 1);
		strcpy(new_seq[0], TC_aln2cons_seq_mat(A, "blosum62mt"));
		
		NS = TC_fill_sequence_struc(1, new_seq, A->file);
		S = TC_add_sequence(NS, S, 0);
		(S->T[S->nseq - 1])->R = R;
		
		TC_free_sequence(NS);
		TC_free_char(new_seq);
		
		return S;
	}
}

#define TC_is_seq_source(Symbol, Mode, SeqMode)\
(\
	Symbol == Mode\
	&& (SeqMode == 0 || TC_strm(SeqMode, "ANY") || (SeqMode[0] != '_' && strchr(SeqMode, Symbol))\
		|| (SeqMode[0] == '_' && !strchr(SeqMode, Symbol)))\
)

/* THE TYPE OF EACH FILE MUST BE INDICATED. */
/* SeqMode indicates the type of file that can be used as sequence sources:
 * ANY: any mode
 * SL: only sequences from Libraries and Sequences
 * _A: anything BUT sequences from A(lignments)
 */
TC_Sequences *AR_read_seq_from_files(std::vector<std::string>::iterator files_beg,
	std::vector<std::string>::iterator files_end, const char *type, const char *SeqMode)
{
	TC_Sequences *S = 0, *S1 = 0;
	TC_Alignment *A = 0;
	char **sequences = 0;
	char **seq_name = 0;
	char mode;
	int nseq = 0;
	int a, b;

	if (files_beg == files_end) {
		std::cerr << '\n' << PROGRAM << ": WARNING: No input files (read_seq_from_files).\n";
		return 0;
	}

	for (; files_beg != files_end; ++files_beg) {
		static char *buf;
		char *lname;

		if (buf)
			AR_free(buf);

		buf = TC_name2type_name(files_beg->c_str());
		mode = buf[0];
		lname = buf + 1;

		if (TC_is_seq_source('A', mode, SeqMode)) {
			A = TC_main_read_aln(lname, 0);
			S1 = TC_aln2seq(A);
			S1 = TC_seq2unique_name_seq(S1);

			S = TC_merge_seq(S1, S);
			if (!S) {
				std::cerr << '\n' << PROGRAM << ": ERROR: File (" << mode << ") '" << lname
						  << "' contains incorrect data.\n";
			}
			
			TC_free_aln(A);
			TC_free_sequence(S1);
		}
		else if(TC_is_seq_source('R', mode, SeqMode)) {
			S = TC_add_prf2seq(lname, S);
		}
		else if (TC_is_seq_source('P', mode, SeqMode)) {
			S1 = TC_get_pdb_sequence(lname);

			S = TC_merge_seq(S1, S);
			if (!S) {
				std::cerr << '\n' << PROGRAM << ": ERROR: File (" << mode << ") '" << lname
						  << "' contains incorrect data.\n";
			}
			else {
				int i;

				i = TC_name_is_in_list(S1->name[0], S->name, S->nseq, 100);
				(S->T[i])->P = TC_fill_P_template(S->name[i], lname);
			}
			
			TC_free_sequence(S1);
		}
		else if (TC_is_seq_source('S', mode, SeqMode)) {
			A = TC_main_read_aln(lname, 0);

			if (A) {
				S1 = TC_aln2seq(A);
				TC_free_aln(A);
			}
			else {
				S1 = TC_main_read_seq(lname);
			}

			if (!S1) {
				std::cerr << '\n' << PROGRAM << ": ERROR: File (" << mode << ") '" << lname
						  << "' contains incorrect data.\n";
			}
			else {
				for (b = 0; b < S1->nseq; ++b)
					TC_ungap(S1->seq[b]);
				S1 = TC_seq2unique_name_seq(S1);
				
				S = TC_merge_seq(S1, S);
				if (!S) {
					std::cerr << '\n' << PROGRAM << ": ERROR: File (" << mode << ") '" << lname
							  << "' contains incorrect data.\n";
				}
				
				TC_free_sequence(S1);
			}
		}
		else if (TC_is_seq_source('L', mode, SeqMode)) {
			TC_read_seq_in_list(lname, &nseq, &sequences, &seq_name);
			S1 = TC_fill_sequence_struc(nseq, sequences, seq_name);
			for (b = 0; b < S1->nseq; b++)
				strncpy(S1->file[b], lname, STRING);
			
			nseq = 0;
			TC_free_char(sequences);
			sequences = 0;
			TC_free_char(seq_name);
			seq_name = 0;
			
			S1 = TC_seq2unique_name_seq(S1);
			S = TC_merge_seq(S1, S);
			if (!S) {
				std::cerr << '\n' << PROGRAM << ": ERROR: File (" << mode << ") '" << lname
						  << "' contains incorrect data.\n";
			}
			
			TC_free_sequence(S1);
		}
		else if (!strchr("ALSMXPRW", mode)) {
			std::cerr << '\n' << PROGRAM << ": ERROR: '" << lname << "' is neither a file nor a method.\n";
		}
	}
	
	if (!S)
		return 0;

	if (type && type[0])
		strncpy(S->type, type, sizeof(S->type) - 1);
	else
		S = TC_get_sequence_type(S);

	if (TC_strm(S->type, "PROTEIN_DNA")) {
		for (a = 0; a < S->nseq; ++a) {
			if (TC_strm(TC_get_string_type(S->seq[a]), "PROTEIN")) {
				S->seq[a] = TC_thread_aa_seq_on_dna_seq(S->seq[a]);
				S->len[a] = strlen(S->seq[a]);
				S->max_len = MAX(S->max_len, S->len[a]);
			}
		}
	}

	return S;
}

TC_Sequences *TC_main_read_seq(const char *name)
{
	char *format = 0;
	TC_Sequences *S = 0;
	TC_Alignment *A = 0;
	int a;

	format = TC_identify_seq_format(name);

	if (format && TC_strm(format, "fasta_seq"))
		S = TC_get_fasta_sequence(name);
	else if (format && TC_strm(format, "pir_seq"))
		S = TC_get_pir_sequence(name);
	else if (format && TC_strm(format, "swissprot_seq"))
		S = TC_get_swissprot_sequence(name);
	else if (format && strstr(format, "aln")) {
		A = TC_main_read_aln(name, 0);
		S = TC_aln2seq(A);
		TC_ungap_seq(S);
		TC_free_aln(A);
	}

	AR_free(format);

	if (S) {
		for (a = 0; a < S->nseq; ++a)
			strncpy(S->file[a], name, STRING);
		TC_ungap_seq(S);
		TC_clean_string(S->nseq, S->seq_comment);
		S->name = TC_clean_seq_names(S->name, S->nseq, GET);
	}

	return S;
}

TC_Sequences *TC_get_fasta_sequence(const char *fname)
{
	TC_Sequences *LS;
	FILE *fp;
	char *buffer;
	char *name;
	char *sub;
	int clen = 0;
	int current = 0;
	int p = 0;
	int max;
	int max_len_seq = 0;
	int min_len_seq = 0;
	int nseq = 0, l = 0;
	int a, c;

	buffer = (char *)AR_allocate(1000, sizeof(char));
	name = (char *)AR_allocate(100, sizeof(char));

	nseq = TC_count_n_char_x_in_file(fname, '>');
	min_len_seq = max = TC_count_n_char_in_file(fname);
	sub = (char *)AR_allocate(max + 1, sizeof(char));

	fp = TC_vfopen(fname, "r");
	c = fgetc(fp);
	while (c != EOF) {
		if (c == '>') {
			fscanf(fp, "%s", name);
			while ((c = fgetc(fp)) != '\n' && c != EOF);

			while ((c = fgetc(fp)) != '>' && c != EOF) {
				if (isalnum(c) || TC_is_gap(c))
					sub[clen++] = c;
			}

			max_len_seq = (clen > max_len_seq) ? clen : max_len_seq;
			min_len_seq = (clen < min_len_seq) ? clen : min_len_seq;
			clen = 0;
		}
		else
			c = fgetc(fp);
	}
	TC_vfclose(fp);

	LS = TC_declare_sequence(min_len_seq, max_len_seq, nseq);
	LS->nseq = nseq;

	fp = TC_vfopen(fname, "r");
	current = 0;
	c = fgetc(fp);
	while (c != EOF) {
		if (c == '>') {
			fscanf(fp, "%s", LS->name[current]);
			l = strlen(LS->name[current]);
			if (LS->name[current][l - 1] == ',' || LS->name[current][l - 1] == ';')
				LS->name[current][l - 1] = '\0';
			TC_translate_name(LS->name[current]);

			a = 0;
			while ((c = fgetc(fp)) != '\n' && c != EOF)
				LS->seq_comment[current][a++] = c;
			LS->seq_comment[current][a] = '\0';

			p = 0;
			while ((c = fgetc(fp)) != '>' && c != EOF) {
				if (isalpha(c) || TC_is_gap(c))
					LS->seq[current][p++] = c;
			}
			LS->seq[current][p] = '\0';

			LS->len[current] = strlen(LS->seq[current]);
			++current;
		}
		else
			c = fgetc(fp);
	}

	TC_vfclose(fp);
	AR_free(sub);
	AR_free(name);
	AR_free(buffer);

	return LS;
}

TC_Sequences *TC_get_pir_sequence(const char *fname)
{
	TC_Sequences *LS;
	FILE *fp;
	char name[100];
	char *buf;
	int clen = 0, l = 0;
	int max_len_seq = 0;
	int min_len_seq = 999999;
	int nseq = 0;
	int current = 0;
	int p = 0, c;

	if ((fp = TC_vfopen(fname, "r")) == 0) {
		std::cerr << '\n' << PROGRAM << ": ERROR: Could not open file '" << fname << "'.\n";
		
		return 0;
	}

	buf = (char *)AR_allocate(1000, sizeof(char));

	c = fgetc(fp);
	while (c != EOF) {
		if (c == '>') {
			if ((c = fgetc(fp)) == 'P')
				while ((c = fgetc(fp)) != ';');
			else
				ungetc(c, fp);
			fscanf(fp, "%s", name);
			buf = fgets(buf, 1000, fp);
			while ((c = fgetc(fp)) != '>' && c != EOF)
				if (isalnum(c) || TC_is_gap(c))
					++clen;
			max_len_seq = (clen > max_len_seq) ? clen : max_len_seq;
			min_len_seq = (clen < min_len_seq) ? clen : min_len_seq;
			++nseq;
			clen = 0;
		}
		else
			c = fgetc(fp);
	}
	TC_vfclose(fp);

	LS = TC_declare_sequence(min_len_seq, max_len_seq, nseq);
	LS->nseq = nseq;

	fp = TC_vfopen(fname, "r");
	current = 0;
	c = fgetc(fp);
	while (c != EOF) {
		if (c == '>') {
			if ((c = fgetc(fp)) == 'P')
				while ((c = fgetc(fp)) != ';');
			else
				ungetc(c, fp);

			fscanf(fp, "%s", LS->name[current]);
			l = strlen(LS->name[current]);
			if (LS->name[current][l - 1] == ',' || LS->name[current][l - 1] == ',')
				LS->name[current][l - 1] = '\0';

			TC_translate_name(LS->name[current]);
			buf = fgets(buf, 1000, fp);
			LS->seq_comment[current] = fgets(LS->seq_comment[current], 1000, fp);
			LS->seq_comment[current][strlen(LS->seq_comment[current]) - 1] = '\0';

			p = 0;
			while ((c = fgetc(fp)) != '>' && c != EOF)
				if (isalpha(c))
					LS->seq[current][p++] = tolower(c);
				else if (!isspace(c) && c != '*')
					LS->seq[current][p++] = c;
			LS->seq[current][p] = '\0';
			LS->len[current] = strlen(LS->seq[current]);
			++current;
		}
		else
			c = fgetc(fp);
	}
	TC_vfclose(fp);
	AR_free(buf);

	return LS;
}

TC_Sequences *TC_get_swissprot_sequence(const char *fname)
{
	TC_Sequences *LS;
	FILE *fp;
	char *buf;
	int nseq = 0;
	int len, max_len_seq = 0, min_len_seq;
	int c;

	if (!AR_check_file_exists(fname)) {
		std::cerr << '\n' << PROGRAM << ": ERROR: Could not open file '" << fname << "'.\n";
		
		return 0;
	}

	buf = (char *)AR_allocate(LONG_STRING + 1, sizeof(char));

	fp = 0;
	while ((fp = TC_find_token_in_file(fname, fp, "\nSQ"))) {
		++nseq;
		fgets(buf, LONG_STRING, fp);
		len = 0;
		while ((c = fgetc(fp)) != '/')
			if (isalpha(c))
				++len;
		if (max_len_seq == 0)
			max_len_seq = min_len_seq = len;
		else {
			max_len_seq = MAX(len, max_len_seq);
			min_len_seq = MIN(len, min_len_seq);
		}
	}

	LS = TC_declare_sequence(min_len_seq, max_len_seq, nseq);
	LS->nseq = 0;

	fp = 0;
	while ((fp = TC_find_token_in_file(fname, fp, "\nID"))) {
		fscanf(fp, "%s", LS->name[LS->nseq]);
		fp = TC_find_token_in_file(fname, fp, "\nSQ");
		fgets(buf, LONG_STRING, fp);
		while ((c = fgetc(fp)) != '/')
			if (isalpha(c))
				LS->seq[LS->nseq][LS->len[LS->nseq]++] = c;
		LS->seq[LS->nseq][LS->len[LS->nseq]] = '\0';
		LS->nseq++;
	}
	AR_free(buf);

	return LS;
}

static TC_Sequences *TC_get_pdb_sequence_from_field(const char *fname, const char *field)
{
	TC_Sequences *S;
	char *tp_name;
	char *pdbid;

	tp_name = AR_tmpname();

	std::ostringstream oss;

	oss << g_conf.extract_pdb() << " -seq_field " << field << " -chain FIRST -infile " << AR_check_file_exists(fname)
		<< " -mode fasta > " << tp_name;
	system(oss.str().c_str());

	S = TC_get_fasta_sequence(tp_name);
	if (!S)
		return 0;

	if ((pdbid = TC_get_pdb_id(fname))) {
		strncpy(S->name[0], pdbid, STRING);
		AR_free(pdbid);
	}

	S->nseq = 1;
	strncpy(S->file[0], fname, STRING);
	S->max_len = S->min_len = S->len[0];

	if (S->len[0] == 0) {
		TC_free_sequence(S);
		S = 0;
	}
	TC_vremove(tp_name);

	return S;
}

TC_Sequences *TC_get_pdb_sequence(const char *fname)
{
	TC_Sequences *S;

	if ((S = TC_get_pdb_sequence_from_field(fname, "SEQRES")) != 0);
	else if ((S = TC_get_pdb_sequence_from_field(fname, "ATOM")) != 0) {
		std::cerr << '\n' << PROGRAM << ": WARNING: Read sequence from ATOM field in '" << fname << "'.\n";
	}
	else {
		S = 0;
		std::cerr << '\n' << PROGRAM << ": ERROR: Failed to extract sequence from '" << fname << "'.\n";
	}

	return S;
}

int TC_read_seq_in_list(const char *fname, int *nseq, char ***sequences, char ***seq_name)
{
	static int max_nseq;
	static int *sn_list;
	FILE *fp;
	char name[1000];
	char *sequence;
	int seq_len, sn;
	int list_nseq;
	int lline;
	int a, l;

	fp = TC_vfopen(fname, "r");
	fp = TC_skip_commentary_line_in_file('!', fp);

	fscanf(fp, "%d\n", &max_nseq);
	for (lline = 0, a = 0; a < max_nseq; a++) {
		fscanf(fp, "%*s %d %*s\n", &l);
		lline = MAX(lline, l);
	}
	TC_vfclose(fp);

	sequence = (char *)AR_allocate(lline + 1, sizeof(char));
	if (seq_name[0] == 0) {
		seq_name[0] = TC_declare_char(max_nseq, 0);
		sequences[0] = TC_declare_char(max_nseq, 0);
	}
	if (sn_list == 0)
		sn_list = (int *)AR_allocate(max_nseq, sizeof(int));
	else
		sn_list = (int *)AR_reallocate(sn_list, max_nseq, sizeof(int));

	fp = TC_vfopen(fname, "r");
	fp = TC_skip_commentary_line_in_file('!', fp);

	if (fscanf(fp, "%d\n", &list_nseq) != 1)
		return 0;
	for (a = 0; a < max_nseq; a++) {
		fp = TC_skip_commentary_line_in_file('!', fp);
		fscanf(fp, "%s %d %s\n", name, &seq_len, sequence);
		TC_lower_string(sequence);

		if ((sn = TC_name_is_in_list(name, seq_name[0], nseq[0], 100)) == -1) {
			seq_name[0][nseq[0]] = (char *)AR_allocate(strlen(name) + 1, sizeof(char));
			strcpy(seq_name[0][nseq[0]], name);

			sequences[0][nseq[0]] = (char *)AR_allocate(strlen(sequence) + 1, sizeof(char));
			strcpy(sequences[0][nseq[0]], sequence);

			sn_list[a] = nseq[0];
			++nseq[0];
		}
		else {
			sn_list[a] = sn;
		}
	}
	TC_vfclose(fp);
	AR_free(sequence);

	return 1;
}

TC_Alignment *TC_main_read_aln(const char *name, TC_Alignment *A)
{
	static char *format;
	TC_Sequences *S = 0;
	TC_Sequences *IN_SEQ;
	int a;

	if (format)
		AR_free(format);

	if (!name)
		return 0;
	else if (!AR_check_file_exists(name)) {
		if (!AR_check_file_exists(name + 1))
			return 0;
		else if (name[0] == 'A' || name[0] == 'S')
			++name;
	}

	if (!A)
		A = TC_declare_aln(0);

	format = TC_identify_seq_format(name);
	IN_SEQ = A->S;

	if (format && TC_strm(format, "saga_aln") || TC_strm(format, "clustal_aln")
		|| TC_strm(format, "t_coffee_aln"))
		TC_read_aln(name, A);
	else if (format && TC_strm(format, "msf_aln"))
		TC_read_msf_aln(name, A);
	else if (format && TC_strm(format, "fasta_aln")) {
		S = TC_get_fasta_sequence(name);
		S->contains_gap = 0;
		TC_seq2aln(S, A, 0);
	}
	else if (format && TC_strm(format, "pir_aln")) {
		S = TC_get_pir_sequence(name);
		S->contains_gap = 0;
		TC_seq2aln(S, A, 0);
	}
	else if (format && TC_strm(format, "fasta_seq") && A) {
		S = TC_get_fasta_sequence(name);
		for (a = 1; a < S->nseq; a++)
			if (strlen(S->seq[a - 1]) != strlen(S->seq[a])) {
				TC_free_sequence(S);
				TC_free_aln(A);
				
				return 0;
			}
		S->contains_gap = 0;
		TC_seq2aln(S, A, 0);
	}
	else if (format && TC_strm(format, "pir_seq") && A) {
		S = TC_get_pir_sequence(name);
		for (a = 1; a < S->nseq; a++)
			if (strlen(S->seq[a - 1]) != strlen(S->seq[a])) {
				TC_free_sequence(S);
				TC_free_aln(A);
				
				return 0;
			}
		S->contains_gap = 0;
		TC_seq2aln(S, A, 0);
	}
	else {
		TC_free_aln(A);
		
		return 0;
	}

	if (TC_check_list_for_dup(A->name, A->nseq)) {
		std::cerr << '\n' << PROGRAM << ": WARNING: Sequence '" << TC_check_list_for_dup(A->name, A->nseq)
				  << "' is duplicated in file '" << A->file[0] << "'.\n";

		if (!getenv("TCOFFEE_ALLOW_DUPLICATE")) {
			exit(EXIT_FAILURE);
		}
	}

	if (IN_SEQ)
		A->S = IN_SEQ;
	else if (!A->S)
		A->S = TC_aln2seq(A);

	A->S = TC_ungap_seq(A->S);
	A = TC_fix_aln_seq(A, A->S);
	TC_compress_aln(A);
	for (a = 0; a < A->nseq; ++a)
		strncpy(A->file[a], name, STRING - 1);

	TC_clean_string(A->nseq, A->seq_comment);
	TC_clean_string(A->nseq, A->aln_comment);
	TC_clean_seq_names(A->name, A->nseq, GET);
	TC_clean_seq_names((A->S)->name, (A->S)->nseq, GET);

	return A;
}

void TC_read_aln(const char *file_name, TC_Alignment *A)
{
	FILE *fp;
	char buf[VERY_LONG_STRING + 1];
	const char *fname;
	int *ptr_aln;
	int tot = 0;
	int flag = 0;
	int n_comment = 0;
	int nseq = 0;
	int max_len = 0;
	int a, b, c, d;

	A->nseq = 0;
	A->len_aln = 0;

	fname = file_name;

	// 1: Count the number of sequences
	fp = TC_vfopen(fname, "r");
	fgets(buf, 1000, fp);
	while (TC_is_blank(buf))
		fgets(buf, 1000, fp);
	while (!TC_is_blank(buf))
		fgets(buf, 1000, fp);
	while (!isalnum(c = fgetc(fp))) {
		ungetc(c, fp);
		fgets(buf, 1000, fp);
	}
	if (c != '\n')
		ungetc(c, fp);
	while (isgraph(c = fgetc(fp))) {
		ungetc(c, fp);
		a = 0;
		while (isgraph(c = fgetc(fp)));
		++nseq;
		fgets(buf, 1000, fp);
	}
	TC_vfclose(fp);

	// 2: Get max_len
	max_len = TC_count_n_char_in_file(fname) / nseq;
	A = TC_realloc_aln2(A, nseq + 1, max_len + 1);

	fp = TC_vfopen(fname, "r");
	fgets(buf, 1000, fp);

	if (!TC_is_blank(buf))
		strncpy(A->aln_comment[n_comment++], buf, STRING - 1);

	while (TC_is_blank(buf))
		fgets(buf, 1000, fp);
	while (!TC_is_blank(buf)) {
		fgets(buf, 1000, fp);
		strncpy(A->aln_comment[n_comment++], buf, STRING - 1);
	}

	while (!isalnum(c = fgetc(fp))) {
		ungetc(c, fp);
		fgets(buf, 1000, fp);
	}
	if (c != '\n')
		ungetc(c, fp);

	while (isgraph(c = fgetc(fp))) {
		ungetc(c, fp);
		a = 0;
		while (isgraph(c = fgetc(fp)))
			A->name[A->nseq][a++] = c;
		A->name[A->nseq][a] = '\0';

		if (TC_name_is_in_list(A->name[A->nseq], A->name, A->nseq, 100) != -1) {
			std::cerr << '\n' << PROGRAM << ": WARNING: Sequence '" << A->name[A->nseq] << "' is duplicated in file '"
					  << file_name << "'.\n";

			if (!getenv("TCOFFEE_ALLOW_DUPLICATE")) {
				exit(EXIT_FAILURE);
			}
		}
		A->nseq++;
		fgets(buf, 1000, fp);
	}
	TC_vfclose(fp);

	if ((fp = TC_vfopen(fname, "r")) == 0)
		std::cerr << '\n' << PROGRAM << ": ERROR: Could not open file '" << fname << "'.\n";

	ptr_aln = (int *)AR_allocate(A->nseq, sizeof(int));

	while (flag == 0) {
		while (!isalnum(c = fgetc(fp)));
		fgets(buf, VERY_LONG_STRING, fp);
		while (TC_is_alpha_line(buf))
			fgets(buf, VERY_LONG_STRING, fp);
		flag = 1;
	}

	while (!isalnum(c = fgetc(fp)));
	ungetc(c, fp);

	while (c != EOF) {
		tot = 0;
		while (tot < A->nseq && c != EOF) {
			b = 0;
			while (isgraph(buf[b++] = fgetc(fp)));

			buf[b - 1] = '\0';
			for (a = -1, d = 0; d < A->nseq; d++)
				if (strcmp(A->name[d], buf) == 0) {
					a = d;
					++tot;
				}

			while (!isgraph(c = fgetc(fp)) && c != EOF && c != '\n');
			ungetc(c, fp);

			if (a == -1)
				while ((c = fgetc(fp)) != '\n' && c != EOF);
			else {
				while ((c = fgetc(fp)) != '\n' && !isspace(c)) {
					if (isgraph(c) || TC_is_gap(c)) {
						if (isalpha(c))
							A->seq_al[a][ptr_aln[a]++] = (A->residue_case == 2) ? c : tolower(c);
						else if (isdigit(c) || TC_is_gap(c))
							A->seq_al[a][ptr_aln[a]++] = c;
					}
				}
			}
			if (c != '\n')
				while ((c = fgetc(fp)) != '\n' && c != EOF);
		}

		while (!isalnum(c = getc(fp)) && c != EOF);
		if (c != EOF)
			ungetc(c, fp);
	}
	TC_vfclose(fp);

	for (a = 0; a < A->nseq; a++) {
		A->seq_al[a][ptr_aln[a]] = '\0';
		A->order[a][0] = a;
		A->order[a][1] = 0;
	}
	A->len_aln = strlen(A->seq_al[0]);

	AR_free(ptr_aln);
}

void TC_read_msf_aln(const char *fname, TC_Alignment *A)
{
	TC_Sequences *S;
	char *tmp_name;
	std::ostringstream oss;

	tmp_name = AR_tmpname();
	oss << "msf_aln2fasta_aln " << fname << " > " << tmp_name;

	if (TC_mysystem(oss.str().c_str()) == EXIT_FAILURE) {
		std::cerr << '\n' << PROGRAM << ": ERROR: File '" << fname << "' does not have a legal MSF format.\n";
		return;
	}

	S = TC_get_fasta_sequence(tmp_name);
	A = TC_seq2aln(S, A, 0);
	TC_vremove(tmp_name);
}

TC_ConstraintList *TC_read_constraint_list(TC_ConstraintList *CL, const char *fname,
	const char *mode, const char *weight_mode)
{
	const char *read_mode = 0;
	const char id = fname[0];

	if (mode && *mode)
		read_mode = mode;
	else if (id == 'X') {
		read_mode = "matrix";
		++fname;
	}
	else if (id == 'A') {
		read_mode = "aln";
		++fname;
	}
	else if (id == 'L') {
		read_mode = "lib";
		++fname;
	}
	else if (id == 'W') {
		read_mode = "struct";
		++fname;
	}
	else if (!(id == 'S' || id == 'P' || id == 'R')) {
		std::cerr << '\n' << PROGRAM << ": ERROR: The descriptor '" << fname
				  << "' could not be identified as a file or a method.\n"
				  << "\tIf it is a method file please indicate it with M" << fname << ".\n";
	}

	if (!read_mode)
		return CL;

	TC_ConstraintList *subCL = 0;
	
	if (CL && TC_strm(read_mode, "matrix")) {
		if (CL->copy_mode == HARD_COPY) {
		    // TC_free_short_int(CL->L);
		    CL->L.free();
		    TC_free_int(CL->M);
		}
		CL->L = 0;
		CL->M = TC_read_matrice(fname);
		CL->extend_jit = 0;
	}
	else if (TC_strm(read_mode, "aln")) {
		subCL = TC_aln_file2constraint_list(fname, subCL, weight_mode);
	}
	else if (TC_strm2(read_mode, "ascii", "lib")) {
		subCL = TC_read_constraint_list_file(0, fname);
	}
	else if (TC_strm(read_mode, "struct")) {
		if (CL && CL->ne > 0) {
			std::cerr << '\n' << PROGRAM << ": ERROR: W<structure> must come before M<method> or A<aln>.\n";
		}
	}
	else {
		subCL = TC_read_constraint_list_file(subCL, fname);
	}

	if (subCL) {
		CL = TC_merge_constraint_list(subCL, CL);
		TC_free_constraint_list_full(subCL);
	}

	return CL;
}

TC_ConstraintList *TC_read_constraint_list_file(TC_ConstraintList *CL, const char *fname)
{
	static char *name;
	static char *mat;
	static char *dp_mode;
	static int *sn_list;
	static int line = 2;
	static int *entry;
	FILE *fp;
	char *stripped_file;
	char *buf;
	char *sequence;
	int lline;
	int seq_len, sn;
	int s1, s2;
	int max_nseq = 0;
	int list_nseq;
	int seq_1_to_n = 0;
	TC_Sequences *S, *small_S;
	TC_Alignment *B = 0;
	int tmp;
	int a, c, n, z;

	std::vector<std::string> tmp_list(1, fname);

	small_S = AR_read_seq_from_files(tmp_list.begin(), tmp_list.end(), 0, 0);

	if (!CL) {
		CL = TC_declare_constraint_list(small_S, 0, 0, 0);
		CL->S = small_S;
	}

	B = TC_seq2aln(small_S, 0, 1);
	B = TC_fix_aln_seq(B, (CL->S));

	if (CL->S != small_S)
		TC_free_sequence(small_S);

	lline = TC_measure_longest_line_in_file(fname) + 1;

	if (!mat)
		mat = (char *)AR_allocate(STRING, sizeof(char));
	if (!dp_mode)
		dp_mode = (char *)AR_allocate(STRING, sizeof(char));

	fp = TC_vfopen(fname, "r");
	while ((c = fgetc(fp)) != '#')
		if (c == '\n')
			++max_nseq;
	TC_vfclose(fp);

	buf = (char *)AR_allocate(lline, sizeof(char));
	sequence = (char *)AR_allocate(lline, sizeof(char));
	
	if (!name)
		name = (char *)AR_allocate(100, sizeof(char));
	if (!entry)
		entry = (int *)AR_allocate(CL->entry_len, sizeof(int));
	if (!sn_list)
		sn_list = (int *)AR_allocate(max_nseq, sizeof(int));
	else
		sn_list = (int *)AR_reallocate(sn_list, max_nseq, sizeof(int));
	
	S = CL->S;

	seq_1_to_n = ((fp = TC_find_token_in_file(fname, 0, "SEQ_1_TO_N")) != 0);
	TC_vfclose(fp);

	stripped_file = TC_strip_file_from_comments("!", fname);

	/* Read constraint list */
	fp = TC_vfopen(stripped_file, "r");

	fscanf(fp, "%d\n", &list_nseq);
	for (a = 0; a < list_nseq; a++) {
		fscanf(fp, "%s %d %s\n", name, &seq_len, sequence);
		++line;
		TC_lower_string(sequence);

		if ((sn = TC_name_is_in_list(name, S->name, S->nseq, 100)) == -1)
			continue;
		else
			sn_list[a] = sn;
	}

	while ((c = fgetc(fp)) != EOF) {
		ungetc(c, fp);
		if (c == '#') {
			fscanf(fp, "#%d %d\n", &s1, &s2);
			++line;
			/* check if the sequence numbering is legal */
			if (seq_1_to_n) {
				--s1;
				--s2;
			}
			if (s1 < 0 || s2 < 0) {
				std::cerr << '\n' << PROGRAM << ": ERROR: Wrong sequence numbering in '" << fname << "'.\n";
				TC_vfclose(fp);
				
				return CL;
			}
			s1 = sn_list[s1];
			s2 = sn_list[s2];
			while (isdigit(c = fgetc(fp))) {
				for (z = 0; z < CL->entry_len; z++)
					entry[z] = 0;
				ungetc(c, fp);
				n = 0;
				entry[n++] = s1;
				entry[n++] = s2;
				while ((c = fgetc(fp)) != '\n') {
					if (isspace(c));
					else {
						ungetc(c, fp);
						fscanf(fp, "%d", &entry[n]);
						++n;
					}
					if (n > CL->entry_len) {
						std::cerr << '\n' << PROGRAM << ": WARNING: Parse error in file '" << fname << "' at line " << line << ".\n";
						TC_vfclose(fp);
						
						return CL;
					}
				}
				if (c == '\n')
					++line;
				if (n <= CONS)
					entry[CONS] = 1;

				// check the legality of the entry
				if (n > 0 && n < 3) {
					std::cerr << '\n' << PROGRAM << ": WARNING: Parse error in file '" << fname << "' at line " << line - 1 << ".\n";
					TC_vfclose(fp);
					
					return CL;
				}
				entry[R1] = (B->seq_cache) ? B->seq_cache[entry[SEQ1]][entry[R1]] : entry[R1];
				entry[R2] = (B->seq_cache) ? B->seq_cache[entry[SEQ2]][entry[R2]] : entry[R2];
				
				if (entry[R1] && entry[R2]) {
					if (entry[R1] <= 0 || entry[R1] > (CL->S)->len[s1]) {
						std::cerr << '\n' << PROGRAM << ": ERROR: Library '" << fname << "', line " << line - 1
								  << ", Field 1: Bad residue numbering (" << entry[R1] << ").\n";
						TC_vfclose(fp);
						
						return CL;
					}
					else if (entry[R2] <= 0 || entry[R2] > (CL->S)->len[s2]) {
						std::cerr << '\n' << PROGRAM << ": ERROR: Library '" << fname << "', line " << line - 1
								  << ", Field 2: Bad residue numbering (" << entry[R2] << ").\n";
						TC_vfclose(fp);
						
						return CL;
					}
					fscanf(fp, "\n");

					if (entry[SEQ1] > entry[SEQ2] || (entry[SEQ1] == entry[SEQ2] && entry[R1] > entry[R2])) {
						SWAPP(entry[SEQ1], entry[SEQ2], tmp);
						SWAPP(entry[R1], entry[R2], tmp);
					}
					for (z = 0; z < CL->entry_len; z++)
						TC_vwrite_clist(CL, CL->ne, z, entry[z]);
					CL->ne++;
				}
			}
			ungetc(c, fp);
		}
		else if (c == '!' || c == 'C')
			while ((c = fgetc(fp)) != '\n');
		else {
			std::cerr << '\n' << PROGRAM << ": WARNING: Parse error in file '" << fname << "' at line " << line << ".\n";
			TC_vfclose(fp);
			
			return CL;
		}
	}
	TC_vfclose(fp);

	TC_free_aln(B);
	AR_free(buf);
	AR_free(sequence);
	remove(stripped_file);

	return CL;
}

TC_ConstraintList *TC_fast_read_constraint_list_file(TC_ConstraintList *CL, const char *fname)
{
	FILE *fp;
	TC_Sequences *NS;
	short int *entry = 0;
	int tmp;
	int **index;
	int *list = 0;
	char buf[1001];
	int c, n;

	if (!CL) {
		return TC_read_constraint_list_file(CL, fname);
	}

	entry = (short int *)AR_allocate(CL->entry_len, sizeof(short int));
	std::vector<std::string> tmp_list(1, fname);

	NS = AR_read_seq_from_files(tmp_list.begin(), tmp_list.end(), 0, 0);
	index = TC_index_seq_name(NS, CL->S);

	/* Read Constraint list */
	fp = TC_vfopen(fname, "r");

	while ((c = fgetc(fp)) != '#');
	ungetc(c, fp);

	while (fgets(buf, 1001, fp) != 0) {
		if (buf[0] == '!')
			continue;
		list = TC_string2num_list2(buf, " #\n");

		if (buf[0] == '#') {
			sscanf(buf, "#%d %d", &entry[SEQ1], &entry[SEQ2]);
			entry[SEQ1] = index[entry[SEQ1] - 1][0];
			entry[SEQ2] = index[entry[SEQ2] - 1][0];
		}
		else {
			sscanf(buf, "%d %d %d %d %d", &entry[R1], &entry[R2], &entry[WE], &entry[CONS], &entry[MISC]);

			if (n <= CONS)
				entry[CONS] = 1;
			if ((entry[SEQ1] > entry[SEQ2]) || (entry[SEQ1] == entry[SEQ2] && entry[R1] > entry[R2])) {
				SWAPP(entry[SEQ1], entry[SEQ2], tmp);
				SWAPP(entry[R1], entry[R2], tmp);
			}
			TC_add_entry2list(entry, CL);
		}
	}
	TC_vfclose(fp);

	AR_free(list);
	TC_free_sequence(NS);
	AR_free(entry);
	AR_free(buf);

	return CL;
}

TC_ConstraintList *TC_aln_file2constraint_list(const char *alname, TC_ConstraintList *CL,
	const char *weight_mode)
{
	TC_Alignment *A;

	A = TC_main_read_aln(alname, 0);
	CL = TC_aln2constraint_list(A, CL, weight_mode);
	TC_free_aln(A);

	return CL;
}

int **TC_read_matrice(const char *mat_name_in)
{
	FILE *fp;
	const char *AA = "abcdefghiklmnpqrstvwxyz";
	const char *mat_name = 0;
	int **matrix = 0;
	int **matrix2 = 0;
	int *vector = 0;
	int a, b, c, l;

	l = strlen(AA);

	if (TC_strm2(mat_name_in, "pam", "PAM"))
		mat_name = "pam250mt";
	else if (TC_strm2(mat_name_in, "blosum", "BLOSUM"))
		mat_name = "blosum62mt";
	else if (TC_strm3(mat_name_in, "id", "ID", "idmat"))
		mat_name = "idmat";
	else
		mat_name = mat_name_in;

	// Read Header Matrices
	if (TC_strm(mat_name, "pam250mt"))
		vector = pam250mt;
	else if (TC_strm(mat_name, "idmat"))
		vector = idmat;
	else if (TC_strm(mat_name, "dna_idmat"))
		vector = idmat;
	else if (TC_strm(mat_name, "est_idmat"))
		vector = est_idmat;
	else if (TC_strm(mat_name, "md_350mt"))
		vector = md_350mt;
	else if (TC_strm(mat_name, "md_250mt"))
		vector = md_250mt;
	else if (TC_strm(mat_name, "md_120mt"))
		vector = md_120mt;
	else if (TC_strm(mat_name, "md_40mt"))
		vector = md_40mt;
	else if (TC_strm(mat_name, "pam350mt"))
		vector = pam350mt;
	else if (TC_strm(mat_name, "pam160mt"))
		vector = pam160mt;
	else if (TC_strm(mat_name, "pam120mt"))
		vector = pam120mt;
	else if (TC_strm(mat_name, "blosum80mt"))
		vector = blosum80mt;
	else if (TC_strm(mat_name, "blosum62mt"))
		vector = blosum62mt;
	else if (TC_strm(mat_name, "exon2mt"))
		vector = blosum62mt;
	else if (TC_strm(mat_name, "blosum62mt3"))
		vector = blosum62mt3;
	else if (TC_strm(mat_name, "blosum62mt2"))
		vector = blosum62mt2;
	else if (TC_strm(mat_name, "blosum55mt"))
		vector = blosum55mt;
	else if (TC_strm(mat_name, "blosum50mt"))
		vector = blosum50mt;
	else if (TC_strm(mat_name, "blosum45mt"))
		vector = blosum45mt;
	else if (TC_strm(mat_name, "blosum40mt"))
		vector = blosum40mt;
	else if (TC_strm(mat_name, "blosum30mt"))
		vector = blosum30mt;
	else if (TC_strm(mat_name, "beta_mat"))
		vector = beta_mat;
	else if (TC_strm(mat_name, "alpha_mat"))
		vector = alpha_mat;
	else if (TC_strm(mat_name, "coil_mat"))
		vector = coil_mat;
	else if (TC_strm(mat_name, "rblosum80mt"))
		vector = rblosum80mt;
	else if (TC_strm(mat_name, "rblosum62mt"))
		vector = rblosum62mt;
	else if (TC_strm(mat_name, "rblosum30mt"))
		vector = rblosum30mt;
	else if (TC_strm(mat_name, "rpam250mt"))
		vector = rpam250mt;
	else if (TC_strm(mat_name, "rpam350mt"))
		vector = rpam350mt;
	else if (TC_strm(mat_name, "rpam160mt"))
		vector = rpam160mt;
	else if (TC_strm(mat_name, "rpam120mt"))
		vector = rpam120mt;
	else if (TC_strm(mat_name, "tmpam250mt"))
		vector = tmpam250mt;
	else if (TC_strm(mat_name, "rtmpam250mt"))
		vector = rtmpam250mt;
	else if (TC_strm(mat_name, "rbeta_mat"))
		vector = rbeta_mat;
	else if (TC_strm(mat_name, "ralpha_mat"))
		vector = ralpha_mat;
	else if (TC_strm(mat_name, "rcoil_mat"))
		vector = rcoil_mat;

	// Header Matrices
	if (vector) {
		matrix = TC_declare_int(256, 256);

		for (a = 0; a < l; ++a) {
			for (b = 0; b <= a; ++b)
				matrix[a][b] = matrix[b][a] = (vector[(a * a + a) / 2 + b]);
		}
	}
	// Hard coded Matrices
	else if (TC_strm(mat_name, "exon_mt")) {
		matrix = TC_declare_int(256, 256);

		AA = "bojx";
		l = 4;
		for (a = 0; a < 3; a++) {
			matrix[a][a] = 10;
			for (b = a + 1; b < 2; b++)
				matrix[a][b] = matrix[b][a] = 5;
		}
	}
	else if (TC_strm(mat_name, "tdamat")) {
		int x, y;

		matrix = TC_declare_int(256, 256);

		AA = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ";
		l = strlen(AA);

		for (x = 0; x < l; x++)
			for (y = 0; y < l; y++)
				matrix[x][y] = -10;
		for (x = 0; x < l; x++) {
			matrix[x][x] = 0;
			matrix[x][GAP_CODE] = -3;
		}

		return matrix;
	}
	// Blast Matrices
	else if (AR_check_file_exists(mat_name) && TC_is_blast_matrix(mat_name)) {
		matrix = TC_read_blast_matrix(mat_name);
		return matrix;
	}
	else if (AR_check_file_exists(mat_name) && TC_is_pavie_matrix(mat_name)) {
		matrix = TC_read_pavie_matrix(mat_name);
		return matrix;
	}
	else if (AR_check_file_exists(mat_name) && TC_is_clustalw_matrix(mat_name)) {
		fp = TC_vfopen(mat_name, "r");
		while ((c = fgetc(fp)) != '$' && c != EOF);

		if (c == EOF) {
			TC_vfclose(fp);
			return 0;
		}

		matrix = TC_declare_int(256, 256);

		fgetc(fp);
		for (a = 0; a < l; a++) {
			for (b = 0; b <= a; b++)
				if (fscanf(fp, "%d,", &matrix[a][b]) == 0) {
					TC_vfclose(fp);
					TC_free_int(matrix);
					TC_free_int(matrix2);
					return 0;
				}
			fscanf(fp, "\n");
		}
		TC_vfclose(fp);
	}
	else
		return 0;

	matrix2 = TC_declare_int(256, 256);

	for (b = 0; b < l; b++)
		for (c = 0; c < l; c++) {
			matrix2[toupper(AA[b]) - 'A'][toupper(AA[c]) - 'A'] = matrix[b][c];
			matrix2[toupper(AA[b]) - 'A'][tolower(AA[c]) - 'A'] = matrix[b][c];
			matrix2[tolower(AA[b]) - 'A'][toupper(AA[c]) - 'A'] = matrix[b][c];
			matrix2[tolower(AA[b]) - 'A'][tolower(AA[c]) - 'A'] = matrix[b][c];
		}

	if (TC_strm("exon2mt", mat_name)) {
		const char *EE = "obj";

		for (a = 0; a < 3; a++) {
			matrix2[toupper(EE[a]) - 'A'][toupper(EE[a]) - 'A'] = 100;
			matrix2[tolower(EE[a]) - 'A'][toupper(EE[a]) - 'A'] = 100;
			matrix2[toupper(EE[a]) - 'A'][tolower(EE[a]) - 'A'] = 100;
			matrix2[tolower(EE[a]) - 'A'][tolower(EE[a]) - 'A'] = 100;
		}
	}
	TC_free_int(matrix);

	return matrix2;
}

int **TC_read_blast_matrix(const char *mat_name)
{
	FILE *fp;
	char sbuf[VERY_LONG_STRING];
	char buf[2];
	char alp[257];
	int **matrix;
	int n_aa, aa1, aa2;
	int value;
	int a, b, c;

	matrix = TC_declare_int(256, 256);
	AR_free(matrix[30]);
	matrix[30] = (int *)AR_allocate(10000, sizeof(int));

	fp = TC_vfopen(mat_name, "r");
	while ((c = fgetc(fp)) == '#' || isspace(c)) {
		fgets(sbuf, VERY_LONG_STRING, fp);
		if (sscanf(sbuf, "ALPHABET=%s", alp) == 1);
	}
	ungetc(c, fp);
	TC_lower_string(alp);
	n_aa = strlen(alp);

	for (a = 0; a < n_aa; a++) {
		fscanf(fp, "%s ", buf);
		aa1 = tolower(buf[0]);
		
		if (aa1 != alp[a]) {
			std::cerr << '\n' << PROGRAM << ": ERROR: Parse error in blast matrix '" << mat_name << "'.\n";
			
			TC_vfclose(fp);
			TC_free_int(matrix);
			
			return 0;
		}
		for (b = 0; b < n_aa; b++) {
			aa2 = tolower((char)alp[b]);
			fscanf(fp, "%d ", &value);
			if (TC_is_gap(aa1) || TC_is_gap(aa2)) {
				int c1, c2;

				c1 = (TC_is_gap(aa1)) ? GAP_CODE : aa1;
				c2 = (TC_is_gap(aa2)) ? GAP_CODE : aa2;
				if (c1 == GAP_CODE && c2 == GAP_CODE)
					matrix[c1][c2] = value;
				else if (c1 == GAP_CODE) {
					matrix[c1][tolower(c2)] = value;
					matrix[c1][toupper(c2)] = value;
				}
				else {
					matrix[tolower(c1)][c2] = value;
					matrix[toupper(c1)][c2] = value;
				}
			}
			else if (aa1 != '*' && aa2 != '*') {
				matrix[tolower(aa1) - 'A'][tolower(aa2) - 'A'] = value;
				matrix[toupper(aa1) - 'A'][toupper(aa2) - 'A'] = value;
				matrix[tolower(aa1) - 'A'][toupper(aa2) - 'A'] = value;
				matrix[toupper(aa1) - 'A'][tolower(aa2) - 'A'] = value;
			}
		}
		fscanf(fp, "\n");
	}
	TC_vfclose(fp);

	return matrix;
}

int **TC_read_pavie_matrix(const char *mat_name)
{
	FILE *fp;
	char sbuf[VERY_LONG_STRING];
	char alp[257];
	char aa1, aa2;
	float v;
	int **matrix;
	int gep = TC_UNDEFINED;
	int c, n_aa;

	matrix = TC_declare_int(256, 256);

	fp = TC_vfopen(mat_name, "r");
	while ((c = fgetc(fp)) == '#' || isspace(c)) {
		fgets(sbuf, VERY_LONG_STRING, fp);
		if (sscanf(sbuf, "ALPHABET=%s", alp) == 1);
	}
	ungetc(c, fp);
	n_aa = strlen(alp);
	while (fgets(sbuf, VERY_LONG_STRING, fp) != 0) {
		aa1 = aa2 = 'Z';
		if (sscanf(sbuf, "%c %c %f", &aa1, &aa2, &v) == 3) {
			v *= PAVIE_MAT_FACTOR;
			if (aa1 == '-' && aa2 == '-')
				gep = (int)v;
			else {
				matrix[tolower(aa1) - 'A'][tolower(aa2) - 'A'] = (int)v;
				matrix[toupper(aa1) - 'A'][toupper(aa2) - 'A'] = (int)v;
				matrix[tolower(aa1) - 'A'][toupper(aa2) - 'A'] = (int)v;
				matrix[toupper(aa1) - 'A'][tolower(aa2) - 'A'] = (int)v;
				matrix[tolower(aa2) - 'A'][tolower(aa1) - 'A'] = (int)v;
				matrix[toupper(aa2) - 'A'][toupper(aa1) - 'A'] = (int)v;
				matrix[tolower(aa2) - 'A'][toupper(aa1) - 'A'] = (int)v;
				matrix[toupper(aa2) - 'A'][tolower(aa1) - 'A'] = (int)v;
			}
		}
	}
	if (gep != TC_UNDEFINED) {
		int a;

		for (a = 0; a < n_aa; a++) {
			if (!matrix[tolower(alp[a]) - 'A'][GAP_CODE]) {
				matrix[tolower(alp[a]) - 'A'][GAP_CODE] = gep;
				matrix[toupper(alp[a]) - 'A'][GAP_CODE] = gep;
			}
		}
	}
	TC_vfclose(fp);

	return matrix;
}

/**
 * @brief Reads the structure id in the header of a pdb file.
 * @param file the name of a pdb file
 * @return the PDB-ID
 */
char *TC_get_pdb_id(const char *file)
{
	FILE *fp;
	char *id;
	char *tmp_name;
	char buf[1000];
	std::ostringstream oss;

	tmp_name = AR_tmpname();

	oss << g_conf.extract_pdb() << " -get_pdb_id " << AR_check_file_exists(file) << " > " << tmp_name;
	system(oss.str().c_str());

	buf[0] = '\0';
	fp = TC_vfopen(tmp_name, "r");
	fscanf(fp, "\n%s\n", buf);
	TC_vfclose(fp);

	id = (char *)AR_allocate(strlen(buf) + 1, sizeof(char));
	strcpy(id, buf);

	return id;
}

TC_Weights *TC_read_seq_weight(char **name, int nseq, const char *seq_weight)
{
	FILE *fp;
	char line[LONG_STRING];
	char sname[MAXNAMES];
	TC_Weights *W = 0;
	float w;
	int a, p;

	/* Read sequence weights:
	 * * comment
	 * name1 weight1
	 * .....
	 *
	 * NOTE:
	 * Weights must be between 0 and 1.
	 * Sequences not in S do not get any weight.
	 * Sequences in S but not in file get a weight of 1. */

	if (!TC_is_single_seq_weight_file(seq_weight)) {
		std::cerr << '\n' << PROGRAM << ": ERROR: File '" << seq_weight
				  << "' is not in 'SINGLE_SEQ_WEIGHT_FORMAT_01' format.\n";
	}
	else {
		W = TC_declare_weights(nseq);

		for (a = 0; a < nseq; a++) {
			strncpy(W->seq_name[a], name[a], 200 - 1);
			W->SEQ_W[a] = 1;
		}
		strncpy(W->mode, seq_weight, FILENAMELEN - 1);

		fp = TC_vfopen(seq_weight, "r");
		while (fgets(line, LONG_STRING - 1, fp)) {
			if (line[0] == '*' || line[0] == '#' || TC_is_blank(line));
			else {
				if (sscanf(line, "%s %f", sname, &w) != 2)
					continue;
				if ((p = TC_name_is_in_list(sname, W->seq_name, nseq, MAXNAMES - 1)) != -1)
					W->SEQ_W[p] = w;
			}
		}
		TC_vfclose(fp);
	}

	return W;
}

static FILE *TC_create_tree(TC_NT_node ptree, TC_NT_node parent, int *nseq, int *ntotal,
	int *nnodes, TC_NT_node **lu, FILE *fp)
{
	float dist = 0;
	float bootstrap = 0;
	char *name;
	int i, type;
	int ch;

	name = (char *)AR_allocate(MAXNAMES + 1, sizeof(char));

	strncpy(name, "   ", MAXNAMES);
	fp = TC_skip_space(fp);

	ch = getc(fp);
	if (ch == '(') {
		type = NODE;
		name[0] = '\0';
		lu[0][ntotal[0]] = lu[1][nnodes[0]] = ptree;
		ntotal[0]++;
		nnodes[0]++;
		TC_create_tree_node(ptree, parent);
		fp = TC_create_tree(ptree->left, ptree, nseq, ntotal, nnodes, lu, fp);
		ch = getc(fp);
		if (ch == ',') {
			fp = TC_create_tree(ptree->right, ptree, nseq, ntotal, nnodes, lu, fp);
			ch = getc(fp);
			if (ch == ',') {
				ptree = TC_insert_tree_node(ptree);
				lu[0][ntotal[0]] = lu[1][nnodes[0]] = ptree;
				ntotal[0]++;
				nnodes[0]++;
				fp = TC_create_tree(ptree->right, ptree, nseq, ntotal, nnodes, lu, fp);
				g_rooted_tree = FALSE;
			}
		}
		fp = TC_skip_space(fp);
		ch = getc(fp);
	}
	else {
		type = LEAF;
		lu[0][ntotal[0]] = lu[2][nseq[0]] = ptree;
		ntotal[0]++;
		nseq[0]++;
		name[0] = ch;
		i = 1;
		
		ch = getc(fp);
		while (ch != ':' && ch != ',' && ch != ')') {
			if (i < MAXNAMES)
				name[i++] = ch;
			ch = getc(fp);
		}
		name[i] = '\0';
		
		if (i >= MAXNAMES + 1) {
			std::cerr << '\n' << PROGRAM << ": ERROR: Tree file contains too long sequence names.\n";
			
			return fp;
		}
	}
	if (ch == ':') {
		fp = TC_skip_space(fp);
		fscanf(fp, "%f", &dist);
		fp = TC_skip_space(fp);
		bootstrap = 0;
	}
	// Tree with bootstrap information
	else if (isdigit(ch)) {
		ungetc(ch, fp);
		fscanf(fp, "%f", &bootstrap);
		if (fscanf(fp, ":%f", &dist) == 1);
		else
			dist = 0;
		fp = TC_skip_space(fp);
	}
	else {
		ungetc(ch, fp);
		TC_skip_space(fp);
	}
	TC_set_info(ptree, parent, type, name, dist, bootstrap);

	AR_free(name);

	return fp;
}

static int TC_tree_file2nseq(const char *fname)
{
	FILE *fp;
	char *string;
	int p, a, b, c, n;

	string = (char *)AR_allocate(TC_count_n_char_in_file(fname) + 1, sizeof(char));

	fp = TC_vfopen(fname, "r");
	n = 0;
	while ((c = fgetc(fp)) != EOF) {
		if (c == '(' || c == ')' || c == ',' || c == ';')
			string[n++] = c;
	}
	TC_vfclose(fp);
	string[n] = '\0';

	for (n = 0, p = 1; string[p]; p++) {
		a = string[p - 1];
		b = string[p];
		if (a == '(' && b == ',')
			n++;
		else if (a == ',' && b == ')')
			n++;
		else if (a == ',' && b == ',')
			n++;
	}
	AR_free(string);

	return n;
}

TC_NT_node **TC_read_tree(const char *treefile, int *tot_node, int nseq, char **seq_names)
{
	FILE *fp;
	TC_NT_node **lu_ptr;
	TC_NT_node seq_tree, root, p;
	int nseq_read = 0;
	int nnodes = 0; // Number of Internal Nodes
	int ntotal = 0; // Number of Internal Nodes + Number of Leaves
	int flag;
	int c_seq;
	char ch;
	int a, b;

	// The Tree Root is in TREE[3][0]
	// TREE[0][ntot]--> pointer to each node and leave

	g_tot_nseq = nseq;
	g_rooted_tree = g_distance_tree = TRUE;

	fp = TC_vfopen(treefile, "r");
	fp = TC_skip_space(fp);
	
	ch = getc(fp);
	if (ch != '(') {
		std::cerr << '\n' << PROGRAM << ": ERROR: Incorrect format of tree file '" << treefile << "'.\n";
		TC_vfclose(fp);
		return 0;
	}

	rewind(fp);

	lu_ptr = (TC_NT_node **)AR_allocate(4, sizeof(TC_NT_node *));
	lu_ptr[0] = (TC_NT_node *)AR_allocate(10 * nseq, sizeof(TC_NT_node));
	lu_ptr[1] = (TC_NT_node *)AR_allocate(10 * nseq, sizeof(TC_NT_node));
	lu_ptr[2] = (TC_NT_node *)AR_allocate(10 * nseq, sizeof(TC_NT_node));
	lu_ptr[3] = (TC_NT_node *)AR_allocate(1, sizeof(TC_NT_node));
	seq_tree = (TC_NT_node)TC_declare_tree_node(nseq);

	TC_set_info(seq_tree, 0, 0, "  ", 0.0, 0);

	fp = TC_create_tree(seq_tree, 0, &nseq_read, &ntotal, &nnodes, lu_ptr, fp);
	fclose(fp);

	if (nseq != g_tot_nseq) {
		std::cerr << '\n' << PROGRAM << ": ERROR: Tree in file '" << treefile << "' is not compatible with alignment.\n"
				  << "There is " << nseq << " sequences in alignment and " << nseq_read << " in tree.\n";
		return 0;
	}

	if (g_distance_tree == FALSE) {
		if (g_rooted_tree == FALSE) {
			std::cerr << '\n' << PROGRAM << ": ERROR: Tree in file '" << treefile << "' is unrooted and has no distances.\n";
			return 0;
		}
	}

	if (g_rooted_tree == FALSE) {
		root = TC_reroot(seq_tree, nseq, ntotal, lu_ptr);
		lu_ptr[1][nnodes++] = lu_ptr[0][ntotal++] = root;
	}
	else {
		root = seq_tree;
	}

	lu_ptr[3][0] = root;
	tot_node[0] = nnodes;

	for (a = 0; a < ntotal; a++) {
		(lu_ptr[0][a])->isseq = (lu_ptr[0][a])->leaf;
		(lu_ptr[0][a])->dp = (lu_ptr[0][a])->dist;
	}
	for (a = 0; a < nseq; a++) {
		if (!seq_names) {
			flag = 1;
			(lu_ptr[2][a])->order = (lu_ptr[2][a])->seq = a;
		}
		else {
			for (flag = 0, b = 0; b < nseq; b++) {
				if (strncmp((lu_ptr[2][a])->name, seq_names[b], MAXNAMES) == 0) {
					flag = 1;
					(lu_ptr[2][a])->order = (lu_ptr[2][a])->seq = b;
					strncpy((lu_ptr[2][a])->name, seq_names[b], MAXNAMES);
				}
			}
		}
	}

	if (seq_names) {
		char *s;
		char **tree_names;
		int fail_flag = 0;
		int tnseq;

		tnseq = TC_tree_file2nseq(treefile);
		tree_names = (char **)AR_allocate(tnseq, sizeof(char *));
		
		for (a = 0; a < tnseq; a++) {
			s = (lu_ptr[2][a])->name;
			tree_names[a] = s;

			if (TC_name_is_in_list(s, seq_names, nseq, MAXNAMES + 1) == -1) {
				std::cerr << '\n' << PROGRAM << ": ERROR: Sequence '" << s << "' in input tree (" << treefile
						  << ") is not in the alignment.\n";
				fail_flag = 1;
			}
		}
		for (a = 0; a < nseq; a++) {
			s = seq_names[a];

			if (TC_name_is_in_list(s, tree_names, nseq, MAXNAMES + 1) == -1) {
				std::cerr << '\n' << PROGRAM << ": ERROR: Input sequence '" << s << "' is not in input tree ("
						  << treefile << ").\n";
				fail_flag = 1;
			}
		}
		AR_free(tree_names);
		
		if (fail_flag == 1)
			return 0;
	}

	for (a = 0; a < nseq; a++) {
		p = lu_ptr[2][a];
		c_seq = p->seq;
		while (p != 0) {
			p->lseq[p->nseq] = c_seq;
			p->nseq++;
			p = p->parent;
		}
	}

	return lu_ptr;
}
