/*
 * File: main.cpp
 * Created: 06.06.2006
 * Last Modified: 16.09.2006
 * Author: Adrian Rospondek
 */


#include <algorithm>
#include <iomanip>
#include <iostream>
#include <numeric>
#include <string>
#include <vector>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "AR_Config.hpp"
#include "AR_node_time.hpp"
#include "JZ_ConstraintList.hpp"
#include "TC_Alignment.hpp"
#include "TC_ConstraintList.hpp"
#include "TC_DistanceMatrix.hpp"
#include "TC_Sequences.hpp"
#include "TC_Templates.hpp"
#include "TC_Weights.hpp"
#include "TC_tree.hpp"
#include "config.hpp"
#include "globals.hpp"
#include "input.hpp"
#include "output.hpp"
#include "par_tree_aln.hpp"
#include "time_log.hpp"
#include "util.hpp"
#include "util_algo.hpp"
#include "util_io.hpp"
#include "util_mem.hpp"
#include "util_mpi.hpp"
#include "util_pdb.hpp"
#include "util_seq.hpp"
#include "util_str.hpp"


#define TC_declare_name(x) x = static_cast<char *>(AR_allocate(FILENAMELEN + 1, sizeof(char)))



const char *TC_get_3dcoffee_defaults()
{
//	return "-in Mslow_pair Msap_pair Mfugue_pair -template_file SELF_P_ ";
    	return "-in Mslow_pair Msap_pair -template_file SELF_P_ ";
}



#if USE_TIMER
yat::TimeLog TL(PROGRAM);
#endif


int main(int argc, char *argv[], char *env[])
{
#if USE_TIMER
	const double start_time = AR_get_time();
	TL.start();
#endif

// 	int provided;
// 	MPI_Init_thread(&argc, &argv, MPI_THREAD_MULTIPLE, &provided);

	MPI_Init(&argc, &argv);

	MPI_Env mpi_env;

	if (mpi_env.size() < 2) {
		std::cerr << "ERROR: " << PROGRAM << " needs at least 2 processors to run.\n";

		MPI_Finalize();
		exit(-1);
	}

	g_conf.read(env);

/* SELECT THE MASTER HOST */
	mpi_env.publish_names();

	if (mpi_env.rank() == 0) {
		int master_rank = 0;

		if (!g_conf.tc_master().empty()) {
			master_rank = mpi_env.rank(g_conf.tc_master());

			if (master_rank == -1)
				master_rank = 0;
		}
		mpi_env.set_root(master_rank);
	}
	mpi_env.ask4root(0);
/* * */

	// cont != 1 causes program to be exited
	// and contains the exit code: 0 if OK or -1 if ERROR occured
	short cont = 1;

/* Parameters */
	char *cache = "";
	const char *seq_source = "";
	const char *inorder = "";
	char *use_tree = "";
	int lalign_n_top;
	const char *lib_list = "";
	int do_normalise;
	int do_extend;
	char *extend_mode = "";
	int max_n_pair;
	const char *compact_mode = "";
	const char *clean_mode = "";
	int maximise;
	int do_self;
	const char *weight = "";
	const char *tree_mode = "";
	char *distance_matrix_mode = "idscore";
	char *dp_mode = "";
	int ktup;
	int fasta_step;
	int diag_mode;
	int diag_threshold;
	char *sim_matrix = "";
	int nomatch;
	int gop, gep;
	int f_gop, f_gep;
	int cosmetic_penalty;
	int tg_mode;
	const char *seq_weight = "t_coffee";
	const char *msa_mode = "";
	char *profile_mode = "";
	char *profile_comparison = "";
	int clean_aln;
	const char *out_aln = "";
	char **out_aln_format = 0;
	int n_out_aln_format;
	const char *residue_case = "";
	const char *outorder = "";
	const char *output_res_num = "";
	int trim;
	int split;
	const char *outseqweight = "";
	const char *evaluate_mode = "";
	char *out_lib = "";
	int lib_only;
	char *tree_file = "";
	int no_report;

	int low_mem = 0;
	int ptc_proc = 1;
	int par_tree = 0;
	int rma_cache = 32;


/* * */
	std::vector<std::string> local_files;
	TC_fname *F = 0;
	char **tot_out_aln = 0;
	char **initial_order = 0;

	TC_Sequences *S = 0;


	argv = AR_app_init(argv, &argc, mpi_env);

	if (mpi_env.am_I_root()) {
		int a;
		bool ok = true;
		int garbage;
		int t_coffee_defaults_flag;
		char *t_coffee_defaults = 0;
		const char *parameters = 0;
		char **special_mode_list = 0;
		int n_special_mode;
		char *extra_params = 0;
		int do_version;
		char **seq_list = 0;
		int n_seq_list;
		char **pdb_list = 0;
		int n_pdb;
		char **profile_list = 0;
		int n_profile_list;
		const char *profile1 = 0;
		const char *profile2 = 0;
		char **aln_file_list = 0;
		int n_aln_file_list;
		char **lib_file_list = 0;
		int n_lib_file_list;
		char **method_list = 0;
		int n_method_list;
		char **list_file = 0;
		int n_list;
		int clean_seq_name;
		int get_type;
		char *type = 0;
		int check_pdb_status;
		char **template_file_list = 0;
		int n_template_file;
		int quicktree;
		const char *run_name = 0;
	/* Parallel T-Coffee */
		int max_n_threads;

		std::cout << "*** " << PROGRAM << " version " << VERSION << " ***\n";

		AR_get_cl_param(argc, argv, ok, "-version", &do_version, "FL", OPTIONAL, 1,
			"Forces the program to output the version number and exit.", &do_version, "0", "1", "0", "1");

		if (do_version) {
			cont = 0;
			goto check_cont;
		}

		TC_declare_name(t_coffee_defaults);
		AR_get_cl_param(argc, argv, ok, "-t_coffee_defaults", &t_coffee_defaults_flag, "R_F", OPTIONAL,
			1, "Get top parameters.", &t_coffee_defaults, "NULL", "NULL", "any", "");

		TC_declare_name(parameters);
		AR_get_cl_param(argc, argv, ok, "-parameters", &garbage, "R_F", OPTIONAL, 1,
			"Get bottom parameters.", &parameters, "NULL", "stdin", "any", "");

		special_mode_list = TC_declare_char(100, STRING);
		n_special_mode = AR_get_cl_param(argc, argv, ok, "-special_mode", &garbage, "S", OPTIONAL, 100,
			"Specifies a special mode.", special_mode_list, "none", "3dcoffee", "none 3dcoffee", "");

/* extra > prompt > special > parameters > defaults */
		argv = TC_break_list(argv, &argc, "=;, \n");
		argv = TC_merge_list(argv, &argc);

		if (argc > 1 && argv[1][0] != '-')
			argv = TC_push_string("-seq ", argv, &argc, 1);

		if (t_coffee_defaults_flag) {
			char *pname = 0;

			if (AR_check_file_exists(t_coffee_defaults))
				pname = t_coffee_defaults;
			else if (pname = const_cast<char *>(g_conf.tc_defaults().c_str())) {
				if (!AR_check_file_exists(pname))
					pname = 0;
			}
			else {
				TC_declare_name(pname);

				snprintf(pname, FILENAMELEN + 1, "%s/.t_coffee_defaults", g_conf.home().c_str());
				if (!AR_check_file_exists(pname)) {
					AR_free(pname);
					pname = 0;
				}
			}

			if (pname) {
				int size = 0;
				char *buf = 0;

				buf = AR_read_file(pname, size);
				argv = TC_push_string(buf, argv, &argc, 1);
				t_coffee_defaults = pname;
			}
			else {
				t_coffee_defaults = 0;
			}
		}

		if (parameters && parameters[0]) {
			int size = 0;
			char *buf = 0;

			buf = AR_read_file(parameters, size);
			argv = TC_push_string(buf, argv, &argc, 1);
		}

		if (n_special_mode) {
			const char *special_mode;
			const char *new_arg;

			for (a = 0; a < n_special_mode; ++a) {
				special_mode = special_mode_list[a];
				new_arg = 0;

				if (special_mode && !special_mode[0]);
				else if (TC_strm(special_mode, "none"));
				else if (TC_strm(special_mode, "3dcoffee"))
					new_arg = TC_get_3dcoffee_defaults();
				else {
					std::cerr << '\n' << PROGRAM << ": WARNING: special mode '" << special_mode << "' is unknown.\n";
					continue;
				}
				argv = TC_push_string(new_arg, argv, &argc, 1);
			}
		}

		AR_free(t_coffee_defaults);
		AR_free(const_cast<char *>(parameters));
		TC_free_char(special_mode_list);

		if (extra_params = const_cast<char *>(g_conf.tc_extra_params().c_str()))
			argv = TC_push_string(extra_params, argv, &argc, argc);

		argv = TC_break_list(argv, &argc, "=;, \n");
		argv = TC_merge_list(argv, &argc);

		TC_declare_name(cache);
		AR_get_cl_param(argc, argv, ok, "-cache", &garbage, "W_F", OPTIONAL, 1,
			"Specifies that a cache must be used to save the structures and their comparison, as well as the blast searches.\nAvailable modes are: use, ignore, update, local, <directory name>",
			&cache, "use", "update", "any", "");

		seq_list = TC_declare_char(1000, STRING);
		n_seq_list = AR_get_cl_param(argc, argv, ok, "-seq", &garbage, "S", OPTIONAL, 1000,
			"List of sequences in any acceptable format.", seq_list, "", "stdin", "any", "");

		pdb_list = TC_declare_char(200, STRING);
		n_pdb = AR_get_cl_param(argc, argv, ok, "-pdb", &garbage, "S", OPTIONAL, 200,
			"Reads/fetch a pdb file: PDBID(PDB_CHAIN)[opt] (FIRST,LAST)[opt],", pdb_list, "", "", "any", "");

		profile_list = TC_declare_char(2000, STRING);
		n_profile_list = AR_get_cl_param(argc, argv, ok, "-profile", &garbage, "S", OPTIONAL, 2000,
			"Input one or many MSA that will be treated as profiles.", profile_list, "", "", "any", "");

		TC_declare_name(profile1);
		AR_get_cl_param(argc, argv, ok, "-profile1", &garbage, "S", OPTIONAL, 1,
			"Input one profile (ClustalW option).", &profile1, "", "", "any", "");

		TC_declare_name(profile2);
		AR_get_cl_param(argc, argv, ok, "-profile2", &garbage, "S", OPTIONAL, 1,
			"Input a profile (ClustalW option).", &profile2, "", "", "any", "");

		aln_file_list = TC_declare_char(1000, STRING);
		n_aln_file_list = AR_get_cl_param(argc, argv, ok, "-aln", &garbage, "S", OPTIONAL, 1000,
			"List of alignments in any acceptable format.", aln_file_list, "", "stdin", "any", "");

		lib_file_list = TC_declare_char(1000, STRING);
		n_lib_file_list = AR_get_cl_param(argc, argv, ok, "-lib", &garbage, "S", OPTIONAL, 1000,
			"List of libraries in any acceptable format.", lib_file_list, "", "stdin", "any", "");

		method_list = TC_declare_char(1000, STRING);
		n_method_list = AR_get_cl_param(argc, argv, ok, "-method", &garbage, "S", OPTIONAL, 1000,
			"List of methods.", method_list, "", "stdin", "any", "");

		list_file = TC_declare_char(2000, STRING);
		n_list = AR_get_cl_param(argc, argv, ok, "-in", &garbage, "S", OPTIONAL, 2000,
			"Reads the Ssequences, Mmethods, Llibraries, Xmatrices, Rprofiles, Pstructures, AAlignments.",
			list_file, (n_method_list > 0 ? "" : "Mlalign_id_pair Mslow_pair"), "stdin", "any", "");

		TC_declare_name(seq_source);
		AR_get_cl_param(argc, argv, ok, "-seq_source", &garbage, "S", OPTIONAL, 1,
			"Indicates the files that will be used as sequence sources, important for dpa. With the default mode alignments must be provided with the Sflag as well as the Aflag if they contribute novel sequences.",
			&seq_source, "ANY", "", "any", "");

		AR_get_cl_param(argc, argv, ok, "-clean_seq_name", &clean_seq_name, "FL", OPTIONAL, 1,
			"Remove Special Char from sequence names.", &clean_seq_name, "0", "1", "0", "1");

		TC_declare_name(inorder);
		AR_get_cl_param(argc, argv, ok, "-inorder", &garbage, "S", OPTIONAL, 1,
			"aligned: sort the sequences in alphabetic order before starting thus making the input order irrelevant but delivering a library in arbitratry order, keep: input order is used in the library but results become input order dependant.",
			&inorder, "aligned", "input", "aligned input", "");

		AR_get_cl_param(argc, argv, ok, "-get_type", &get_type, "FL", OPTIONAL, 1,
			"Forces t_coffee to get the type of the sequences.", &get_type, "0", "1", "0", "1");

		TC_declare_name(type);
		AR_get_cl_param(argc, argv, ok, "-type", &garbage, "S", OPTIONAL, 1,
			"Automatically set, but can be forced with this flag.", &type, "", "", "PROTEIN DNA RNA", "");

		AR_get_cl_param(argc, argv, ok, "-check_pdb_status", &check_pdb_status, "FL", OPTIONAL, 1,
			"Reports the existance of a PDB file.", &check_pdb_status, "0", "1", "0", "1");

		template_file_list = TC_declare_char(100, STRING);
		n_template_file = AR_get_cl_param(argc, argv, ok, "-template_file", &garbage, "S", OPTIONAL, 1000,
			"List of templates file for the sequences.", template_file_list, "", "stdin", "any", "");

		TC_declare_name(use_tree);
		AR_get_cl_param(argc, argv, ok, "-usetree", &garbage, "R_F", OPTIONAL, 1,
			"Use an existing guide tree.", &use_tree, "NULL", "NULL", "any", "");

		AR_get_cl_param(argc, argv, ok, "-lalign_n_top", &garbage, "D", OPTIONAL, 1,
			"Number of local alignments reported by the local method (lalign) when building the library.",
			&lalign_n_top, "10", "10", "any", "");

		TC_declare_name(lib_list);
		AR_get_cl_param(argc, argv, ok, "-lib_list", &garbage, "S", OPTIONAL, 1,
			"A File that contains every pair/group of sequence to process when computing the lib.\nFormat: <nseq> <index1><index2>",
			&lib_list, "", "default", "any", "");

		AR_get_cl_param(argc, argv, ok, "-do_normalise", &garbage, "D", OPTIONAL, 1,
			"Normalisation factor when computing scores.", &do_normalise, "1000", "1000",
			"-10000", "10000");

		AR_get_cl_param(argc, argv, ok, "-extend", &do_extend, "D", OPTIONAL, 1,
			"Do Library extention on the fly.", &do_extend, "1", "1", "any", "");

		TC_declare_name(extend_mode);
		AR_get_cl_param(argc, argv, ok, "-extend_mode", &garbage, "S", OPTIONAL, 1,
			"Library extension mode.", &extend_mode, "very_fast_triplet", "",
			"triplet fast_triplet very_fast_triplet slow_triplet matrix", "");

		AR_get_cl_param(argc, argv, ok, "-max_n_pair", &garbage, "D", OPTIONAL, 1,
			"Indicates the Number of Pairs to Compare when making prf vs prf (0 = every pair).",
			&max_n_pair, "10", "3", "any", "");

		compact_mode = "default";

		clean_mode = "no";

		maximise = 1;

		do_self = 0;

		TC_declare_name(weight);
		AR_get_cl_param(argc, argv, ok, "-weight", &garbage, "S", OPTIONAL, 1,
			"Defines the library weight: sim OR  sim_(matrix).", &weight, "default", "sim", "any", "");

		tree_mode = "nj";


		TC_declare_name(distance_matrix_mode);
		strncpy(distance_matrix_mode, "idscore", FILENAMELEN);

/*
		AR_get_cl_param(argc, argv, ok, "-distance_matrix_mode", &garbage, "S", OPTIONAL, 1,
			"Computation of the distances for the tree.", &distance_matrix_mode, "idscore", "slow",
			"idscore", "");

		AR_get_cl_param(argc, argv, ok, "-quicktree", &quicktree, "FL", OPTIONAL, 0,
			"Use distance_matrix_mode=very_fast.", &quicktree, "0", "1", "0", "1");

		if (quicktree)
			strncpy(distance_matrix_mode, "very_fast", FILENAMELEN);
*/

		TC_declare_name(dp_mode);
		AR_get_cl_param(argc, argv, ok, "-dp_mode", &garbage, "S", OPTIONAL, 1,
			"Type of alignment algorithm used by T-Coffee", &dp_mode, "gotoh_pair_wise", "gotoh_pair_wise",
			"myers_miller_pair_wise gotoh_pair_wise fasta_pair_wise cfasta_pair_wise very_fast_pair_wise", "");

		AR_get_cl_param(argc, argv, ok, "-ktuple", &garbage, "D", OPTIONAL, 1,
			"Word size when using the heursitic dynamic programming modes fasta_pair_wise and cfasta_pair_wise.",
			&ktup, "1", "1", "any", "");

		AR_get_cl_param(argc, argv, ok, "-ndiag", &garbage, "D", OPTIONAL, 1,
			"Number of diagonals to consider when using the heursitic dynamic programming modes fasta_pair_wise and cfasta_pair_wise.",
			&fasta_step, "0", "10", "any", "");

		AR_get_cl_param(argc, argv, ok, "-diag_mode", &garbage, "D", OPTIONAL, 1,
			"0: Use the whole Diag, 1: Use the best match.", &diag_mode, "0", "1", "any", "");

		AR_get_cl_param(argc, argv, ok, "-diag_threshold", &garbage, "D", OPTIONAL, 1, "ND",
			&diag_threshold, "0", "10", "any", "");

		TC_declare_name(sim_matrix);
		AR_get_cl_param(argc, argv, ok, "-sim_matrix", &garbage, "S", OPTIONAL, 1,
			"Degenerated matrix used to compute a similarity.", &sim_matrix, "vasiliky", "idmat",
			"idmat simple polarity vasiliky clustalw clustalw_col clustalw_dot", "");

		nomatch = 0;

		AR_get_cl_param(argc, argv, ok, "-gapopen", &garbage, "D", OPTIONAL, 1,
			"Gap opening penalty. Cosmetic penalty is set to 0 when Gapopen is set. Negative values penalize gaps, positive values reward them.",
			&gop, "0", "", "any", "");

		AR_get_cl_param(argc, argv, ok, "-gapext", &garbage, "D", OPTIONAL, 1,
			"Gap Extension Penalty. Positive values give rewards to gaps and prevent the alignment of unrelated segments.",
			&gep, "0", "", "any", "");

		f_gop = 0;
		f_gep = 0;

		AR_get_cl_param(argc, argv, ok, "-cosmetic_penalty", &garbage, "D", OPTIONAL, 1,
			"A very low Gap Opening Penalty. It only affects the non stable portions of the alignmnent. Negative values penalize gaps, positive values reward them.",
			&cosmetic_penalty, "-50", "-50", "any", "");

		AR_get_cl_param(argc, argv, ok, "-tg_mode", &garbage, "D", OPTIONAL, 1,
			"0: Penalise term gap with gapopen and gapext, 1: gapopen only, 2: No penalty.",
			&tg_mode, "1", "0", "0 1 2", "");

// 		TC_declare_name(seq_weight);
// 		AR_get_cl_param(argc, argv, ok, "-seq_weight", &garbage, "S", OPTIONAL, 1,
// 			"Defines the sequences weighting scheme", &seq_weight, "t_coffee", "t_coffee", "any", "");

		msa_mode = "tree";

		TC_declare_name(profile_mode);
		AR_get_cl_param(argc, argv, ok, "-profile_mode", &garbage, "S", OPTIONAL, 1,
			"Function used to compute profile-to-profile scores.", &profile_mode, "cw_profile_profile", "cw_profile_profile",
			"cw_profile_profile muscle_profile_profile channel_profile_profile", "");

		TC_declare_name(profile_comparison);
		AR_get_cl_param(argc, argv, ok, "-profile_comparison", &garbage, "S", OPTIONAL, 1,
			"Method used to compare two profiles: full<N>: compares <every | N best> pair of sequence and every pair of structure if a structure method is used, profile: compares only the profiles.",
			&profile_comparison, "full50", "profile", "any", "");

		clean_aln = 0;

		TC_declare_name(out_aln);
		tot_out_aln = TC_declare_char(200, STRING);
		AR_get_cl_param(argc, argv, ok, "-outfile", &garbage, "W_F", OPTIONAL, 1,
			"Name of the output alignment.", &out_aln, "default", "", "any", "");

		out_aln_format = TC_declare_char(200, STRING);
		n_out_aln_format = AR_get_cl_param(argc, argv, ok, "-output", &garbage, "S", OPTIONAL, 200,
			"Specifies one or many formats that must be output. The file extension is the output format.",
			out_aln_format, "clustalw", "clustalw", "tcoffee clustalw msa fasta_aln pir_aln", "");

		TC_declare_name(residue_case);
		AR_get_cl_param(argc, argv, ok, "-case", &garbage, "S", OPTIONAL, 1,
			"The case of residues in output file.", &residue_case, "keep", "upper", "keep lower upper", "");

		TC_declare_name(outorder);
		AR_get_cl_param(argc, argv, ok, "-outorder", &garbage, "S", OPTIONAL, 1,
			"Specifies the order of the sequences in the msa.", &outorder, "aligned", "input", "input aligned", "");

		TC_declare_name(output_res_num);
		AR_get_cl_param(argc, argv, ok, "-seqnos", &garbage, "S", OPTIONAL, 1,
			"Adds residue numbers to the MSA.", &output_res_num, "off", "on", "on off", "");

		trim = 0;

		split = 0;

		TC_declare_name(outseqweight);
		AR_get_cl_param(argc, argv, ok, "-outseqweight", &garbage, "W_F", OPTIONAL, 1,
			"Prompts the program to write the sequuence weight values.", &outseqweight, "no", "default", "any", "");

		TC_declare_name(evaluate_mode);
		AR_get_cl_param(argc, argv, ok, "-evaluate_mode", &garbage, "S", OPTIONAL, 1,
			"Mode used to produce the color output.", &evaluate_mode, "tcoffee_fast", "tcoffee_fast",
			"no tcoffee_fast tcoffee_slow matrix categories sar boxshade", "");

/*
		TC_declare_name(out_lib);
		AR_get_cl_param(argc, argv, ok, "-out_lib", &garbage, "W_F", OPTIONAL, 1,
			"Prompts the program to write the computed library file.", &out_lib, "no", "default", "any", "");


		AR_get_cl_param(argc, argv, ok, "-lib_only", &lib_only, "FL", OPTIONAL, 1,
			"Only compute the library.", &lib_only, "0", "1", "0", "1");
*/

		TC_declare_name(tree_file);
		AR_get_cl_param(argc, argv, ok, "-newtree", &garbage, "W_F", OPTIONAL, 1,
			"Name of the output guide tree.", &tree_file, "default", "", "any", "");

		TC_declare_name(run_name);
		AR_get_cl_param(argc, argv, ok, "-run_name", &garbage, "S", OPTIONAL, 1,
			"Sets the prefix of all the output files.", &run_name, "NULL", "", "any", "");

		AR_get_cl_param(argc, argv, ok, "-no_error_report", &no_report, "FL", OPTIONAL, 1,
			"Do not generate the error report file.", &no_report, "0", "1", "0 1", "");

/* ***************************** PARALLEL T-COFFEE PARAMETERS: START ***************************** */

		AR_get_cl_param(argc, argv, ok, "-low_mem", &low_mem, "FL", OPTIONAL, 1,
				"Set low memory mode.", &low_mem, "0", "1", "0 1", "");
		

		AR_get_cl_param(argc, argv, ok, "-par_tree", &par_tree, "FL", OPTIONAL, 1,
				"Perform progressive alignment in parallel.", &par_tree, "0", "1", "0 1", "");

		AR_get_cl_param(argc, argv, ok, "-threads", &garbage, "D", OPTIONAL, 1,
			"Max number of threads used to download remote PDB files.", &max_n_threads, "4", "4", "2", "8");

		AR_get_cl_param(argc, argv, ok, "-ptc_proc", &garbage, "D", OPTIONAL, 1,
				"Number of MPI TCoffee processes executed on single cluster node.", &ptc_proc, "1", "1", "1", "4");

		AR_get_cl_param(argc, argv, ok, "-rma_cache", &garbage, "D", OPTIONAL, 1,
				"Size in MB of cache memory utilised by RMA", &rma_cache, "32", "32", "any", "");

/* ***************************** PARALLEL T-COFFEE PARAMERERS: END   ***************************** */

		if (argc == 1 || TC_name_is_in_list("-help", argv, argc, STRING) != -1) {
			cont = 0;
			goto check_cont;
		}
		AR_get_cl_param(argc, argv, ok, 0, 0, 0, 0, 0, 0);

		if (!ok) {
			cont = -1;
			goto check_cont;
		}

		AR_prepare_cache(cache);

/* ************************* FILL list_file (contains seq, aln and meth) ************************* */

/* Expand -in */
/* Introduce the sequences from the -profile flag */
		if (profile1 && profile1[0]) {
			snprintf(list_file[n_list++], STRING, "R%s", profile1);
		}
		if (profile2 && profile2[0]) {
			snprintf(list_file[n_list++], STRING, "R%s", profile2);
		}
		for (a = 0; a < n_profile_list; a++) {
			FILE *fp;

			if ((fp = TC_find_token_in_file(profile_list[a], 0, "FILE_LIST")) != 0) {
				int z;
				char rname[1000];

				TC_vfclose(fp);
				fp = TC_vfopen(profile_list[a], "r");

				while ((z = fgetc(fp)) != EOF) {
					ungetc(z, fp);
					fscanf(fp, "%s\n", rname);
					if (AR_check_file_exists(rname))
						snprintf(list_file[n_list++], STRING, "R%s", rname);
				}
				TC_vfclose(fp);
			}
			else {
				snprintf(list_file[n_list++], STRING, "R%s", profile_list[a]);
			}
		}
/* Introduce the sequences from the -seq flag */
		for (a = 0; a < n_seq_list; a++) {
			if (AR_check_file_exists(seq_list[a]))
				snprintf(list_file[n_list++], STRING, "S%s", seq_list[a]);
			else if (AR_check_file_exists(seq_list[a] + 1))
				snprintf(list_file[n_list++], STRING, "%s", seq_list[a]);
		}
/* Introduce the alignments from the -aln flag */
		for (a = 0; a < n_aln_file_list; a++) {
			snprintf(list_file[n_list++], STRING, "A%s", aln_file_list[a]);
		}
/* Introduce the methods from the -method flag */
		for (a = 0; a < n_method_list; a++) {
			snprintf(list_file[n_list++], STRING, "M%s", method_list[a]);
		}
/* Introduce the libraries from the -lib flag */
		for (a = 0; a < n_lib_file_list; a++) {
			snprintf(list_file[n_list++], STRING, "L%s", lib_file_list[a]);
		}

		if (n_list == 0 || argc < 2) {
			std::cerr << '\n' << PROGRAM << ": ERROR: You have not provided enough arguments.\n";
			cont = -1;
			goto check_cont;
		}

/* *************************************** INPUT SEQUENCES *************************************** */

		std::cout << "\nREADING SEQUENCES.....\n";

/* GROUP PDB BY LOCATION */
		std::vector<AR_pdb_chain *> all_pdb;
		std::vector<std::string> remote_pdb;
		std::vector<AR_pdb_coords> pdb_coords;
		int location;

		if (n_pdb > 0) {
			std::vector<AR_pdb_chain *>::iterator it;

			AR_parse_pdb_list(pdb_list, n_pdb, all_pdb);

			for (it = all_pdb.begin(); it != all_pdb.end(); ++it) {
				location = AR_check_pdb_location((*it)->pdb_id);

				if (location == REMOTE_PDB) {
					remote_pdb.push_back((*it)->pdb_id);
					pdb_coords.push_back((*it)->coords);
					delete *it;
					*it = 0;
				}
				else if (location == NOT_PDB) {
					delete *it;
					*it = 0;
				}
			}
		}

		for (a = 0; a < n_list; ++a) {
			if (list_file[a][0] == 'P' && !AR_check_file_exists(list_file[a])) {
				std::string pdb(list_file[a] + 1);

				location = AR_check_pdb_location(pdb);
				if (location == REMOTE_PDB)
					remote_pdb.push_back(pdb);
				else if (location == LOCAL_PDB)
					local_files.push_back("P" + pdb);
			}
			else
				local_files.push_back(list_file[a]);
		}

		TC_free_char(seq_list);
		TC_free_char(pdb_list);
		TC_free_char(profile_list);
		AR_free(const_cast<char *>(profile1));
		AR_free(const_cast<char *>(profile2));
		TC_free_char(aln_file_list);
		TC_free_char(lib_file_list);
		TC_free_char(method_list);
		TC_free_char(list_file);
/* * */

		if (!remote_pdb.empty())
			AR_begin_mt_download_pdb(remote_pdb, max_n_threads);

/* READ LOCAL FILES */
		std::vector<AR_pdb_chain *>::iterator it = all_pdb.begin();

		for (; it != all_pdb.end(); ++it)
			if (*it) {
				local_files.push_back("P" + AR_extract_pdb((*it)->pdb_id, ((*it)->coords).start, ((*it)->coords).end));
				delete *it;
				*it = 0;
			}
		all_pdb.clear();

		AR_identify_list_format(local_files.begin(), local_files.end());

		if (type && *type && TC_strm(type, "RNA"))
			type[0] = '\0';

		if (clean_seq_name == 1)
			TC_clean_seq_names(0, 0, ON);

		S = AR_read_seq_from_files(local_files.begin(), local_files.end(), type, seq_source);

/* READ DOWNLOADED FILES */
		if (!remote_pdb.empty()) {
			const int n_remote = remote_pdb.size(), n_local = local_files.size();

			AR_end_mt_download_pdb(remote_pdb);

			std::vector<std::string>::iterator it_rem = remote_pdb.begin() + n_remote;
			std::vector<AR_pdb_coords>::iterator it_crd = pdb_coords.begin();
			TC_Sequences *S1 = 0;

			for (; it_crd != pdb_coords.end(); ++it_crd, ++it_rem)
				local_files.push_back("P" + AR_extract_pdb(*it_rem, it_crd->start, it_crd->end));
			for (; it_rem != remote_pdb.end(); ++it_rem)
				local_files.push_back("P" + *it_rem);

			remote_pdb.clear();
			pdb_coords.clear();

			S1 = AR_read_seq_from_files(local_files.begin() + n_local, local_files.end(), type, seq_source);
			if (S1) {
				// TODO: check if S::type is different from S1::type !
				S = TC_merge_seq(S1, S);
				TC_free_sequence(S1);
			}
		}

		std::cout << "[Done]\n";

/* FATAL: NO SEQUENCES */
		if (!S || S->nseq < 3) {
		    std::cerr << '\n' << PROGRAM << ": ERROR: You have not provided enough sequences.\n";
			cont = -1;
			goto check_cont;
		}

		if (get_type) {
			std::cout << "\nTYPE OF SEQUENCES: " << S->type << '\n';

			TC_free_sequence(S);
			cont = 0;
			goto check_cont;
		}

/* GET THE TEMPLATES */
		std::vector<TC_X_template *> remote_templates;

		if (n_template_file) {
			std::cout << "\nLOOKING FOR SEQUENCE TEMPLATES\n";

			for (a = 0; a < n_template_file; a++) {
				std::cout << "\tTemplate Type: " << std::setw(10) << TC_template_type2type_name(template_file_list[a])
						  << ", Mode or File: " << template_file_list[a] << ".....";
				S = AR_seq2template_seq(S, template_file_list[a], remote_templates);
				std::cout << "[Done]\n";
			}

			if (TC_seq2n_X_template(S, "_R_")) {
				std::cout << "\nLOOKING FOR PROFILE TEMPLATES\n";

				for (a = 0; a < n_template_file; a++) {
					std::cout << "\tTemplate Type: " << std::setw(10) << TC_template_type2type_name(template_file_list[a])
							  << ", Mode or File: " << template_file_list[a] << ".....";
					S = AR_profile_seq2template_seq(S, template_file_list[a], remote_templates);
					std::cout << "[Done]\n";
				}
			}
		}

		TC_free_char(template_file_list);

		if (!remote_templates.empty()) {
			std::vector<TC_X_template *>::iterator it = remote_templates.begin();

			for (; it != remote_templates.end(); ++it)
				remote_pdb.push_back((*it)->template_name);
			AR_begin_mt_download_pdb(remote_pdb, max_n_threads);
		}
/* * */

		std::cout << "\nINPUT FILES:\n";
		for (std::vector<std::string>::iterator it = local_files.begin(); it != local_files.end(); ++it) {
			const char *file = it->c_str();
			char *f = 0;

			std::cout << "\tFile: (" << file[0] << ") " << file + 1;
			if (file[0] == 'A' || file[0] == 'S' || file[0] == 'P' || file[0] == 'R') {
				f = TC_identify_seq_format(file + 1);
				std::cout << "  Format: " << f << '\n';
				AR_free(f);
			}
			else
				std::cout << '\n';
		}

		TC_display_sequences_names(S, stdout, check_pdb_status, TEMPLATES);

/* Fill The F Structure (contains information for output names for the defaults) */
		if (run_name)
			F = TC_parse_fname(run_name);
		else if (argv[1][0] != '-' && (AR_check_file_exists(argv[1]) || AR_check_file_exists(argv[1] + 1))) {
			if (AR_check_file_exists(argv[1]))
				F = TC_parse_fname(argv[1]);
			else if (AR_check_file_exists(argv[1] + 1))
				F = TC_parse_fname(argv[1] + 1);
		}
		else {
			std::vector<std::string>::iterator it = local_files.begin();
			const char *file = 0;

			for (; it != local_files.end(); ++it) {
				file = it->c_str();

				if (!TC_is_method(file)) {
					if (AR_check_file_exists(file)) {
						F = TC_parse_fname(file);
						break;
					}
					else if (TC_is_in_set(file[0], "ASLXRP") && AR_check_file_exists(file + 1)) {
						F = TC_parse_fname(file + 1);
						break;
					}
				}
			}
		}

		if (!run_name)
			F->path[0] = '\0';

		AR_free(const_cast<char *>(run_name));

/* Set Default Names */
		if (F && TC_strm(out_lib, "default"))
			snprintf(out_lib, FILENAMELEN + 1, "%s%s.tc_lib", F->path, F->name);

		if (F && TC_strm(tree_file, "default"))
			snprintf(tree_file, FILENAMELEN + 1, "%s%s.dnd", F->path, F->name);
		else if (TC_strm(tree_file, "no")) {
			AR_free(tree_file);
			tree_file = 0;
		}

		if (!use_tree)
			TC_vremove(tree_file);
		else if (!AR_check_file_exists(use_tree)) {
			AR_free(use_tree);
			use_tree = 0;
		}

		if (F && TC_strm(out_aln, "default")) {
			for (a = 0; a < n_out_aln_format; ++a)
				snprintf(tot_out_aln[a], STRING, "%s%s.%s", F->path, F->name, (TC_strm(out_aln_format[a], "gcg") ?
						"msf" : (TC_strm(out_aln_format[a], "clustalw") ? "aln" : out_aln_format[a])));
		}
		else if (n_out_aln_format == 1)
			strncpy(tot_out_aln[0], out_aln, STRING - 1);
		else {
			for (a = 0; a < n_out_aln_format; ++a)
				snprintf(tot_out_aln[a], STRING, "%s%s.%s", F->path, out_aln, (TC_strm(out_aln_format[a], "gcg") ?
						"msf" : (TC_strm(out_aln_format[a], "clustalw") ? "aln" : out_aln_format[a])));
		}

/* REORDER SEQUENCES */
		initial_order = TC_duplicate_char(S->name, -1, -1);
		if (TC_strm(inorder, "aligned")) {
			char **new_order;

			new_order = TC_duplicate_char(S->name, -1, -1);
			new_order = TC_sort_string_array(new_order, S->nseq);
			S = TC_reorder_seq(S, new_order, S->nseq);

			TC_free_char(new_order);
		}

/* GET DOWNLOADED TEMPLATES */
		if (!remote_templates.empty()) {
			const int n = remote_pdb.size();

			AR_end_mt_download_pdb(remote_pdb);

			std::vector<std::string>::const_iterator it = remote_pdb.begin() + n;
			std::vector<TC_X_template *>::iterator it2 = remote_templates.begin();

			for (; it != remote_pdb.end(); ++it, ++it2)
				AR_fill_P_template(*it2, it->c_str());

			remote_templates.clear();
			remote_pdb.clear();
		}

		S = TC_seq2template_type(S);
	} // am_I_root()

check_cont:
/* BROADCAST THE PARAMS */
	int ibuf_size = 28;
	int *ibuf = new int[ibuf_size];

	if (mpi_env.am_I_root()) {
		ibuf[0] = cont;

		ibuf[1] = lalign_n_top;
		ibuf[2] = lib_only;

		ibuf[3] = strlen(cache) + 1;
		ibuf[4] = strlen(profile_mode) + 1;
		ibuf[5] = strlen(profile_comparison) + 1;
		ibuf[6] = strlen(distance_matrix_mode) + 1;
		ibuf[7] = strlen(dp_mode) + 1;
		ibuf[8] = strlen(sim_matrix) + 1;
		ibuf[9] = strlen(extend_mode) + 1;
		ibuf[10] = ktup;
		ibuf[11] = cosmetic_penalty;
		ibuf[12] = gop;
		ibuf[13] = gep;
		ibuf[14] = maximise;
		ibuf[15] = do_normalise;
		ibuf[16] = do_self;
		ibuf[17] = nomatch;
		ibuf[18] = tg_mode;
		ibuf[19] = diag_mode;
		ibuf[20] = fasta_step;
		ibuf[21] = max_n_pair;
		ibuf[22] = diag_threshold;
		ibuf[23] = ptc_proc;
		ibuf[24] = par_tree;
		ibuf[25] = low_mem;
		ibuf[26] = do_extend;
		ibuf[27] = rma_cache;
	}

	MPI_Bcast(ibuf, ibuf_size, MPI_INT, mpi_env.get_root(), mpi_env.comm());

	if (!mpi_env.am_I_root()) {
		cont = ibuf[0];
		lalign_n_top = ibuf[1];
		lib_only = ibuf[2];
		ktup = ibuf[10];
		cosmetic_penalty = ibuf[11];
		gop = ibuf[12];
		gep = ibuf[13];
		maximise = ibuf[14];
		do_normalise = ibuf[15];
		do_self = ibuf[16];
		nomatch = ibuf[17];
		tg_mode = ibuf[18];
		diag_mode = ibuf[19];
		fasta_step = ibuf[20];
		max_n_pair = ibuf[21];
		diag_threshold = ibuf[22];
		ptc_proc = ibuf[23];
		par_tree = ibuf[24];
		low_mem = ibuf[25];
		do_extend = ibuf[26];
		rma_cache = ibuf[27];
	}

	if (cont != 1) {
		delete[] ibuf;

		MPI_Finalize();
		exit(cont);
	}

	int cbuf_size = 0;
	char *cbuf = 0;

	cbuf_size = std::accumulate(ibuf + 3, ibuf + ibuf_size, cbuf_size);
	cbuf = new char[cbuf_size];

	if (mpi_env.am_I_root()) {
		char *p = cbuf;
		int *l = ibuf + 3;

		p = std::copy(cache, cache + *l, p); ++l;
		p = std::copy(profile_mode, profile_mode + *l, p); ++l;
		p = std::copy(profile_comparison, profile_comparison + *l, p); ++l;
		p = std::copy(distance_matrix_mode, distance_matrix_mode + *l, p); ++l;
		p = std::copy(dp_mode, dp_mode + *l, p); ++l;
		p = std::copy(sim_matrix, sim_matrix + *l, p); ++l;
		p = std::copy(extend_mode, extend_mode + *l, p); ++l;
	}

	MPI_Bcast(cbuf, cbuf_size, MPI_CHAR, mpi_env.get_root(), mpi_env.comm());

	if (!mpi_env.am_I_root()) {
		cache = static_cast<char *>(AR_allocate(ibuf[3], sizeof(char)));
		profile_mode = static_cast<char *>(AR_allocate(ibuf[4], sizeof(char)));
		profile_comparison = static_cast<char *>(AR_allocate(ibuf[5], sizeof(char)));
		distance_matrix_mode = static_cast<char *>(AR_allocate(ibuf[6], sizeof(char)));
		dp_mode = static_cast<char *>(AR_allocate(ibuf[7], sizeof(char)));
		sim_matrix = static_cast<char *>(AR_allocate(ibuf[8], sizeof(char)));
		extend_mode = static_cast<char *>(AR_allocate(ibuf[9], sizeof(char)));

		char *p = cbuf;
		int *l = ibuf + 3;

		std::copy(p, p + *l, cache);
		p += *l; ++l;
		std::copy(p, p + *l, profile_mode);
		p += *l; ++l;
		std::copy(p, p + *l, profile_comparison);
		p += *l; ++l;
		std::copy(p, p + *l, distance_matrix_mode);
		p += *l; ++l;
		std::copy(p, p + *l, dp_mode);
		p += *l; ++l;
		std::copy(p, p + *l, sim_matrix);
		p += *l; ++l;
		std::copy(p, p + *l, extend_mode);

		AR_prepare_cache(cache);
	}

	delete[] ibuf;
	delete[] cbuf;
	AR_free(cache);

/* BROADCAST THE SEQUENCES */
	S = AR_bcast_seq(S, mpi_env);

/* ********************************* PREPARE THE CONSTRAINT LIST ********************************* */

	TC_ConstraintList *CL;
	std::vector<AR_node_time> nd_times;

	CL = TC_declare_constraint_list(S, 0, 0, 0);

#if USE_TIMER
	TL.add("Initialization");
#endif

/* ESTIMATE THE DISTANCE MATRIX */
	CL->DM = AR_par_compute_distance_matrix(CL, "idscore", mpi_env, nd_times);
	broadcast_DM(CL, mpi_env);


#if USE_TIMER
	TL.add("Distance Matrix");
#endif

	if (mpi_env.am_I_root()) {
		std::vector<int> map(nd_times.size());

		// force root to be last in the map
		nd_times[mpi_env.rank()].time = -100.0;

		std::sort(nd_times.begin(), nd_times.end());
		for (unsigned int i = 0; i < nd_times.size(); ++i) {
		    map[i] = nd_times[i].node;
		}

		mpi_env.set_map(map);
	}
/* * */

	strncpy(CL->lib_list, lib_list, 100 - 1);
	CL->lalign_n_top = lalign_n_top;

/* Set the evaluation functions */
	CL->profile_mode = TC_get_profile_mode_function(profile_mode, 0);
	strncpy(CL->profile_comparison, profile_comparison, 100 - 1);

	CL->STRUC_LIST = 0;
	CL->align_pdb_param_file[0] = '\0';
	CL->align_pdb_hasch_mode[0] = '\0';

/* Split parameters */
	CL->split = split;

/* CREATE THE PRIMARY LIBRARY */
#if USE_RMA
	CL->residue_index.cache_size = rma_cache * 1024 * 1024;
	CL = JZ_par_compile_constraint_list(CL, local_files, weight, mpi_env);
#else
	CL = AR_par_compile_constraint_list(CL, local_files, weight, mpi_env);
#endif


#if USE_TIMER
	TL.add("Constraint List");
#endif

	if (mpi_env.am_I_root()) {
	    // if the list is empty
	    if (!CL->M && CL->ne == 0 && (CL->S)->nseq > 1) {
		std::cerr << '\n' << PROGRAM << ": ERROR: No entries in library.\n"
			  << "\tMake sure you provide at least TWO sequences, a method\n"
			  << "\tand there are suitable sequences for this method.\n";
		AR_print_command_line(std::cerr, PROGRAM);
		cont = -1;
	    }
	    
	    if (CL->M) clean_aln = 0;

	    CL->extend_jit = (do_extend > 0) ? 1 : 0;
	    CL->extend_threshold = (do_extend == 1) ? 0 : do_extend;

	    if (CL->extend_jit && CL->extend_threshold != 0) {
		TC_filter_list(CL, 0, CL->ne, CL->extend_threshold);
	    }
	} // am_I_root
#if USE_RMA
	else {
	    if (!CL->M && CL->ne == 0 && (CL->S)->nseq > 1) {
		std::cerr << '\n' << PROGRAM << ": ERROR: No entries in library.\n";
		AR_print_command_line(std::cerr, PROGRAM);
		cont = -1;
	    }
	    
	    if (CL->M) clean_aln = 0;

	    CL->extend_jit = (do_extend > 0) ? 1 : 0;
	    CL->extend_threshold = (do_extend == 1) ? 0 : do_extend;

	    if (CL->extend_jit && CL->extend_threshold != 0) {
		TC_filter_list(CL, 0, CL->ne, CL->extend_threshold);
	    }
	}

	short int r_cont = 0;
	MPI_Allreduce(&cont, &r_cont, 1, MPI_SHORT, MPI_MIN, mpi_env.comm());
	cont = r_cont;
#endif

/* * */

	strncpy(CL->distance_matrix_mode, distance_matrix_mode, 100 - 1);

	AR_free(profile_mode);
	AR_free(profile_comparison);
	AR_free(distance_matrix_mode);

	if (cont != 1) goto final;

	if (!gop && !gep && CL->M) {
	    const char *alph = TC_strm((CL->S)->type, "PROTEIN") ? AA_ALPHABET : "gcuta";

	    gop = static_cast<int>(TC_get_avg_matrix_mm(CL->M, alph) * 10);
	    gep = gop / 10;

	    if (mpi_env.am_I_root())
		std::cout << "\nAUTOMATIC PENALTIES: gapopen= " << gop << " gapext= " << gep << '\n';
	}
	else if (!CL->M && cosmetic_penalty && !gop && !gep) {
	    gop = cosmetic_penalty;
	    gep = 0;
	}

	if (TC_strm((CL->S)->type, "DNA"))
	    ktup = MAX(2, ktup);

	CL->pw_parameters_set = 1;
	CL->normalise = do_normalise;
	CL->do_self = do_self;
	CL->nomatch = nomatch;
	CL->maximise = maximise;
	CL->gop = gop;
	CL->gep = gep;
	CL->f_gop = f_gop;
	CL->f_gep = f_gep;
	CL->ktup = ktup;
	CL->TG_MODE = tg_mode;
	CL->use_fragments = diag_mode;
	CL->fasta_step = fasta_step;
	CL->max_n_pair = max_n_pair;
	CL->diagonal_threshold = diag_threshold;

	mpi_env.set_proc(ptc_proc);
	mpi_env.set_mem(low_mem);

	strncpy(CL->matrix_for_aa_group, sim_matrix, 100 - 1);
	strncpy(CL->dp_mode, dp_mode, 100 - 1);

	CL = TC_choose_extension_mode(extend_mode, CL);

	if (mpi_env.am_I_root()) {
/* WEIGHT CONSTRAINT LIST */
		strncpy(CL->tree_mode, tree_mode, 100 - 1);
		strncpy(CL->evaluate_mode, evaluate_mode, 100 - 1);
		strncpy(CL->extend_clean_mode, clean_mode, 100 - 1);
		strncpy(CL->extend_compact_mode, compact_mode, 100 - 1);

/* Prepare quadruplets, Not supported */
		CL->nseq_for_quadruplet = 0;
		AR_free(CL->seq_for_quadruplet);
		CL->seq_for_quadruplet = 0;

/* ************************************ PREPARE THE ALIGNMENT ************************************ */

		TC_Alignment *A;
		TC_NT_node **T = 0;

		A = TC_seq2aln(CL->S, 0, 1);
		TC_ungap_array(A->seq_al, A->nseq);

#if USE_TIMER
		TL.add("Extended Library");
#endif

		TC_NT_node *SNL = 0;
		int tot_node;

		if (use_tree) {
		    std::cout << "\nREADING PRECOMPUTED TREE '" << use_tree << "'.....";
		    T = TC_read_tree(use_tree, &tot_node, (CL->S)->nseq, (CL->S)->name);
		    std::cout << "[Done]\n";
		}
		else {
		    T = TC_make_tree(CL, CL->S, tree_file);
		    AR_display_output_filename(std::cout, "GUIDE_TREE", "newick", tree_file, STORE);
		}

		if (T) {
		    // HERE WE GO WITH TREE ALIGNMENT
		    if (par_tree == 0) {
#if USE_RMA
			CL->residue_index.mpi_env = &mpi_env;
#endif
			SNL = AR_seq_tree_aln(T, A, CL, seq_weight);
		    }
		    else {
			SNL = AR_par_tree_aln(T, A, CL, mpi_env);
		    }
		}
		else {
		    std::cerr << '\n' << PROGRAM << ": ERROR: Tree could not be made/read.\n";
		}

#if USE_TIMER
		TL.add("Progressive Alignment");
#endif

		A->nseq = (CL->S)->nseq;

/* PREPARE THE ALIGNMENT FOR OUTPUT */
		if (AR_check_file_exists(outorder)) {
			TC_Sequences *OS;

			OS = TC_get_fasta_sequence(outorder);

			if (TC_prf_in_seq(CL->S)) {
				A->expanded_order = OS->name;
				A = TC_reorder_aln(A, (CL->S)->name, (CL->S)->nseq);
			}
			else
				A = TC_reorder_aln(A, OS->name, A->nseq);
		}
		else if (TC_strm(outorder, "aligned") && T) {
			A = TC_reorder_aln(A, A->tree_order, A->nseq);
		}
		else {
			A = TC_reorder_aln(A, initial_order, (CL->S)->nseq);
		}

		A->output_res_num = TC_strm(output_res_num, "on");

		if (TC_strm(residue_case, "keep"))
			A->residue_case = KEEP_CASE;
		else if (TC_strm(residue_case, "lower"))
			A->residue_case = LOWER_CASE;
		else if (TC_strm(residue_case, "upper"))
			A->residue_case = UPPER_CASE;
		else
			A->residue_case = UPPER_CASE;

		print_A(A);


/* OUTPUT THE ALIGNMENT */
 		TC_Alignment *EA = A;

		for (int a = 0; a < n_out_aln_format; ++a) {
			TC_output_format_aln(out_aln_format[a], A, EA, tot_out_aln[a]);
			AR_display_output_filename(std::cout, "MSA", out_aln_format[a], tot_out_aln[a], STORE);
		}

		TC_free_aln(A);
	} // am_I_root()
	else { // WORKER
 	    // HERE IS WORKER PART FOR TREE ALIGNMENT
  	    if (par_tree == 1) {
		TC_Alignment *A;

   		A = TC_seq2aln(CL->S, 0, 1);
    		TC_ungap_array(A->seq_al, A->nseq);
		
    		AR_par_tree_aln(0, A, CL, mpi_env);
	    }
#if USE_RMA
	    else {
		CL->residue_index.mpi_env = &mpi_env;
		CL = JZ_weight_constraint_list(CL, seq_weight, mpi_env);
		CL = AR_index_res_constraint_list(CL, WE);
	    }
#endif
 	} // WORKER


#if USE_RMA
	MPI_Barrier(MPI_COMM_WORLD);
	G_STORAGE.free();
#endif

final:
	S = TC_free_constraint_list(CL);
	TC_free_sequence(S);


#if USE_TIMER
	const int total_time = static_cast<int>(AR_get_time() - start_time);

	TL.stop();

	if (mpi_env.am_I_root()) {
		AR_print_command_line(std::cout, PROGRAM);
		std::cout << PROGRAM << ": total time:  "
				  << (total_time / 3600) << "h:"
				  << ((total_time % 3600) / 60) << "m:"
				  << ((total_time % 3600) % 60) << "s\n";
		TL.print(std::cout);
	}
#endif

	MPI_Finalize();
	return 0;
}
