/*
 * File: methods.cpp
 * Created: 06.06.2006
 * Last Modified: 13.09.2006
 * Author: Adrian Rospondek
 */


#include <iostream>
#include <sstream>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "AR_Config.hpp"
#include "AR_LibJob.hpp"
#include "TC_Alignment.hpp"
#include "TC_ConstraintList.hpp"
#include "TC_Method.hpp"
#include "TC_Profile.hpp"
#include "TC_Sequences.hpp"
#include "TC_Templates.hpp"
#include "dp_drivers.hpp"
#include "globals.hpp"
#include "input.hpp"
#include "methods.hpp"
#include "util.hpp"
#include "util_algo.hpp"
#include "util_io.hpp"
#include "util_mem.hpp"
#include "util_seq.hpp"
#include "util_str.hpp"



TC_ConstraintList *AR_seq2list(AR_LibJob *job)
{
	TC_Alignment *A = 0;
	TC_ConstraintList *CL, *PW_CL, *RCL;
	TC_Method *M;
	const char *mode;
	const char *weight;
	const int *seq_list;
	int full_prf, nprf;

	M = job->param.TCM;
	mode = M->executable;
	weight = M->weight;
	PW_CL = M->PW_CL;
	seq_list = job->param.seq_c; // [2]
	CL = job->io.CL;

	// Proteins
	if (strncmp(CL->profile_comparison, "full", 4) == 0) {
		full_prf = 1;
		if (CL->profile_comparison[4])
			nprf = atoi(CL->profile_comparison + 4);
		else
			nprf = 0;
	}
	else {
		full_prf = 0;
	}

	if ((TC_method_uses_structure(M) || full_prf) && (TC_seq2R_template_profile(CL->S, seq_list[0])
			|| TC_seq2R_template_profile(CL->S, seq_list[1])) && !TC_method_uses_profile(M)) {
		RCL = AR_profile2list(job, nprf);
	}
	else if (TC_strm(mode, "slow_pair") || TC_strm(mode, "lalign_id_pair")) {
		A = AR_fast_pair(job);
		RCL = TC_aln2constraint_list((A->A) ? A->A : A, CL, weight);

		TC_free_aln(A);
	}
	// Structural methods
	else if (TC_strm(mode, "sap_pair")) {
		RCL = AR_sap_pair(seq_list, CL);
	}
	else if (TC_strm(mode, "fugue_pair")) {
		RCL = AR_fugue_pair(seq_list, CL);
	}
	else {
		RCL = 0;
	}

	if (!RCL)
		RCL = CL;

	return RCL;
}

static int *TC_get_name_index(char **l1, int n1, char **l2, int n2)
{
	int *r;
	int a;
	
	r = (int *)AR_allocate(n1, sizeof(int));
	
	for (a = 0; a < n1; ++a)
		r[a] = TC_name_is_in_list(l1[a], l2, n2, 100);
	
	return r;
}

TC_ConstraintList *AR_profile2list(AR_LibJob *job, int nprf)
{
	static short int *entry;
//	static short int **L;
	static int max_L_len;
	TC_Alignment *A1, *A2, *A;
	TC_ConstraintList *CL;
	TC_ConstraintList *SCL, *SCL2;
	TC_Method *M;
	const int *seq_list;
	const char *weight;
	const char *command;
	int *cache1, *cache2;
	int *cons_table;
	int *iA1, *iA2;
	int **score, n, s1, s2, si, r1, r2;
	int max, n_pairs;
	int cons, cons_thres, max_n_pairs, tot_n_pairs;
	int a, b, c;
	
	CL = job->io.CL;
	M = job->param.TCM;
	command = M->executable;
	weight = M->weight;
	seq_list = job->param.seq_c; // [2]
	
	if (!entry)
		entry = (short int *)AR_allocate(CL->entry_len, sizeof(short int));
	
	entry[SEQ1] = seq_list[0];
	entry[SEQ2] = seq_list[1];
	
	A1 = TC_seq2profile(CL->S, seq_list[0]);
	A2 = TC_seq2profile(CL->S, seq_list[1]);
	
	SCL = TC_duplicate_constraint_list_soft(CL);
//	SCL->L = L;
	SCL->max_L_len = max_L_len;
	SCL->M = 0;
	SCL->ne = 0;
	
	SCL2 = TC_duplicate_constraint_list_soft(CL);
//	SCL2->L = L;
	SCL2->max_L_len = max_L_len;
	SCL2->M = 0;
	SCL2->ne = 0;
	
	SCL->S = TC_merge_seq(A1->S, 0);
	SCL->S = TC_merge_seq(A2->S, SCL->S);
	
	// 1: Compare the two profiles and identify the N master pairs
	n = (A1->nseq * A2->nseq);
	max = (nprf == 0 || TC_method_uses_structure(M)) ? n : MIN(nprf, n);

	if (max < n) {
		TC_vsrand(0);

		A = TC_align_two_aln(A1, A2, "blosum62mt", -4, -1, "myers_miller_pair_wise");
		score = TC_declare_int(A1->nseq * A2->nseq, 3);
		
		for (n = 0, a = 0; a < A1->nseq; a++)
			for (b = 0; b < A2->nseq; b++, n++) {
				score[n][0] = a;
				score[n][1] = b;
				score[n][2] = (int)TC_get_seq_fsim(A->seq_al[a], A->seq_al[b + A1->nseq], "-", 0, 0, AVERAGE_POSITIONS);
			}
		
		TC_free_aln(A);
		TC_sort_int_inv(score, 3, 2, 0, n - 1);
	}
	else {
		score = TC_declare_int(n, 3);
		
		for (n = 0, a = 0; a < A1->nseq; a++)
			for (b = 0; b < A2->nseq; b++, n++) {
				score[n][0] = a;
				score[n][1] = b;
			}
	}
	
	iA1 = TC_get_name_index(A1->name, A1->nseq, (SCL->S)->name, (SCL->S)->nseq);
	iA2 = TC_get_name_index(A2->name, A2->nseq, (SCL->S)->name, (SCL->S)->nseq);
	
	// Submit the N pairs
	for (n_pairs = 1, a = 0; a < max; ++a) {
		s1 = score[a][0];
		s2 = score[a][1];

		job->param.seq_c[0] = iA1[s1];
		job->param.seq_c[1] = iA2[s2];
		
		// 2: Compute the pairewise library
		job->io.CL = SCL;
		SCL = AR_seq2list(job);
		
		// 3: Update the main library with the pairwise library
		cache1 = TC_seq2inv_pos(A1->seq_al[s1]);
		cache2 = TC_seq2inv_pos(A2->seq_al[s2]);
		
		n_pairs += (SCL->ne > 0) ? 1 : 0;
		for (c = 0; c < SCL->ne; c++) {
			si = TC_vread_clist(SCL, c, SEQ1);
			r1 = TC_vread_clist(SCL, c, (si == iA1[s1]) ? R1 : R2);
			r2 = TC_vread_clist(SCL, c, (si == iA1[s1]) ? R2 : R1);
			entry[R1] = cache1[r1];
			entry[R2] = cache2[r2];
			entry[WE] = TC_vread_clist(SCL, c, WE);
			entry[CONS] = 1;
			TC_add_entry2list(entry, SCL2);
		}
		SCL->ne = 0;
		
		AR_free(cache1);
		AR_free(cache2);
		
		TC_compact_list(SCL2, 0, SCL2->ne, "default");
	}

	TC_free_sequence(SCL->S);
	AR_free(iA1);
	AR_free(iA2);
	
	TC_compact_list(SCL2, 0, SCL2->ne, "default");
	
	// Get the concistency distribution
	cons_table = (int *)AR_allocate(101, sizeof(int));
	
	for (c = 0; c < SCL2->ne; c++) {
		entry[R1] = TC_vread_clist(SCL2, c, R1);
		entry[R2] = TC_vread_clist(SCL2, c, R2);
		entry[WE] = TC_vread_clist(SCL2, c, WE);
		entry[CONS] = TC_vread_clist(SCL2, c, CONS);
		cons = (entry[CONS] * 100) / n_pairs;
		cons_table[cons]++;
	}
	
	// Identify the threshold
	max_n_pairs = (MIN(A1->len_aln, A2->len_aln) * 2);
	for (cons_thres = 0, tot_n_pairs = 0, c = 100; c >= 0; c--) {
		tot_n_pairs += cons_table[c];
		if (tot_n_pairs >= max_n_pairs) {
			cons_thres = c;
			c = -1;
		}
	}
	
	AR_free(cons_table);
	
	// Produce the library
	for (c = 0; c < SCL2->ne; c++) {
		entry[R1] = TC_vread_clist(SCL2, c, R1);
		entry[R2] = TC_vread_clist(SCL2, c, R2);
		entry[WE] = TC_vread_clist(SCL2, c, WE);
		entry[CONS] = TC_vread_clist(SCL2, c, CONS);
		entry[WE] /= entry[CONS];
		cons = (entry[CONS] * 100) / n_pairs;
		if (cons >= cons_thres)
			TC_add_entry2list(entry, CL);
	}
	
	if (!TC_seq2R_template_profile(CL->S, seq_list[0]))
		TC_free_aln(A1);
	if (!TC_seq2R_template_profile(CL->S, seq_list[1]))
		TC_free_aln(A2);
	
//	TC_free_short_int(SCL->L);
	SCL->L.free();
	TC_free_constraint_list(SCL);
//	TC_free_short_int(SCL2->L);
	SCL2->L.free();
	TC_free_constraint_list(SCL2);
	
	job->io.CL = CL;
	
	return CL;
}

TC_Alignment *AR_fast_pair(const AR_LibJob *job)
{
	static int *ns;
	static int **l_s;
	TC_Alignment *A;
	TC_ConstraintList *CL, *PW_CL;
	TC_Sequences *S;
	TC_Method *M;
	const int *seq_list;
	char **buf;
	int s, n, a;

	M = job->param.TCM;
	PW_CL = M->PW_CL;
	seq_list = job->param.seq_c; // [2]
	CL = job->io.CL;
	A = job->io.A;

	n = 2; // number of seq, currently always 2
	if (n != 2) {
		std::cerr << '\n' << PROGRAM << ": ERROR: Method '" << M->executable
				  << "' can handle only two sequences at time.\n";
		return A;
	}

	S = CL->S;

	if (!A)
		A = TC_declare_aln(S);

	if (!ns) {
		ns = (int *)AR_allocate(2, sizeof(int));
		l_s = TC_declare_int(2, S->nseq);
	}
	buf = (char **)AR_allocate(S->nseq, sizeof(char *));

	for (a = 0; a < n; ++a) {
		s = seq_list[a];

		if (TC_strm(M->seq_type, "G")) {
			buf[s] = S->seq[s];
			S->seq[s] = (((S->T[s])->G)->S)->seq[0];
		}
		else
			buf[s] = 0;

		strncpy(A->seq_al[a], S->seq[s], A->declared_len);
		strncpy(A->name[a], S->name[s], STRING - 1);
		A->order[a][0] = s;
	}

	A->S = S;
	PW_CL->S = S;
	A->CL = CL;
	ns[0] = ns[1] = 1;
	l_s[0][0] = 0;
	l_s[1][0] = 1;

	TC_pair_wise(A, ns, l_s, PW_CL);
	A->nseq = n;

	for (a = 0; a < S->nseq; ++a) {
		if (buf[a])
			S->seq[a] = buf[a];
	}
	AR_free(buf);

	return A;
}

static int TC_is_sap_file(const char *name)
{
	FILE *fp;
	
	if (!AR_check_file_exists(name))
		return 0;
	
	if ((fp = TC_find_token_in_file(name, 0, "Percent")) != 0) {
		if ((fp = TC_find_token_in_file(name, fp, "Percent")) != 0) {
			TC_vfclose(fp);
			return 1;
		}
		else
			return 0;
	}
	else
		return 0;
}

static char *TC_normalize_pdb_file(const char *name, const char *seq, const char *out_file)
{
	TC_Sequences *S;
	TC_Alignment *A;
	int start, end, npdb, r1, r2;
	int a;
	
	if (!TC_is_pdb_file(name))
		std::cerr << '\n' << PROGRAM << ": ERROR: '" << name << "' is not a PDB file.\n";
	
	S = TC_get_pdb_sequence(name);
	A = TC_align_two_sequences(S->seq[0], seq, "idmat", -3, 0, "fasta_pair_wise");
	
	for (start = -1, end = -1, npdb = 0, a = 0; a < A->len_aln; ++a) {
		r1 = 1 - TC_is_gap(A->seq_al[0][a]);
		r2 = 1 - TC_is_gap(A->seq_al[1][a]);
		npdb += r1;
		
		if (r1 && r2 && start == -1)
			start = npdb;
		if (r1 && r2)
			end = npdb;
	}
	TC_free_aln(A);
	TC_free_sequence(S);
	
	std::ostringstream oss;

	oss << g_conf.extract_pdb() << " -infile " << AR_check_file_exists(name) << " -atom ALL -chain FIRST -coor "
		<< start << ' ' << end << " -nodiagnostic > " << out_file;
	system(oss.str().c_str());
	
	return const_cast<char *>(out_file);
}

TC_ConstraintList *AR_sap_pair(const int seq[2], TC_ConstraintList *CL)
{
	TC_X_template *P = 0;
	const char *template1, *template2;
	int s1, s2;

	s1 = seq[0];
	s2 = seq[1];
	template1 = (P = TC_seq_has_template(CL->S, s1, "_P_")) ? P->template_name : 0;
	template2 = (P = TC_seq_has_template(CL->S, s2, "_P_")) ? P->template_name : 0;

	if (!template1 || !template2)
		return CL;

	char tmp_name1[128], tmp_name2[128];
	const char *tmp_name = 0;
	char *tmp_pdb1, *tmp_pdb2;

	snprintf(tmp_name1, 128, "%s_%s.sap_results", template1, template2);
	snprintf(tmp_name2, 128, "%s_%s.sap_results", template2, template1);

	if (TC_is_sap_file(tmp_name1)) {
		tmp_name = tmp_name1;
	}
	else if (TC_is_sap_file(tmp_name2)) {
		int x;

		tmp_name = tmp_name2;
		SWAPP(s1, s2, x);
	}
	else {
		char full_name[FILENAMELEN + 128];
		std::ostringstream oss;

		tmp_name = tmp_name1;

		tmp_pdb1 = TC_normalize_pdb_file(TC_seq2P_template_file(CL->S, s1), (CL->S)->seq[s1], AR_tmpname());
		tmp_pdb2 = TC_normalize_pdb_file(TC_seq2P_template_file(CL->S, s2), (CL->S)->seq[s2], AR_tmpname());

		snprintf(full_name, FILENAMELEN + 128, "%s%s", AR_get_cache_dir().c_str(), tmp_name);
		oss << g_conf.sap() << ' ' << tmp_pdb1 << ' ' << tmp_pdb2 << " > " << full_name;
		system(oss.str().c_str());

		if (!TC_is_sap_file(full_name)) {
			std::cerr << '\n' << PROGRAM << ": WARNING: SAP failed to align " << TC_seq2P_template_file(CL->S, s1)
					  << " against " << TC_seq2P_template_file(CL->S, s2) << ".\n";
			if (AR_check_file_exists(full_name))
				TC_add2file2remove_list(full_name);

			return CL;
		}

		remove("super.pdb");
	}

	unsigned int max_struc_len = 10000;
	unsigned int max_buf_len = 4096;
	char *sap_seq1, *sap_seq2;
	FILE *fp;
	char buf[max_buf_len];
	int r1, r2;
	int sim = 0, tot = 0, score = 0;

	sap_seq1 = (char *)AR_allocate(max_struc_len, sizeof(char));
	sap_seq2 = (char *)AR_allocate(max_struc_len, sizeof(char));

	fp = TC_find_token_in_file(tmp_name, 0, "Percent");
	fp = TC_find_token_in_file(tmp_name, fp, "Percent");
	while (fgetc(fp) != '\n');

	while (fgets(buf, max_buf_len, fp)) {
		if (!strstr(buf, "eighted")) {
			TC_remove_charset(buf, "!alnum");
			r1 = buf[0];
			r2 = buf[strlen(buf) - 1];
		}

		sim += (r1 == r2) ? 1 : 0;
		if (tot > max_struc_len) {
			max_struc_len += max_struc_len;
			sap_seq1 = (char *)AR_reallocate(sap_seq1, max_struc_len, sizeof(char));
			sap_seq2 = (char *)AR_reallocate(sap_seq2, max_struc_len, sizeof(char));
		}

		sap_seq1[tot] = r1;
		sap_seq2[tot] = r2;
		++tot;
	}
	TC_vfclose(fp);

	score = (sim * 100) / tot;
	sap_seq1[tot] = sap_seq2[tot] = '\0';

	const char *sap_lib;

	fp = TC_vfopen(sap_lib = AR_tmpname(), "w");
	fprintf(fp, "! TC_LIB_FORMAT_01\n");
	fprintf(fp, "2\n");
	fprintf(fp, "%s %d %s\n", (CL->S)->name[s2], (int)strlen(sap_seq1), sap_seq1);
	fprintf(fp, "%s %d %s\n", (CL->S)->name[s1], (int)strlen(sap_seq2), sap_seq2);
	fprintf(fp, "#1 2\n");

	score = 100;
	for (int a = 0; a < tot; ++a)
		fprintf(fp, "%d %d %d 1 0\n", a + 1, a + 1, score);

	fprintf(fp, "! CPU 0\n");
	fprintf(fp, "! SEQ_1_TO_N\n");
	TC_vfclose(fp);

	CL = TC_read_constraint_list_file(CL, sap_lib);
	TC_vremove(sap_lib);

	AR_free(sap_seq1);
	AR_free(sap_seq2);

	return CL;
}

static TC_ConstraintList *TC_fugue_pair2(int s1, int s2, TC_ConstraintList *CL)
{
	TC_X_template *P = 0;
	const char *pdb_id;

	P = TC_seq_has_template(CL->S, s1, "_P_");
	if (!P || (pdb_id = P->pdb_id)[0] == '\0')
		return CL;

	char fugue_result[STRING + 21];

	snprintf(fugue_result, STRING + 21, "%s_%s.fugue_results", pdb_id, (CL->S)->name[s2]);
	
	if (!AR_check_file_exists(fugue_result)) {
		const char *seq_file;
		FILE *fp;

		seq_file = AR_tmpname();
		fp = TC_vfopen(seq_file, "w");
		fprintf(fp, ">%s\n%s\n", (CL->S)->name[s2], (CL->S)->seq[s2]);
		TC_vfclose(fp);

		std::ostringstream oss;

		oss << "fugue_client -pdb " << pdb_id << " -pep " << seq_file << " > " << AR_get_cache_dir() << fugue_result;
		TC_mysystem(oss.str().c_str());
	}
	
	TC_Alignment *F;

	F = TC_main_read_aln(fugue_result, 0);
	if (F) {
		strncpy(F->name[0], (CL->S)->name[s1], STRING - 1);
		strncpy(F->name[1], (CL->S)->name[s2], STRING - 1);

		F = TC_fix_aln_seq(F, CL->S);
		CL = TC_aln2constraint_list(F, CL, "sim");
		
		TC_free_aln(F);
	}
	else
		std::cerr << '\n' << PROGRAM << ": WARNING: Method 'fugue_pair' failed.\n";

	return CL;
}

TC_ConstraintList *AR_fugue_pair(const int seq[2], TC_ConstraintList *CL)
{
	int s1 = seq[0];
	int s2 = seq[1];
	
	CL = TC_fugue_pair2(s1, s2, CL);
	CL = TC_fugue_pair2(s2, s1, CL);
	
	return CL;
}
