/*
 * File: output.cpp
 * Created: 06.06.2006
 * Last Modified: 13.09.2006
 * Author: Adrian Rospondek
 */


#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "AR_Config.hpp"
#include "TC_Alignment.hpp"
#include "TC_ConstraintList.hpp"
#include "TC_Sequences.hpp"
#include "TC_Weights.hpp"
#include "config.hpp"
#include "globals.hpp"
#include "output.hpp"
#include "util.hpp"
#include "util_io.hpp"
#include "util_mem.hpp"
#include "util_seq.hpp"
#include "util_str.hpp"



int TC_is_out_format_list(const char *name)
{
	static const char *out_formats[] = {
		"tc_aln", "tcoffee_aln", "tcoffee",
		"aln", "clustalw_aln", "clustalw", "msa",
		"fasta_aln", "fasta",
		"pir_aln", "pir"
	};
	static const int n_out_formats = sizeof(out_formats) / sizeof(out_formats[0]);
	int i;

	if (name == 0 || name[0] == '\0')
		return 0;

	for (i = 0; i < n_out_formats; ++i)
		if (strcmp(name, out_formats[i]) == 0)
			break;

	return (i < n_out_formats);
}

void AR_main_output_aln(TC_Alignment *A, const char *format, const char *fname)
{
	if (A == 0 || format == 0 || format[0] == '\0')
		return;

	if (TC_strm3(format, "tcoffee", "tc_aln", "tcoffee_aln"))
		AR_output_tc_aln(fname, A);
	else if (TC_strm3(format, "clustalw", "aln", "clustalw_aln") || TC_strm(format, "msa"))
		TC_output_clustal_aln(fname, A);
	else if (TC_strm2(format, "fasta_aln", "fasta"))
		TC_output_fasta_aln(fname, A);
	else if (TC_strm2(format, "pir_aln", "pir"))
		TC_output_pir_aln(fname, A);
	else
		std::cerr << '\n' << PROGRAM << ": ERROR: Unknown aln output format '" << format << "'.\n";
}

static FILE *TC_output_string_wrap(int wrap, const char *string, FILE *fp)
{
	int b;

	for (b = 1; *string; ++string, ++b) {
		fprintf(fp, "%c", *string);
		if (b == wrap) {
			fprintf(fp, "\n");
			b = 0;
		}
	}

	return fp;
}

static FILE *TC_output_aln_without_header(TC_Alignment *B, FILE *fp)
{
	char s;
	int *n_residues;
	int max_len = 0;
	int line;
	int a, b, c;

	if (fp == 0)
		return fp;

	for (a = 0; a < B->nseq; a++) {
		if (strlen(B->name[a]) > max_len)
			max_len = strlen((B->name[a]));
	}
	max_len = MAX(max_len + 2, 16);

	line = g_conf.aln_width();

	n_residues = (int *)AR_allocate(B->nseq + 1, sizeof(int));
	for (a = 0; a < B->nseq; a++)
		n_residues[a] = (B->output_res_num == 2) ? B->order[a][1] : 0;

	fprintf(fp, "\n");
	for (a = 0; a < B->len_aln; a += line) {
		for (b = 0; b <= B->nseq; b++) {
			fprintf(fp, "%-*s", max_len, B->name[b]);
			if (B->output_res_num)
				fprintf(fp, " %4d ", n_residues[b] + 1);
			for (c = a; c < a + line && c < B->len_aln; c++) {
				if (b == B->nseq) {
					n_residues[b]++;
					s = TC_analyse_aln_column(B, c);
				}
				else {
					n_residues[b] += !TC_is_gap(B->seq_al[b][c]);
					s = GET_CASE(B->residue_case, B->seq_al[b][c]);
				}

				fprintf(fp, "%c", s);
			}
			if (B->output_res_num)
				fprintf(fp, " %4d", n_residues[b]);
			fprintf(fp, "\n");
		}
		fprintf(fp, "\n");
	}
	fprintf(fp, "\n\n");
	AR_free(n_residues);

	return fp;
}

void AR_output_tc_aln(const char *fname, TC_Alignment *A)
{
	FILE *fp;

	if (!A)
		return;

	fp = TC_vfopen(fname, "w");
	fprintf(fp, "%s, %s (%s)\n%s\nCPU   %d sec\nSCORE %d\nNSEQ  %d\nLEN   %d\n", PROGRAM, VERSION,
		DATE, AUTHOR, (A->cpu + TC_get_time()) / 1000, A->score_aln, A->nseq, A->len_aln);
	fp = TC_output_aln_without_header(A, fp);
	TC_vfclose(fp);
}

void TC_output_clustal_aln(const char *fname, TC_Alignment *A)
{
	FILE *fp;
	int *n_residues;
	int max_len = 0;
	int line = 0;
	int a, b, c;

	if (getenv("TCOFFEE_SEP")) {
		while (line < A->len_aln && A->seq_al[0][line] != 'o' && A->seq_al[0][line] != 'O')
			line++;
		if (A->seq_al[0][line] == 'O' || A->seq_al[0][line] == 'o')
			line++;
	}
	else {
		while (line < A->len_aln)
			line++;
	}

	if (line == A->len_aln)
		line = g_conf.aln_width();

	n_residues = (int *)AR_allocate(A->nseq + 1, sizeof(int));
	for (a = 0; a < A->nseq; a++) {
		if (strlen(A->name[a]) > max_len)
			max_len = strlen((A->name[a]));
		n_residues[a] = A->order[a][1];
	}
	max_len = MAX(max_len + 2, 16);

	fp = TC_vfopen(fname, "w");

	fprintf(fp, "CLUSTAL FORMAT for %s %s, Nseq=%d, Len=%d ", PROGRAM,
		VERSION, A->nseq, A->len_aln);
	fprintf(fp, "\n\n");

	if (A->len_aln == 0) {
		for (b = 0; b <= A->nseq; b++)
			fprintf(fp, "%-*s -\n", max_len, A->name[b]);
	}
	else {
		for (a = 0; a < A->len_aln; a += line) {
			for (b = 0; b <= A->nseq; b++) {
				if (b != A->nseq) {
					fprintf(fp, "%-*s", max_len, A->name[b]);
					for (c = a; c < a + line && c < A->len_aln; c++) {
						if (TC_is_gap(A->seq_al[b][c]))
							fprintf(fp, "%c", '-');
						else {
							n_residues[b]++;
							fprintf(fp, "%c", GET_CASE(A->residue_case, A->seq_al[b][c]));
						}
					}
					if (A->output_res_num)
						fprintf(fp, " %d", n_residues[b]);
					fprintf(fp, "\n");
				}
				else if (b == A->nseq) {
					fprintf(fp, "%-*s", max_len, " ");
					for (c = a; c < a + line && c < A->len_aln; c++) {
						fprintf(fp, "%c", TC_analyse_aln_column(A, c));
					}
					fprintf(fp, "\n");
				}
			}
			fprintf(fp, "\n");
		}
	}
	fprintf(fp, "\n\n");

	AR_free(n_residues);
	TC_vfclose(fp);
}

void TC_output_fasta_aln(const char *fname, const TC_Alignment *A)
{
	FILE *fp;
	int line = 0;
	int a;

	line = g_conf.aln_width();
	fp = TC_vfopen(fname, "w");

	for (a = 0; a < A->nseq; a++) {
		fprintf(fp, ">%s", A->name[a]);
		if (A->seq_comment[a][0] && !TC_is_blank(A->seq_comment[a]))
			fprintf(fp, "\t%s", A->seq_comment[a]);
		fprintf(fp, "\n");
		fp = TC_output_string_wrap(line, A->seq_al[a], fp);
		fprintf(fp, "\n");
	}
	TC_vfclose(fp);
}

void TC_output_pir_aln(const char *fname, const TC_Alignment *A)
{
	FILE *fp;
	const char *type = 0;
	int a;

	fp = TC_vfopen(fname, "w");
	for (a = 0; a < A->nseq; a++) {
		if (TC_strm(TC_get_string_type(A->seq_al[a]), "DNA"))
			type = "DL";
		else if (TC_strm(TC_get_string_type(A->seq_al[a]), "PROTEIN"))
			type = "P1";

		fprintf(fp, ">%s;%s\n%s\n", type, A->name[a], A->seq_comment[a]);
		fp = TC_output_string_wrap(50, A->seq_al[a], fp);
		fprintf(fp, "\n*\n");
	}
	TC_vfclose(fp);
}

void TC_output_fasta_seq(const char *fname, const TC_Alignment *A)
{
	FILE *fp;
	int a;

	fp = TC_vfopen(fname, "w");

	for (a = 0; a < A->nseq; a++) {
		TC_ungap(A->seq_al[a]);

		fprintf(fp, ">%s", A->name[a]);
		if (A->seq_comment[a][0] && !TC_is_blank(A->seq_comment[a]))
			fprintf(fp, "\t%s\n", A->seq_comment[a]);
		else
			fprintf(fp, "\n");
		fp = TC_output_string_wrap(50, A->seq_al[a], fp);
		fprintf(fp, "\n");
	}
	TC_vfclose(fp);
}

const char *TC_output_fasta_seqX(const char *name, const char *mode, const TC_Sequences *S,
	const TC_Alignment *A, int i)
{
	FILE *fp;

	if (!name)
		name = AR_tmpname();
	
	fp = TC_vfopen(name, mode);
	if ((S && S->nseq <= i) || (A && S->nseq <= i) || !A && !S) {
		std::cerr << '\n' << PROGRAM << ": ERROR: No sequence and alignment (output_fasta_seqX).\n";
	}
	else if (S)
		fprintf(fp, ">%s %s\n%s\n", S->name[i], S->seq_comment[i], S->seq[i]);
	else if (A) {
		TC_ungap(A->seq_al[i]);
		fprintf(fp, ">%s %s\n%s\n", A->name[i], A->seq_comment[i], A->seq_al[i]);
	}
	TC_vfclose(fp);

	return name;
}

static FILE *TC_save_list_header(FILE *OUT, const TC_ConstraintList *CL)
{
	int nseq = 0;
	int a;

	for (a = 0; a < (CL->S)->nseq; a++)
		nseq += ((CL->S)->len[a] != -1);

	fprintf(OUT, "! TC_LIB_FORMAT_01\n%d\n", nseq);
	for (a = 0; a < (CL->S)->nseq; a++)
		if ((CL->S)->len[a] != -1)
			fprintf(OUT, "%s %d %s\n", (CL->S)->name[a], (CL->S)->len[a], (CL->S)->seq[a]);

	return OUT;
}

static FILE *TC_save_list_footer(FILE *OUT, const TC_ConstraintList *CL)
{
	if (CL->cpu)
		fprintf(OUT, "! CPU %d\n", TC_get_time());
	fprintf(OUT, "! SEQ_1_TO_N\n");

	return OUT;
}

static FILE *TC_save_constraint_list_ascii(FILE *OUT, const TC_ConstraintList *CL, const int start,
	const int len, const int *translation)
{
	int a, b, s1, s2;
	int x1, x2;

	if (len == start && CL->cpu != -1) {
		fprintf(OUT, "! CPU %d\n", TC_get_time());
		return OUT;
	}
	else {
		s1 = translation[TC_vread_clist(CL, start, SEQ1)];
		s2 = translation[TC_vread_clist(CL, start, SEQ2)];
		if (s1 != -1 && s2 != -1)
			fprintf(OUT, "#%d %d\n", s1 + 1, s2 + 1);

		for (a = start; a < (len + start); a++) {
			x1 = translation[TC_vread_clist(CL, a, SEQ1)];
			x2 = translation[TC_vread_clist(CL, a, SEQ2)];

			if (x1 != -1 && x2 != -1) {
				if (x1 != s1 || x2 != s2) {
					s1 = x1;
					s2 = x2;
					fprintf(OUT, "#%d %d\n", s1 + 1, s2 + 1);
				}

				for (b = 2; b < CL->entry_len; b++)
					fprintf(OUT, "%5d ", TC_vread_clist(CL, a, b));
				fprintf(OUT, "\n");
			}
		}
	}

	return TC_save_list_footer(OUT, CL);
}

FILE *TC_save_constraint_list(const TC_ConstraintList *CL, const char *fname, FILE *fp,
	const char *mode, const TC_Sequences *S)
{
	static int *translation;
	int a, b;

	if (fp == 0) {
		if (translation != 0)
			AR_free(translation);
		translation = (int *)AR_allocate((CL->S)->nseq + 1, sizeof(int));
		
		for (b = 0, a = 0; a < (CL->S)->nseq; a++) {
			if (TC_name_is_in_list((CL->S)->name[a], S->name, S->nseq, 100) == -1) {
				(CL->S)->len[a] = -1;
				translation[a] = -1;
			}
			else {
				translation[a] = b++;
			}
		}
	}
	if (TC_strm2(mode, "lib", "ascii")) {
		if (fp == 0)
			fp = TC_vfopen(fname, "w");
		fp = TC_save_list_header(fp, CL);
		fp = TC_save_constraint_list_ascii(fp, CL, 0, CL->ne, translation);
	}
	else {
		std::cerr << '\n' << PROGRAM << ": ERROR: Unknown output mode '" << mode << "'.\n";
	}

	return fp;
}

int TC_output_seq_weights(const TC_Weights *W, const char *wfile)
{
	FILE *fp;
	int a;

	if (W == 0)
		return 0;

	fp = TC_vfopen(wfile, "w");
	if (fp == 0)
		return 0;

	for (a = 0; a < W->nseq; ++a) {
		fprintf(fp, "%s %.2f\n", W->seq_name[a], W->SEQ_W[a]);
	}
	TC_vfclose(fp);

	return 1;
}
