/*
 * File: util.cpp
 * Created: 06.06.2006
 * Last Modified: 19.09.2006
 * Author: Adrian Rospondek
 */


#include <fstream>
#include <sstream>
#include <math.h>
#include <signal.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/times.h>
#include <time.h>
#include <unistd.h>
#include "AR_Config.hpp"
#include "config.hpp"
#include "globals.hpp"
#include "perl_scripts.hpp"
#include "util.hpp"
#include "util_io.hpp"
#include "util_mem.hpp"
#include "util_mpi.hpp"
#include "util_str.hpp"



std::string g_cache;
static int g_rank = -1;
static bool g_do_report = true;
static std::string g_report_file;
static char *g_in_cl = 0;
static const long g_msec_conv = 1000;
static long g_ticks_per_sec = 0, g_time_ref = 0;


static void AR_app_finalize()
{
	AR_display_output_filename(std::cout, 0, 0, 0, FLUSH);
	TC_rm_tmp_files();
	AR_free_all();
}

static void TC_error_function(int signum = -1)
{
	static const char *nd = "";
	static const char *sig_desc[] = {
		"MPI Error",
		nd, nd, nd, // 1-3
		"Illegal instruction",
		nd, nd, // 5,6
		"Buss error (bad memory access)",
		"Floating point exception",
		nd, nd, // 9,10
		"Invalid memory reference"
	};

	std::cerr << "\n ****************************************************************************";
	std::cerr << "\n * " << PROGRAM << ": PROGRAM CRASHED (" << (signum > -1 ? sig_desc[signum] : nd) << ')';
	std::cerr << "\n *   Node " << g_rank;
	std::cerr << "\n ****************************************************************************";
	std::cerr << "\n * Please CHECK:";
	std::cerr << "\n *   1. The format of your Input Files.";
	std::cerr << "\n *   2. The parameters.";
	std::cerr << "\n *   3. The use of special characters in sequence names:";
	std::cerr << "\n *      (@, |, %, ...).";
	std::cerr << "\n *   4. The Online Doc (" << URL << ").";

	if (g_do_report) {
		std::ofstream f(g_report_file.c_str(), std::ios_base::app);

		f << "\n</ErrorReport>\n";
		f.close();

		std::ostringstream oss;

		oss << PROGRAM << '.' << getpid() << ".report";
		std::string name = oss.str();

		oss.str("");
		oss << "cp " << g_report_file << ' ' << name;
		system(oss.str().c_str());

		std::cerr << "\n *   5. Send the file " << name;
		std::cerr << "\n *      to " << MAIL;
		std::cerr << "\n * Please NOTE:";
		std::cerr << "\n *   File " << name << " may contain your PERSONNAL DATA.";
		std::cerr << "\n *   Remove all confidential data from this file before sending it.";
	}

	std::cerr << "\n ****************************************************************************\n";
	AR_print_command_line(std::cerr, PROGRAM);

	exit(-1);
}

static void AR_mpi_error_function(MPI_Comm *comm, int *x, ...)
{
	TC_error_function(0);
}

char **AR_app_init(char **in_argv, int *in_argc, const MPI_Env &mpi_env)
{
	static bool init_done = false;

	if (in_argv == 0)
		return 0;
	else if (init_done)
		return in_argv;
	else
		init_done = true;

	g_rank = mpi_env.rank();

/*
	// signal handlers
	signal(SIGSEGV, TC_error_function);
	signal(SIGFPE, TC_error_function);
	signal(SIGILL, TC_error_function);
	signal(SIGBUS, TC_error_function);

	// MPI error handler
	mpi_env.set_error_handler(AR_mpi_error_function);

	// standard exit
	atexit(AR_app_finalize);
*/

	// prepare the tmp directory
	if (AR_mkdir(g_conf.tc_tmp()) == false)
		g_conf.tc_tmp("./");

	char **out_argv = in_argv;

	if (mpi_env.am_I_root()) {
		int a, c;

		for (c = 0, a = 0; a < in_argc[0]; ++a)
			c += strlen(in_argv[a]) + 3;

		g_in_cl = (char *)AR_allocate(c, sizeof(char));

		for (a = 0; a < in_argc[0]; ++a) {
			strcat(g_in_cl, in_argv[a]);
			strcat(g_in_cl, " ");
		}

		out_argv = TC_break_list(in_argv, in_argc, "=;, \n");		
		g_do_report = TC_name_is_in_list("-no_error_report", out_argv, in_argc[0], 20) == -1;

		// create report file: contains every input file
		g_report_file = AR_tmpname();

		if (g_do_report) {
		    std::ofstream f(g_report_file.c_str());

		    f << "<ErrorReport>\n";
		    f << "\n<Program name=\"" << PROGRAM << "\" version=\"" << VERSION << "\"/>\n";
		    f << "\n<MpiEnv size=\"" << mpi_env.size() << "\">\n";
		    f << "\t<Master name=\"" << mpi_env.name() << "\" rank=\"" << mpi_env.rank() << "\"/>\n";
		    f << "</MpiEnv>\n";

		    f << "\n<CommandLine>\n\t";
		    for (a = 0; a < in_argc[0]; ++a)
			f << out_argv[a] << ' ';
		    f << "\n</CommandLine>\n";

		    for (a = 1; a < in_argc[0]; ++a) {
			if (AR_check_file_exists(out_argv[a]) && !TC_filename_is_special(out_argv[a])) {
			    f << "\n<InputFile name=\"" << out_argv[a] << "\">\n";
			    
			    int size;
			    char *buf = AR_read_file(out_argv[a], size);
			    
			    f << buf << '\n';
			    f << "</InputFile>\n";
			    
			    AR_free(buf);
			}
		    }

		    f.close();
		}
	}

	// initialize the time
	TC_get_time();

	return out_argv;
}

static void TC_ignore_cache()
{
	std::ostringstream oss;

	oss << "rm -rf " << AR_get_cache_dir();
	system(oss.str().c_str());
}

static void TC_update_cache()
{
	const std::string old_cache = AR_get_cache_dir();
	std::ostringstream oss;

	AR_prepare_cache("use");
	oss << "mv " << old_cache << "* " << AR_get_cache_dir()
		<< "; rm -rf " << old_cache;

	system(oss.str().c_str());
}

const std::string &AR_prepare_cache(const char *mode)
{
	if (TC_strm(mode, "local")) {
		g_cache = "./";
	}
	else {
		if (TC_strm(mode, "use")) {
			g_cache = g_conf.tc_cache();
		}
		else if (TC_strm(mode, "ignore") || TC_strm(mode, "no")) {
			g_cache = AR_tmpname();
			atexit(TC_ignore_cache);
		}
		else if (TC_strm(mode, "update")) {
			g_cache = AR_tmpname();
			atexit(TC_update_cache);
		}
		else {
			g_cache = mode;
		}

		if (*g_cache.rbegin() != '/')
			g_cache += '/';

		if (AR_mkdir(g_cache) == false)
			g_cache = "./";
	}

	return g_cache;
}

static std::ostream &AR_print_array_char(std::ostream &out, char **array, int n, const char *sep)
{
	if (array == 0 || TC_read_array_size(array, sizeof(char *)) < n)
		return out;

	for (int a = 0; a < n; ++a)
		out << array[a] << sep;

	return out;
}

/**
 * @param argc n_ arg
 * @param argv list
 * @param ok false if error occured
 * @param para_name param
 * @param set_flag set to 1 if param set
 * @param type FL, D, F, S, R_F (read file_name), W_F (written file_name)
 * @param optional 1 for yes, 0 for no
 * @param max_n_val maximum number of values
 * @param usage usage list with optional value
 */
/* @param val pointer to the varaible holding the value(s)
 * @param default1 default value (if value id not there)
 * @param default2 default value if the flag is there but no value set ("") indicates an error
 * @param range_left min value ("any" for any)
 * @param range_right max value (must be "" if range_left is "any")
 */
int AR_get_cl_param(int argc, char **argv, bool &ok, const char *para_name, int *set_flag, const char *type,
	int optional, int max_n_val, const char *usage, ...)
{
	va_list ap;
	int *int_val;
	float *float_val;
	char **string_val;
	char *range_right;
	char *range_left;
	char *default_value1;
	char *default_value2;
	int n_para = 0;
	double max, min;
	char **para_name_list;
	int n_para_name;
	char **para_val;
	int n_para_val;
	char **pv_l = 0;
	int n_pv_l;
	char **pv_r = 0;
	int n_pv_r;
	char value[STRING];
	int pos;
	int a;

	static char **parameter_list;
	static int number_of_parameters;

	// check that all params in argv exist
	if (para_name == 0) {
		for (a = 1; a < argc; ++a) {
			if (TC_is_parameter(argv[a]) && TC_name_is_in_list(argv[a], parameter_list, number_of_parameters, STRING) == -1)
				std::cerr << '\n' << PROGRAM << ": WARNING: '" << argv[a] << "' is not a parameter of " << PROGRAM << ".\n";
		}
		TC_free_char(parameter_list);

		return 0;
	}

	if (parameter_list == 0)
		parameter_list = TC_declare_char(MAX_N_PARAM, STRING);

	para_name_list = TC_get_list_of_tokens(para_name, 0, &n_para_name);
	for (a = 0; a < n_para_name; ++a) {
		strncpy(parameter_list[number_of_parameters++], para_name_list[a], STRING - 1);
	}
	TC_free_char(para_name_list);

	set_flag[0] = 0;
	va_start(ap, usage);

	if (TC_strm3(type, "S", "R_F", "W_F"))
		string_val = va_arg(ap, char **);
	else if (TC_strm2(type, "D", "FL"))
		int_val = va_arg(ap, int *);
	else if (TC_strm(type, "F"))
		float_val = va_arg(ap, float *);
	else
		return 0;

	default_value1 = va_arg(ap, char *);
	default_value2 = va_arg(ap, char *);
	range_left = va_arg(ap, char *);
	range_right = va_arg(ap, char *);

	va_end(ap);

	para_name_list = TC_get_list_of_tokens(para_name, 0, &n_para_name);
	for (a = 0; a < n_para_name; a++) {
		if ((pos = TC_name_is_in_list(para_name_list[a], argv, argc, STRING)) != -1)
			break;
	}
	TC_free_char(para_name_list);

	if (TC_name_is_in_list("-help", argv, argc, STRING) != -1
			&& (argc == 2 || TC_name_is_in_list(para_name, argv, argc, STRING) != -1)) {
		std::cout << "PARAMETER   : " << para_name << '\n';
		std::cout << "USAGE       : " << usage << '\n';
		std::cout << "MAX_N_VALUES: " << max_n_val << '\n';
		std::cout << "DEFAULT     : " << default_value1 << " OR " << default_value2 << " (when flag set)\n";
		std::cout << "RANGE       : [" << range_left << ", " << (*range_right ? range_right : "\b\b") << "]\n";
		std::cout << "TYPE        : " << type << "\n\n";

		return 0;
	}
	else if (TC_name_is_in_list("-help", argv, argc, STRING) != -1) {
		return 0;
	}
	else if (para_name[0] != '-') {
		std::cerr << "\nWrong parameter definition: '" << para_name << "' must start with a dash.\n";
		return 0;
	}
	else if (pos == -1) {
		if (optional == OPTIONAL) {
			set_flag[0] = 0;
			para_val = TC_get_list_of_tokens(default_value1, 0, &n_para_val);

			for (n_para = 0; n_para < n_para_val && !TC_strm(default_value1, "NULL"); ++n_para) {
				if (TC_strm(para_val[n_para], "")) {
					set_flag[0] = 0;
					break;
				}
				else if (TC_strm(type, "FL")) {
					set_flag[0] = atoi(para_val[n_para]);
					break;
				}
				else if (TC_strm3(type, "S", "R_F", "W_F")) {
					strcpy(string_val[n_para], para_val[n_para]);
				}
				else if (TC_strm(type, "D"))
					int_val[n_para] = atoi(para_val[n_para]);
				else if (TC_strm(type, "F"))
					float_val[n_para] = atof(para_val[n_para]);
			}
			TC_free_char(para_val);

			if (n_para == 0 && TC_strm3(type, "S", "W_F", "R_F") && TC_strm(default_value1, "NULL")) {
				AR_free(string_val[0]);
				string_val[0] = 0;
			}
			else if (n_para == 0 && TC_strm(type, "D") && TC_strm(default_value1, "NULL"))
				int_val[0] = 0;
			else if (n_para == 0 && TC_strm(type, "F") && TC_strm(default_value1, "NULL"))
				float_val[0] = 0;
		}
		else {
			std::cerr << '\n' << PROGRAM << ": ERROR: Parameter '" << para_name << "' is not optional.\n";
			ok = false;
			return 0;
		}
	}
	else if (pos != -1) {
		set_flag[0] = 1;

		for (a = pos + 1; a < argc; ++a) {
			if (TC_is_parameter(argv[a]))
				break;
			else {
				if (n_para >= max_n_val) {
					n_para = max_n_val - 1;
				}
				if (!TC_strm(argv[a], "NULL"))
					if (TC_strm3(type, "S", "R_F", "W_F")) {
						strcpy(string_val[n_para], argv[a]);
					}
					else if (TC_strm(type, "D")) {
						int_val[n_para] = atoi(argv[a]);
					}
					else if (TC_strm(type, "F")) {
						float_val[n_para] = atof(argv[a]);
					}
				n_para++;
			}
		}

		if (n_para == 0 && !TC_strm2(default_value2, "", "NULL") && !TC_strm(type, "FL")) {
			para_val = TC_get_list_of_tokens(default_value2, 0, &n_para_val);

			for (n_para = 0; n_para < n_para_val; ++n_para) {
				if (TC_strm3(type, "S", "R_F", "W_F"))
					strcpy(string_val[n_para], para_val[n_para]);
				else if (TC_strm(type, "D"))
					int_val[n_para] = atoi(para_val[n_para]);
				else if (TC_strm(type, "F"))
					float_val[n_para] = atof(para_val[n_para]);
			}
			TC_free_char(para_val);
		}
		else if (n_para == 0 && TC_strm(type, "FL"));
		else if (n_para == 0 && TC_strm3(type, "S", "W_F", "R_F") && TC_strm(default_value2, "NULL")) {
			AR_free(string_val[0]);
			string_val[0] = 0;
		}
		else if (n_para == 0 && TC_strm(type, "D") && TC_strm(default_value2, "NULL"))
			int_val[0] = 0;
		else if (n_para == 0 && TC_strm(type, "F") && TC_strm(default_value2, "NULL"))
			float_val[0] = 0;
		else if (n_para == 0 && TC_strm(default_value2, "")) {
			std::cerr << '\n' << PROGRAM << ": ERROR: Parameter '" << para_name << "' needs a value.\n";
			ok = false;
			return 0;
		}
	}

	// Check that parameters are in good range
	pv_l = TC_get_list_of_tokens(range_left, 0, &n_pv_l);
	pv_r = TC_get_list_of_tokens(range_right, 0, &n_pv_r);

	if (!pv_l) {
		pv_l = TC_declare_char(1, 4);
		strncpy(pv_l[0], "any", 4 - 1);
		n_pv_l = 1;
	}
	if (!pv_r) {
		pv_r = TC_declare_char(1, 1);
		pv_r[0][0] = '\0';
		n_pv_r = 1;
	}

	for (a = 0; a < n_para; ++a) {
		if (TC_strm(pv_l[0], "any"));
		else if (TC_strm(type, "R_F") && !AR_check_file_exists(string_val[a])
				&& !AR_check_file_exists(string_val[a] + 1)) {
			std::cerr << '\n' << PROGRAM << ": ERROR: File '" << string_val[a] << "' does not exists.\n";
			ok = false;
			return 0;
		}
		else if (TC_strm(type, "D")) {
			if (n_pv_l == 1) {
				min = static_cast<double>(atoi(pv_l[0]));
				max = static_cast<double>(atoi(pv_r[0]));

				if (int_val[a] < min || int_val[a] > max) {
					std::cerr << '\n' << PROGRAM << ": ERROR: Parameter '" << para_name << "' is out of range ["
							  << static_cast<int>(min) << ", " << static_cast<int>(max) << "].\n";
					ok = false;
					return 0;
				}
			}
			else {
				sprintf(value, "%d", int_val[a]);

				if (TC_name_is_in_list(value, pv_l, n_pv_l, STRING) == -1) {
					std::cerr << '\n' << PROGRAM << ": ERROR: Parameter '" << para_name << "' is out of set {";
					AR_print_array_char(std::cerr, pv_l, n_pv_l, ", ");
					std::cerr << "\b\b}.\n";
					ok = false;
					return 0;
				}
			}
		}
		else if (TC_strm(type, "F")) {
			if (n_pv_l == 1) {
				min = static_cast<double>(atof(range_left));
				max = static_cast<double>(atof(range_right));

				if (float_val[a] < min || float_val[a] > max) {
					std::cerr << '\n' << PROGRAM << ": ERROR: Parameter '" << para_name << "' is out of range ["
							  << static_cast<float>(min) << ", " << static_cast<float>(max) << "].\n";
					ok = false;
					return 0;
				}
			}
			else {
				sprintf(value, "%f", float_val[a]);

				if (TC_name_is_in_list(value, pv_l, n_pv_l, STRING) == -1) {
					std::cerr << '\n' << PROGRAM << ": ERROR: Parameter '" << para_name << "' is out of set {";
					AR_print_array_char(std::cerr, pv_l, n_pv_l, ", ");
					std::cerr << "\b\b}.\n";
					ok = false;
					return 0;
				}
			}
		}
		else if (TC_strm(type, "S")) {
			if (TC_name_is_in_list(string_val[a], pv_l, n_pv_l, STRING) == -1) {
				std::cerr << '\n' << PROGRAM << ": ERROR: Parameter '" << para_name << "' is out of set {";
				AR_print_array_char(std::cerr, pv_l, n_pv_l, ", ");
				std::cerr << "\b\b}.\n";
				ok = false;
				return 0;
			}
		}
	}

	TC_free_char(pv_l);
	TC_free_char(pv_r);

	return n_para;
}

std::ostream &AR_print_command_line(std::ostream &out, const char *comment)
{
	out << '\n' << comment << ": cmd:  " << g_in_cl << '\n';
	return out;
}

std::ostream &AR_print_cpu_usage(std::ostream &out, const char *comment)
{
	out << '\n' << comment << ": CPU usage:  " << TC_get_time() << "ms\n";
	return out;
}

static void TC_add_file2error_report(const char *name)
{
	if (!g_do_report)
		return;

	std::ofstream f(g_report_file.c_str(), std::ios_base::app);
	int size;
	char *buf;

	f << "\n<OutputFile name=\"" << name << "\">\n";

	buf = AR_read_file(name, size);
	f << buf << '\n';
	f << "</OutputFile>\n";

	AR_free(buf);
	f.close();
}

std::ostream &AR_display_output_filename(std::ostream &out, const char *type, const char *format, const char *name,
	int check_output)
{
	static int nbuf = 0;
	static char ***buf = 0;

	if (TC_strm(name, "stdout") || TC_strm(name, "stderr"))
		return out;

	if (check_output == STORE) {
		if (buf == 0)
			buf = (char ***)AR_allocate(1000, sizeof(char **));

		buf[nbuf] = TC_declare_char(3, 1000);
		strncpy(buf[nbuf][0], type, 1000 - 1);
		strncpy(buf[nbuf][1], format, 1000 - 1);
		strncpy(buf[nbuf][2], name, 1000 - 1);
		nbuf++;
	}
	else if (check_output == FLUSH) {
		if (nbuf > 0)
			out << "\nOUTPUT FILES:\n";

		for (int a = 0; a < nbuf; ++a) {
			AR_display_output_filename(out, buf[a][0], buf[a][1], buf[a][2], CHECK);
			TC_free_char(buf[a]);
		}
		nbuf = 0;
	}
	else if (check_output == CHECK) {
		out << "\tFile: (" << type << ") " << name << "  Format: " << format;

		if (AR_check_file_exists(name))
			TC_add_file2error_report(name);
		else
			out << "  [Not Produced]";
		out << '\n';
	}

	return out;
}

int TC_vsrand(int val)
{
	static int initialized;

	if (initialized)
		return 0;
	else {
		srand((val == 0) ? time(0) : val);
		initialized = 1;
	}

	return 1;
}

int TC_get_time()
{
	static long time;
	struct tms time_buf;
	long tms_stime, tms_utime;

	if (g_ticks_per_sec == 0)
		g_ticks_per_sec = sysconf(_SC_CLK_TCK);

	times(&time_buf);
	tms_stime = (long)time_buf.tms_stime * g_msec_conv;
	tms_utime = (long)time_buf.tms_utime * g_msec_conv;

	if (g_time_ref == 0) {
		g_time_ref = (tms_stime + tms_utime);
		return 0;
	}
	else {
		time = (tms_utime + tms_stime) - g_time_ref;
		return (int)(time / g_ticks_per_sec);
	}
}

int TC_mysystem(const char *cmd)
{
	static const int n_perl_scripts = sizeof(PerlScriptName) / sizeof(PerlScriptName[0]);
	static int unpacked_scripts[n_perl_scripts];
	const char *params = 0;
	int n, i;

	for (params = cmd; *params && *params != ' '; ++params);
	n = params - cmd;

	for (i = 0; i < n_perl_scripts; ++i)
		if (strncmp(PerlScriptName[i], cmd, n) == 0)
			break;

	std::string new_cmd;

	if (i < n_perl_scripts) {
		new_cmd = g_conf.tc_tmp() + PerlScriptName[i];

		if (unpacked_scripts[i] == FALSE) {
			const char *p = new_cmd.c_str();
			std::ofstream f(p);

			f << PERL_HEADER << "\n\n" << PerlScriptFile[i] << '\n';
			f.close();

			chmod(p, S_IRUSR | S_IXUSR | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH);

			TC_add2file2remove_list(p);
			unpacked_scripts[i] = TRUE;
#if DEBUG_MYSYSTEM
			std::cout << "\nSCRIPT '" << PerlScriptName[i] << "' UNPACKED\n";
#endif
		}
		new_cmd += params;
	}
	else
		new_cmd = cmd;

#if DEBUG_MYSYSTEM
	std::cout << "\nRUNNING " << new_cmd << '\n';
#endif

	return system(new_cmd.c_str());
}

static char *TC_clean_seq_name(char *name, int sw)
{
	const std::string orig = name;
	bool modif = false;

	for (char *s = name; *s; ++s) {
		if (!strrchr(NAMES_ALPHABET, *s)) {
			modif = true;
			if (sw == ON)
				*s = '_';
		}
	}

	if (modif) {
		if (sw == ON)
			std::cerr << '\n' << PROGRAM << ": WARNING: Renamed sequence '" << orig << "' into '" << name << "'.\n"
					  << "\tUse -clean_seq_name 0 to keep special characters in original names.\n";
		else
			std::cerr << '\n' << PROGRAM << ": WARNING: Sequence '" << orig << "' contains special characters.\n"
					  << "\tUse -clean_seq_name 1 to make them removed.\n";
	}

	return name;
}

char **TC_clean_seq_names(char **name, int n, int mode)
{
	static int sw;
	int a;

	if (mode != GET)
		sw = mode;
	if (!name)
		return 0;

	for (a = 0; a < n; a++) {
		name[a] = TC_clean_seq_name(name[a], sw);
	}

	return name;
}
