/*
 * File: util_algo.cpp
 * Created: 06.06.2006
 * Last Modified: 26.08.2006
 * Author: Adrian Rospondek
 */


#include <limits.h>
#include <math.h>
#include <stdarg.h>
#include <string.h>
#include "globals.hpp"
#include "util_algo.hpp"
#include "util_mem.hpp"



typedef int (*TC_cmp_func_type)(const void *, const void *);
typedef int (*TC_cmp_func_2_type)(const void *, const void *, const int, const int);
typedef void *(*TC_cpy_func_type)(void *, const void *, size_t);


static int g_sort_field = 0;


/* Exchange a, b */
static void TC_exchange(void *a, void *b, size_t size)
{
	size_t i;
	int *ia;
	int *ib;
	char *ca;
	char *cb;

	ia = (int *)a;
	ib = (int *)b;
	for (i = sizeof(int); i <= size; i += sizeof(int)) {
		int t = ia[0];

		ia[0] = ib[0];
		ib[0] = t;
		ia++;
		ib++;
	}

	ca = (char *)ia;
	cb = (char *)ib;
	for (i = i - sizeof(int) + 1; i <= size; i++) {
		char t = ca[0];

		ca[0] = cb[0];
		cb[0] = t;
		ca++;
		cb++;
	}
}

#define MAXSTACK (sizeof(size_t) * CHAR_BIT)

/* ANSI-C qsort() */
static void TC_qsort(void *base, size_t nmemb, size_t size, TC_cmp_func_type compare)
{
	void *lbStack[MAXSTACK], *ubStack[MAXSTACK];
	unsigned int offset;
	int sp;

	lbStack[0] = (char *)base;
	ubStack[0] = (char *)base + (nmemb - 1) * size;

	for (sp = 0; sp >= 0; sp--) {
		char *lb, *ub, *m;
		char *P, *i, *j;

		lb = (char *)lbStack[sp];
		ub = (char *)ubStack[sp];
		while (lb < ub) {
			/* select pivot and exchange with 1st element */
			offset = (ub - lb) >> 1;
			P = lb + offset - offset % size;
			TC_exchange(lb, P, size);
			/* partition into two segments */
			i = lb + size;
			j = ub;
			while (1) {
				while (i < j && compare(lb, i) > 0)
					i += size;
				while (j >= i && compare(j, lb) > 0)
					j -= size;
				if (i >= j)
					break;
				TC_exchange(i, j, size);
				j -= size;
				i += size;
			}
			/* pivot belongs in A[j] */
			TC_exchange(lb, j, size);
			m = j;
			/* keep processing smallest segment, and stack largest */
			if (m - lb <= ub - m) {
				if (m + size < ub) {
					lbStack[sp] = m + size;
					ubStack[sp++] = ub;
				}
				ub = m - size;
			}
			else {
				if (m - size > lb) {
					lbStack[sp] = lb;
					ubStack[sp++] = m - size;
				}
				lb = m + size;
			}
		}
	}
}

static int TC_cmp_int(const int **a, const int **b)
{
	if (a[0][g_sort_field] < b[0][g_sort_field])
		return -1;
	else if (a[0][g_sort_field] == b[0][g_sort_field])
		return 0;
	else
		return 1;
}

void TC_sort_int(int **V, int F, int left, int right)
{
	g_sort_field = F;
	TC_qsort(V, (right - left) + 1, sizeof(int **), (TC_cmp_func_type)TC_cmp_int);
}

void TC_sort_int_inv(int **V, int N_F, int F, int left, int right)
{
	int **list;
	int a, b;
	
	g_sort_field = F;
	TC_qsort(V, (right - left) + 1, sizeof(int **), (TC_cmp_func_type)TC_cmp_int);
	
	list = TC_declare_int((right - left) + 1, N_F);
	for (a = left; a < (right - left) + 1; a++)
		for (b = 0; b < N_F; b++)
			list[a - left][b] = V[a][b];
	for (a = left; a < (right - left) + 1; a++)
		for (b = 0; b < N_F; b++)
			V[a][b] = list[(right - left) - a][b];
	
	TC_free_int(list);
}

static int TC_cmp_str(const char *p1, const char *p2)
{
	return strcmp(*(const char **)p1, *(const char **)p2);
}

char **TC_sort_string_array(char **V, int n)
{
	TC_qsort(V, n, sizeof(char *), (TC_cmp_func_type)TC_cmp_str);
	return V;
}

static void **TC_hsort_array(void **ra, int n, int len, size_t size, int first_comp_field,
		int n_comp_fields, TC_cmp_func_2_type compare, TC_cpy_func_type copy)
{
	void *rra;
	unsigned long i, ir, j, l;
	int FE = 1;
	
	if (FE == 1)
		--ra;
	else
		--n;
	
	rra = AR_allocate(len, size);
	
	if (n < 2)
		return ra;
	
	l = (n >> 1) + 1;
	ir = n;
	
	for (;;) {
		if (l > FE) {
			copy(rra, ra[--l], len);
		}
		else {
			copy(rra, ra[ir], len);
			copy(ra[ir], ra[FE], len);
			if (--ir == FE) {
				copy(ra[FE], rra, len);
				break;
			}
		}
		i = l;
		j = l + l;
		while (j <= ir) {
			if (j < ir && compare(ra[j], ra[j + 1], first_comp_field, n_comp_fields) < 0)
				++j;
			if (compare(rra, ra[j], first_comp_field, n_comp_fields) < 0) {
				copy(ra[i], ra[j], len);
				i = j;
				j <<= 1;
			}
			else
				j = ir + 1;
		}
		copy(ra[i], rra, len);
	}
	
	AR_free(rra);
	ra += FE;
	
	return ra;
}

static int TC_compare(const void *a, const void *b, const int first, const int clen)
{
	const short int *ax, *bx;
	int p;
	
	ax = (const short int *)a;
	bx = (const short int *)b;
	
	for (p = first; p < clen + first; ++p) {
		if (ax[p] < bx[p])
			return -1;
		else if (ax[p] == bx[p]);
		else
			return 1;
	}
	
	return 0;
}

static void *TC_copy(void *to, const void *from, size_t size)
{
	short int *ax;
	const short int *bx;
	size_t p;
	
	ax = (short int *)to;
	bx = (const short int *)from;
	
	for (p = 0; p < size; ++p)
		ax[p] = bx[p];
	
	return to;
}

void **TC_hsort_list_array(void **L, int len, size_t size, int entry_len, int first_comp_field,
		int n_comp_fields)
{
	return TC_hsort_array(L, len, entry_len, size, first_comp_field, n_comp_fields, TC_compare,
			TC_copy);
}

#define SWAP_FUNCTION(type, function)\
static void function(type *a, type *b, int n)\
{\
	type t;\
	int c;\
\
	for (c = 0; c < n; ++c) {\
		t = a[c];\
		a[c] = b[c];\
		b[c] = t;\
	}\
}

SWAP_FUNCTION(char, TC_swap_char)

#define INVERT_LIST(type, function, swap_function)\
type *function(type *list, int len)\
{\
	int a, b;\
\
	for (a = 0, b = len - 1; a < b; ++a, --b)\
		swap_function(&list[a], &list[b], 1);\
\
	return list;\
}

INVERT_LIST(char, TC_invert_list_char, TC_swap_char)

static void *TC_bsearch_array(const void *key, int *p, int comp_first, int comp_len, void **list,
		int len, TC_cmp_func_2_type compare)
{
	void *key2;
	int upper, lower, c, i;
	
	upper = -1;
	lower = len;
	
	while ((lower - upper) > 1) {
		i = (lower + upper) >> 1;
		key2 = list[i];
		c = compare(key2, key, comp_first, comp_len);
		if (c == 0) {
			p[0] = i;
			return key2;
		}
		else if (c < 0)
			upper = i;
		else if (c > 0)
			lower = i;
	}
	
	return 0;
}

void **TC_search_in_list_array(void *key, int *p, int comp_len, void **L, int len)
{
	static void *l[1];
	
	l[0] = TC_bsearch_array(key, p, 0, comp_len, L, len, TC_compare);
	
	if (l[0] == 0)
		return 0;
	else
		return l;
}

#define UNDEFINED_INT TC_UNDEFINED

#define RETURN_MAX(type, function, cmp_op, undef)\
type function(type **array, int len_array, int field)\
{\
	type max;\
	int a;\
\
	if (array == 0 || len_array == 0)\
		return 0;\
	else {\
		if (len_array == -1)\
			len_array = TC_read_array_size(array, sizeof(type *));\
		max = array[0][field];\
		for (a = 1; a < len_array; ++a)\
			if (max == undef)\
				max = array[a][field];\
			else if (array[a][field] != undef)\
				max = (array[a][field] cmp_op max) ? array[a][field] : max;\
	}\
\
	return (max == undef) ? 0 : max;\
}

RETURN_MAX(int, TC_return_max_int, >, UNDEFINED_INT)

#define RETURN_SD(type, function, undef)\
type function(type **array, int len, int field, type mean)\
{\
	int a;\
	double c = 0;\
\
	if (len == 0 || array == 0 || len > TC_read_array_size(array, sizeof(type *)))\
		return 0;\
	else {\
		for (a = 0; a < len; ++a) {\
			if ((array[a][field] != undef) && (mean - array[a][field]) != 0)\
				c += (double)((mean - array[a][field]) * (mean - array[a][field]));\
		}\
		c = sqrt(c) / (double)len;\
		return (type)MAX(c, 1.0);\
	}\
}

RETURN_SD(int, TC_return_sd_int, UNDEFINED_INT)

#define RETURN_MEAN(type, function, undef)\
double function(type **array, int len, int field)\
{\
	int a, c;\
	double b;\
\
	if (len == 0 || array == 0 || len > TC_read_array_size(array, sizeof(type *)))\
		return 0;\
	else {\
		for (b = 0, c = 0, a = 0; a < len; ++a) {\
			if (array[a][field] != undef) {\
				b += (double)array[a][field];\
				++c;\
			}\
		}\
	}\
\
	return (c == 0) ? 0 : (b / c);\
}

RETURN_MEAN(int, TC_return_mean_int, UNDEFINED_INT)

/*
 * 1st arg: number of values,
 * 2nd arg: 1 (maximise)/0 (minimise),
 * 3rd arg: (int *) contains the index of the best value,
 * Nth arg: n value
 */
#define BEST_OF_MANY(type, function, undef)\
type function(int n, ...)\
{\
	va_list ap;\
	type v, best;\
	int *fop, a;\
	int maximise;\
\
	va_start(ap, n);\
	maximise = va_arg(ap, int);\
	fop = va_arg(ap, int *);\
	best = va_arg(ap, type);\
	fop[0] = 0;\
\
	for (a = 1; a < n; ++a) {\
		v = va_arg(ap, type);\
		if (best == undef) {\
			best = v;\
			fop[0] = a;\
		}\
		if (best == undef || v == undef);\
		else if (maximise == 1 && v > best) {\
			fop[0] = a;\
			best = v;\
		}\
		else if (maximise == 0 && v < best) {\
			fop[0] = a;\
			best = v;\
		}\
	}\
	va_end (ap);\
\
	return best;\
}

BEST_OF_MANY(int, TC_best_int, UNDEFINED_INT)
