/*
 * File: util_hasch.cpp
 * Created: 06.06.2006
 * Last Modified: 12.09.2006
 * Author: Adrian Rospondek
 */


#include <iostream>
#include "globals.hpp"
#include "util.hpp"
#include "util_hasch.hpp"
#include "util_mem.hpp"



static TC_hasch_entry *TC_allocate_hasch_entry(TC_hasch_entry *e, int action,
	TC_declare_hasch_func_type declare_data, TC_free_hasch_func_type free_data);

/**
 * @brief Inserts entry e between entry p and entry n and returns e.
 */
static TC_hasch_entry *TC_insert_hasch_entry_in_list(TC_hasch_entry *p, TC_hasch_entry *e,
	TC_hasch_entry *n, TC_declare_hasch_func_type declare_data, TC_free_hasch_func_type free_data)
{
	if (!e)
		e = TC_allocate_hasch_entry(0, DECLARE, declare_data, free_data);

	if (!p && !n);
	else if (!p)
		p = n->p;
	else if (!n)
		n = p->n;
	e->p = p;
	if (p)
		p->n = e;
	e->n = n;
	if (n)
		n->p = e;

	return e;
}

/**
 * @brief Extracts entry e and returns p or next if is NULL.
 */
static TC_hasch_entry *TC_extract_hasch_entry_from_list(TC_hasch_entry *e)
{
	TC_hasch_entry *p = 0, *n = 0;

	if (!e);
	else {
		p = e->p;
		n = e->n;
		if (p)
			p->n = n;
		if (n)
			n->p = p;
		e->p = e->n = 0;
	}

	return e;
}

static TC_hasch_entry *TC_allocate_hasch_entry(TC_hasch_entry *e, int action,
	TC_declare_hasch_func_type declare_data, TC_free_hasch_func_type free_data)
{
	static TC_hasch_entry *s;
	TC_hasch_entry *ns;

	if (!s)
		s = (TC_hasch_entry *)AR_allocate(1, sizeof(TC_hasch_entry));

	if (action == DECLARE) {
		ns = s->p;
		e = TC_extract_hasch_entry_from_list(s);
		if (e->free_data)
			(e->free_data)(e->data);
		e->declare_data = declare_data;
		e->free_data = free_data;
		e->declare_data(e);
		e->k = TC_UNDEFINED;
		s = ns;
	}
	else if (action == FREE) {
		TC_extract_hasch_entry_from_list(e);
		e->k = TC_UNDEFINED;
		if (e->free_data)
			e->data = (e->free_data)(e->data);
		e->free_data = 0;
		e->declare_data = 0;
		s = TC_insert_hasch_entry_in_list(s, e, 0, declare_data, free_data);
	}
	else if (action == FREE_STACK) {
		while (s) {
			e = s->p;
			TC_allocate_hasch_entry(s, FREE, declare_data,free_data);
			AR_free(s);
			s = e;
		}
	}
	else
		std::cerr << '\n' << PROGRAM << ": ERROR: Unknown action " << action << " (allocate_hasch_entry).\n";

	return e;
}

TC_hasch_t *TC_hcreate(int n_elements, TC_declare_hasch_func_type declare_data,
	TC_free_hasch_func_type free_data)
{
	TC_hasch_t *T;
	int a;

	n_elements = n_elements * 2 + 1;
	T = (TC_hasch_t *)AR_allocate(1, sizeof(TC_hasch_t));
	T->ne = n_elements;
	T->p = (TC_hasch_entry **)AR_allocate(n_elements, sizeof(TC_hasch_entry *));
	for (a = 0; a < n_elements; a++)
		T->p[a] = TC_allocate_hasch_entry(0, DECLARE, declare_data, free_data);

	return T;
}

TC_hasch_t *TC_hdestroy(TC_hasch_t *T, TC_declare_hasch_func_type declare_data,
	TC_free_hasch_func_type free_data)
{
	TC_hasch_entry *p, *pp;
	int a;

	if (T == 0)
		return 0;

	for (a = 0; a < T->ne; a++) {
		p = T->p[a];
		while (p) {
			pp = p;
			p = p->n;
			TC_allocate_hasch_entry(pp, FREE, declare_data, free_data);
		}
	}

	AR_free(T->p);
	AR_free(T);

	return 0;
}

/* action: FIND, ADD, REMOVE */
/* find the key: k->h */
TC_hasch_entry *TC_hsearch(TC_hasch_t *T, int k, int action, TC_declare_hasch_func_type declare_data,
	TC_free_hasch_func_type free_data)
{
	TC_hasch_entry *p, *pi;
	int h;

	h = k % T->ne;

	if (action == ADD || action == FIND) {
		p = pi = T->p[h];
		while (p && p->k != k) {
			p = p->n;
		}
		if (action == ADD && !p) {
			p = TC_insert_hasch_entry_in_list(pi, 0, 0, declare_data, free_data);
			p->k = k;
		}
		else if (action == FIND && !p)
			p = 0;
		
		return p;
	}
	else if (action == REMOVE) {
		TC_allocate_hasch_entry(TC_hsearch(T, k, FIND, declare_data, free_data), FREE, declare_data,
				free_data);

		return 0;
	}

	return 0;
}

static TC_char_node *TC_declare_char_node(int action)
{
	static TC_char_node **heap;
	static int key;  
	static int heap_size, free_heap, a;

	if (action == DECLARE) {
		if (free_heap == 0) {
			free_heap = 100;
			heap = (TC_char_node **)AR_reallocate(heap, (heap_size + free_heap), sizeof(TC_char_node *));
			for (a = heap_size; a < (heap_size + free_heap); a++) {
				heap[a] = (TC_char_node *)AR_allocate(1, sizeof(TC_char_node));
				(heap[a])->c = (TC_char_node **)AR_allocate(256, sizeof(TC_char_node *));
				(heap[a])->key = key++;
			}
			heap_size += free_heap;
		}
		
		return heap[heap_size - (free_heap--)];
	}
	else if (action == FREE_STACK) {
		for (a = 0; a < heap_size; a++) {
			heap[a]->key = key++;
			AR_free(heap[a]->c);
			(heap[a])->c = (TC_char_node **)AR_allocate(256, sizeof(TC_char_node *));
		}
		free_heap = heap_size;
		
		return 0;
	}

	return 0;
}

int TC_string2key(const char *s, TC_char_node *n)
{
	static TC_char_node *root;

	if (!root)
		root = TC_declare_char_node(DECLARE);

	if (n == 0 && s == 0)
		TC_declare_char_node(FREE_STACK);
	else if (n == 0)
		return TC_string2key(s, root);
	else if (s[0] == '\0')
		return n->key;
	else
		return TC_string2key(s + 1, (n->c[s[0]]) ? (n->c[s[0]])
				: (n->c[s[0]] = TC_declare_char_node(DECLARE)));

	return 0;
}
