/*
 * File: util_io.cpp
 * Created: 06.06.2006
 * Last Modified: 13.09.2006
 * Author: Adrian Rospondek
 */


#include <fstream>
#include <iostream>
#include <sstream>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include "AR_Config.hpp"
#include "config.hpp"
#include "globals.hpp"
#include "util.hpp"
#include "util_io.hpp"
#include "util_mem.hpp"



struct TC_tmpname {
	char *name;
	TC_tmpname *next;
};


static FILE *g_NFP = 0; // Null file pointer: should only be open once
static TC_tmpname *g_tmpname_list = 0, *g_ntmpname = 0;
static int g_n_tmpname = 0;


char *AR_check_file_exists(const std::string &fname)
{
	if (fname.empty())
		return 0;

	std::string expanded;

	if (fname[0] == '~') {
		expanded = g_conf.home();
		expanded.append(fname, 1, fname.size());
	}
	else
		expanded = fname;

	const std::string cached = AR_get_cache_dir() + expanded;
	const char *fname1 = expanded.c_str();
	const char *fname2 = cached.c_str();
	const char *path = 0;

	if (TC_filename_is_special(fname1))
		path = fname1;
	else if (TC_strm5(fname1, "no", "NO", "No", "NO_FILE", "no_file"))
		path = 0;
	else {
		std::ifstream f(fname1);

		if (!f == false) {
			f.close();
			path = fname1;
		}
		else {
			f.clear();
			f.open(fname2);

			if (!f == false) {
				f.close();
				path = fname2;
			}
		}
	}

#if DEBUG_FILE
	std::cout << "\nFULL NAME OF FILE '" << fname << "' IS '" << (path ? path : "") << "'\n";
#endif

	if (path) {
		char *buf = static_cast<char *>(AR_allocate(strlen(path) + 1, sizeof(char)));

		strcpy(buf, path);
		path = buf;
	}

	return const_cast<char *>(path);
}

bool AR_mkdir(const std::string &dir_in)
{
	const char dir_sep = '/';
	std::string expanded;
	
	if (dir_in[0] == '~') {
		expanded = g_conf.home();
		expanded.append(dir_in, 1, dir_in.size());
	}
	else
		expanded = dir_in;

	if (*expanded.rbegin() != dir_sep)
		expanded += dir_sep;

	bool ok = true;
	char *dir = const_cast<char *>(expanded.c_str());
	char tmp;

	for (char *d = dir; *d; ++d) {
		if (*d == dir_sep) {
			tmp = *(d + 1);
			*(d + 1) = '\0';
			
			if (access(dir, F_OK) == -1) {
				mkdir(dir, S_IRUSR | S_IWUSR | S_IXUSR | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH);
				
				if (access(dir, F_OK) == -1) {
					std::cerr << '\n' << PROGRAM << ": ERROR: Could not create directory '" << dir << "'.\n";
					ok = false;
					break;
				}
#if DEBUG_FILE
				std::cout << "\nDIRECTORY '" << dir << "' CREATED\n";
#endif
			}
			*(d + 1) = tmp;
		}
	}

	return ok;
}

FILE *TC_vfopen(const char *name_in, const char *mode)
{
	if (!name_in)
		return 0;

	const std::string expanded = (name_in[0] == '~') ? (g_conf.home() + (name_in + 1)) : name_in;
	const char *name = expanded.c_str();

	bool get_new_name = false;
	bool tolerate_mistake = false;
	bool cache_used = false;

	if (mode[0] == 'g') {
		get_new_name = true;
		mode++;
	}
	else if (mode[0] == 't') {
		tolerate_mistake = true;
		mode++;
	}
	// use cached version from TCOFFEE_CACHE
	else if (mode[0] == 'c') {
		cache_used = true;
		mode++;
	}

	if (name == 0 || TC_strm5(name, "no", "NO", "No", "NULL", "/dev/null") || TC_strm2(name, "no_file", "NO_FILE")) {
		if (g_NFP == 0)
			g_NFP = fopen(NULL_DEVICE, mode);
		return g_NFP;
	}
	else if (TC_strm3(name, "stdout", "STDOUT", "Stdout"))
		return stdout;
	else if (TC_strm3(name, "stderr", "STDERR", "Stderr"))
		return stderr;
	else if (TC_strm3(name, "stdin", "STDIN", "Stdin")) {
		static const char *stdin_file = 0;
		FILE *tmp_fp;
		int c;

		if (!stdin_file) {
			stdin_file = AR_tmpname();
			tmp_fp = fopen(stdin_file, "w");

			while ((c = fgetc(stdin)) != EOF)
				fprintf(tmp_fp, "%c", c);
			fclose(tmp_fp);
		}
		return fopen(stdin_file, "r");
	}
	else if (!name[0] && (*mode == 'w' || *mode == 'a'))
		return stdout;
	else if (!name[0] && *mode == 'r')
		return 0;

	FILE *fp = 0;

	fp = fopen(name, mode);
	if (fp == 0) {
		if (*mode == 'r' && cache_used == false) {
			const std::string name2 = AR_get_cache_dir() + name;

			fp = TC_vfopen(name2.c_str(), "cr");
		}
		else if (*mode == 'r' && cache_used == true) {
			std::cerr << '\n' << PROGRAM << ": ERROR: Could not open file '" << name << "'.\n";
			fp = 0;
		}
		else if (*mode == 'a' && cache_used == false) {
			const std::string name2 = AR_get_cache_dir() + name;

			fp = TC_vfopen(name2.c_str(), "ca");
		}
		else if (*mode == 'a' && cache_used == true) {
			std::cerr << '\n' << PROGRAM << ": ERROR: Could not open file '" << name << "'.\n";
			fp = 0;
		}
		else if (*mode == 'w') {
			std::cerr << '\n' << PROGRAM << ": ERROR: Could not open file '" << name << "'.\n";
			fp = 0;
		}
	}

	return fp;
}

int TC_vfclose(FILE *fp)
{
	if (fp == 0 || fp == g_NFP || fp == stdin || fp == stdout || fp == stderr)
		return 0;
	return fclose(fp);
}

void TC_vremove(const char *s)
{
	const char *path;
	
	if (s && (path = AR_check_file_exists(s)))
		remove(path);
}

char *TC_add2file2remove_list(const char *name)
{
	if (!g_tmpname_list)
		g_ntmpname = g_tmpname_list = (TC_tmpname *)AR_allocate(1, sizeof(TC_tmpname));
	else
		g_ntmpname = g_ntmpname->next = (TC_tmpname *)AR_allocate(1, sizeof(TC_tmpname));

	g_ntmpname->name = (char *)AR_allocate(strlen(name) + 1, sizeof(char));
	strcpy(g_ntmpname->name, name);

	return g_ntmpname->name;
}

void TC_rm_tmp_files()
{
	TC_tmpname *b;

	while (g_tmpname_list) {
		TC_vremove(g_tmpname_list->name);
		b = g_tmpname_list;
		g_tmpname_list = g_tmpname_list->next;

		AR_free(b->name);
		AR_free(b);
	}
}

static unsigned int AR_get_tmpname_root()
{
	const unsigned int MAX_TMPNAME_ROOT = 1000;
	const std::string file = g_conf.tc_tmp() + "tmpname_number";
	std::fstream f;
	unsigned int val = 0;

	f.open(file.c_str(), std::ios_base::in);
	if (!f == false) {
		f >> val;

		if (val > MAX_TMPNAME_ROOT)
			val = 0;

		f.close();
	}
	f.clear();

	++val;

	f.open(file.c_str(), std::ios_base::out);
	f << val;
	f.close();

	return val;
}

static char *AR_make_tmpname()
{
	static const unsigned int root = AR_get_tmpname_root();
	static unsigned int file = 0;

	std::ostringstream oss;

	oss << g_conf.tc_tmp() << root << '_' << file++ << ".tmp";
	std::string str = oss.str();

	char *tmp_name = static_cast<char *>(AR_allocate(str.size() + 1, sizeof(char)));

	memcpy(tmp_name, str.data(), str.size());

	return tmp_name;
}

char *AR_tmpname()
{
	char *tmp_name = 0;
	char *s = 0;

	++g_n_tmpname;
	tmp_name = AR_make_tmpname();

	s = TC_add2file2remove_list(tmp_name);

	if (s != tmp_name) {
		AR_free(tmp_name);
		tmp_name = s;
	}

	return tmp_name;
}

int TC_filename_is_special(const char *fname)
{
	if (TC_strm5(fname, "default", "stdin", "stdout", "stderr", "/dev/null"))
		return 1;
	if (TC_strm3(fname, "STDIN", "STDOUT", "STDERR"))
		return 1;
	return 0;
}

static TC_fname *TC_declare_fname()
{
	TC_fname *F;

	F = (TC_fname *)AR_allocate(1, sizeof(TC_fname));
	F->path = (char *)AR_allocate(FILENAMELEN, sizeof(char));
	F->name = (char *)AR_allocate(FILENAMELEN, sizeof(char));
	F->suffix = (char *)AR_allocate(FILENAMELEN, sizeof(char));

	return F;
}

static void TC_free_fname(TC_fname *F)
{
	AR_free(F->path);
	AR_free(F->name);
	AR_free(F->suffix);
}

char *TC_path2filename(const char *array)
{
	TC_fname *F;
	char *name;

	name = (char *)AR_allocate(strlen(array) + 2, sizeof(char));
	F = TC_parse_fname(array);

	if (F->suffix)
		sprintf(name, "%s.%s", F->name, F->suffix);
	else
		strcpy(name, F->name);

	TC_free_fname(F);

	return name;
}

TC_fname *TC_parse_fname(const char *array)
{
	TC_fname *F;
	int l;

	F = TC_declare_fname();

	strncpy(F->path, array, FILENAMELEN - 1);
	l = strlen(array);
	while (l != -1 && (F->path)[l] != '/')
		(F->path)[l--] = '\0';

	strncpy(F->name, array + l + 1, FILENAMELEN - 1);
	l = strlen(F->name);
	while (l != -1) {
		if ((F->name)[l] == '.') {
			F->name[l] = '\0';
			strncpy(F->suffix, F->name + l + 1, FILENAMELEN - 1);
			break;
		}
		else
			--l;
	}

	return F;
}

char **TC_file2lines(const char *name)
{
	// lines[0] -> nlines; lines[1] -> first line
	int size = 0;
	char *string = 0;
	char **lines = 0;

	string = AR_read_file(name, size);
	if (string) {
		lines = TC_string2list2(string, "\n");
		AR_free(string);
	}

	return lines;
}

/* Token - any string. If Token[0]=='\n' then token only from the beginning of the line. */
FILE *TC_find_token_in_file(const char *fname, FILE *fp, const char *token)
{
	int c;
	char *name;
	int token_len;
	int only_start;

	if (!fp && !AR_check_file_exists(fname))
		return 0;
	if (token[0] == '\n') {
		token++;
		only_start = 1;
	}
	else
		only_start = 0;

	token_len = strlen(token);
	name = (char *)AR_allocate((fname ? TC_measure_longest_line_in_file(fname) : 10000) + 1,
		sizeof(char));

	if (!fp)
		fp = TC_vfopen(fname, "r");

	while (fscanf(fp, "%s", name) != EOF) {
		if (name[0] == '*')
			while ((c = fgetc(fp)) != '\n' && c != EOF);
		else if (strncmp(name, token, token_len) == 0) {
			AR_free(name);
			return fp;
		}
		else if (only_start)
			while ((c = fgetc(fp)) != '\n' && c != EOF);
	}
	AR_free(name);
	fclose(fp);

	return 0;
}

/**
 * @brief Finds the string 'token' (as a single word) in the n_line first lines of fname.
 * @return NULL if not found or the position of the fp
 */
FILE *TC_find_token_in_file_nlines(const char *fname, FILE *fp, const char *token, int n_line)
{
	FILE *fp1;
	FILE *fp2;
	char *tmp_name;
	char buffer[10001];
	int a;

	if (!fp && !AR_check_file_exists(fname))
		return 0;

	if (fp == 0) {
		tmp_name = AR_tmpname();
		fp1 = TC_vfopen(fname, "r");
		fp2 = TC_vfopen(tmp_name, "w");
		for (a = 0; a < n_line && fgets(buffer, 10000, fp1) != 0; a++)
			fprintf(fp2, "%s", buffer);
		TC_vfclose(fp1);
		TC_vfclose(fp2);
	}

	return TC_find_token_in_file(tmp_name, fp, token);
}

int TC_token_is_in_file(const char *fname, const char *token)
{
	FILE *fp;
	char buf[10001];

	if (!fname || !AR_check_file_exists(fname))
		return 0;
	else {
		fp = TC_vfopen(fname, "r");
		while (fgets(buf, 10000, fp) != 0) {
			if (strstr(buf, token)) {
				TC_vfclose(fp);
				return 1;
			}
		}
		TC_vfclose(fp);
	}
	return 0;
}

int TC_count_n_char_in_file(const char *name)
{
	int c, n;
	FILE *fp;

	n = 0;
	fp = TC_vfopen(name, "r");

	while ((c = fgetc(fp)) != EOF)
		n++;
	TC_vfclose(fp);

	return n;
}

int TC_count_n_char_x_in_file(const char *name, const char x)
{
	FILE *fp;
	int n = 0, c;

	fp = TC_vfopen(name, "r");
	while ((c = fgetc(fp)) != EOF)
		n += (c == x);
	TC_vfclose(fp);

	return n;
}

int TC_measure_longest_line_in_file(const char *name)
{
	int c;
	FILE *fp;
	int longest = 0;
	int current = 0;

	fp = TC_vfopen(name, "r");

	while ((c = fgetc(fp)) != EOF) {
		if (c == '\n') {
			longest = MAX(longest, current + 1);
			current = 0;
		}
		else
			current++;
	}
	longest = MAX(longest, current + 1);
	TC_vfclose(fp);

	return longest;
}

FILE *TC_skip_space(FILE *fp)
{
	int c;

	do
		c = getc(fp);
	while (isspace(c));
	ungetc(c, fp);

	return fp;
}

FILE *TC_skip_commentary_line_in_file(const char com, FILE *fp)
{
	int c = 0;

	if (fp == 0)
		return 0;

	while ((c = fgetc(fp)) == com) {
		while ((c = fgetc(fp)) != '\n' && c != EOF);
	}
	if (c != EOF && c != '\n')
		ungetc(c, fp);

	return fp;
}

/**
 * @brief Removes in file in_file every portion of line to the right of one of the symbols
 * included in com.
 * Writes the striped file into a vtmpnam file.
 */
char *TC_strip_file_from_comments(const char *com, const char *in_file)
{
	FILE *fp1;
	FILE *fp2;
	char *out_file;
	int c;

	out_file = AR_tmpname();
	fp1 = TC_vfopen(in_file, "r");
	fp2 = TC_vfopen(out_file, "w");

	while ((c = fgetc(fp1)) != EOF) {
		if (strchr(com, c)) {
			while ((c = fgetc(fp1)) != '\n' && c != EOF);
		}
		else {
			fprintf(fp2, "%c", c);
			while ((c = fgetc(fp1)) != '\n' && c != EOF)
				fprintf(fp2, "%c", c);
			if (c != EOF)
				fprintf(fp2, "%c", c);
		}
	}
	TC_vfclose(fp1);
	TC_vfclose(fp2);

	return out_file;
}

void AR_write_file(const std::string &name, const char *buf, const int size)
{
	if (name.empty() || !buf)
		return;

	std::ofstream f(name.c_str());

	f.write(buf, size);
	f.close();
}

char *AR_read_file(const std::string &name, int &size)
{
	char *buf = 0;
	size = 0;

	if (name.empty())
		return buf;

	std::ifstream f(name.c_str());

	if (!f)
		return buf;

	f.seekg(0, std::ios_base::end);
	size = static_cast<int>(f.tellg()) + 1;
	f.seekg(0);

	buf = static_cast<char *>(AR_allocate(size, sizeof(char)));

	f.get(buf, size, EOF);
	size = f.gcount();

	return buf;
}
