/*
 * File: util_mem.cpp
 * Created: 06.06.2006
 * Last Modified: 12.09.2006
 * Author: Adrian Rospondek
 */


#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include "globals.hpp"
#include "util.hpp"
#include "util_mem.hpp"



struct AR_mem_control {
	int nobj;
	size_t obj_size;
	AR_mem_control *prev, *next;
};


static AR_mem_control *g_memlast = 0;
static float g_alloc_mem = 0, g_tot_mem = 0;



static void TC_verify_memory(int s)
{
	g_alloc_mem += s;
	if (g_alloc_mem > g_tot_mem)
		g_tot_mem = g_alloc_mem;
}

static AR_mem_control *AR_add_mc_to_list(AR_mem_control *M)
{
	M[0].prev = g_memlast;
	M[0].next = 0;
	if (g_memlast)
		g_memlast->next = M;
	g_memlast = M;

	return M;
}

static AR_mem_control *AR_rm_mc_from_list(AR_mem_control *M)
{
	AR_mem_control *p, *n;
	
	p = M[0].prev;
	n = M[0].next;
	
	if (p)
		p->next = n;
	if (n)
		n->prev = p;
	if (g_memlast == M)
		g_memlast = p;

	return M;
}

void *AR_allocate(int nobj, size_t obj_size)
{
	int size;
	void *ptr = 0;

	if (nobj <= 0 || obj_size <= 0)
		return ptr;

	size = nobj * obj_size + sizeof(AR_mem_control);
	TC_verify_memory(size);
	ptr = malloc(size);

	if (ptr == 0) {
		std::cerr << '\n' << PROGRAM << ": ERROR: Memory allocation (allocate).\n";
		//AR_free_all();
		exit(-1);
	}
	else {
		AR_mem_control *M;
		
		memset(ptr, 0, size);
		
		M = (AR_mem_control *)ptr;
		M[0].nobj = nobj;
		M[0].obj_size = obj_size;

		AR_add_mc_to_list(M);

		++M;
		ptr = M;
	}

	return ptr;
}

void *AR_reallocate(void *ptr, int nobj, size_t obj_size)
{
	if (ptr == 0)
		return AR_allocate(nobj, obj_size);
	else if (nobj <= 0 || obj_size <= 0) {
		AR_free(ptr);
		return 0;
	}
	else {
		int size, old_size;
		AR_mem_control *M;

		M = (AR_mem_control *)ptr;
		--M;

		old_size = M[0].nobj * M[0].obj_size + sizeof(AR_mem_control);
		AR_rm_mc_from_list(M);

		size = nobj * obj_size + sizeof(AR_mem_control);
		TC_verify_memory(size - old_size);
		ptr = realloc(M, size);

		if (ptr == 0) {
			std::cerr << '\n' << PROGRAM << ": ERROR: Memory allocation (reallocate).\n";
			//AR_free_all();
			exit(-1);
		}
		else {
			for (int i = old_size; i < size; ++i)
				((char *)ptr)[i] = 0;
			
			M = (AR_mem_control *)ptr;
			M[0].nobj = nobj;
			M[0].obj_size = obj_size;

			AR_add_mc_to_list(M);

			++M;
			ptr = M;
		}
	}

	return ptr;
}

void AR_free(void *ptr)
{
	if (ptr == 0)
		return;
	else {
		int size;
		AR_mem_control *M;

		M = (AR_mem_control *)ptr;
		--M;

		size = M[0].nobj * M[0].obj_size + sizeof(AR_mem_control);
		AR_rm_mc_from_list(M);

		TC_verify_memory(-size);
		free(M);
	}
}

void AR_free_all()
{
	int size = 0;
	AR_mem_control *p;

	while (g_memlast) {
		size += g_memlast->nobj * g_memlast->obj_size + sizeof(AR_mem_control);
		p = g_memlast->prev;
		free(g_memlast);
		g_memlast = p;
	}
	
	TC_verify_memory(-size);
}

static void *TC_declare_arrayN2(int ndim, int *A, size_t size)
{
	int a;
	void **p;
	
	if (ndim > 1) {
		p = (void **)AR_allocate(A[0], sizeof(void *));
		
		for (a = 0; a < A[0]; ++a)
			p[a] = TC_declare_arrayN2(ndim - 1, A + 1, size);
	}
	else
		p = (void **)AR_allocate(A[0], size);
	
	return p;
}

void *TC_declare_arrayN(int ndim, size_t size, ...)
{
	va_list ap;
	void **p;
	int *array;
	int a;
	
	array = (int *)AR_allocate(ndim, sizeof(int));
	
	va_start(ap, size);
	for (a = 0; a < ndim; ++a) {
		array[a] = va_arg(ap, int);
		if (array[a] < 0) {
			va_end(ap);
			return 0;
		}
	}
	va_end(ap);
	
	if (ndim == 2) {
		p = (void **)AR_allocate(array[0], sizeof(void *));
		
		for (a = 0; a < array[0]; ++a)
			p[a] = AR_allocate(array[1], size);
	}
	else
		p = (void **)TC_declare_arrayN2(ndim, array, size);
	
	AR_free(array);
	
	return p;
}

static void *TC_realloc_arrayN2(int ndim, void *p, int *A, size_t size)
{
	int o, n;
	int a;
	
	if (ndim == 0)
		return 0;
	
	if (ndim > 1) {
		o = TC_read_array_size(p, sizeof(void *));
		if (A[0] > o)
			p = AR_reallocate(p, A[0], sizeof(void *));
		n = (A[0] == -1) ? o : A[0];
		
		for (a = 0; a < n; ++a)
			((void **)p)[a] = TC_realloc_arrayN2(ndim - 1, ((void **)p)[a], A + 1, size);
	}
	else {
		o = TC_read_array_size(p, size);
		if (A[0] > o)
			p = AR_reallocate(p, A[0], size);
	}
	
	return p;
}

void *TC_realloc_arrayN(int ndim, void *main_array, size_t size, ...)
{
	va_list ap;
	void *p;
	int *array;
	int a;
	
	array = (int *)AR_allocate(ndim, sizeof(int));
	
	va_start(ap, size);
	for (a = 0; a < ndim; ++a) {
		array[a] = va_arg(ap, int);
		if (array[a] < -1) {
			va_end(ap);
			return 0;
		}
	}
	va_end(ap);
	
	p = TC_realloc_arrayN2(ndim, main_array, array, size);
	AR_free(array);
	
	return p;
}

void TC_free_arrayN(void *p, int n)
{
	void **i;
	int a, s;
	
	if (p == 0)
		return;
	else if (n == 1)
		AR_free(p);
	else {
		i = (void **)p;
		s = TC_read_array_size(p, sizeof(void *));
		
		for (a = 0; a < s; ++a)
			TC_free_arrayN(i[a], n - 1);
		AR_free(p);
	}
}

#define DECLARE_ARRAY(type, function)\
type **function(int first, int second)\
{\
	return (type **)TC_declare_arrayN(2, sizeof(type), first, second);\
}

DECLARE_ARRAY(char, TC_declare_char)
DECLARE_ARRAY(short int, TC_declare_short_int)
DECLARE_ARRAY(int, TC_declare_int)
DECLARE_ARRAY(float, TC_declare_float)
DECLARE_ARRAY(double, TC_declare_double)

static void **TC_declare_array(int first, int second, size_t size)
{
	return (void **)TC_declare_arrayN(2, size, first, second);
}

static void **TC_realloc_array(void **array, size_t size, int first, int second, int ext1, int ext2)
{
	int d1, d2;
	int a;
	
	if (array == 0)
		return TC_declare_array(((first == -1) ? 0 : first) + ext1,
				((second == -1) ? 0 : second) + ext2, size);
	else if (first == -1)
		first = TC_read_array_size(array, sizeof(void *));
	
	if (second == -1)
		second = TC_read_array_size(array[0], size);
	
	d1 = first + ext1;
	d2 = second + ext2;
	
	for (a = d1; a < first; ++a)
		AR_free(array[a]);
	
	array = (void **)AR_reallocate(array, d1, sizeof(void *));
	
	if (d2 != second) {
		for (a = 0; a < d1 && a < first; ++a)
			array[a] = AR_reallocate(array[a], d2, size);
	}
	for (a = first; a < d1; ++a)
		array[a] = AR_reallocate(array[a], d2, size);
	
	return array;
}

#define REALLOC_ARRAY(type, function)\
type **function(type **array, int first, int second, int ext1, int ext2)\
{\
	return (type **)TC_realloc_array((void **)array, sizeof(type), first, second, ext1, ext2);\
}

REALLOC_ARRAY(char, TC_realloc_char)
REALLOC_ARRAY(short int, TC_realloc_short_int)
REALLOC_ARRAY(int, TC_realloc_int)

#define NEW_REALLOC_ARRAY(type, function, free_function)\
type **function(type **array, int ext1, int ext2)\
{\
	type **new_array;\
	int first, l1;\
	int second, l2;\
	int a, b;\
\
	first = TC_read_array_size(array, sizeof(type *));\
	second = TC_read_array_size(array[0], sizeof(type));\
\
	if (ext1 == -1)\
		ext1 = first;\
	if (ext2 == -1)\
		ext2 = second;\
\
	l1 = MIN(ext1, first);\
	l2 = MIN(ext2, second);\
\
	new_array = (type **)TC_declare_arrayN(2, sizeof(type), ext1, ext2);\
\
	for (a = 0; a < l1; ++a)\
		for (b = 0; b < l2; ++b)\
			new_array[a][b] = array[a][b];\
\
	free_function(array);\
\
	return new_array;\
}

NEW_REALLOC_ARRAY(char, TC_new_realloc_char, TC_free_char)
NEW_REALLOC_ARRAY(short int, TC_new_realloc_int, TC_free_short_int)
NEW_REALLOC_ARRAY(int, TC_new_realloc_int, TC_free_int)
NEW_REALLOC_ARRAY(float, TC_new_realloc_float, TC_free_float)
NEW_REALLOC_ARRAY(double, TC_new_realloc_double, TC_free_double)

#define FREE_ARRAY(type, function)\
void function(type **array)\
{\
	TC_free_arrayN(array, 2);\
}

FREE_ARRAY(char, TC_free_char)
FREE_ARRAY(short int, TC_free_short_int)
FREE_ARRAY(int, TC_free_int)
FREE_ARRAY(float, TC_free_float)
FREE_ARRAY(double, TC_free_double)

int TC_read_array_size(void *array, size_t size)
{
	AR_mem_control *M;

	if (array == 0)
		return 0;

	M = (AR_mem_control *)array;
	--M;

	if (size == 0 || size == M->obj_size)
		return M->nobj;
	else {
		std::cerr << '\n' << PROGRAM << ": WARNING: read_array_size: param 'size' is different from mem_control::obj_size.\n";
		return (M->nobj * M->obj_size) / size;
	}
}

int *TC_ga_memcpy_int(const int *array1, int *array2, int n)
{
	int a;
	
	for (a = 0; a < n; ++a)
		array2[a] = array1[a];
	
	return array2;
}

float *TC_ga_memcpy_float(const float *array1, float *array2, int n)
{
	int a;
	
	for (a = 0; a < n; ++a)
		array2[a] = array1[a];
	
	return array2;
}

double *TC_ga_memcpy_double(const double *array1, double *array2, int n)
{
	int a;
	
	for (a = 0; a < n; ++a)
		array2[a] = array1[a];
	
	return array2;
}

char **TC_duplicate_char(char **array, int len, int field)
{
	return TC_copy_char(array, TC_declare_char(len, field), len, field);
}

char *TC_duplicate_string(const char *string)
{
	char *buf = 0;
	int l;
	
	if (string && (l = strlen(string))) {
		buf = (char *)AR_allocate(l + 1, sizeof(char));
		strcpy(buf, string);
	}
	
	return buf;
}

int **TC_duplicate_int(int **array, int len, int field)
{
	return TC_copy_int(array, TC_declare_int(len, field), len, field);
}

float **TC_duplicate_float(float **array, int len, int field)
{
	return TC_copy_float(array, TC_declare_float(len, field), len, field);
}

double **TC_duplicate_double(double **array, int len, int field)
{
	return TC_copy_double(array, TC_declare_double(len, field), len, field);
}

char **TC_copy_char(char **array1, char **array2, int len, int number_field)
{
	int a;
	
	if (array1 == 0)
		return 0;
	
	if (len == -1)
		len = TC_read_array_size(array1, sizeof(char *));
	
	if (number_field == -1) {
		number_field = TC_read_array_size(array1[0], sizeof(char));
		for (a = 0; a < len; ++a)
			number_field = MAX(number_field, strlen(array1[a])) + 1;
	}
	if (array2)
		TC_free_char(array2);
	
	array2 = TC_declare_char(len, number_field);
	
	for (a = 0; a < len; ++a)
		strcpy(array2[a], array1[a]);
	
	return array2;
}

int **TC_copy_int(int **array1, int **array2, int len, int number_field)
{
	int a;
	
	if (array1 == 0)
		return 0;
	
	if (len == -1)
		len = TC_read_array_size(array1, sizeof(int *));
	if (number_field == -1)
		number_field = TC_read_array_size(array1[0], sizeof(int));
	
	if (array2)
		TC_free_int(array2);
	
	array2 = TC_declare_int(len, number_field);
	
	for (a = 0; a < len; ++a)
		TC_ga_memcpy_int(array1[a], array2[a], number_field);
	
	return array2;
}

float **TC_copy_float(float **array1, float **array2, int len, int number_field)
{
	int a;
	
	if (array1 == 0)
		return 0;
	
	if (len == -1)
		len = TC_read_array_size(array1, sizeof(float *));
	if (number_field == -1)
		number_field = TC_read_array_size(array1[0], sizeof(float));
	
	if(array2)
		TC_free_float(array2);
	
	array2 = TC_declare_float(len, number_field);
	
	for (a = 0; a < len; ++a)
		TC_ga_memcpy_float(array1[a], array2[a], number_field);
	
	return array2;
}

double **TC_copy_double(double **array1, double **array2, int len, int number_field)
{
	int a;
	
	if (array1 == 0)
		return 0;
	
	if (len == -1)
		len = TC_read_array_size(array1, sizeof(double *));
	if (number_field == -1)
		number_field = TC_read_array_size(array1[0], sizeof(double));
	
	if (array2)
		TC_free_double(array2);
	
	array2 = TC_declare_double(len, number_field);
	
	for (a = 0; a < len; ++a)
		TC_ga_memcpy_double(array1[a], array2[a], number_field);
	
	return array2;
}

std::ostream &AR_print_mem_usage(std::ostream &out, const char *comment)
{
	const float div = 1024.f * 1024.f;
	std::streamsize prec;

	prec = out.precision(3);
	showpoint(out);
	
	out << '\n' << comment << ": memory usage:  current= " << (g_alloc_mem / div) << "MB, max= "
		<< (g_tot_mem / div) << "MB\n";

	out.precision(prec);
	noshowpoint(out);

	return out;
}
