/*
 * File: util_pdb.cpp
 * Created: 09.05.2006
 * Last Modified: 13.09.2006
 * Author: Adrian Rospondek
 */


#include <iostream>
#include <sstream>
#include <utility>
#include <pthread.h>
#include <stdlib.h>
#include <string.h>
#include "AR_Config.hpp"
#include "config.hpp"
#include "input.hpp"
#include "util.hpp"
#include "util_io.hpp"
#include "util_mem.hpp"
#include "util_pdb.hpp"
#include "util_str.hpp"



typedef std::vector<std::string> AR_pdb_list_type;
typedef std::pair<AR_pdb_list_type::const_iterator, AR_pdb_list_type::const_iterator> AR_range_type;


static int g_n_threads = 0;
static pthread_t *g_thread_id = 0;



bool AR_parse_pdb_list(char *pdb_list[], int n_pdb, std::vector<AR_pdb_chain *> &out_list)
{
	const int len = 64;
	char buf[len];
	char *p = 0;
	AR_pdb_chain *chain = 0;
	
	if (pdb_list == 0 || n_pdb < 1)
		return false;

	for (int i = 0; i < n_pdb; ++i) {
		if (!TC_is_number(pdb_list[i])) {
			chain = new AR_pdb_chain;
			
			(chain->coords).start = (chain->coords).end = 0;
			if (i + 1 < n_pdb && TC_is_number(pdb_list[i + 1]))
				(chain->coords).start = atoi(pdb_list[i + 1]);
			if (i + 2 < n_pdb && TC_is_number(pdb_list[i + 2]))
				(chain->coords).end = atoi(pdb_list[i + 2]);
			
			if (strlen(pdb_list[i]) > 5) {
				strncpy(buf, pdb_list[i], len);
				chain->pdb_id = strtok(buf, "_");

				if ((p = strtok(0, "_")) && (chain->coords).start == 0)
					(chain->coords).start = atoi(p);
				if ((p = strtok(0, "_")) && (chain->coords).end == 0)
					(chain->coords).end = atoi(p);
			}
			else
				chain->pdb_id = pdb_list[i];

			out_list.push_back(chain);
		}
	}

	return true;
}

int AR_check_pdb_location(std::string &pdb_id)
{
	int location;

	if (AR_is_pdb_file(pdb_id))
		location = LOCAL_PDB;
	else if (AR_is_pdb_file(pdb_id + ".pdb")) {
		pdb_id += ".pdb";
		location = LOCAL_PDB;
	}
	else if (AR_is_pdb_name(pdb_id))
		location = REMOTE_PDB;
	else
		location = NOT_PDB;

	return location;
}

std::string AR_download_pdb(const std::string &pdb_id)
{
	if (pdb_id.empty())
		return std::string();

	std::ostringstream buf;

	buf << pdb_id << ".pdb";
	std::string pdb_file = buf.str();

	buf.str("");
	buf << g_conf.extract_pdb() << " -netfile " << pdb_id << " > " << AR_get_cache_dir() << pdb_file << " 2>/dev/null";
	std::string cmd = buf.str();

#if DEBUG_PDB
	std::cout << "\nDOWNLOADING PDB FILE (id: " << pdb_id << ")\n\tcmd: " << cmd << '\n';
#endif
	system(cmd.c_str());

	if (!AR_is_pdb_file(pdb_file))
		pdb_file.clear();

#if DEBUG_PDB
	std::cout << "\nPDB FILE (id: " << pdb_id << ") SAVED AS:\n\t" << pdb_file << '\n';
#endif
	return pdb_file;
}

std::string AR_extract_pdb(const std::string &pdb_file, const int start, const int end)
{
	if (pdb_file.empty() || (start == 0 && end == 0))
		return pdb_file;

	std::string pdb_file2 = AR_check_file_exists(pdb_file);
	std::string pdb_id = AR_get_pdb_id(pdb_file2);

	std::ostringstream buf;

	buf << pdb_id << '_' << start << '_' << end << ".pdb";
	std::string pdb_file3 = buf.str();

	buf.str("");
	buf << g_conf.extract_pdb() << " -infile " << pdb_file2 << " -chain FIRST -coor " << start << ' ' << end << " > "
		<< AR_get_cache_dir() << pdb_file3;
	std::string cmd = buf.str();

#if DEBUG_PDB
	std::cout << "\nEXTRACTING PDB FILE '" << pdb_file << "'\n\tcmd: " << cmd << '\n';
#endif
	system(cmd.c_str());

	if (!AR_is_pdb_file(pdb_file3))
		pdb_file3.clear();

#if DEBUG_PDB
	std::cout << "\nPDB FILE '" << pdb_file << "' EXTRACTED TO:\n\t" << pdb_file3 << '\n';
#endif
	return pdb_file3;
}

static void *AR_thread_func(void *arg)
{
	AR_range_type *range = (AR_range_type *)arg;
	const int n_pdb = range->second - range->first;
	std::vector<std::string> *pdb_files = new std::vector<std::string>(n_pdb);

	for (int i = 0; range->first != range->second; ++(range->first))
		(*pdb_files)[i++] = AR_download_pdb(*(range->first));

	delete range;
	pthread_exit(pdb_files);
}

int AR_begin_mt_download_pdb(const std::vector<std::string> &pdb_list, const int max_n_thr)
{
	const int n_pdb = pdb_list.size();
	int pdb_per_thr;

	if (n_pdb > max_n_thr) {
		pdb_per_thr = n_pdb / max_n_thr;
		g_n_threads = max_n_thr;
	}
	else {
		pdb_per_thr = 1;
		g_n_threads = n_pdb;
	}

	g_thread_id = new pthread_t[g_n_threads];
	const int r = n_pdb - (g_n_threads * pdb_per_thr);
	std::vector<std::string>::const_iterator it = pdb_list.begin();
	AR_range_type *range = 0;
	int n_thr = 0, i;

	for (++pdb_per_thr, i = 0; i < r; ++i, it += pdb_per_thr) {
		range = new AR_range_type(it, it + pdb_per_thr);

		if (pthread_create(g_thread_id + n_thr, 0, AR_thread_func, range) == 0)
			++n_thr;
	}
	for (--pdb_per_thr; i < g_n_threads; ++i, it += pdb_per_thr) {
		range = new AR_range_type(it, it + pdb_per_thr);

		if (pthread_create(g_thread_id + n_thr, 0, AR_thread_func, range) == 0)
			++n_thr;
	}
	g_n_threads = n_thr;

	return g_n_threads;
}

int AR_end_mt_download_pdb(std::vector<std::string> &out_list)
{
	std::vector<std::string> *pdb_files = 0;
	std::vector<std::string>::const_iterator it;
	int n_thr = 0;
	
	for (int i = 0; i < g_n_threads; ++i) {
		if (pthread_join(g_thread_id[i], (void **)&pdb_files) == 0) {
			++n_thr;
			for (it = pdb_files->begin(); it != pdb_files->end(); ++it)
				out_list.push_back(*it);

			delete pdb_files;
		}
	}

	delete[] g_thread_id;
	g_thread_id = 0;
	g_n_threads = 0;

	return n_thr;
}
