/*
 * File: util_str.cpp
 * Created: 06.06.2006
 * Last Modified: 13.09.2006
 * Author: Adrian Rospondek
 */


#include <iostream>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "globals.hpp"
#include "util.hpp"
#include "util_mem.hpp"
#include "util_str.hpp"



char *TC_generate_string(int len, char x)
{
	int a;
	char *string;

	string = (char *)AR_allocate(len + 1, sizeof(char));

	for (a = 0; a < len; a++)
		string[a] = x;
	string[a] = '\0';

	return string;
}

int TC_vstrcmp(const char *s1, const char *s2)
{
	if (!s1 && !s2)
		return 0;
	else if (!s1 || !s2)
		return 1;
	else
		return strcmp(s1, s2);
}

int TC_vstrncmp(const char *s1, const char *s2, int n)
{
	if (!s1 && !s2)
		return 0;
	else if (!s1 || !s2)
		return 1;
	else
		return strncmp(s1, s2, n);
}

/**
 * @brief Compares two strings.
 * @return One if s1 matches s2 and zero otherwise.
 */
int TC_case_insensitive_strcmp(const char *s1, const char *s2)
{
	if (s1 == s2)
		return 1;
	else if (!s1 || !s2)
		return 0;
	else {
		while (*s1 && *s2 && tolower(*s1) == tolower(*s2)) {
			++s1;
			++s2;
		}
	}

	return !(*s1 || *s2);
}

char *TC_after_strstr(const char *string, const char *token)
{
	char *p;

	if (!string || !token)
		return 0;
	else if ((p = (char*)strstr(string, token)))
		return p + strlen(token);

	return 0;
}

char *TC_upper_string(char *string)
{
	if (string) {
		char *s;

		for (s = string; *s; ++s)
			*s = toupper(*s);
	}

	return string;
}

char *TC_lower_string(char *string)
{
	if (string) {
		char *s;
		
		for (s = string; *s; ++s)
			*s = tolower(*s);
	}

	return string;
}

/**
 * @brief Replaces every occurence of token t with token r in string_in.
 */
char *TC_substitute(char *string_in, const char *t, const char *r)
{
	char *string_out;
	char *p, *heap_in;
	int delta, l;

	heap_in = string_in;
	l = TC_read_array_size(string_in, 0) + 1;
	string_out = (char *)AR_allocate(l, sizeof(char));
	delta = strlen(r) - strlen(t);
	delta = (delta < 0) ? 0 : delta;

	while ((p = strstr(string_in, t)) != 0) {
		p[0] = '\0';
		if (delta) {
			l += delta;
			string_out = (char *)AR_reallocate(string_out, l, sizeof(char));
		}
		strcat(string_out, string_in);
		strcat(string_out, r);
		string_in = p + strlen(t);
	}
	strcat(string_out, string_in);

	if (l < strlen(string_out)) {
		string_in = (char *)AR_reallocate(string_in, strlen(string_out), sizeof(char));

		if (string_in != heap_in) {
			std::cerr << '\n' << PROGRAM << ": WARNING: Change of pointer address (substitute).\n";
		}
	}
	strcpy(heap_in, string_out);
	AR_free(string_out);

	return heap_in;
}

char **TC_clean_string(int n, char **string)
{
	int a, b, c, l;

	if (!string)
		return string;

	for (a = 0; a < n; a++) {
		if (!string[a])
			continue;
		l = strlen(string[a]);
		for (b = 0; b < l; b++) {
			c = string[a][b];
			if (!isgraph(c) && c != '\n' && c != '\t' && c != ' ')
				string[a][b] = ' ';
		}
	}

	return string;
}

void TC_translate_name(char *name)
{
	char *s;

	for (s = name; *s; ++s) {
		if (isspace(*s))
			*s = '\0';
		else if (!isgraph(*s) || *s == ';' || *s == ':' || *s == '(' || *s == ')')
			*s = '_';
	}
}

void TC_remove_charset(char *seq, const char *set)
{
	int a, b, l;
	char *set2;

	set2 = (char *)AR_allocate(256, sizeof(char));

	if (TC_strm(set, "!alnum")) {
		for (b = 0, a = 1; a < 256; a++)
			if (!isalnum(a))
				set2[b++] = a;
	}
	else if (TC_strm(set, "ungap")) {
		strncpy(set2, GAP_LIST, 256 - 1);
	}
	else {
		strncpy(set2, set, 256 - 1);
	}

	l = strlen(seq);
	for (b = 0, a = 0; a <= l; a++) {
		if (strchr(set2, seq[a]) == 0)
			seq[b++] = seq[a];
	}
	seq[b] = '\0';

	AR_free(set2);
}

char **TC_string2list(const char *string)
{
	return TC_string2list2(string, SEPARATORS);
}

/**
 * @brief Breaks down a list of words separated by any legal separator
 * and put them in an array of the right size.
 * @return list a list of char with the size written in list[0].
 */
char **TC_string2list2(const char *string, const char *separators)
{
	char *buf, *s;
	char **list;
	int n, max_len;

	if (string == 0)
		return 0;
	buf = (char *)AR_allocate(strlen(string) + 1, sizeof(char));

	n = max_len = 0;
	strcpy(buf, string);

	s = strtok(buf, separators);
	while (s != 0) {
		n++;
		max_len = MAX(max_len, (strlen(s)));
		s = strtok(0, separators);

	}

	if (n == 0) {
		AR_free(buf);
		return 0;
	}

	list = (char **)TC_declare_arrayN(2, sizeof(char), n + 1, max_len + 1);

	n = 1;
	strcpy(buf, string);

	s = strtok(buf, separators);
	while (s != 0) {
		strncpy(list[n++], s, max_len);
		s = strtok(0, separators);
	}

	sprintf(list[0], "%d", n);
	AR_free(buf);

	return list;
}

int *TC_string2num_list(const char *string)
{
	return TC_string2num_list2(string, SEPARATORS);
}

/**
 * @brief Breaks down a list of numbers separated by any legal separator
 * and put them in an array of the right size.
 * Skips non numbers.
 * @return list a list of integer.
 */
int *TC_string2num_list2(const char *string, const char *separators)
{
	char **nlist;
	int *list;
	int a, m, n;

	nlist = TC_string2list2(string, separators);
	if (nlist == 0)
		return 0;

	n = atoi(nlist[0]);
	list = (int *)AR_allocate(n, sizeof(int));

	for (m = 1, a = 1; a < n; a++) {
		if (TC_is_number(nlist[a]))
			list[m++] = atoi(nlist[a]);
	}
	list[0] = m;
	TC_free_char(nlist);

	if (m == 0) {
		AR_free(list);
		return 0;
	}
	else {
		return list;
	}

	return 0;
}

void TC_string_array_upper(char **string, int n)
{
	int a;

	for (a = 0; a < n; a++)
		TC_upper_string(string[a]);
}

void TC_string_array_lower(char **string, int n)
{
	int a;

	for (a = 0; a < n; a++)
		TC_lower_string(string[a]);
}

/**
 * @brief Pads strings so that they all have the same length.
 */
char **TC_padd_strings(char **string, int n, char pad)
{
	int max_len, a;
	char *buf;

	max_len = TC_get_longest_string(string, n, 0, 0);
	for (a = 0; a < n; a++) {
		buf = TC_generate_string(max_len - strlen(string[a]), pad);
		strcat(string[a], buf);
		AR_free(buf);
	}

	return string;
}

/**
 * @brief Breaks down the argv command line in smaller units, breaking at every separator.
 */
char **TC_break_list(char **argv, int *argc, const char *separators)
{
	int a, b;
	int n_in;
	char **out;
	char **ar;
	int n_ar;

	n_in = argc[0];
	argc[0] = 0;

	if (n_in >= MAX_N_PARAM) {
		std::cerr << '\n' << PROGRAM << ": ERROR: Too many parameters (break_list).\n"
				  << "\tRecompile with MAX_N_PARAM set at a higher value.\n";
		return 0;
	}

	out = (char **)AR_allocate(MAX_N_PARAM, sizeof(char *));

	for (a = 0; a < n_in; a++) {
		ar = TC_get_list_of_tokens(argv[a], separators, &n_ar);
		for (b = 0; b < n_ar; b++) {
			out[argc[0]] = (char *)AR_allocate(strlen(ar[b]) + 1, sizeof(char));
			strcpy(out[argc[0]++], ar[b]);
		}
		TC_free_char(ar);
		ar = 0;
	}
	TC_free_char(ar);

	return out;
}

char **TC_merge_list(char **argv, int *argc)
{
	int a, b;
	int n_in;
	char **out;
	char current[STRING];

	out = TC_declare_char(argc[0], STRING);
	n_in = argc[0];
	argc[0] = 0;

	a = 0;
	while (a < n_in && !TC_is_parameter(argv[a])) {
		strncpy(out[argc[0]++], argv[a], STRING - 1);
		argv[a][0] = '\0';
		a++;
	}

	for (a = 0; a < n_in; a++) {
		if (TC_is_parameter(argv[a])) {
			strncpy(out[argc[0]++], argv[a], STRING - 1);
			strncpy(current, argv[a], STRING - 1);

			for (b = 0; b < n_in;) {
				if (TC_is_parameter(argv[b]) && TC_strm(current, argv[b])) {
					argv[b][0] = '\0';
					b++;

					while (b < n_in && !TC_is_parameter(argv[b])) {
						if (argv[b][0]) {
							strncpy(out[argc[0]++], argv[b], STRING - 1);
							argv[b][0] = '\0';
						}
						b++;
					}
				}
				else
					b++;
			}
		}
	}

	TC_free_char(argv);

	return out;
}

char **TC_get_list_of_tokens(const char *in_string, const char *separators, int *n_tokens)
{
	char **list = 0;
	char *p = 0;
	char *string;

	n_tokens[0] = 0;

	if (!in_string || !in_string[0]);
	else if (in_string[0] == '[') {
		list = TC_declare_char(1, strlen(in_string) + 1);

		strcpy(list[n_tokens[0]], in_string);
		n_tokens[0]++;
	}
	else {
		list = TC_declare_char(strlen(in_string) + 1, 1);
		string = (char *)AR_allocate(strlen(in_string) + 1, sizeof(char));

		strcpy(string, in_string);

		if (!separators)
			separators = SEPARATORS;

		while ((p = strtok((p == 0) ? string : 0, separators)) != 0) {
			list[n_tokens[0]] = (char *)AR_reallocate(list[n_tokens[0]], strlen(p) + 1, sizeof(char));

			strcpy(list[n_tokens[0]], p);
			n_tokens[0]++;
		}

		AR_free(string);
	}

	return list;
}

char **TC_push_string(const char *val, char **stack, int *nval, int position)
{
	char **new_stack;
	int a;

	if (!val || nval[0] <= position)
		return stack;

	nval[0]++;
	new_stack = (char **)AR_allocate(nval[0], sizeof(char *));
	// XXX: it may cause the problem !!
	new_stack[position] = const_cast<char *>(val);

	for (a = 0; a < position; ++a)
		new_stack[a] = stack[a];
	for (a = position + 1; a < nval[0]; ++a)
		new_stack[a] = stack[a - 1];

	AR_free(stack);

	return new_stack;
}

/**
 * @brief Returns the offset of the last occurence of name in name_list.
 */
int TC_name_is_in_list(const char *name, char **name_list, int n_name, int len)
{
	int a;
	int pos = -1;

	if (name_list == 0 || name == 0)
		return -1;

	for (a = 0; a < n_name; a++) {
		if (len != -1) {
			if (strncmp(name, name_list[a], len) == 0)
				pos = a;
		}
		else if (TC_strm(name, name_list[a]))
			pos = a;
	}

	return pos;
}

int TC_is_in_set(char c, const char *list)
{
	if (strchr(list, c) != 0)
		return 1;

	return 0;
}

char *TC_check_list_for_dup(char **list, const int ne)
{
	int a, b;

	for (a = 0; a < ne - 1; a++)
		for (b = a + 1; b < ne; b++)
			if (TC_strm(list[a], list[b]))
				return list[a];

	return 0;
}

int TC_get_longest_string(char **array, int n, int *len, int *index)
{
	int max_len, max_index;
	int a, l;

	if (array == 0 || n < 1)
		return -1;

	max_len = strlen(array[0]);
	max_index = 0;
	for (a = 1; a < n; a++) {
		l = strlen(array[a]);
		if (l > max_len) {
			max_len = l;
			max_index = a;
		}
	}

	if (index != 0)
		index[0] = max_index;
	if (len != 0)
		len[0] = max_len;

	return max_len;
}

int TC_get_shortest_string(char **array, int n, int *len, int *index)
{
	int min_len, min_index;
	int a, l;

	if (array == 0 || n < 1)
		return -1;

	min_len = strlen(array[0]);
	min_index = 0;
	for (a = 1; a < n; a++) {
		l = strlen(array[a]);
		if (l < min_len) {
			min_len = l;
			min_index = a;
		}
	}

	if (index != 0)
		index[0] = min_index;
	if (len != 0)
		len[0] = min_len;

	return min_len;
}

int TC_is_alpha_line(const char *buf)
{
	const char *s;
	
	if (buf == 0)
		return 0;

	for (s = buf; *s; ++s)
		if (isalpha(*s))
			return 1;

	return 0;
}

int TC_is_blank(const char *buf)
{
	const char *s;

	if (buf == 0)
		return 0;

	for (s = buf; *s; ++s)
		if (isalnum(*s))
			return 0;

	return 1;
}

int TC_is_number(const char *num)
{
	while (*num && strchr("0123456789.-+", *num))
		++num;

	return *num == '\0';
}
